// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda;

import java.util.Properties;
import java.util.Enumeration;
import org.isam.perv.IsamClassLoader;
import org.isam.util.xml.*;

public class Exehda
{
    private static String CFG_URL_PROP = "config.url";
    private static ServiceManager serviceManager;
    private static ProfileManager profileManager;
    private static Properties exehdaProps;
    private static boolean    exehdaIsUp = false;
    private static String     bootProfile = null;

    static {
            // enable ISAM URL protocol handlers
        setupProtocolHandlers();
            // init middleware properties
        exehdaProps = new Properties();
            // init profile manager
        profileManager = new ProfileManager();
            // init services manager
        serviceManager = new ServiceManager(profileManager);
    }

    public static void
    bootstrap( String profileName )
        {
//             do {
            shutdown();

            logInfoMessage("");
            logInfoMessage("Bootstraping EXEHDA services [profile:"+ profileName+"]...");
            logInfoMessage("");
            
            if ( serviceManager.bootstrapServices(profileName) ) {
                logMessage("EXEHDA is ready.");
                exehdaIsUp = true;
            }
            else {
                logMessage("EXEHDA initialization failed! Exiting...");
                throw new RuntimeException(
                    "Bootstraping of some middleware service has failed. Aborting...");
//                 System.exit(-1);
            }
//             }
//             while ( bootProfile != profileName );

                //
                // XXX: force local HostId initilization (if not yet inialized during
                // services bootstrap)
                //
            HostId.getLocalHost();
        }

    public static void
    shutdown()
        {
            if ( exehdaIsUp ) {
                logInfoMessage("Shuting down EXEHDA services...");
                if ( serviceManager.shutdownServices() ) {
                        //
                        // XXX: force clean-up of unused objects
                        //
                    System.gc();
                    logMessage("...shutdown completed.");
                    exehdaIsUp = false;
                }
                else {
                    logMessage("EXEHDA Shutdown failed! Aborting...");
                    System.exit(-1);
                }
            }
        }

    public static Object
    getService( String serviceName )
        {
            serviceManager.startService(serviceName);
            return serviceManager.getServiceProxy(serviceName);
        }

    public static String
    getProperty( String propName, String defaultValue )
        {
            String r = profileManager.getProperty(propName);

            if ( r != null ) {
                return r;
            }
            else {
                r = exehdaProps.getProperty(propName);
                return r == null ? defaultValue : r;
            }
        }

    public static String
    getServiceProperty( String serviceName, String propName, String defValue)
        {
            String r = profileManager.getServiceProperty(serviceName, propName);

            return r == null ? defValue : r;
        }

    public static int
    getServiceProperty( String serviceName, String propName, int defValue)
        {
            String r = profileManager.getServiceProperty(serviceName, propName);

            return r == null ? defValue : Integer.parseInt(r);
        }
    
    public static String
    getActiveProfile()
        {
            return serviceManager.getActiveProfile();
        }
    
    public static void
    setProperty( String propName, String propValue )
        {
            exehdaProps.put(propName, propValue);
        }

    public static void
    addProfile( XmlElement profileData )
        {
            profileManager.addProfileData(profileData);
        }

    
        /**
         * Facility for logging core messages in a standardized way.
         *
         * <b>Should not be used outside the core!</b>
         *
         * @param msg an <code>Object</code> value
         */
    static void
    logMessage(Object msg)
        {
            System.err.println(">>>> "+msg);
        }

        /**
         * Facility for logging core messages in a standardized way.
         *
         * <b>Should not be used outside the core!</b>
         *
         * @param msg an <code>Object</code> value
         */
    static void
    logInfoMessage(Object msg)
        {
            System.err.println("**** "+msg);
        }

    public static void
    main( String[] args ) throws Exception
        {
            String optProfile = "node";

            try {
                for ( int idx = 0; idx < args.length; idx++ ) {
                    String opt = args[idx];

                    if ( "--profile".equals(opt) ) {
                        optProfile = args[++idx];
                    }
                }
            }
            catch (ArrayIndexOutOfBoundsException e) {
                logMessage("Missing parameter: " + e);
                System.exit(-1);   
            }
            
                // load any profiles defined in exehda-services.xml
            loadExehdaServices();

            bootstrap(optProfile);

            try {
                Object lock = new Object();

                synchronized (lock) {
                    lock.wait();
                }
            }
            catch (Exception e) {
                    // empty
            }
        }

    
        /**
         * Enable ISAM-defined URL protocol handlers
         *
         */
    private static void
    setupProtocolHandlers()
        {
            final String PROP_PROT_HANDLERS = "java.protocol.handler.pkgs";
            final String ISAM_PROT_HANDLERS = "org.isam.perv.protocols";
            
                // enable isam protocol handlers
            try {
                
                try {
                        // XXX: for Java Web Start
                    System.setProperty(PROP_PROT_HANDLERS, ISAM_PROT_HANDLERS);
                }
                catch (Exception e) {
                        // XXX: for Personal Java 
                    System.getProperties().put(PROP_PROT_HANDLERS, ISAM_PROT_HANDLERS);
                }
            }
            catch (Exception e) {
                System.err.println("Could not setup protocol handlers due to " + e
                                   +". Be advised that some things may not work propertly.");
                System.err.println("Property '"+PROP_PROT_HANDLERS+"' is currently set to '"
                                   +System.getProperty(PROP_PROT_HANDLERS)+"'");
            }
        }

    
        /**
         * Load profiles defined in file exehda-services.xml
         *
         */
    private static void
    loadExehdaServices()
        {
            try {            
                    // load middleware profiles
                String customURL = System.getProperty("exehda."+CFG_URL_PROP);
                java.net.URL configURL = null;
                if ( customURL == null ) {
                
                    ClassLoader cl = IsamClassLoader.getInstance();
                    configURL = cl.getResource(
                        "org/isam/exehda/exehda-services.xml");
                }
                else if ( "".equals(customURL.trim()) ) {
                        // do nothing
                    return;
                }
                else {
                    configURL = new java.net.URL(customURL);
                }
            
                Exehda.setProperty(CFG_URL_PROP, configURL.toString());

                XmlElement docroot = XmlElement.parse(
                    new java.io.InputStreamReader(configURL.openStream()));

                XmlElement configDoc = (XmlElement)
                    XmlElement.selectByPath(docroot, "/EXEHDA", null).elementAt(0);

                if ( ! "EXEHDA".equals(configDoc.getTag()) ) {
                    throw new java.io.IOException("exehda-services.xml corrupted!");
                }
                
                for ( Enumeration profileEnum = configDoc.getChildren();
                      profileEnum.hasMoreElements();
                      ) {
                    profileManager.addProfileData((XmlElement)profileEnum.nextElement());
                }
                
            }
            catch (Exception e) {
                logMessage("EXEHDA initialization failed.");
                e.printStackTrace();
//             System.exit(-1);
            }
        }
}


