// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services;

import org.isam.exehda.ApplicationId;
import org.isam.exehda.ResourceName;
import java.util.Hashtable;

public interface Collector
{
    public static final String SERVICE_NAME = "collector";

    
        /**
         * Represents an elementary source of monitoring data.
         *
         */
    public static class Sensor
        implements java.io.Serializable
    {
            /**
             * Creates a new <code>Sensor</code> instance.
             *
             * Note: the key must be serializable in order to allow the sensor object to
             * be transmited between hosts.
             *
             * @param n the sensor name
             * @param t the sensor type
             * @param k key 
             */
        public Sensor(ResourceName n, Class t, Object k)
            {
                if ( n == null ) throw new IllegalArgumentException("null name not allowed");
                if ( t == null ) throw new IllegalArgumentException("null type not allowed");
                
                this.name = n;
                this.type = t;
                this.key = k;
                this.hash = n.hashCode();
            }
        
            /**
             * Returns the name of this sensor in the forma of a Resource Name.
             *
             * @return this sensor name
             */
        public ResourceName getName()
            {
                return name;
            }
 
            /**
             * Returns the type of the data produced by this sensor.
             *
             * @return a Class representing the type of the sensor.
             */
        public Class getType()
            {
                return type;
            }
 
            /**
             * Returns the unique key associated with this sensor. This key is typically
             * used to link the sensor object to a peer raw representation used
             * internally by the sensor provider.
             *
             * Note: the key must be serializable in order to allow the sensor object to
             * be transmited between hosts.
             *
             * @return the key assigned to this sensor by the sensor provider.
             */
        public Object getKey()
            {
                return key;
            }

        public String toString()
            {
                return name.toString();
            }

        public int hashCode()
            {
                return hash;
            }

        public boolean equals(Object o)
            {
                if ( o != null ) {
                    if ( o instanceof Sensor ) {
                        Sensor s2 = (Sensor) o;

                        return ( (hash == s2.hash)
                                 && name.equals(s2.name)
                                 && type.equals(s2.type)
                                 && key.equals(s2.key) );
                    }
                }
                return false;
            }

        private ResourceName name;
        private Class type;
        private Object key;
        private int hash;
    }

        /**
         * Represents a sensor configuration parameter.
         *
         */
    public static class SensorParameter
        implements java.io.Serializable
    {
            /**
             * Creates a new <code>SensorParameter</code> instance.
             *
             * @param n parameter name
             * @param t parameter type
             * @param s sensor who owns this parameter
             */
        public SensorParameter(String n, Class t, Sensor s)
            {
                this.name = n;
                this.type = t;
                this.sensor = s;
                this.hash = n.hashCode();
            }
            /**
             * Returns the name of this parameter.
             * @return parameter name
             */
        public String getName()
            {
                return this.name;
            }
            
            /**
             * Returns the type of this parameter.
             * @return parameter type
             */
        public Class getType()
            {
                return this.type;
            }
            
            /**
             * Returns the sensor that owns this parameter.
             * @return Onwer sensor
             */
        public Sensor getSensor()
            {
                return this.sensor;
            }

        public int hashCode()
            {
                return hash;
            }

        public boolean equals(Object o)
            {
                if ( o != null ) {
                    if ( o instanceof SensorParameter ) {
                        SensorParameter p2 = (SensorParameter) o;
                        
                        return ( (hash == p2.hash)
                                 && name.equals(p2.name)
                                 && type.equals(p2.type)
                                 && sensor.equals(p2.sensor) );
                    }
                }
                return false;
            }

        private String name;
        private Class type;
        private Sensor sensor;
        private int hash;
    }

    
        /**
         * Represents a component that exports/manages a group of sensors.
         *
         */
    public static interface Monitor
    {
            /**
             * Returns the name of the monitor.
             *
             * The monitor's name defines the namespace for sensors exported by the
             * monitor. Its also used to control which kind of sensors the monitor whould
             * be able to install on-demand.
             *
             * @return the monitor's name
             */
        public ResourceName getName();

            /**
             * Returns an array containing the sensors this monitor currently provides.
             *
             * The set of supported sensors may grow due to new sensor install requests.
             *
             * @return currently available sensors
             */
        public Sensor[] getSensors();
        
            /**
             * Returns the sensor described by <code>name</code>, if it does exist.
             *
             * @param name sensor name
             * @return a Sensor object or null if there is no such sensor
             */
        public Sensor getSensor( String name );
        
            /**
             * Returns the set of parameters supported by the given sensor.
             *
             * @param s a valid sensor object
             * @return an array of SensorParameter objects describing the supported
             *         parameters.
             */
        public SensorParameter[] getSensorParameters( Sensor s );
        
            /**
             * Queries the current value of a sensor parameter.
             *
             * @param p a <code>SensorParameter</code> object
             * @return the current parameter value or null if there is no such parameter
             */
        public Object getSensorParameter( SensorParameter p );
        
            /**
             * Sets the value of a sensor parameter.
             *
             * @param p a <code>SensorParameter</code> object
             * @param v the new parameter value
             */
        public void setSensorParameter( SensorParameter p, Object v );
        
            /**
             * Enables/disables a sensor.
             *
             * Only enabled sensors produce monitoring data.
             *
             * @param s a <code>Sensor</code> value
             * @param enable a <code>boolean</code> value
             */
        public void setSensorEnabled( Sensor s, boolean enable );
        
            /**
             * Returns the current enabling status of a sensor (i.e. enabled or disabled)
             *
             * @param s a <code>Sensor</code> object
             * @return true if the sensor is enabled, false otherwise.
             */
        public boolean isSensorEnabled( Sensor s );
        
            /**
             * Queries the current value of a sensor (i.e. the monitoring data extracted
             * by the sensor).
             *
             * In order to minimize memory consumption, if a Monitoring data is provided,
             * that object is updated with the current sensor value and returned. If null
             * is provided, a new MonitoringData is allocated for hold the sensor value.
             *
             * @param s a <code>Sensor</code> object
             * @param d a <code>MonitoringData</code> object or null
             * @return a <code>MonitoringData</code> object describind the current sensor value
             */
        public MonitoringData probeSensor( Sensor s, MonitoringData d );
        
            /**
             * Asks for a sensor to be installed. This operation would fail if the
             * underlaying monitor does not support on-demand installation of sensors.
             *
             * @param name a <code>String</code> value
             * @param impl a <code>String</code> value
             * @param params a <code>Hashtable</code> value
             */
        public void installSensor( String name, String impl, Hashtable params );
        
            /**
             * Notifies the monitor that a monitoring quantum has expired, so sensors'
             * current values must be updated.
             *
             */
        public void quantumExpired();
    }

    public static abstract class AppMonitor
        implements Monitor
    {
            /**
             * Describe <code>getName</code> method here.
             *
             * @return a <code>ResourceName</code> value
             */
        public ResourceName getName(){
            throw new NullPointerException("Not yet implemented");
        }
        
            /**
             * Creates a new <code>AppMonitor</code> instance.
             *
             * @param app an <code>ApplicationId</code> value
             */
        public AppMonitor( ApplicationId app ){
            throw new NullPointerException("Not yet implemented");
        }
            
            /**
             * Describe <code>defineSensor</code> method here.
             *
             * @param s an <code>AppSensor</code> value
             */
        public void defineSensor( AppSensor s ){
            throw new NullPointerException("Not yet implemented");
        }
            
            /**
             * Describe <code>removeSensor</code> method here.
             *
             * @param s an <code>AppSensor</code> value
             */
        public void removeSensor( AppSensor s ){
            throw new NullPointerException("Not yet implemented");
        }
    }

    public static interface AppSensor
    {
            /**
             * Describe <code>getName</code> method here.
             *
             * @return a <code>String</code> value
             */
        public String getName();
        
            /**
             * Describe <code>probe</code> method here.
             *
             * @return an <code>Object</code> value
             */
        public Object probe();
            
            /**
             * Describe <code>getParameters</code> method here.
             *
             * @return a <code>Hashtable</code> value
             */
        public Hashtable getParameters();
            
            /**
             * Describe <code>setParameters</code> method here.
             *
             * @param params a <code>Hashtable</code> value
             */
        public void setParameters( Hashtable params );
    }
    
    public static interface MonitoringConsumer
    {
            /**
             * Describe <code>update</code> method here.
             *
             * @param timestamp a <code>long</code> value
             * @param data a <code>MonitoringData[]</code> value
             */
        public void update ( long timestamp, MonitoringData[] data );
    }

    
        /**
         * Helper class to store and manipulate monitoring data in an efficient way.
         *
         */
    public static class MonitoringData
        // TODO: Serializable -> Externalizable
        implements java.io.Serializable
    {
            /**
             * Creates a new <code>MonitoringData</code> instance.
             *
             * @param sname a <code>Sensor</code> name
             */
        public MonitoringData( ResourceName sname )
            {
                this.sname = sname;
                this.timeStamp = -1;
                this.intValue = 0;
                this.doubleValue = 0;
                this.longValue = 0;
                this.stringValue = "";
                this.dataType = -1;
            }
        
            /**
             * Returns the name of the sensor to which this data is linked (formelly, the
             * sensor who produced this data).
             *
             * @return a <code>ResourceName</code> value
             *
             * @deprecated Use method getSensorName() instead
             */
        public ResourceName getSensor()
            {
                return this.sname;
            }

            /**
             * Returns the name of the sensor to which this data is linked (formelly, the
             * sensor who produced this data).
             *
             * @return a <code>ResourceName</code> value
             */
        public ResourceName getSensorName()
            {
                return this.sname;
            }

            /**
             * Returns a time stamp describing the last time this data was modified.
             *
             * @return millis since 1970
             * @see java.lang.System#currentTimeMillis()
             */
        public long getTimeStamp()
            {
                return this.timeStamp;
            }

            /**
             * Returns this data as an int value. Conversion may take place if int is
             * not the original data format.
             *
             * @return this data as an int value
             */
        public int getInt()
            {
                return this.intValue;
                    // TODO: type conversions
            }

            /**
             * Returns this data as a long value. Conversion may take place if long is
             * not the original data format.
             *
             * @return this data as a long value
             */
        public long getLong()
            {
                return this.longValue;
                    // TODO: type conversions
            }

            /**
             * Returns this data as a double value. Conversion may take place if double
             * is not the original data format.
             *
             * @return this data as a double value
             */
        public double getDouble()
            {
                return this.doubleValue;
            }

            /**
             * Returns this data as a String. Conversion may take place if String is not
             * the original data format.
             *
             * @return this data as a String
             */
        public String getString()
            {
                switch (dataType) {
                    case T_STRING:
                        return this.stringValue;
                    case T_INT:
                        return Integer.toString(intValue);
                    case T_DOUBLE:
                        return Double.toString(doubleValue);
                    case T_LONG:
                        return Long.toString(longValue);
                    case T_VOID:
                        return "Not initialized";
                    default:
                        throw new NullPointerException("Corrupted MonitoringData structure");
                }
            }

            /**
             * Returns the type of the data stored in this MonitoringData object.
             *
             * @return a <code>Class</code> value; Void.TYPE if this object was not yet
             * initialized.
             */
        public Class getDataType()
            {
                switch (dataType) {
                    case T_STRING:
                        return String.class;
                    case T_INT:
                        return Integer.TYPE;
                    case T_DOUBLE:
                        return Double.TYPE;
                    case T_LONG:
                        return Long.TYPE;
                    case T_VOID:
                        return Void.TYPE;
                    default:
                        throw new NullPointerException("Corrupted MonitoringData structure");
                }
            }

            /**
             * Set this data last modification time-stamp.
             *
             * @param ts millis since 1970
             */
        public void setTimeStamp(long ts)
            {
                this.timeStamp = ts;
            }

            /**
             * Sets this data value as an int value.
             *
             * @param v new value
             */
        public void setInt ( int v )
            {
                this.dataType = T_INT;
                this.intValue = v;
            }

            /**
             * Sets this data value as a long value.
             *
             * @param v new value
             */
        public void setLong ( long v )
            {
                this.dataType = T_LONG;
                this.longValue = v;
            }

            /**
             * Sets this data value as a double value
             *
             * @param v new value
             */
        public void setDouble ( double v )
            {
                this.dataType = T_DOUBLE;
                this.doubleValue = v;
            }

            /**
             * Sets this data vaue as a String.
             *
             * @param s new value.
             */
        public void setString ( String s )
            {
                this.dataType = T_STRING;
                this.stringValue = s;
            }

        public String toString()
            {
                String type = "STRING: ";
                switch (dataType) {
                    case T_INT:
                        type = "INT: ";
                        break;
                    case T_DOUBLE:
                        type = "DOUBLE: ";
                        break;
                    case T_LONG:
                        type = "LONG: ";
                        break;
                }

                return "SENSOR{name="+sname+", value="+type + getString()+"}";
            }

        public void copy(MonitoringData d)
            {
                if ( d == null ) throw new IllegalArgumentException("null source not allowed");
                
                this.sname = d.sname;
                this.timeStamp = d.timeStamp;
                this.dataType = d.dataType;
                
                switch (dataType) {
                    case T_STRING:
                        this.stringValue = d.stringValue;
                        break;
                    case T_INT:
                        this.intValue = d.intValue;
                        break;
                    case T_DOUBLE:
                        this.doubleValue = d.doubleValue;
                        break;
                    case T_LONG:
                        this.longValue = d.longValue;
                        break;
                    case T_VOID:
                            // do nothing
                        break;
                    default:
                        throw new NullPointerException("Corrupted MonitoringData structure");
                }
            }

        
            /**
             * Compares for equality two MonitoringData objects ignoring their timestamps.
             *
             * @param d a <code>MonitoringData</code> value
             * @return a <code>boolean</code> value
             */
        public boolean valueEquals( MonitoringData d )
            {
                if ( d != null && dataType == d.dataType ) {
                    switch (dataType) {
                        case T_STRING: return stringValue.equals(d.stringValue);
                        case T_INT: return intValue == d.intValue;
                        case T_DOUBLE: return doubleValue == d.doubleValue;
                        case T_LONG: return this.longValue == d.longValue;
                    }
                }
                return false;
            }

        private ResourceName sname;
        private long timeStamp;
        private int intValue;
        private double doubleValue;
        private long longValue;
        private String stringValue;
        private int dataType;
        private final int T_VOID = -1;
        private final int T_INT = 0;
        private final int T_LONG = 1;
        private final int T_DOUBLE = 2;
        private final int T_STRING = 3;
    }

        /**
         * Monitoring consumer identifier.
         *
         */
    public static class ConsumerId
        implements java.io.Serializable
    {
            /**
             * Creates a new <code>ConsumerId</code> instance.
             *
             * @param k an <code>Object</code> value
             * @param c a <code>ResourceName</code> value
             * @param ch a <code>String</code> value
             */
        public ConsumerId(Object k, ResourceName c, String ch)
            {
                if ( k == null )
                    throw new IllegalArgumentException("Null key not allowed");
                if ( c == null )
                    throw new IllegalArgumentException("Null collector name  not allowed");
                
                this.key = k;
                this.collector = c;
                this.channel = ch;
            }
        
            /**
             * Returns the unique key associated with this consumer id. This key is
             * typically used to link the consumerid object to a peer internal
             * representation used by the collector.
             *
             * Note: the key must be serializable in order to allow the ConsumerId to be
             * transmitted between hosts.
             *
             * @return an <code>Object</code> value
             */
        public Object getKey ()
            {
	        return this.key;
            }
        
            /**
             * Returns the name of the Collector instance which have generated this
             * consumer id.
             *
             * @return the resource name of the collector instance
             */
        public ResourceName getCollector()
            {
	        return this.collector;
            }
        
            /**
             * Returns the name of the multicast channel where the monitoring data for
             * this consumer will be made available.
             *
             * Typically , this value will be <code>null</code> the consumer provided a
             * non-null call back by the time it requested the id.
             *
             * @return a <code>String</code> value
             */
        public String getMulticastChannel()
            {
                throw new NullPointerException("Not yet implemented");
            }

        public int hashCode()
            {
                return key.hashCode();
            }

        public boolean equals(Object o)
            {
                if ( o != null ) {
                    if ( o instanceof ConsumerId ) {
                        ConsumerId c2 = (ConsumerId) o;

                        return ( key.equals(c2.key)
                                 && collector.equals(c2.collector)) ;
                    }
                }
                return false;
            }

        private Object key;
        private ResourceName collector;
        private String channel;
    }
    
        /**
         * Describe <code>getName</code> method here.
         *
         * @return a <code>ResourceName</code> value
         */
    public ResourceName getName();
    
        /**
         * Describe <code>getSensors</code> method here.
         *
         * @return a <code>Sensor[]</code> value
         */
    public Sensor[] getSensors();

        /**
         * Describe <code>getSensor</code> method here.
         *
         * @param name a <code>String</code> value
         * @return a <code>Sensor</code> value
         */
    public Sensor getSensor( String name );

        /**
         * Describe <code>getSensorParameters</code> method here.
         *
         * @param s a <code>Sensor</code> value
         * @return a <code>SensorParameter[]</code> value
         */
    public SensorParameter[] getSensorParameters( Sensor s );

        /**
         * Describe <code>getSensorParameter</code> method here.
         *
         * @param p a <code>SensorParameter</code> value
         * @return an <code>Object</code> value
         */
    public Object getSensorParameter( SensorParameter p );

        /**
         * Describe <code>setSensorParameter</code> method here.
         *
         * @param c a <code>ConsumerId</code> value
         * @param p a <code>SensorParameter</code> value
         * @param v an <code>Object</code> value
         */
    public void setSensorParameter( ConsumerId c, SensorParameter p, Object v );

        /**
         * Describe <code>setSensorEnabled</code> method here.
         *
         * @param c a <code>ConsumerId</code> value
         * @param s a <code>Sensor</code> value
         * @param enable a <code>boolean</code> value
         */
    public void setSensorEnabled( ConsumerId c, Sensor s, boolean enable );

        /**
         * Describe <code>isSensorEnabled</code> method here.
         *
         * @param c a <code>ConsumerId</code> value
         * @param s a <code>Sensor</code> value
         * @return a <code>boolean</code> value
         */
    public boolean isSensorEnabled( ConsumerId c, Sensor s );

        /**
         * Describe <code>probeSensor</code> method here.
         *
         * @param c a <code>ConsumerId</code> value
         * @param s a <code>Sensor</code> value
         * @param d a <code>MonitoringData</code> value
         * @return a <code>MonitoringData</code> value
         */
    public MonitoringData probeSensor( ConsumerId c, Sensor s, MonitoringData d );

        /**
         * Describe <code>installSensor</code> method here.
         *
         * @param c a <code>ConsumerId</code> value
         * @param name a <code>String</code> value
         * @param impl a <code>String</code> value
         * @param params a <code>Hashtable</code> value
         */
    public void installSensor( ConsumerId c, String name, String impl, Hashtable params );

        /**
         * Describe <code>addMonitor</code> method here.
         *
         * @param m a <code>Monitor</code> value
         */
    public void addMonitor( Monitor m );

        /**
         * Describe <code>removeMonitor</code> method here.
         *
         * @param m a <code>Monitor</code> value
         */
    public void removeMonitor( Monitor m );

        /**
         * Describe <code>addConsumer</code> method here.
         *
         * @param c a <code>MonitoringConsumer</code> value
         * @return a <code>ConsumerId</code> value
         */
    public ConsumerId addConsumer( MonitoringConsumer c );

        /**
         * Describe <code>removeConsumer</code> method here.
         *
         * @param c a <code>ConsumerId</code> value
         */
    public void removeConsumer( ConsumerId c );
}

