// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services;

import org.isam.exehda.ObjectId;

import java.io.IOException;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.util.Hashtable;

public interface ObjectSeed
{
    public static final String SERVICE_NAME = "objectseed";


    public static interface Activator
        extends java.io.Serializable
    {            
            /**
             * "standard" attribute reserved to use of the ObjectSeed implementation
             * ifself and should not be used by custom Activators (will be overwritten)
             */
        public static final String OXATT_ID = "ox.id";
            /**
             * "standard" attribute reserved to use of the ObjectSeed implementation
             * ifself and should not be used by custom Activators (will be overwritten)
             */
        public static final String OXATT_STATE = "ox.state";
            /**
             * "standard" attribute reserved to use of the ObjectSeed implementation
             * ifself and should not be used by custom Activators (will be overwritten)
             */
        public static final String OXATT_HANDLE = "ox.handle";
            /**
             * "standard" attribute reserved to use of the ObjectSeed implementation
             * ifself and should not be used by custom Activators (will be overwritten)
             */
        public static final String OXATT_ACTIVATOR = "ox.activator";
        
            /**
             * Proceed ox activation.
             *
             * @param oxID id of the ox being activated
             * @param obj the ox instance being activated 
             * @param extState if non-null, additional (extended) state information
             * @exception Exception if an error occurs
             */
        void activate( ObjectId oxID, Object obj, MarshaledOX extState )
            throws Exception;

            /**
             * Proceed ox deactivation. The <code>extState</code> object may be used to
             * stored extended information describing the ox current state.
             *
             * @param oxID id of the ox being deactivated
             * @param obj the ox instance being deactivated 
             * @param extState container for extended attributes
             * @exception Exception if an error occurs
             */
        void deactivate( ObjectId oxID, Object obj, MarshaledOX extState )
            throws Exception;
    }

        /**
         * Specialized container for minimizing serialization overhead when transmitting
         * OX state between nodes.
         *
         * <p>It uses ahead-of-time serialization, i.e. attributes are serialized or
         * de-serialized on-access instead of on-transmission.
         *
         */
    public static class MarshaledOX
        implements java.io.Serializable //,java.io.ObjectOutput //,java.io.ObjectInput
    {
        private static final int  OUT_BUFF_SIZE = 512;
        
        public MarshaledOX()
            {
                this.attrs = new Hashtable();
            }
        
        public void setAttribute ( String attr, Object val )
            throws IOException
            {
                    // XXX: Hashtable do not accepts null values
                if ( val == null ) {
                    attrs.remove(attr);
                }
                else {
                    ByteArrayOutputStream rout = new ByteArrayOutputStream(OUT_BUFF_SIZE);
                    ObjectOutputStream out = new ObjectOutputStream(rout);

                    out.writeObject(val);
                    out.flush();
                    out.close();
                
                    attrs.put(attr, rout.toByteArray());
                }
            }
        
        public Object getAttribute ( String attr )
            throws IOException, ClassNotFoundException
            {
                byte[] rdata = (byte[]) attrs.get(attr);

                if ( rdata != null ) {
                        //
                        // XXX: ensure the application class loader will be used to
                        // resolve attribute classes.
                        //
                    ObjectInputStream in =
                        new ObjectInputStream(
                            new ByteArrayInputStream(rdata))
                        {
                            protected Class
                            resolveClass(java.io.ObjectStreamClass v)
                                throws IOException, ClassNotFoundException
                                {
                                    return org.isam.perv.IsamClassLoader
                                        .getInstance().loadClass(v.getName());
                                }
                        };

                    Object val = in.readObject();
                    in.close();

                    return val;
                }
                else {
                    return null;
                }
            }

        private Hashtable attrs;
    }

    
    public static class OXActivationException
        extends java.lang.RuntimeException
    {
        public OXActivationException(String msg)
            {
                super(msg);
            }
    }
    
    public ObjectId
    newInstance(Class cls, Object[] params, Activator a);
    
    public MarshaledOX
    deactivateObject( ObjectId ox, MarshaledOX oxState )
        throws OXActivationException;
    
    public void
    activateObject( MarshaledOX oxState )
        throws OXActivationException;


        //
        // TO DO: attaching unmanaged objects to exehda
        //
    
//     public ObjectId
//     attachObject( Object obj, Activator a);

//     public void
//     dettachObject( ObjectId ox);

///////////////////////////////////////////////////////////////////////////////////
    
//     public static interface Activatable
//     {
//         void activate() throws Exception;
//         void deactivate() throws Exception;
//     }
    


}
