// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.avu;

import java.io.*;
import org.isam.exehda.Exehda;
import org.isam.exehda.services.HttpService;
import org.isam.exehda.services.Logger.Sev;
import org.isam.perv.protocols.avu.Handler;

/**
 * This class implements ISAM AVU service, the user virtual environment of the ISAM platform.
 *
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-07-15 10:09:49 -0300 (Fri, 15 Jul 2005) $ $Revision: 1001 $
 */
public class AvuImpl
    extends HttpService
{
    static final String SERVICE_NAME = "avu";
    static final String DEFAULT_DOCROOT = "./";
    static final int    DEFAULT_PORT = 1977;
    
    static final String PROP_AVU_OPERATION = "Isam-Avu-Operation";
    static final String PROP_AVU_RESOURCE  = "Resource";
    static final String PROP_AVU_USER      = "Isam-User";
    static final String CONTENT_LENGTH     = "Content-Length";

    static final int    AVU_OP_READ        = 0;
    static final int    AVU_OP_WRITE       = 1;

    static final int    AVU_CELL_INDEX        = 1;
    static final int    AVU_USER_INDEX        = 0;

    private String docRoot;

    
    public void
    start() 
        throws IOException
        {
            final String METHOD_NAME = "start()";
                        
            docRoot = Exehda.getServiceProperty(
                SERVICE_NAME, "docroot", DEFAULT_DOCROOT);            
            log(Sev.INFO, METHOD_NAME, "Set up file system docroot to \"{0}\"", docRoot);
            
            int port = Exehda.getServiceProperty(
                SERVICE_NAME, "port", DEFAULT_PORT);

                // configure avu: protocol handler
            Handler.config("localhost", port);

            setPort(port);
            super.start();
            log(Sev.INFO, METHOD_NAME, "Listening at port "+port);
        }

    protected void
    handleRequest( ServiceRequest req )
        throws IOException
        {
            switch ( getAvuOperation(req) ) {
                case AVU_OP_READ:
                    handleAvuRead(req);
                    break;
                case AVU_OP_WRITE:
                    handleAvuWrite(req);
                    break;
                default:
                    log( Sev.ERROR, "handleRequest()", "Opps! Unknown AVU operation.");
            }
                // op read
                // check if is a valid file name( do not start with .. for example)
                // open local file
                // output HTTP response header
                // output file contents
        }

    protected File
    checkAccess( String fileName, String userName, int op )
        {
            final String METHOD_NAME = "checkAccess()";
            
            int atIndex = userName.indexOf("@");
            String userPath = userName.substring(atIndex+1)
                + File.separator
                + userName.substring(0,atIndex);

            try {

                File userDir = new File(docRoot+File.separator+userPath);

                if ( userDir.exists() && userDir.isDirectory() ) {
                    File realFilePath = new File(userDir, fileName);
                    File realFileDir = new File(realFilePath.getParent());

                        // check subtree protection
                    if ( realFileDir.getCanonicalPath()
                         .startsWith(userDir.getCanonicalPath()) ) {
                    
                        switch (op) {
                            case AVU_OP_WRITE:
                                if ( (realFileDir.mkdirs() || realFileDir.exists())
                                     && realFileDir.isDirectory() ) {
                                    return  realFilePath;
                                }
                                break;
                            case AVU_OP_READ:
                                if ( realFilePath.exists() && realFilePath.isFile() ) {
                                    return realFilePath;
                                }
                                break;
                        }
                    }
                }
            }
            catch (IOException ioe) {
                log(Sev.ERROR, METHOD_NAME, ioe);
            }
                // access denied
            return null;
        }

    protected int
    getAvuOperation( ServiceRequest req )
        {
                // defaults to 'read'
            int op = AVU_OP_READ;
            String opString = req.getProperty(PROP_AVU_OPERATION);
            System.out.println("Operation == " + opString);
            if ( "write".equals(opString) ) {
                op = AVU_OP_WRITE;
            }

            return op;
        }

//     protected String[]
//     getUserCellPair( String userName)
//         {
//             String []userCellPair = new String[2];

//             int atIndex = userName.indexOf("@");
//                 //TODO: Throw Exception
//             userCellPair[AVU_USER_INDEX] = userName.substring(0,atIndex);
//             userCellPair[AVU_CELL_INDEX] = userName.substring(atIndex+1);

//             return(userCellPair);

//         }

    protected void
    handleAvuRead( ServiceRequest req )
        throws IOException
        {
            final String METHOD_NAME = "handleAvuRead()";
            
            String fileName = req.getProperty(PROP_AVU_RESOURCE);
            String userName = req.getProperty(PROP_AVU_USER);

            File realPath = checkAccess(fileName, userName, AVU_OP_READ);
            if ( realPath != null ) {

                log(Sev.INFO, METHOD_NAME,
                    "Serving {0} to user {1}, real path is {2}",
                    fileName, userName, realPath);
                
                ServiceResponse resp =
                    new ServiceResponse(req, ServiceResponse.OK);

                resp.putHeader(CONTENT_LENGTH, ""+realPath.length());

                byte[] buff = new byte[4096];
                InputStream ins = new FileInputStream(realPath);
                
                int nr;
                while ( (nr = ins.read(buff)) != -1 ) {
                    resp.putData(buff, 0, nr);
                }
                
                resp.close();
            }
            else {
                log(Sev.ERROR, METHOD_NAME,
                    "Access denied to {0}, real path is {1}", fileName, realPath);
                ServiceResponse resp =
                    new ServiceResponse( req, ServiceResponse.BAD_REQUEST);
                resp.close();
            }
        }

    protected void
    handleAvuWrite( ServiceRequest req )
        throws IOException
        {

            final String METHOD_NAME = "handleAvuWrite()";
	
            String fileName = req.getProperty(PROP_AVU_RESOURCE);
            String userName = req.getProperty(PROP_AVU_USER);
            int bytesToRead = Integer.parseInt(req.getProperty(CONTENT_LENGTH));
        
            File realPath = checkAccess(fileName, userName, AVU_OP_WRITE);
            if ( realPath != null ) {
                log(Sev.INFO, METHOD_NAME,
                    "Writing to {0} in behalf of {1}, real path is {2}",
                    fileName, userName, realPath);
	
                java.net.Socket clientSocket = req.getClientSocket();
                InputStream ins = clientSocket.getInputStream();
	    
                final byte[] buff = new byte[4096];
                OutputStream outs = new FileOutputStream(realPath);
            
                int nr = 0;
                int bytesRead = 0;
	   
                while (bytesRead < bytesToRead) {
                    nr = ins.read(buff);
                    outs.write(buff, 0, nr);
                    bytesRead += nr;
                }
	    
                outs.close();
                    //ins.close();
	   
                ServiceResponse resp =
                    new ServiceResponse(req, ServiceResponse.OK);
                resp.close();
            }
            else {
                log(Sev.ERROR, METHOD_NAME,
                    "Access denied to {0}, real path is {1}", fileName, realPath);
                ServiceResponse resp =
                    new ServiceResponse( req, ServiceResponse.BAD_REQUEST);
                resp.close();
            }
        }
}
    
