// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.cib.backend.ldap;

import org.isam.exehda.ResourceName;
import org.isam.exehda.ResourceName.NameComponent;
import org.isam.exehda.ResourceName.NameSpace;

import javax.naming.NamingException;
import javax.naming.directory.Attributes;
import javax.naming.directory.Attribute;

/**
 * Provides a mapping beteween the CIB type <code>service</code> and the LDAP type
 * <code>exehdaService</code>.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-07-08 13:35:00 -0300 (Fri, 08 Jul 2005) $ $Revision: 970 $
 */
class ServiceLRTH
    extends BaseLRTH
{
    private static final String CIB_ATT_CADDR = "contactAddress";
    
    private static final int ATT_OBJCLASS = 0;
    private static final int ATT_CN = 1;
    private static final int ATT_CADDR = 2;

    private static final String[] DFLT_ATTR;
    private static final String[] DFLT_ATTR_VAL;

    private static final String[] HANDLED_CIB_TYPES = { "service" }; 
    private static final String[] HANDLED_DN_PREFIXES = { "cn=service:" };

    static {
        
        DFLT_ATTR = new String[] {
            "objectClass",     // ATT_OBJCLASS
            "cn",              // ATT_CN
            "xContactAddress"  // ATT_CADDR
        };
        
        DFLT_ATTR_VAL = new String[] {
            "exehdaService",  // ATT_OBJCLASS
            null,             // ATT_CN
            null,             // ATT_CADDR
        };
    }

    public ServiceLRTH()
        {
            super(HANDLED_CIB_TYPES, HANDLED_DN_PREFIXES);
        }

    public String
    createNode( LdapSession s, ResourceName rn )
        throws NamingException
        {
            if ( !"service".equals(rn.getType()) ) {
                throw new IllegalArgumentException(
                    "Invalid resource type: "+rn.getType());
            }
                        
                // map the given resource name to an ldap DN
            String dn = encode(rn, null).toString();
                // fill in mandatory attributes
            String[] avs = (String[]) DFLT_ATTR_VAL.clone();
                // overwrite default values with user-provided values
            avs[ATT_CN] = encodeAttrValue("service",
                                          CIB_ATT_RES_INSTANCE,
                                          rn.getSimpleName());
            avs[ATT_CADDR] = "/"+rn.getSimpleName();
                // commit to LDAP
            internalCreateNode(s, rn, dn, DFLT_ATTR, avs);

            return dn;
        }

    public NameComponent
    decodeNameComponent( String dnc )
        {
//            System.out.println(dnc);
            
            int idx = dnc.indexOf('=');
            if ( idx <= 0 || idx == (dnc.length()-1) ) {
                throw new IllegalArgumentException("DN component:"+dnc);
            }

                //
                // FIX ME: not checking if dnc starts with "cn=service:"
                //
            
            String t = "service";
            String i = dnc.substring(idx+9);

            return new NameComponent(t,i);
        }

    public String
    getDefaultSearchFilter()
        {
            return "(objectClass=exehdaService)";
        }

    public LdapResourceInfo
    toLdapResourceInfo(String dn, Attributes atts)
        {
            ResourceName rn = decodeResourceName(dn);            
                
            return new LdapResourceInfo(rn, atts) {
                    public String getAttribute(String a) {
                        try {
                            final String LDAP_ATT_CN = "cn";
                            final String LDAP_ATT_CADDR = "xContactAddress";
                            final String LDAP_ATT_XATTR = "xAttr";
                            
                            if ( CIB_ATT_RES_TYPE.equals(a) ) {
                                return "service";
                            }
                            else if ( CIB_ATT_RES_INSTANCE.equals(a)) {
                                String cn = (String) atts.get(LDAP_ATT_CN).get();
                                return cn.substring(cn.indexOf(':')+1);
                            }
                            else if ( CIB_ATT_CADDR.equals(a)) {
                                return (String) atts.get(LDAP_ATT_CADDR).get();
                            }
                            else {
                                if ( this.cache == null ) {
                                    this.cache = new java.util.Hashtable();
                                        
                                        // parse xAttrs only once
                                    Attribute att = atts.get(LDAP_ATT_XATTR);
                                    for ( int i=0; i<att.size(); i++ ) {
                                        String s = (String) att.get(i);
                                        int p = s.indexOf(':');
                                        cache.put(s.substring(0,p),
                                                  s.substring(p+1));
                                    }
                                }
                                return (String) cache.get(a);
                            }
                        }
                        catch ( NamingException ne ) {
                                //
                                // TO DO: use logger
                                //
                            ne.printStackTrace();
                            return null;
                        }
                    }
                };
        }
    
        /////////////////////////////////////////////////////////
        // PROTECTED STUFF
        /////////////////////////////////////////////////////////
    
    protected String
    encodeAttrName( String ct, String ca )
        {
            if ( CIB_ATT_RES_TYPE.equals(ca) ) {
                return "objectClass";
            }
            else if ( CIB_ATT_RES_INSTANCE.equals(ca) ) {
                return "cn";
            }
            else if ( CIB_ATT_CADDR.equals(ca) ) {
                return "xContactAddress";
            }
            else {
                return "xAttr";
            }
        }

    protected String
    encodeAttrValue( String ct, String ca, String cav )
        {
            if ( CIB_ATT_RES_TYPE.equals(ca) ) {
                return "exehdaService";
            }
            else if ( CIB_ATT_RES_INSTANCE.equals(ca) ) {
                return "service:"+cav;
            }
            else if ( CIB_ATT_CADDR.equals(ca) ) {
                return cav;
            }
            else {
                return ca+":"+cav;
            }
        }
    
    protected String
    decodeAttrValue( String ca, String la, String lav )
        {
            if ( "xAttr".equals(la) ) {
                return lav.startsWith(ca+":") ? lav.substring(lav.indexOf(':')+1)
                    : null;
            }
            else if ( CIB_ATT_CADDR.equals(ca)
                      && "xContactAddress".equals(la) ) {
                return lav;
            }
            else if ( CIB_ATT_RES_INSTANCE.equals(ca)
                      && "cn".equals(la)) {
                return lav.substring(lav.indexOf(':')+1);
            }
            else if ( CIB_ATT_RES_TYPE.equals(ca) ) {
                return "service";
            }
            else {
                throw new NullPointerException(
                    "Unknown encoding: ca="+ca+" la="+ca+" lav="+lav);
            }
        }
}
