// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda;



/**
 * Matches resources that define a given attribute with a given value.
 *
 * <p> A restricted regular-expression syntax is supported for specifying valid values
 * for the attribute. Basically, "*" is allowed at the beginning or end of a word and "|"
 * may be used to specify match alternatives (i.e. its an "or").
 *
 * <p> For example the expression "a|ab*|*c" will match "a", "abc", "ac", but neither
 * "bC" nor " ab". <b>Notice:</b> the matching is blank(space) and case sensitive.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-02-24 02:29:43 -0300 (Thu, 24 Feb 2005) $ $Revision: 584 $
 */
public class AttributeSimpleMatchConstraint
    extends AttributeConstraint
{
    
        /**
         * Creates a new <code>AttributeSimpleMatchConstraint</code> instance.
         *
         * @param att the name of the attribute to be checked
         * @param rexp the restricted regexp used in matching the attribute's value 
         */
    public AttributeSimpleMatchConstraint( String att, String rexp)
        {
            super(att);

            if ( rexp == null || "".equals(rexp.trim()) ) {
                throw new IllegalArgumentException(
                    "Illegal matching expression:\""+rexp+"\"");
            }


            this.fullExp = rexp;
            this.expv = null;

            try {
                splitExp();
            }
            catch ( StringIndexOutOfBoundsException e ) {
                throw new IllegalArgumentException(
                    "Malformed expression:\""+rexp+"\"");
            }
        }


        /**
         * Returns the expression used when matching the attribute's value.
         *
         * @return a <code>String</code> value
         */
    public String
    getMatchingExp()
        {
            return this.fullExp;
        }

    
        /**
         * Checks whether the selected attribute's value matches the given restricted
         * regular expression.
         *
         * @param r a <code>ResourceInfo</code> value
         * @return ResourceConstraint.UNDEFINED if the attribute is not defined for the
         * given resource, ResourceConstraint.OK if the attribute exists and its value
         * matches the expression, ResourceConstraint.BROKEN otherwise.
         */
    public int
    check( ResourceInfo r )
        {
                // ensure expv is initialized
            splitExp();

                // lookup the value of the target attribute
            String v = r.getAttribute(attrName);

            if ( v == null ) {
                    // no such attribute
                return ResourceConstraint.UNDEFINED;
            }
            else {
                for (int i=0; i<expv.length; i++ ) {
                    String e = expv[i];
                    if ( e.endsWith("*") ) {
                            // match "----*"
                        if ( e.regionMatches(0, v, 0, e.length()-1) ) {
                               return ResourceConstraint.OK;
                        }
                    }
                    else if ( e.startsWith("*") ) {
                            // match "*----"
                        int p = v.length() - e.length() + 1;
                        if ( p > 0 && v.regionMatches(p, e, 1, e.length()-1) ) {
                            return ResourceConstraint.OK;
                        }
                    }
                    else {
                            // match "----"
                        if ( v.equals(e) ) {
                            return ResourceConstraint.OK;
                        }
                    }
                }
                return ResourceConstraint.BROKEN;
            }
        }

        /////////////////////////////////////////////////////////
        // PRIVATE STUFF
        /////////////////////////////////////////////////////////
    
    private final void
    splitExp()
        {
                // if expv not initialized
            if ( expv == null ) {
                    // count '|'
                int n = 0;
                int p = 0;
                while ( (p = fullExp.indexOf('|', p+1)) > 0 ) {
                    n++;
                    if ( p == fullExp.length()) break;
                }
                    // alloc array for storing sub-expressions
                expv = new String[n+1];
                
                    // split
                p = 0;
                for ( int i=0; i<=n; i++) {
                        // check not last string
                    if ( i < n ) {
                            //
                            // FIX ME: need to check for malformed expressions
                            //         like "||||" or "|a|"
                            //
                        int q = fullExp.indexOf('|', p);
                        expv[i] = fullExp.substring(p,q);
                        p = q+1;
                    }
                    else {
                            // consumes the last string
                        expv[n] = fullExp.substring(p);
                    }
                }
            }
        }

    private String fullExp;
    private transient String[] expv;
}
