// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.primos;

import org.isam.exehda.services.Collector.Sensor;
import org.isam.exehda.services.Collector.SensorParameter;
import org.isam.exehda.services.Collector.MonitoringData;
import org.isam.exehda.services.Collector.ConsumerId;
import org.isam.exehda.services.Collector.Monitor;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Stack;
import java.util.Vector;

/**
 * Facility to manipulate sensors.
 *
 * Each sensor is broken in a public version and another private one. The private version
 * is used to interact with the underlaying monitor (actual sensor's provider). The
 * public sensor is made publically available to applications and may be used with no
 * risk of opening security flaws or having unwanted side-effects.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-18 23:01:05 -0300 (Sat, 18 Jun 2005) $ $Revision: 934 $
 */
class MetaSensor
{
        /**
         * Creates a new <code>MetaSensor</code> instance for the given private Sensor
         * and monitor.
         *
         * @param pvtSensor an internal Sensor
         * @param provider the underlaying sensor provider
         */
    public MetaSensor(Sensor pvtSensor, Monitor provider)
        {
            this.pvtSensor = pvtSensor;
            this.provider = provider;
            this.enabled = provider.isSensorEnabled(pvtSensor);
            this.nConsumers = 0;

            initPubSensor();

                // FIX ME: pvtSensor or pubSensor
            this.currentData = new MonitoringData(pubSensor.getName());
            this.lastData = null;
            
            initParamTable();
        }
    
        /**
         * Returns the public sensor object associated with this sensor.
         *
         * @return the public Sensor object
         */
    public Sensor
    getPublicSensor()
        {
            return this.pubSensor;
        }

        /**
         * Returns the public key os this sensor.
         *
         * @return the public Key.
         */
    public Object
    getKey()
        {
            return this.pubSensor.getKey();
        }


        /**
         * Updates last published state.
         *
         */
    public void
    update(long timeStamp)
        {
            currentData.setTimeStamp(timeStamp);
            if (lastData == null) {
                lastData = new MonitoringData(pubSensor.getName());
            }
            lastData.copy(currentData);
        }

        /**
         * Returns whether the sensor value has significantly changed since last
         * publication.
         *
         * @return a <code>boolean</code> value
         */
    public boolean
    checkStateChanged()
        {
                // TO DO: compare currentData with lastData using threashold
            return !currentData.valueEquals(lastData);
        }
    
        /**
         * Register a new consumer for the sensor.
         *
         * @todo review synchronized
         *
         * @param c a new Consumer
         */
    public synchronized void
    addConsumer(ConsumerId c)
        {            
            this.nConsumers++;

                // XXX: force event generation for the new consumer
            this.lastData = null;
        }

        /**
         * Unregisters the given object as a consumer of the data produced by this
         * sensor.
         *
         * Any parameter configurations made by the consumer are also canceled.
         *
         * @todo review synchronized
         *
         * @param c a <code>ConsumerId</code> value
         */
    public synchronized void
    removeConsumer(ConsumerId c)
        {
                // clear parameter configurations
            Enumeration csenum = confsByParam.elements();
            while (csenum.hasMoreElements()) {
                ConfStack cs = (ConfStack) csenum.nextElement();
                cs.clean(c);
            }
                // if there are no consumers the sensor is disabled
            if ( --nConsumers == 0 ) {
                setEnabled(false);
            }
        }

        /**
         * Enables/disables this sensor.
         *
         * @param enable a <code>boolean</code> value
         */
    public synchronized void
    setEnabled(boolean enable)
        {
                // XXX: force event generation whenever the sensor enabling state changes
            if (enable) lastData = null;
            
            provider.setSensorEnabled(pvtSensor, enable);
                // conservative approach
            this.enabled = provider.isSensorEnabled(pvtSensor);
        }
    
        /**
         * Returns whether this sensor is currently enabled or not.
         *
         * @return true if enabled, false otherwise.
         */
    public synchronized boolean
    isEnabled()
        {
            return this.enabled;
         //    return provider.isSensorEnabled(pvtSensor);
        }

        /**
         * Returns the list of parameters supported by this sensor.
         *
         * Such a list includes the original sensor parameters and also meta-parameters
         * implemented on top of those ones by the MetaSensor class itself..
         *
         * @return a <code>SensorParameter[]</code> value
         */
    public synchronized SensorParameter[]
    getParameters()
        {
                // TO DO: not yet implemented
            throw new NullPointerException("Not yet implemented");
        }

        /**
         * Returns the current value assigned to the given parameter.
         *
         * @param p a <code>SensorParameter</code> value
         * @return an <code>Object</code> value
         */
    public Object
    getParameter(SensorParameter p)
        {
                // TO DO: not yet implemented
            throw new NullPointerException("Not yet implemented");
        }

        /**
         * Assigns a new value to the given parameter.
         *
         * @param c a <code>ConsumerId</code> value
         * @param p a <code>SensorParameter</code> value
         * @param v an <code>Object</code> value
         */
    public synchronized void
    setParameter(ConsumerId c, SensorParameter p, Object v)
        {
                // TODO: what about the consumer?
//            p = getPrivateParameter(p);
            ConfStack conf = (ConfStack) confsByParam.get(p);
            conf.push(c, v);
            provider.setSensorParameter(p,v);
        }

        /**
         * Probes the current value of the sensor.
         *
         * @param d a <code>MonitoringData</code> value
         * @return a <code>MonitoringData</code> value
         */
    public synchronized MonitoringData
    probe( MonitoringData d )
        {
                // FIX ME:
            if ( d == null ) {
                return currentData = provider.probeSensor(pvtSensor, currentData);
            }
            else {
                return provider.probeSensor(pvtSensor, d);
            }
        }

//         /**
//          * Returns the private Sensor instance used to access the underlaying monitor who
//          * actually provides the sensor.
//          *
//          * @return the private Sensor instance
//          */
//     private Sensor
//     getPrivateSensor()
//         {
//             return this.pvtSensor;
//         }
    
        /**
         * Returns the underlaying monitor who actually provides this sensor.
         *
         * @return the sensor provider
         */
    public Monitor
    getProvider()
        {
            return this.provider;
        }

        /**
         * Returns a string representation of this MetaSensor object (mostly for
         * debugging purpouses).
         *
         * @return a <code>String</code> value
         */
    public String toString()
        {
            return pvtSensor.toString();
        }

    public boolean
    equals( Object o )
        {
            if ( o != null ) {
                if ( o instanceof MetaSensor ) {
                    return this.pubSensor.equals(((MetaSensor) o).pubSensor);
                }
//                 else if ( o instanceof Sensor ) {
//                     Sensor s2 = (Sensor) o;
//                     System.out.println("MetaSensor equals to Sensor");
//                     System.out.println("s1.name="+pubSensor.getName()+"\"");
//                     System.out.println("s1.type="+pubSensor.getType()+"\"");
//                     System.out.println("s1.key="+pubSensor.getKey()+"\"");
//                     System.out.println("s1.hash="+pubSensor.hashCode()+"\"");
//                     System.out.println("s2.name="+s2.getName()+"\"");
//                     System.out.println("s2.type="+s2.getType()+"\"");
//                     System.out.println("s2.key="+s2.getKey()+"\"");
//                     System.out.println("s2.hash="+s2.hashCode()+"\"");
//                     System.out.println("equasl()="+pubSensor.equals(o));
//                     return pubSensor.equals(o);
//                 }
//                 else {
//                     System.out.println("equals() neither instance of MetaSensor, nor Sensor");
//                     System.out.println(this);
//                     System.out.println(o);
                                       
//                 }
            }
            return false;
        }

    public int
    hashCode()
        {
            return pubSensor.hashCode();
        }

        /**
         * Creates the parameter configuration table by allocating a ConfStack object for
         * each defined parameter and initializing it with the paramter's current value.
         *
         */
    private void
    initParamTable()
        {
            this.confsByParam = new Hashtable();
            SensorParameter[] params = provider.getSensorParameters(pvtSensor);
            
            for (int i=0; i<params.length; i++) {
                ConfStack cs = new ConfStack();
                cs.push(null, provider.getSensorParameter(params[i]));
                
                confsByParam.put(params[i], cs);
            }
        }

        /**
         * Creates the public sensor object (the one that will be visible outside the
         * collector).
         *
         * @todo better key generation
         * @todo fix pub sensor name
         */
    private void
    initPubSensor()
        {
                // TODO: better key generation
            Object key =  new Long(nextKey++);

            this.pubSensor =  new Sensor(pvtSensor.getName(), // TODO: fix pub sensor name
                                         pvtSensor.getType(),
                                         key);
        }
    

    
        /** The private sensor object as provided by the underlaying monitor. */
    private Sensor pvtSensor;
        /** Whether the underlaying sensor is enabled or diasbled (cache) */
    private boolean enabled;
        /** The monitor who actually provides the sensor. */
    private Monitor provider;
        /** The public visible sensor object */
    private Sensor pubSensor;
        /** Table mapping parameters to current configurations */
    private Hashtable confsByParam;
        /** Number of consumers attached to the sensor */
    private int nConsumers;
        /** Last value broadcast to consumers */
    private MonitoringData lastData;
        /** Current value (last update) */
    private MonitoringData currentData;

    private static long nextKey = System.currentTimeMillis();
}

/**
 * A "smart" stack for storing parameter configurations.
 *
 * @todo does it need synchronized somewhere?
 */
class ConfStack
{
        /**
         * Helper class used to store parameter configurations.
         *
         */
    private static final class ParamConf
    {
        public ParamConf(ConsumerId c, Object v)
            {
                this.owner = c;
                this.value = v;
            }
        ConsumerId owner;
        Object value;
    }

        /**
         * Creates a new <code>ConfStack</code> instance.
         *
         */
    public ConfStack()
        {
            this.confs = new Stack();
        }
        
        /**
         * Activates a new parameter configuration by pushing it on top of the
         * configuration stack.
         *
         * @param c owner of the new configuration
         * @param v new parameter value
         */
    public final void
    push(ConsumerId c, Object v)
        {
            confs.push(new ParamConf(c,v));
        }
        
        /**
         * Removes all configurations pushed by the given consumer.
         *
         * This method should tipically be called after an application exits as part
         * of the application cleanup procedure.
         *
         * @todo review synchronized
         *
         * @param c a <code>ConsumerId</code> value
         */
    public final void
    clean(ConsumerId c)
        {
            Vector v = (Vector) confs.clone();
            confs.clear();
            for (int i=0; i<v.size(); i++) {
                ParamConf pc = (ParamConf)v.elementAt(i);

                if (! c.equals(pc.owner) ) {
                    confs.addElement(pc);
                }
            }
        }

        /**
         * Returns the current active configuration.
         *
         * @return current configuration value or null if there is no active
         *         configuration.
         */
    public final Object
    currentValue()
        {
            ParamConf pc = (ParamConf) confs.peek();

            return (pc == null) ? null : pc.value;
        }

        /**
         * Returns the owner of the current configuration value.
         *
         * @return current configuration owner or null it there is no active
         * configuration.
         */
    public final ConsumerId
    currentOwner()
        {
            ParamConf pc = (ParamConf) confs.peek();
                   
            return (pc == null) ? null : pc.owner;
        }

        /**
         * Storage for configurations
         */
    private Stack confs;
}

// class MetaSensor {
//     Sensor privDesc;
//     Sensor pubDesc;
//     Monitor monitor;
//     MonitoringData data;
//     Vector consumers;
//     boolean enabled;
//     SensorParameter[] parameters;
//     int intThreadhold;
//     long longThreashold;
//     double doubleThreadshold;
// }
