// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.worb;

import org.isam.exehda.HostId;
import java.net.Socket;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;


/**
 * Extends the stardard Java sockets (TCP/IP) implementation with a connection-time ID handshake
 * protocol.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-04-12 16:55:18 -0300 (Tue, 12 Apr 2005) $ $Revision: 744 $
 */
class WorbSocket
    extends Socket
{   
        /**
         * Creates a unconnected <code>WorbSocket</code> instance.
         *
         */
    WorbSocket()
        {
            super();
        }

        /**
         * Creates a <code>WorbSocket</code> instance connected to the specified host
         * address and port.
         *
         * @param addr a <code>String</code> value
         * @param port an <code>int</code> value
         * @exception IOException if an error occurs
         */
    WorbSocket( String addr, int port )
        throws IOException
        {
            super(addr, port);
            
            doHostHandshake();
        }

        /**
         * Exchange ID's with the remote end-point.
         *
         * @exception IOException if an error occurs
         */
    void
    doHostHandshake()
        throws IOException
        {
            try {
                writeLocalHostId();
                readRemoteHostId();
            }
            catch (ClassNotFoundException cnfe) {
                throw new IOException("Host hand-shake failed: "+cnfe);
            }
        }
    
        /**
         * Returns the ID of the host to which this socket is connected.
         *
         * @return a <code>HostId</code> value
         */
    HostId
    getRemoteHost()
        {
            return this.remoteHost;
        }


        //////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        //////////////////////////////////////////////////////////////

    
    private final void
    readRemoteHostId()
        throws IOException, ClassNotFoundException
        {
                // read size
            final byte[] nb = new byte[4];
            readBytes(nb, 0, nb.length);
            
            int n = 0;
            for ( int i=0,j=0; j<nb.length; i+=8,j++ ) {
                n |= ((nb[j] & 0x00FF) << i);
            }
            
                // read data
            byte[] buf = new byte[n];
            readBytes(buf, 0, buf.length);
            
                // de-serialize the host ID 
            ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(buf));
            this.remoteHost = (HostId) ois.readObject();
        }

    private final void
    writeLocalHostId()
        throws IOException
        {
            OutputStream os = getOutputStream();
            
            int n = LOCAL_HOST_BYTES.length;

                // write data size
            for (int i=0; i<32; i+=8) os.write( ( n >> i ) & 0x00FF);
            
                // write data
            os.write(LOCAL_HOST_BYTES);
            os.flush();
        }

    private final void
    readBytes( byte[] buf, int offset, int n )
        throws IOException
        {
            InputStream is = getInputStream();
            
            while ( n > 0 ) {
                int num = is.read(buf, offset, n);
                
                if ( num > 0 ) {
                    n -= num;
                    offset += num;
                }
                else {
                    throw new IOException("Unexpected end-of-stream");
                }
            }
        }

    static {

        try {
            ByteArrayOutputStream bout = new ByteArrayOutputStream(4096);
            ObjectOutputStream out = new ObjectOutputStream(bout);

            out.writeObject(HostId.getLocalHost());
            out.close();

            LOCAL_HOST_BYTES = bout.toByteArray();
        }
        catch (IOException ioe) {
            throw new ExceptionInInitializerError(ioe);
        }
    }

    private HostId remoteHost = null;
    
    private static final byte[] LOCAL_HOST_BYTES;
}
