// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.cib.backend.ldap;

import java.util.*;
import javax.naming.*;
import javax.naming.directory.*;
import javax.naming.ldap.*;


/**
 * Facility for accessing the LDAP server.
 *
 * @author frainer, last modified by $Author: lucc $
 * @version $Date: 2005-06-21 14:10:15 -0300 (Tue, 21 Jun 2005) $ $Revision: 951 $
 */
class LdapSession
{
        /**
         * Creates a new <code>LdapSession</code> instance that will connect to the
         * given server address using the specified user and password
         *
         * @param sAddr server address
         * @param rootDn DN of the root node
         * @param user user name used when connecting to the server
         * @param passwd the user's password
         */
    public LdapSession(String sAddr, String rootDn, String user, String passwd)
        {
            this.serverURL = "ldap://"+sAddr+"/"+rootDn;
            this.user = user;
            this.passwd = passwd;
            this.dirCtx = null;
        }
        
        /**
         * Activates this session by openning a connection to the ldap server previously
         * defined.
         *
         */
    public void
    open()
        throws NamingException
        {

            Hashtable env = new Hashtable();

            env.put( Context.INITIAL_CONTEXT_FACTORY,
                     "com.sun.jndi.ldap.LdapCtxFactory" );
                
            env.put( Context.PROVIDER_URL,
                     this.serverURL);
                
            env.put( Context.SECURITY_PRINCIPAL,
                     this.user);
                
            env.put( Context.SECURITY_CREDENTIALS,
                     this.passwd);

                // Create the initial context
            this.dirCtx = new InitialDirContext(env);
        }

    
        /**
         * Deactivates this session and releasing allocated resources.
         *
         */
    public void
    close() throws NamingException
        {
            dirCtx.close();
            dirCtx = null;
        }

    
        /**
         * Creates a new LDAP node for the given distinguished name using the given
         * initial attribute values.
         *
         * <p><b>Note</b> No checking is done if the parent node already exists in
         * LDAP. Ensuring this precondition is left up to the LdapSession user.
         *
         * @param dn a <code>String</code> value
         * @param ans initial attribute names
         * @param avs initial attribute values
         * @exception NameNotFoundException if parent node do not exists
         * @exception NameAlreadyBoundException if the given node already exists in the
         * directory service
         * @exception NamingException if some other error occurred.
         */
    public void
    createNode(String dn, String[] ans, String[] avs)
        throws NameNotFoundException, NameAlreadyBoundException, NamingException
        {
                // parameter checking
            if ( ans.length != avs.length ) {
                throw new IllegalArgumentException(
                    "Lengths of attribute names and values arrays do not match.");
            }
            

                // fill in initAtts
            Attributes initAtts = new BasicAttributes(true);
            for ( int i=0; i< ans.length; i++ ) {
                initAtts.put(ans[i],avs[i]);
            }
                // create the node
            dirCtx.createSubcontext(dn,initAtts);
        }

        /**
         * Returns the value of the given LDAP node attribute.
         *
         * <p> Concerning to performance, when acessing several attributes of the same
         * node consider using <code>getNodeAttributes(...),</code>.
         *
         * @param dn DN of the target LDAP node
         * @param attr the name of the attribute being queried
         * @return current attribute value
         */
    public String
    getNodeAttribute( String dn, String attr )
        throws NamingException
        {
            return getNodeAttributes(dn, new String[] {attr})[0];
        }

    
        /**
         * Returns all the values of the given LDAP node multi-valued attribute.
         *
         * @param dn a <code>String</code> value
         * @param attr a <code>String</code> value
         * @return a <code>String[]</code> value
         */
    public String[]
    getMultiValuedNodeAttribute( String dn, String attr )
        throws NamingException
        {            
            Attribute result = dirCtx.getAttributes(
                dn, new String[] {attr}).get(attr);

            if ( result == null ) {
                return new String[0];
            }
            else {
                String[] tmp = new String[result.size()];
            
                for (int i=0; i<tmp.length; i++ ) {
                    tmp[i] = (String) result.get(i);
                }
                
                return tmp;
            }
        }

        /**
         * Returns the value of the given LDAP node attributes.
         *
         * @param dn distinguished name of the target LDAP node
         * @param atts list of attributes of interest
         * @return the attribute current values
         */
    public String[]
    getNodeAttributes( String dn, String[] atts )
        throws NamingException
        {
            Attributes result = dirCtx.getAttributes(dn, atts);

            String[] tmp = new String[atts.length];

            for (int i=0; i<atts.length; i++ ) {
                tmp[i] = (String) result.get(atts[i]).get();
            }
                
            return tmp;
        }

        /**
         * Sets the value of the given resource attribute.
         *
         * <p> Concerning to performance, when acessing several attributes of the same
         * node consider using <code>setNodeAttributes(...),</code>.
         *
         * @param dn DN of the LDAP node being updated
         * @param an name of the attribute being updated
         * @param av attribute value
         */
    public void
    setNodeAttribute( String dn, String an, String av )
        throws NamingException
        {
            setNodeAttributes(dn, new String[] {an}, new String[] {av});
        }


    
        /**
         * Updates an LDAP multi-valued attribute entry
         *
         * @param dn DN of the LDAP node being updated
         * @param an name of the attribute being updated
         * @param avs attribute values
         */
    public void
    setMultiValuedNodeAttribute(String dn, String an, String[] avs)
        throws NamingException
        {
                // create a new attribute an initilize it to the given values
            Attribute newAtt = new BasicAttribute(an);
            for (int i=0; i<avs.length; i++) newAtt.add(avs[i]);
            
                // create an update attribute set containing the attribute to be updated
            Attributes atts = new BasicAttributes(true);
            atts.put(newAtt);
                // commit attribute update
            dirCtx.modifyAttributes(dn, DirContext.REPLACE_ATTRIBUTE, atts);
        }

    
        /**
         * Modifies the value of the given resource attributes.
         *
         * @param dn distiguished name of the LDAP node being updated
         * @param ans a list of attribute names
         * @param avs a list of attribute values
         */
    public void
    setNodeAttributes( String dn, String[] ans, String[] avs )
        throws NamingException
        {
            if ( ans.length != avs.length ) {
                throw new IllegalArgumentException(
                    "Size of attribute names and values do not match"); 
            }
                    
            Attributes atts = new BasicAttributes(true);

                // for each provided attribute name/value pair
            for (int i=0; i< ans.length; i++) {
                String an = ans[i];
                String av = avs[i];
                        
                    // create a new attribute an initilize it to the given value
                Attribute newAtt = new BasicAttribute(an);
                newAtt.add(av);

                    //
                    // FIX ME: need better handling of multi-valued attributes
                    //

                    //
                    // XXX: multi-value currently enabled on for xAttr
                    //
                if ( "xAttr".equals(an) ) {
                        // retrieve current attribute value
                    Attribute oldAtt = dirCtx.getAttributes(dn).get(an);
                    if ( oldAtt != null ) {
                            // append old values to the new one
                            //
                            // FIX ME: should replace the old value, otherwise the
                            // encoded cib attribute will be duplicated after the first
                            // update
                            //
                        Enumeration ave = oldAtt.getAll();
                        while ( ave.hasMoreElements() ) {
                            newAtt.add((String) ave.nextElement());
                        }
                    }
                }
                    // add the attribute to the update set
                atts.put(newAtt);
            }
                // commit attributes update
            dirCtx.modifyAttributes(dn, DirContext.REPLACE_ATTRIBUTE, atts);            
        }

        /**
         * Returns the attributes defined for the given ldap node
         *
         * @param dn DN of the target LDAP node
         * @return a <code>String[]</code> value
         */
    public String[]
    listNodeAttributes( String dn )
        throws NamingException
        {
                //
                // FIX ME: need to handle multiple xAttr
                //
                // get node attributes from ldap
            Attributes atts = dirCtx.getAttributes(dn);         
                // pack attributes in a String[]
            String[] results = new String[atts.size()];
            NamingEnumeration all = atts.getAll();
            for (int i=0; all.hasMore(); i++) {
                results[i] = (String)(((Attribute)all.next()).get());
            }
            return results;
        }

        /**
         * Retries from the given LDAP node an object previosly save with setObject().
         *
         * @param dn DN of the LDAP where the object is stored
         * @return an <code>Object</code> value
         */
    public Object
    getObject( String dn )
        throws NamingException
        {
            return dirCtx.lookup(dn);
        }

        /**
         * Stores a serialized version of the given object in the given LDAP node.
         *
         * @param dn DN of the LDAP node where the object is to be stored
         * @param o the object to be stored
         */
    public void
    setObject( String dn, Object o )
        throws NamingException
        {
            dirCtx.bind(dn, o);
        }

        /**
         * Search for LDAP nodes that match the given search filter. If
         * <code>subtree</code> is non-null, the search is restricted to the scope of the
         * given directory subtree.
         *
         * @param subtree a valid LDAP DN or null for searching from the root
         * @param filter a <code>String</code> value
         * @param attributesToReturn attributes to be returned, 'null' means all
         * attributes
         * @return a <code>javax.naming.NamingEnumeration</code> of
         * <code>javax.naming.directory.SearchResult</code> objects
         */
    public NamingEnumeration
    search( String subtree, String filter, String[] attributesToReturn)
        throws NamingException
        {
            if ( subtree == null ) subtree = "";
            
            SearchControls ctls = new SearchControls();
            ctls.setSearchScope(SearchControls.SUBTREE_SCOPE);
            ctls.setReturningAttributes(attributesToReturn);
                
            return dirCtx.search(subtree, filter, ctls);
        }
    

        ///////////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        ///////////////////////////////////////////////////////////////////
    
    private transient String serverURL;
    private transient String rootDn;
    private transient String user;
    private transient String passwd;
    private transient DirContext dirCtx;
}

