// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.ctxm;

import org.isam.exehda.Exehda;
import org.isam.exehda.services.Collector;
import org.isam.exehda.services.Collector.ConsumerId;
import org.isam.exehda.services.Collector.MonitoringConsumer;
import org.isam.exehda.services.Collector.Sensor;
import org.isam.exehda.services.Collector.MonitoringData;
import java.util.Hashtable;

/**
 * Smart storage for variables describing the current environment state.
 *
 * The implementation is based on a Hashtable. Additionally, access behavior is optimized
 * so that successive operations over a same environment variable are done faster by
 * caching the last accessed Hashtable entry.
 *
 * Environment variables representing underlaying raw sensors are in the form
 * "sensor:<rawsensor>", where <rawsensor> is the name of the underlaying sensor of
 * interest.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-16 18:17:53 -0300 (Thu, 16 Jun 2005) $ $Revision: 921 $
 */
public class Environment
    implements MonitoringConsumer
{
    static final long ENTRY_CLEANUP_AGE = 1000;

        /**
         * An environment variable.
         *
         */
    private class Entry
        {
                // each entry will stay in the environment at least 10s;
            static final long MIN_TTL = 10 * ENTRY_CLEANUP_AGE;
            
            static final int T_UNDEF  = 0;
            static final int T_STRING = 1;
            static final int T_DOUBLE = 2;
            static final int T_LONG   = 3;
            static final int T_INT    = 4;

            Entry( String name )
                {
                    this.entryName = name;
                    this.entryType = T_UNDEF;
                    this.entryLastModified = System.currentTimeMillis();
                    this.entryLastAccessed = entryLastModified + MIN_TTL;
                }

            final String getName()
                {
                    return this.entryName;
                }

            final long getLastModified()
                {
                    return this.entryLastModified;
                }

            final long getLastAccessed()
                {
                    return this.entryLastAccessed;
                }

            final boolean isDefined()
                {
                    return entryType != T_UNDEF;
                }

            final boolean isDefinedAs(int type)
                {
                    return entryType == type;
                }

            final Class getType()
                {
                    switch (entryType) {
                        case T_DOUBLE: return Double.TYPE;
                        case T_STRING: return String.class;
                        case T_INT: return Integer.TYPE;
                        case T_LONG: return Long.TYPE;
                        default: return Void.TYPE;   
                    }
                }

            final void setUndefined(long ts)
                {
                    this.entryLastModified = ts;
                    this.entryType = T_UNDEF;
                }

            final void update(String val, long ts)
                {
                    this.entryType = T_STRING;
                    this.entryLastModified = ts;
                    this.stringValue = val;
                }

            final void update(double val, long ts)
                {
                    this.entryType = T_DOUBLE;
                    this.entryLastModified = ts;
                    this.doubleValue = val;
                }

            final void update(long val, long ts)
                {
                    this.entryType = T_LONG;
                    this.entryLastModified = ts;
                    this.longValue = val;
                }

            final void update(int val, long ts)
                {
                    this.entryType = T_INT;
                    this.entryLastModified = ts;
                    this.intValue = val;
                }

            final String getString( boolean typeconv )
                {
                    updateAccessTime();

                    if ( typeconv ) {
                        switch (entryType) {
                            case T_DOUBLE: return Double.toString(doubleValue);
                            case T_INT: return Integer.toString(intValue);
                            case T_LONG: return Long.toString(longValue);
                            default: return stringValue;
                        }
                    }
                    else {
                        return stringValue;
                    }
                }

            final double getDouble( boolean typeconv )
                {
                    updateAccessTime();

                    if ( typeconv ) {
                        switch (entryType) {
                            case T_INT: return (double) intValue;
                            case T_LONG: return (double) longValue;
                            case T_STRING: return Double.valueOf(stringValue).doubleValue();
                            default: return doubleValue;
                        }
                    }
                    else {
                        return doubleValue;
                    }
                }

            final long getLong( boolean typeconv )
                {
                    updateAccessTime();

                    if ( typeconv ) {
                        switch (entryType) {
                            case T_INT: return (long) intValue;
                            case T_DOUBLE: return (long) doubleValue;
                            case T_STRING: return Long.parseLong(stringValue);
                            default: return longValue;
                        }
                    }
                    else {
                        return longValue;
                    }
                }

            final int getInt( boolean typeconv )
                {
                    updateAccessTime();
                    
                    if ( typeconv ) {
                        switch (entryType) {
                            case T_LONG: return (int) longValue;
                            case T_DOUBLE: return (int) doubleValue;
                            case T_STRING: return Integer.parseInt(stringValue);
                            default: return intValue;
                        }
                    }
                    else {
                        return intValue;
                    }
                }

                ///////////////////////////////////////////////////////
                // PRIVATE STUFF
                ///////////////////////////////////////////////////////
            private final void
            updateAccessTime()
                {
                    this.entryLastAccessed = getSnapshotTimeStamp();
                }
                
            private String entryName;
            private int entryType;
            private long entryLastModified;
            private long entryLastAccessed;
            
            private double doubleValue;
            private long   longValue;
            private int    intValue;
            private String stringValue;
        }

    public Environment()
        {
            this.envVars = new Hashtable();
                // XXX: findEntry() requires lastEntry to be non-null so we create a fake entry.
            this.lastAccessedEntry = new Entry("");

                // register to receive monitoring data
            this.cid = getCollector().addConsumer(this);
        }

    public void
    finalize()
        {
            ((Collector) Exehda.getService(Collector.SERVICE_NAME))
                .removeConsumer(this.cid);
        }
    
    public final double
    getDouble ( String var, double defVal, boolean typeconv )
        {
            Entry e = findEntry(var);

            return ( e.isDefinedAs(Entry.T_DOUBLE) ) ? e.getDouble(false)
                : ( e.isDefined() && typeconv) ? e.getDouble(true)
                : defVal;
        }
 
    public final int
    getInt ( String var, int defVal, boolean typeconv )
        {
            Entry e = findEntry(var);

            return ( e.isDefinedAs(Entry.T_INT) ) ? e.getInt(false)
                : ( e.isDefined() && typeconv) ? e.getInt(true)
                : defVal;
        }
 
    public final long
    getLong ( String var, long defVal, boolean typeconv )
        {
            Entry e = findEntry(var);

            return ( e.isDefinedAs(Entry.T_LONG) ) ? e.getLong(false)
                : ( e.isDefined() && typeconv) ? e.getLong(true)
                : defVal;
        }
 
    public final String
    getString ( String var, String defVal, boolean typeconv )
        {
            Entry e = findEntry(var);

            return ( e.isDefinedAs(Entry.T_STRING) ) ? e.getString(false)
                : ( e.isDefined() && typeconv) ? e.getString(true)
                : defVal;
        }
 
    public final boolean
    isDefined ( String var )
        {
            return ( envVars.get(var) != null ) ;
        }
 
    public final boolean
    isModifiedSince ( String var, long ts )
        {
            return ( findEntry(var).getLastModified() > ts );
        }
 
    public final long
    getLastModified ( String var )
        {
            return findEntry(var).getLastModified();
        }

    public final Class
    getTypeOf( String var )
        {
            Entry e = (Entry) envVars.get(var);

            return ( e != null ) ? e.getType() : Void.TYPE;
        }

        /**
         * Callback for monitoring data (from <code>MonitoringConsumer</code> interface.
         *
         * @param timeStamp a <code>long</code> value
         * @param data a <code>MonitoringData[]</code> value
         */
    public void
    update( long timeStamp, MonitoringData[] data )
        {
                // update environment variables with the new monitoring data
            synchronized (this) {
                for (int i = 0; i< data.length; i++ ) {
                    if ( data[i] == null ) break;

                    Entry ee = (Entry) envVars.get(data[i].getSensor());
                    if ( ee != null ) {
                        
                        Class type = data[i].getDataType();

                        if ( Integer.TYPE.equals(type) ) {
                            ee.update(data[i].getInt(), data[i].getTimeStamp());
                        }
                        else if ( Double.TYPE.equals(type) ) {
                            ee.update(data[i].getDouble(), data[i].getTimeStamp());
                        }
                        else if ( Long.TYPE.equals(type) ) {
                            ee.update(data[i].getLong(), data[i].getTimeStamp());
                        }
                        else if ( String.class.equals(type) ) {
                            ee.update(data[i].getString(), data[i].getTimeStamp());
                        }
                        else {
                                //
                                // TO DO: log warning
                                //
//                         log(Sev.WARNING, "update",
//                             "Unsupported data type {0} for sensor {1}", type, name);
//                             System.out.println("\n\nOOPS entry type not suported: "+type+"\n\n");
                        }

//                         System.out.println(ee.getName()+" last modified "+ee.getLastModified());
                    }
                    else {
                            //
                            // TO DO: log warning
                            //
//                         System.out.println("\n\nOOPS no such entry: "+data[i].getSensor()+"\n\n");
//                         System.out.println(envVars);
                    }
                }
                    // wake up context manager main thread
                notifyAll();
            }

                // garbage-collect old unused entries
            cleanEnvironment();
        }

    public void
    lockSnapshot()
        {
                //
                // TO DO: either freeze a snapshot of current envVars state or block
                // updates to envVars while locked
                //

            this.snapshotTimeStamp = System.currentTimeMillis();
        }

    public void
    unlockSnapshot()
        {
                //
                // TO DO: allow envVars to be modified again
                //
        }

    public long
    getSnapshotTimeStamp()
        {
            return this.snapshotTimeStamp;
        }

        ///////////////////////////////////////////////////////////
        // PRIVATE STUFF
        ///////////////////////////////////////////////////////////
 
    private final void
    undefine ( String var, long ts )
        {
            findEntry(var).setUndefined(ts);
        }
    
        /**
         * Returns the Entry object corresponding to the given environment variable.
         *
         * Entry objects are allocated on-demand. So are enabled the corresponding
         * underlaying sensors.
         *
         * @param var a <code>String</code> value
         * @return an <code>Entry</code> value
         */
    private final Entry
    findEntry( String var )
        {
            if ( lastAccessedEntry.getName().equals(var) ) {
                return lastAccessedEntry;
            }
            else if ( (lastAccessedEntry = (Entry) envVars.get(var)) == null ) {
                envVars.put(var, lastAccessedEntry = new Entry(var));

                    //
                    // XXX: enable underlaying sensors on first access; perhaps this
                    // should be done somewhere else outside findEntry
                    //
                if ( var.startsWith("sensor:") ) {
                    enableSensor(var.substring(7), lastAccessedEntry);
                }
            }
            return lastAccessedEntry;
        }
    
        /**
         * Contacts the collector to enable the given sensor, binding the sensor to the
         * given environment entry.
         *
         * @param sn sensor's name
         * @param ee environemnt entry to which the sensor must be bound
         */
    private final void
    enableSensor( String sn, Entry ee )
        {
//             System.out.println("Enabling sensor "+sn);
            
                // lookup corresponding sensor object
            Sensor s = getCollector().getSensor(sn);

            if ( s == null ) {
                    //
                    // FIX ME: perhaps should silently catch the exception; the filter
                    // will behave like the sensor is not avaiable, returning the default
                    // value
                    //
                throw new IllegalArgumentException("No such sensor :"+sn);
            }   
                // bind the entry to the sensor object
                //
                // XXX: MonitoringData.getSensor(), as used in method update(), returns a
                // ResourceName and not a Sensor object
            envVars.put(s.getName(), ee);
                // request the collector to enable the sensor
            getCollector().setSensorEnabled(this.cid, s, true);
        }

    private final void
    disableSensor( String s )
        {
                //
                // TO DO: add implementation
                //
            throw new NullPointerException("Not yet implemented.");
        }

    
        /**
         * Cleans up the environment by removing no longer used entries and disabling the
         * respective underlaying sensors.
         *
         */
    private final void
    cleanEnvironment()
        {
                //
                // TO DO: clean-up entries whose the difference from the current time to
                // its last access time is greather than ENTRY_CLEANUP_AGE
                //
        }

        /**
         * Returns a reference to the middleware's Collector service.
         *
         * @return a <code>Collector</code> value
         */
    private final Collector
    getCollector()
        {
            return (Collector) Exehda.getService(Collector.SERVICE_NAME);
        }

    private ConsumerId cid;
        /** environment vars indexed by variable's name and raw sensor*/
    private Hashtable envVars;
    private Entry lastAccessedEntry;
    private long  snapshotTimeStamp;
}

