package dimi.channel;

import java.util.Set;
import java.util.Iterator;
import java.util.Collections;
import java.util.TreeSet;
import java.rmi.RemoteException;

import dimi.DBase;
import dimi.network.NetworkDistanceMeasurer;
import dimi.network.NetworkRandomDistanceMeasurer;
import dimi.channel.neighbour.NeighbourPointer;
import dimi.channel.neighbour.NeighbourPointerComparator;

/**
 * Classe auxiliar para o trato vizinhos em um canal
 * @author mcmoraes
 */
public class NeighbourhoodManager {
    
    private Set neighbourPointerSet = null;
    private NetworkDistanceMeasurer networkDistanceMeasurer = null;
    
    /** Creates a new instance of PeerManager */
    public NeighbourhoodManager() {
        this.neighbourPointerSet = Collections.synchronizedSortedSet(new TreeSet(new NeighbourPointerComparator()));
        this.networkDistanceMeasurer = new NetworkRandomDistanceMeasurer();
    }
    
    /**
     * Obtem todos os vizinhos deste manager.
     * @return todos os vizinhos deste manager
     */    
    public final synchronized DBase[] getAllNeighbours(){
        DBase[] result = new DBase[neighbourPointerSet.size()];
        int i = 0;
        Iterator it = neighbourPointerSet.iterator();
        while(it.hasNext()){
            result[i++] = ((NeighbourPointer)it.next()).getTarget();
        }
        return result;
    }
    
    /**
     * Obtem todos os ponteiros para vizinhos deste manager
     * @return todos os ponteiros para vizinhos deste manager
     */    
    public final synchronized NeighbourPointer[] getAllPointers(){
        return (NeighbourPointer[])neighbourPointerSet.toArray();
    }

    /**
     * Adiciona um conjunto de vizinhos a este manager. Este
     * metodo pode ser lento, visto que ele calcula a distancia
     * ate todos os elementos do conjunto.
     * @param peers o conjunto de unidades de disseminacao a tornarem-se
     * vizinhas em um canal
     */    
    public final void addAllNeighbours(DBase[] dbases){
        if(dbases!=null){
            for(int i=0;i<dbases.length;i++){
                try{
                    addNeighbour((DBase)dbases[i]);
                }
                catch(RemoteException re){}
            }
        }
    }
    
    /**
     * Adiciona um conjunto de ponteiros para vizinhos a este manager.
     * @param pointers os ponteiros para os vizinhos deste manager
     */    
    public final synchronized void addAllPointers(NeighbourPointer[] pointers){
        for(int i=0;i<pointers.length;i++){
            neighbourPointerSet.add(pointers[i]);
        }
    }

    /**
     * Adiciona um vizinho. Este metodo calcula a distancia ate o novo vizinho.
     * @param dbase o novo vizinho
     * @throws RemoteException indicate that either a communication failure or a protocol error has occurred
     */    
    public void addNeighbour(DBase dbase) throws RemoteException{
        long distance = networkDistanceMeasurer.measureDistanceTo(dbase);
        NeighbourPointer neighbourPointer = new NeighbourPointer(dbase,distance);
        synchronized(this){
            if(neighbourPointerSet.contains(neighbourPointer)){
                neighbourPointerSet.remove(neighbourPointer);
            }
            neighbourPointerSet.add(neighbourPointer);
        }
    }

    /**
     * Adiciona um ponteiro para um vizinho
     * @param neighbourPointer o ponteiro a ser adicionado
     */    
    public final synchronized void addPointer(NeighbourPointer neighbourPointer){
        if(neighbourPointer != null){
            if(neighbourPointerSet.contains(neighbourPointer)){
                neighbourPointerSet.remove(neighbourPointer);
            }
            neighbourPointerSet.add(neighbourPointer);
        }
    }
    
    /**
     * Obtem um vizinho aleatorio.
     * @return o vizinho aleatorio
     */    
    public final synchronized NeighbourPointer getRandomNeighbour(){
        return null;
    }
    
    /**
     * Obtem o vizinho mais proximo.
     * @return o vizinho mais proximo
     */    
    public final synchronized DBase getClosestNeighbour(){
        Iterator it = neighbourPointerSet.iterator();
        while(it.hasNext()){
            return ((NeighbourPointer)it.next()).getTarget();
        }
        return null;
    }
        
}
