// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.primos;

import org.isam.exehda.HostId;
import org.isam.exehda.ResourceName;
import org.isam.exehda.ResourceName.NameSpace;
import org.isam.exehda.services.Collector.Monitor;
import org.isam.exehda.services.Collector.Sensor;
import org.isam.exehda.services.Collector.SensorParameter;
import org.isam.exehda.services.Collector.MonitoringData;

import java.util.Vector;
import java.util.Hashtable;


/**
 * Primos's java runtime monitor, i.e. information is gathered from sources directly
 * available in the Java Runtime, do not demand direct native host accesses.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-19 20:03:45 -0300 (Sun, 19 Jun 2005) $ $Revision: 943 $
 */
public class JavaRuntimeMonitor
    implements Monitor
{
    private final String MONITOR_NAME = "PrimosJavaRuntimeMonitor";

    public JavaRuntimeMonitor()
        {
            NameSpace nsLocalHost = HostId.getLocalHost().toNameSpace();
            this.name = new ResourceName("monitor",MONITOR_NAME, nsLocalHost);
            this.nspace = new NameSpace(name.getCommonName(), nsLocalHost);
            this.sensorsByName = new Hashtable();
            loadSensors();
        }

    public ResourceName
    getName()
        {
            return this.name;
        }

    public Sensor[]
    getSensors()
        {
            Sensor[] sa  = new Sensor[sensors.length];    
            for (int i=0; i< sa.length; i++) {
                sa[i] = makeSensor(sensors[i]);
            }
            return sa;
        }

    public Sensor
    getSensor( String sname)
        {
            return makeSensor(findRuntimeSensor(sname));
        }

    public SensorParameter[]
    getSensorParameters( Sensor s )
        {
            return new SensorParameter[0];
        }

    public Object
    getSensorParameter( SensorParameter p )
        {
                // XXX: currently there are no parameters
            throw new IllegalArgumentException("No such sensor parameter");
        }

    public void
    setSensorParameter( SensorParameter p, Object v )
        {
                // XXX: do nothing
        }
    
    public void
    setSensorEnabled( Sensor s, boolean enable )
        {
            findRuntimeSensor(s).setEnabled(enable);
        }

    public boolean isSensorEnabled( Sensor s )
        {
            return findRuntimeSensor(s).isEnabled();
        }

    public MonitoringData
    probeSensor( Sensor s, MonitoringData d )
        {
                // TODO: check if d is valid and reuse d
            if ( d == null ) {
                d = new MonitoringData(s.getName());
            }

            findRuntimeSensor(s).probe(d);
            return d;
        }

    public void
    installSensor( String name, String impl, Hashtable params )
        {
                // TO DO: 
            throw new NullPointerException("Operation not supported.");
        }

    public void
    quantumExpired()
        {
            for ( int i=0; i<sensors.length; i++ ) {
                sensors[i].update();
            }
        }
    
        ///////////////////////////////////////////////////
        // PRIVATE STUFF
        ///////////////////////////////////////////////////
    
    private final void
    loadSensors()
        {            
            RuntimeSensor rs = null;
            Vector sset = new Vector();
            
            sset.addElement(rs = new SysPropSensor("java.version"));
            sensorsByName.put(rs.getName(), rs);
            
            sset.addElement(rs = new SysPropSensor("java.vendor"));
            sensorsByName.put(rs.getName(), rs);
            
            sset.addElement(rs = new SysPropSensor("os.name"));
            sensorsByName.put(rs.getName(), rs);
            
            sset.addElement(rs = new SysPropSensor("os.version"));
            sensorsByName.put(rs.getName(), rs);
            
            sset.addElement(rs = new SysPropSensor("os.arch"));
            sensorsByName.put(rs.getName(), rs);
            
            sset.addElement(rs = new RuntimeFreeMemSensor());
            sensorsByName.put(rs.getName(), rs);
            
            sset.addElement(rs = new RuntimeTotalMemSensor());
            sensorsByName.put(rs.getName(), rs);
            
            sset.addElement(rs = new GuiModeSensor());
            sensorsByName.put(rs.getName(), rs);

            this.sensors = new RuntimeSensor[sset.size()];
            sset.copyInto(this.sensors);
        }

    private final Sensor
    makeSensor(RuntimeSensor rsen)
        {
                // FIX ME: use indirect key instead of rsen better isolation
            return  new Sensor(
                new ResourceName("sensor", rsen.getName(), this.nspace),
                rsen.getType(),
                rsen); 
        }

    private final RuntimeSensor
    findRuntimeSensor(String name)
        {
            RuntimeSensor rsen =  (RuntimeSensor) sensorsByName.get(name);
            if ( rsen != null ) return rsen;
            else throw new NullPointerException("No such sensor");
        }

    private final RuntimeSensor
    findRuntimeSensor(Sensor s)
        {
            RuntimeSensor rsen = (RuntimeSensor) s.getKey();
            if ( rsen != null ) return rsen;
            else throw new NullPointerException("No such sensor");
        }

    private ResourceName name;
    private NameSpace nspace;
    private RuntimeSensor[] sensors;
    private Hashtable sensorsByName;

        //////////////////////////////////////////////////////////
        // INNER CLASSES
        //////////////////////////////////////////////////////////

        /**
         * Abstract super-class of all standalone Java Runtime Sensors
         *
         * @author last modified by $Author: lucc $
         * @version $Date: 2005-06-19 20:03:45 -0300 (Sun, 19 Jun 2005) $ $Revision: 943 $
         */
    static abstract class RuntimeSensor
    {
        public RuntimeSensor(String name) { this.name = name; }
        public String getName() { return this.name; }
        public boolean isEnabled() { return enabled; }
        public void setEnabled( boolean e) { this.enabled = e; }
        public abstract Class getType();
        public abstract void update();
        public abstract void probe( MonitoringData data );
        protected final String name;
        protected boolean enabled = false;
    }
    
        /**
         * Probes the value of a given system property.
         *
         * @author last modified by $Author: lucc $
         * @version $Date: 2005-06-19 20:03:45 -0300 (Sun, 19 Jun 2005) $ $Revision: 943 $
         */
    static class SysPropSensor extends RuntimeSensor
    {
        public SysPropSensor( String p ) {
            super("runtime."+p);
            propName = p; propValue = null;
        }
        public Class getType() { return String.class; }
        public void update() { propValue = System.getProperty(propName); }
        public void probe( MonitoringData data ) { data.setString(propValue); }
        private final String propName;
        private String propValue;
    }

        /**
         * Probes java.lang.Runtime.freeMemory()
         *
         * @author last modified by $Author: lucc $
         * @version $Date: 2005-06-19 20:03:45 -0300 (Sun, 19 Jun 2005) $ $Revision: 943 $
         */
    static class RuntimeFreeMemSensor extends RuntimeSensor
    {
        public RuntimeFreeMemSensor() { super("runtime.freeMemory"); }
        public Class getType() { return Long.TYPE; }
        public void update() { freeMem = Runtime.getRuntime().freeMemory(); }
        public void probe( MonitoringData data ) { data.setLong(freeMem); }
        private long freeMem;
    }

        /**
         * Probes java.lang.Runtime.totalMemory()
         *
         * @author last modified by $Author: lucc $
         * @version $Date: 2005-06-19 20:03:45 -0300 (Sun, 19 Jun 2005) $ $Revision: 943 $
         */
    static class RuntimeTotalMemSensor extends RuntimeSensor
    {
        public RuntimeTotalMemSensor() { super("runtime.totalMemory"); }
        public Class getType() { return Long.TYPE; }
        public void update() { totalMem = Runtime.getRuntime().totalMemory(); }
        public void probe( MonitoringData data ){ data.setLong(totalMem);}
        private long totalMem;
    }

        /**
         * Provides information about the graphic user interface capabilites of the
         * current runtime.
         *
         * @author last modified by $Author: lucc $
         * @version $Date: 2005-06-19 20:03:45 -0300 (Sun, 19 Jun 2005) $ $Revision: 943 $
         */
    static class GuiModeSensor extends RuntimeSensor
    {
        public GuiModeSensor() { super("runtime.gui.awt.mode"); update(); }
        public Class getType() { return String.class; }
        public void update()
            {
                    // check system only once
                if ( enabled && mode == null ) {
                    try {
                        Class.forName("java.awt.Frame");
//                     java.awt.Dimension sres =
//                         java.awt.Toolkit.getDefaultToolkit().getScreenResolution();
                        java.awt.Dimension ssz =
                            java.awt.Toolkit.getDefaultToolkit().getScreenSize();

                        mode = ( ssz.width>500 && ssz.height>500 ) ? "high res"
                            : "low res";

                        try {
                            Class.forName("javax.swing.JFrame");                    
                            mode = "swing-"+mode;
                        }
                        catch (ClassNotFoundException e) { mode = "awt-"+mode; }
                    }
                    catch (Exception e) { mode = "none"; }
                }
            }
    
        public void probe( MonitoringData data ) {data.setString(mode);}
        private String mode;
    }

}



