// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.primos.exec;

import org.isam.perv.IsamClassLoader;
import org.isam.exehda.Exehda;
import org.isam.exehda.ResourceName;
import org.isam.exehda.HostId;
import org.isam.exehda.ObjectId;
import org.isam.exehda.ApplicationId;
import org.isam.exehda.services.Worb;
import org.isam.exehda.services.Executor;
import org.isam.exehda.services.CellInformationBase;
import org.isam.exehda.services.ObjectSeed;
import org.isam.exehda.services.OXManager;
import org.isam.exehda.services.OXManager.OXHandle;
import org.isam.exehda.services.ObjectSeed.Activator;
import org.isam.exehda.services.ObjectSeed.MarshaledOX;
import org.isam.exehda.services.Logger.Sev;
import org.isam.exehda.services.AbstractService;
import org.isam.util.xml.XmlElement;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.io.StringReader;
import java.io.IOException;
import java.net.UnknownHostException;



/**
 * PRIMOS' Executor implementation
 *
 * <p>Accepted profile parameters
 *
 * <ul>
 *
 * <li><b>defaultHeuristic</b>: defines the scheduling (placement) heuristic to use in
 * the absence of a custom, application provided, scheduling heuristic.
 *
 *    <p>The default value is <code>SysSchedulerSchedHeuristic</code>, which behaves as
 *    follows: if the hint is of type HostId its interpreted as a direct host
 *    specification; otherwise if it is of type ResourceConstraint the request is
 *    forwarded to the Scheduler service, if such service is avaiable. If anything fails,
 *    it selects the localhost.
 *
 * <li><b>maxRetries</b>: defines how many times the executor service will retry a remote
 * object instantiation or migration before returning an error to the application.
 *
 *    <p>The default value is <code>5</code>.
 *
 * </ul>
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-08-16 18:09:07 -0300 (Tue, 16 Aug 2005) $ $Revision: 1045 $
 */
public class ExecutorImpl
    extends AbstractService
    implements Executor
{
    private final static String DFLT_SCHED_HEURISTIC = SysSchedulerSchedHeuristic.class.getName();
    private final static int DFLT_MAX_RETRY = 5;

    private final static String CIB_ATT_APP_CODEBASE = "codeBase";
    private final static String CIB_ATT_APP_LDESC = "lauchingDescriptor";
    private final static String CIB_ATT_APP_STATUS = "status";
    private final static String APP_STATUS_RUNNING = "running";
    private final static String APP_STATUS_DONE = "done";

    
        /**
         * Creates a new <code>ExecutorImpl</code> instance.
         *
         */
    public ExecutorImpl()
        {
            this.appByThrGrp = new Hashtable();
            this.appById = new Hashtable();
            this.thisHost = HostId.getLocalHost();
        }
    
    public void
    start()
        throws ClassNotFoundException, InstantiationException,
        IllegalAccessException, UnknownHostException
        {
                // setup default scheduling heuristic
            String pSchedHeuristic = Exehda.getServiceProperty(
                SERVICE_NAME, "defaultHeuristic", DFLT_SCHED_HEURISTIC);
            this.defaultHeuristic = (SchedulingHeuristic)
                Class.forName(pSchedHeuristic).newInstance();

                // configure max retries for createObject and moveObject
            this.maxRetryies = Exehda.getServiceProperty(
                SERVICE_NAME, "maxRetryies", DFLT_MAX_RETRY);

                // export service to WORB
            Worb worb = (Worb) Exehda.getService(Worb.SERVICE_NAME);
            worb.exportService(this, Executor.class, Executor.SERVICE_NAME);

                //
                // XXX: remeber thread group of kernel threads
                //
            this.kernelThrGroup = Thread.currentThread().getThreadGroup();
        }
    
    public void
    stop() throws java.lang.Exception
        {
                // TODO: unexport service
        }

    
        /**
         * Describe <code>startApplication</code> method here.
         *
         * @param launchDescriptor a <code>XmlElement</code> value
         */
    public void
    startApplication(XmlElement launchDescriptor)
        {
            final String MTD = "startApplication()";
            
            log(Sev.TRACE, MTD, "entering");
            try {
                ApplicationId newID = nextApplicationId();
                Application app = new Application(newID, launchDescriptor, false);
                
                log(Sev.INFO, MTD, "Launching application codebase={0}", app.getCodeBase());

                    //
                    // XXX: need to register the launched application before calling
                    // start() otherwise currentApplication() would be inconsistent for a
                    // few moments
                    //
                registerApplication(app, true);

                    // start main thread
                app.start();
            }
            catch (Exception e) {
                log(Sev.ERROR, MTD, e);

                    //
                    // TO DO: cancel registration of the app and release allocated resources
                    //
            }
            log(Sev.TRACE, MTD, "leaving");
        }

    

    public ObjectId
    createObject( Class cls, Object[] params, Activator a, Object hint)
        {
            final String MTD = "createObject()";
            try {
                log(Sev.TRACE, MTD, "entering");


                if ( thisHost.equals(hint) ) {
                    return localCreateObject(cls, params, a);
                }
                else {                
                    Vector avoidedHosts = new Vector();
                    int n = maxRetryies;

                    do {
                        HostId targetHost = getHeuristic().chooseCreationHost(
                            cls.getName(), params, hint, avoidedHosts);

                        log(Sev.DEBUG, MTD, "heuristic has selected host {0}", targetHost);

                        try {
                            if ( targetHost != null ) {
                                return thisHost.equals(targetHost)
                                    ? localCreateObject(cls, params, a)
                                    : remoteCreateObject(cls, params, a, targetHost);
                            }
                        }
                        catch (Exception e) {
                            log(Sev.ERROR, MTD, "failed on {0} due to exception:", targetHost);
                            log(Sev.ERROR, MTD, e);
                        }
                    }
                    while ( --n > 0 );

                        //
                        // TO DO: throw exception
                        // 
                }
            }            
            catch ( Exception e ) {
                log(Sev.ERROR, MTD, e);
            }
            finally {
                log(Sev.TRACE, MTD, "leaving");
            }

                //
                // TO DO: throw exception on failure
                // 
            return null;
        }
    

    public void
    moveObject( ObjectId oxID, Object placementHint )
        {
            final String MTD = "moveObject()";
            try {
                log(Sev.DEBUG, MTD, "entering");
            
                    // where is it?
                HostId loc = getOXLocation(oxID);
                            
                    // check the ox is hosted in this node
                if ( thisHost.equals(loc) ) {

                    log(Sev.DEBUG, MTD, "target OX is HERE! Deactivating it...");
                       
                        // deactivate the ox
                    MarshaledOX mox = deactivateObject(oxID);

                    log(Sev.DEBUG, MTD, "OX {0} deactivated.", oxID);

                    Vector avoidedHosts = new Vector();
                    int n = maxRetryies;
                    do {
                            // select the target host for receiving it
                        HostId targetHost = getHeuristic().chooseMigrationHost(
                            oxID,  placementHint, avoidedHosts);

                        log(Sev.DEBUG, MTD, "Selected host is {0}", targetHost);

                        try {
                            if ( targetHost != null ) {

                                if ( thisHost.equals(targetHost) ) {
                                        // just reactivate the OX
                                    activateObject(mox);
                                }
                                else {
                                        // contact remote Executor for re-activating the OX
                                    remoteActivateObject(mox, targetHost);
                                }
                                    // exit (success)
                                return;
                            }
                            else {
                                    // scheduling heuristic had given up
                                break;
                            }
                        }
                        catch (Exception e) {
                            log(Sev.ERROR, MTD, "failed on {0} due to exception:", targetHost);
                            log(Sev.ERROR, MTD, e);
                        }
                    }
                    while ( --n > 0 );
                
                    log(Sev.ERROR, MTD,
                        "max retries exceeded. Restoring the object in the local host.:");

                        //
                        // XXX: restore locally or throw exception?
                        //
                    activateObject(mox);

                }
                    // the ox is not here
                else {
                        // forward the request to the remote node
                    getRemoteExecutor(loc).moveObject(oxID, placementHint);
                }
            }
            finally {
                log(Sev.DEBUG, "moveObject()", "leaving");
            }
        }


    public MarshaledOX
    deactivateObject( ObjectId oxID )
        {
                //
                // TO DO: check the OX is hosted in this node
                //
            
                // just forward the request to the local objectseed
            MarshaledOX oxState = new MarshaledOX();

            getObjectSeed().deactivateObject(oxID, oxState);

            return oxState;
        }


    public void
    activateObject( MarshaledOX oxState )
        {
                // just forward the request to the local objectseed
            getObjectSeed().activateObject(oxState);
        }


    public void
    exitApplication()
        {
            final String MTD = "exitApplication()";
            
            final ApplicationId appId = currentApplication();

            if ( appId != null ) {
                log(Sev.INFO, MTD, "finalizing app {0}", appId);
                
                runKernelAction( new Runnable() {
                        public void run() {
                            try {
                                    // XXX: delays kill all threads a few seconds
                                Thread.currentThread().sleep(3000);
                                leaveApplication(appId);

                                    //
                                    // TODO: propagate the request to other nodes running the application
                                    //
                            }
                            catch(Exception e) {
                                log(Sev.WARNING, "exitApplication()", e);
                            }
                        }
                    });
            }

            Thread.currentThread().stop();
        }

    public void
    setHeuristic( SchedulingHeuristic h )
        {
            ThreadGroup tg = Thread.currentThread().getThreadGroup();

            if ( tg instanceof Application ) {
                log(Sev.DEBUG, "setHeuristic", "current thread group is instance of Application");
                ((Application) tg).setHeuristic(h);
            }
            else {
                    // TODO: follow up tg hierarchy
            }
        }

    
        /**
         * Returns the ID of the current application according to the current execution
         * context (formelly, the current thread).
         *
         * @return an <code>ApplicationId</code> value
         *
         * @exception NullPointerException if the current execution context could not
         * determined because the caller thread is an un-managed thread.
         */
    public ApplicationId
    currentApplication()
        {
            Application app = getApplicationForCurrentThread();

            if ( app != null ) {
                return app.getId();
            }
            else if ( isKernelThread() ) {
                return null;
            }
            else {
                throw new NullPointerException(
                    "Could not determine current execution context because"
                    +" the current thread is not actally managed by EXEHDA");
            }
        }

    
        /**
         * Synchronously run the provided action in behalf of the given application.
         *
         * @param aid an <code>ApplicationId</code> value
         * @param action a <code>Runnable</code> value
         */
    public void
    runAction(ApplicationId aid, Runnable action)
        {
            final String MTD = "runAction()";
            
            log(Sev.TRACE, MTD, "entering appid={0}", aid);
            try {
                    // if inside the kernel, allow ID switching
                if ( isKernelThread() ) {
                    if ( aid == null ) {
                        runKernelAction(action).join();
                    }
                    else {
                            // impersonate the specified application
                        Application app = getApplicationForId(aid);
                        app.runAction(action);
                    }
                }
                else {
                        // otherwise, enforce application identity

                        // XXX: in the case that currentApplication fails(), just belive
                        // the provided ID is correct.
                    if ( currentApplication() != null ) {
                        aid = currentApplication();
                    }
                    Application app = getApplicationForId(aid);
                    app.runAction(action);
                }
                
            }
            catch (Exception e) {
                log(Sev.ERROR, MTD, e);
            }
            finally {
                log(Sev.TRACE, MTD, "leaving");
            }
        }

        ////////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        ////////////////////////////////////////////////////////////////

        /**
         * Returs a reference to the Executor service running on the specified remote
         * host.
         *
         * @param targetHost target remote host
         * @return an <code>Executor</code> value
         */
    private final Executor
    getRemoteExecutor(HostId targetHost)
        {
            Worb worb = (Worb) Exehda.getService(Worb.SERVICE_NAME);

                //
                // XXX: is it safe or should we use contactAddress and CIB to build URI?
                //
            String executorUri = "worb://"+targetHost.toExternalForm()+"/"+SERVICE_NAME;

            return (Executor) worb.lookupService(executorUri, Executor.class);
        }

    
        /**
         * Returns whether the invoking thread is a middleware kernel thread.
         *
         * @return a <code>boolean</code> value
         */
    private final boolean
    isKernelThread()
        {
                // FIX ME: there should be a specific thread group for kernel threads
            
            ThreadGroup tg = Thread.currentThread().getThreadGroup();
            
            while ( tg != null ) {
                if ( tg.equals(kernelThrGroup) ) {
                    return true;
                }
                else {
                    tg = tg.getParent();
                }
            }

            return false;
        }
    
        /**
         * Runs the specified action in a privileged kernel thread.
         *
         * @param action a <code>Runnable</code> value
         * @return the kernel thread in which the action will be run
         * 
         */
    private final Thread
    runKernelAction(Runnable action) 
        {
            Thread kt = new Thread(kernelThrGroup, action);
            kt.start();
            return kt;
        }
    
        /**
         * Returns the scheduling heurist active for the aplication the owns the current
         * thread.
         *
         * @return a <code>SchedulingHeuristic</code> value
         */
    private final SchedulingHeuristic
    getHeuristic()
        {
            final String MTD = "getHeuristic()";
            
                //
                // TODO: proceed with parent thread group if app not found
                //
            Application app = getApplicationForCurrentThread();

            if ( app == null ) {
                log(Sev.WARNING, MTD, "calling thread not registered, will use the default heuristic");
                return defaultHeuristic;
            }
            
            SchedulingHeuristic h = app.getHeuristic();

            if ( h != null ) {
                log(Sev.DEBUG, MTD, "using app configured heuristic {0}", h);
                return  h;
            }
            else {
                log(Sev.DEBUG, MTD, "using default heuristic {0}", defaultHeuristic);
                return defaultHeuristic;
            }
        }

    
        /**
         * Returns the Application (local thread group) corresponding to the given ID.
         *
         * <p>An Application object is created if it does not yet exists for the given ID.
         *
         * @param appID an <code>ApplicationId</code> value
         * @return an <code>Application</code> value
         * @exception Exception if an error occurs
         */
    private final Application
    getApplicationForId( ApplicationId appID ) throws Exception
        {
            final String MTD = "getApplicationForId()";
            
            log(Sev.TRACE, MTD, "entering id={0}", appID);
            
            Application app = (Application) appById.get(appID);
            
            if ( app == null ) {
                    // the given application is not yet running on this node
                
                log(Sev.INFO, MTD, "Localhost is joining the execution of application {0}", appID);

                app = startApplicationWorker(appID);
            }
            log(Sev.TRACE, MTD, "leaving");
            
            return app;
        }

        /**
         * Returns the Application (thread group) to which the  invoking thread belongs.
         *
         * @return an <code>Application</code> value or null if the owning application
         * could not be determined
         */
    private final Application
    getApplicationForCurrentThread()
        {
            ThreadGroup tg = Thread.currentThread().getThreadGroup();

                //
                // TO DO: cache detected applications for faster access; only significant
                // if the thread group hierachy has too many levels.
                //
            
            while ( tg != null ) {
                if ( tg instanceof Application ) {
                    return ((Application) tg);
                }
                else if ( tg.equals(kernelThrGroup) ) {
                    return null;
                }
                tg = tg.getParent();
            }

            return null;
        }
    
        /**
         * Forwards the createObject() request to the localhost object seed.
         *
         * @param cls the class to be instantiated
         * @param params arguments to the class constructor
         * @param a custom object activator
         * @return an <code>ObjectId</code> value
         * @exception Exception if an error occurs
         */
    private final ObjectId
    localCreateObject(Class cls, Object[] params, Activator a)
        throws Exception
        {
            final String MTD = "localCreateObject()";
            
            log(Sev.DEBUG, MTD, "Invoking newInstance for class={0}", cls);
            return getObjectSeed().newInstance(cls, params, a);
        }

    
        /**
         * Forwards the createObject() request to the specified remote host.
         *
         * @param cls the class to be instantiated
         * @param params arguments to the class constructor
         * @param a custom object activator
         * @param targetHost target remote host
         * @return an <code>ObjectId</code> value
         */
    private final ObjectId
    remoteCreateObject(Class cls, Object[] params, Activator a, HostId targetHost)
        {
            final String MTD = "remoteCreateObject()";
                
            log(Sev.DEBUG, MTD, "Forwarding request to host {0}", targetHost);
            return getRemoteExecutor(targetHost).createObject(cls, params, a, targetHost);
        }

    private final void
    remoteActivateObject(MarshaledOX mox, HostId targetHost)
        {
            final String MTD = "remoteActivateObject()";
            
            log(Sev.DEBUG, MTD, "Selected target host is {0}", targetHost);            
            getRemoteExecutor(targetHost).activateObject(mox);
        }

    private final void
    registerApplication(Application app, boolean master)
        {
            final String MTD = "registerApplication()";
            
            appByThrGrp.put(app, app);
            appById.put(app.getId(), app);

            if ( master ) {
                log(Sev.DEBUG, MTD, "updating CIB");
                
                CellInformationBase cib = getCib();

                    //
                    // TO DO: include cell id in app resource name
                    //
                ResourceName apprn = app.getId().toResourceName();
                cib.addResource(apprn);
                cib.setAttribute(apprn, CIB_ATT_APP_CODEBASE, app.getCodeBase());
                cib.setAttribute(apprn, CIB_ATT_APP_LDESC, app.getLaunchingDescriptor());
                cib.setAttribute(apprn, CIB_ATT_APP_STATUS, APP_STATUS_RUNNING);                
            }
            else {
                    //
                    // TO DO: register worker node
                    //
            }
        }
    
    private final Application
    startApplicationWorker( ApplicationId appID )
        throws Exception
        {
            final String MTD = "startApplicationWorker()";
            
                // fetch the application attributes from the CIB
            CellInformationBase cib = getCib();
                //
                // FIX ME: just toResourceName(), what if the application is not from the localcell?
                //
            ResourceName apprn = appID.toResourceName();
                
            String codeBase = cib.getAttribute(apprn, CIB_ATT_APP_CODEBASE);
            String ldesc = cib.getAttribute(apprn, CIB_ATT_APP_LDESC);
            
            Application app = new Application(appID, ldesc, true);
            registerApplication(app, false);
            
                // setup the new environment
            log(Sev.DEBUG, MTD,
                "Worker joined app {0}, codebase={1}", appID, app.getCodeBase());

            return app;
        }

    
        /**
         * Force all threads running in behalf of the given application to cease their execution.
         *
         * @param app an <code>Application</code> value
         */
    private final void
    leaveApplication(ApplicationId appId) throws Exception
        {
            final Application app = getApplicationForId(appId);

                // mark the application as finished in the cib to allow distributed
                // cleanup
                ResourceName apprn = app.getId().toResourceName();
                getCib().setAttribute(apprn, CIB_ATT_APP_STATUS, APP_STATUS_DONE);

                //
                // TO DO: remove application from internal tables and block new tryies to
                // joing the killed application.
                //

                while ( app.activeCount() > 0 ) {

                    app.interrupt();
                
                    try { Thread.sleep(1000); }
                    catch (Exception e) {};

                    app.stop();

                    try { Thread.sleep(1000); }
                    catch (Exception e) {};
                }

                if ( ! app.isDestroyed() ) {
                    app.destroy();
                }

                System.gc();
        }

    private final ApplicationId
    nextApplicationId()
        {
                // FIX ME: need stronger ID generation strategy
            return new ApplicationId(System.currentTimeMillis());
        }


    private final CellInformationBase
    getCib()
        {
            return (CellInformationBase)
                Exehda.getService(CellInformationBase.SERVICE_NAME);
        }

    private final ObjectSeed
    getObjectSeed()
        {
            return (ObjectSeed)
                Exehda.getService(ObjectSeed.SERVICE_NAME);
        }

    private final HostId
    getOXLocation( ObjectId oxID )
        {
                // if OX Manager available, ask it where the given OX is
            OXManager oxm = (OXManager) Exehda.getService(OXManager.SERVICE_NAME);
            if ( oxm != null ) {
                    //
                    // XXX: need to create a new handle every time? Well, leave any
                    // handle creation optimizations to the OX Manager itself
                    //
                OXHandle oxh = oxm.createHandle(oxID);

                return oxh.getLocation();
            }
                // otherwise, try to guess from the ObjectId
            else {
                return oxID.getHost();
            }
        }


    private int maxRetryies;
    private HostId thisHost;
    private SchedulingHeuristic defaultHeuristic;
    private final Hashtable appByThrGrp;
    private final Hashtable appById;
    private ThreadGroup kernelThrGroup;


}

