// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services;


/**
 * The interface Monitor is a facade (design pattern) for accessing underlying
 * monitoring sensors. The main concerns which lead us to follow this
 * organization were:
 *
 *     - The monitor would, eventually, be accessed from a remote node; whereas
 *     - raw sensor implementations should not ever be individually accessed.
 *
 * The adoption of the facade design pattern allows the monitor to enforce
 * sensor configuration policies which are consistent for all the monitoring
 * consumers.
 *
 *
 * In order to reduce memory consumption, the flyweight design pattern is used
 * for managing Sensor and SensorParameter objects.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2004-04-29 11:22:57 -0300 (Thu, 29 Apr 2004) $ $Revision: 192 $
 */
public interface Monitor
{   
        /**
         * Describes a sensor supported by this monitor.
         *
         * Operations on sensors are done through the monitor object in order to
         * enforce configuration policies, keeping the sensor operation
         * consistent with all the registered consumers.
         *
         */
    static final class Sensor
    {
            // TO DO: should the sensor keep a reference to the monitor?
        private String name;
        private Class type;

        public Sensor(String n, Class t)
            {
                name = n; type = t;
            }
        
        public String getName() { return name; }
        public Class getType() { return type; }
    }

    
        /**
         * Describes a sensor parameter.
         *
         * Operations on sensor parameters are done through the monitor object
         * in order to enforce configuration policies, keeping the sensor
         * operation consistent with all the registered consumers.
         *
         */
    static final class SensorParameter
    {
        private Sensor sensor;
        private String name;
        private Class  type;

        public SensorParameter(Sensor s, String n, Class t)
            {
                sensor = s; name=n; type=t;
            }
        
        public Sensor getSensor() { return sensor; }
        public String getName() { return name; }
        public Class getType() { return type; }
    }
    
        /**
         * Lists sensors made available by this monitor.
         *
         * TO DO: consider using Enumeration for return type. How does it
         * reflect in remote accesses?
         *
         * @return the currently available sensors.
         */
    Sensor[]
    getSensors();

    
        /**
         * Enable/disables the given sensor.
         *
         * @param s a Sensor
         * @param enable true for enabling the sensor, false otherwise.
         */
    void
    setSensorEnabled(Sensor s, boolean enable);
    
        /**
         * Returns the enable state of the given sensor.
         *
         * @param s a sensor
         * @return true if the sensor is enabled, false otherwise.
         */
    boolean
    isSensorEnabled(Sensor s);
    
        /**
         * Configures the given sensor parameter to have the value
         * <code>v</code>.
         *
         * This operation may be vetoed by the monitor if the sensor is already
         * configured to operate in a finner grain mode. In that case, the
         * effect of calling this method is void. No error condition is signed.
         *
         * @param p a <code>SensorParameter</code> value
         * @param v new value to be assigned to the parameter
         */
    void
    setSensorParameter(SensorParameter p, Object v);
    
        /**
         * Returns the current value of the given sensor parameter.
         *
         * @param p a sensor parameter
         * @return the current value of the given parameter
         */
    Object
    getSensorParameter(SensorParameter p);

    
        /**
         * Lists the parameters supported by the given sensor.
         *
         * TO DO: consider using enumeration for return type. How does it
         * reflect in remote accesses?
         *
         * @param s a sensor
         * @return an array describing the supported parameters
         */
    SensorParameter[]
    getSensorParameters(Sensor s);

        /**
         * Gets the current value of the given sensor.
         *
         * If the sensor is not currently enabled, null is returned meaning the
         * sensor state is unknown.
         *
         * The returned value may be safely cast to the type returned by the
         * <code>getType()</code> method of the given sensor.
         *
         * @param s an enabled sensor
         * @return an Object representing the current value of the sensor or
         *         null if the given sensor is not enabled.
         */
    Object
    probeSensor(Sensor s);
        
//     void installSensor() ou registerAppSensor ou ambos ??
    

}
