// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Foobar; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services;

import java.net.*;
import java.io.*;
import java.util.*;



/**
 * Base class for HTTP-based ISAM services
 *
 * @author lucc
 * @version $Date: 2003-12-11 17:37:34 -0200 (Thu, 11 Dec 2003) $ $Revision: 61 $
 */
public class HttpService
    implements org.isam.exehda.services.Service,java.lang.Runnable
{
    private final int BUFF_SIZE = 512;

    private int port;
    private Thread serverThread;
    private ServerSocket serverSocket;

    
    protected static class ServiceRequest
    {
        public final static int GET  = 0;
        public final static int POST = 1;
        public final static int HEAD = 2;

        private int requestType;
        private HashMap reqProps;
        private Socket  reqSocket;

        ServiceRequest(int type, Socket s)
            {
                this.requestType = type;
                this.reqProps = new HashMap();
                this.reqSocket = s;

                System.out.println("New request");
            }
    
        public String
        getProperty( String propName )
            {
                return (String) reqProps.get( propName );
            }

        public int
        getType()
            {
                return requestType;
            }

        public Socket
        getClientSocket()
            {
                return reqSocket;
            }

        public void
        setProperty( String propName, String propValue )
            {
                System.out.println("      prop:"+ propName +"="+propValue);
                reqProps.put(propName, propValue);
            }

        public String
        toString()
            {
                String str = null;

                switch (requestType) {
                    case GET:  str = "GET: "; break;
                    case POST: str = "POST: "; break;
                    default: str = "Unknown: "; break;
                }

                return str + reqProps;
            }
    
    }

    protected static class ServiceResponse
    {
        public final static int SUCCESS_MASK         = 200;
        public final static int CLIENT_ERROR_MASK    = 400;
        public final static int OK           = 200;
        public final static int CREATED      = 201;
        public final static int ACCEPTED     = 202;
        public final static int NO_CONTENT   = 203;
        public final static int BAD_REQUEST  = 400;
        public final static int UNAUTHORIZED = 401;
        public final static int FORBIDDEN    = 403;
        public final static int NOT_FOUND    = 404;
        private final static String CRLF     = "\r\n";

        private OutputStream outs;
        private HashMap      headers;
        private boolean      inHeader;

        public ServiceResponse(ServiceRequest req, int responseCode)
            throws IOException
            {
                outs = req.getClientSocket().getOutputStream();
                headers = null;
                inHeader = true;
                outs.write(("HTTP/1.0 "+responseCode + CRLF).getBytes());
            }

        public void
        putHeader(String attrName, String attrValue)
            {
                if ( inHeader ) {
                    if ( headers == null ){
                        headers = new HashMap();
                    }

                    headers.put(attrName, attrValue);
                }
            }

        private void
        flushHeaders()
            throws IOException
            {
                if ( inHeader ) {
                    if ( headers != null ) {
                        Iterator hi = headers.entrySet().iterator();

                        while ( hi.hasNext() ) {
                            Map.Entry entry = (Map.Entry) hi.next();
                            String key = (String) entry.getKey();
                            String val = (String) entry.getValue();

                            outs.write((key+": "+val+CRLF).getBytes());
                        }
                        
                    }
                    outs.write(CRLF.getBytes());
                    inHeader = false;
                }
            }

        public void
        putData(byte[] data, int offset, int size)
            throws IOException
            {
                flushHeaders();
                outs.write( data, offset, size );
            }

        public void
        close()
            throws IOException
            {
                outs.flush();
            }
    }
    
    public HttpService( int port )
        {
            this.port = port;
            this.serverThread = null;
            this.serverSocket = null;
        }
    
    private ServiceRequest
    getRequest( Socket clientSocket )
        throws java.io.IOException
        {
            InputStream in = clientSocket.getInputStream();
            
            byte[] buff = new byte[BUFF_SIZE];
            buff[0] = (byte) '\n';
            
            int pos = 0;
            int c = -1;
            int nprops = 0;
            ServiceRequest req = null;

            while ((c=in.read()) != -1) {
                switch (c) {
                    case '\r':
                        break;
                    case '\n':
                        String prop = new String(buff, 1, pos);
                        
                        System.out.println("++++"+prop);

                            // if first line
                        if ( nprops == 0 ) {
                            if (prop.startsWith("GET ")) {
                                req = new ServiceRequest( ServiceRequest.GET,
                                                          clientSocket );
                                
                                req.setProperty("Resource",
                                                prop.substring(4,
                                                               prop.lastIndexOf(" HTTP/")));
                            }
                            else if ( prop.startsWith("POST ") ) {
                                req = new ServiceRequest( ServiceRequest.POST,
                                                          clientSocket );
                                
                                req.setProperty("Resource",
                                                prop.substring(5,
                                                               prop.lastIndexOf(" HTTP/")));
                                
                            }
                            else {
                                System.out.println("Unknown or unsupported request: "+prop);
                            }
                        }
                        else {
                            if ( buff[pos] == c ) {
                                return req;
                            }
                            else {
                                int sepIdx = prop.indexOf(':');
                                if ( sepIdx < 0 ) System.out.println(prop);
                                req.setProperty(prop.substring(0, sepIdx),
                                                prop.substring(sepIdx+1).trim());
                            }
                        }
                        pos = 0;
                        nprops++;
                        break;
                    default:
                        buff[++pos] = (byte) c;
                        break;
                }
            }
            
            return null;
        }

    protected void
    handleRequest( ServiceRequest req )
        throws IOException
        {
            System.out.println(req);
        }
    
    public void
    run() {

        try {
            serverSocket = new ServerSocket(port);

            while (! Thread.interrupted() ) {

                Socket clientSocket = serverSocket.accept();
                
                handleRequest( getRequest(clientSocket) );

                clientSocket.close();
            }
        }
        catch (Exception e){
            e.printStackTrace();
        }
    }

    public void
    start()
        {
            if (serverThread == null) {
                serverThread = new Thread(this);
                serverThread.start();
          //       System.err.println("Service "+this.getClass()
//                                    +" started at port "+port);
            }
            else {
//                 System.err.println("Service already running");
            }
        }

    public void
    stop()
        {
            try {
                if (serverThread != null) {
                    serverThread.interrupt();
                    serverSocket.close();
                    serverThread = null;
                    System.err.println("Service shutdown completed");
                }
                else {
                    System.out.println("Error: service not running");
                }
            }
            catch (Exception e) {
                System.out.println(e);
            }
        }

    public static void
    main( String[] args )
        throws Exception
        {
            HttpService service = new HttpService(1977);

            service.start();

            Thread.currentThread().wait();
            
            service.stop();
        }
    
}

