// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Foobar; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.perv.desktop;


import org.isam.perv.IsamClassLoader;
import java.lang.reflect.Method;
import java.util.Properties;
import java.util.jar.Attributes;
import java.util.jar.Manifest;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.URL;


/**
 * Helper tool used to launch ISAM applications.
 *
 * @author lucc
 * @version $Id: Launcher.java 41 2003-11-25 20:45:33Z lucc $
 */
public class Launcher
{    
        /** 
         * Set to 'true' if debugging output should be generated.
         */
    static final boolean NDEBUG = false;
    
        /**
         * Exit code if a command-line argument is missing.
         */
    static final int ERR_ARG_MISSING = 65;
    
        /**
         * Exit code if an unexpected error occured.
         */
    static final int ERR_UNEXPECTED  = 66;
    
        /**
         * Program name displayed by the --help command-line option.
         */
    static final String PROGRAM_NAME = "isam-run" ;

    
        /**
         * Name of the properties file that holds user-specific ISAM
         * configurations.
         *
         */
    static final String USER_PROPS_FILE = ".isam_config";
    
    static {        
            // initialize the environment
        Properties sysprop = System.getProperties();
        
        sysprop.put("java.protocol.handler.pkgs", "org.isam.perv.protocols");
        sysprop.put("isam.bdaHost", "localhost");
        sysprop.put("isam.avuHost", "localhost");
        sysprop.put("isam.gatekeeper.host", "localhost");
        sysprop.put("isam.gatekeeper.port", "29901");

        loadUserConf();
	
	debug(sysprop.get("isam.bdaHost")+"----------------------------->");
    }

        /**
         * ISAM-enabled classloader used in running ISAM pervasive applications.
         *
         */
    private static IsamClassLoader isamClassLoader;
    
        /**
         * Code base URL extracted from the application launch descriptor.
         *
         */
    private static String          appCodeBaseURL;
    
        /**
         * Application's main class name extrated from the launch descriptor.
         *
         */
    private static String          appMainClassName;
    
        /**
         * Prints the provided debug information to the System.err if not in
         * quiet mode. If in quiet mode (NDEBUG set to true) it is a NO-OP call.
         *
         * @param msg the debug message
         */
    private static final void
    debug( Object msg )
        {
            if ( !NDEBUG ) {
                System.err.println("Launcher: "+msg);
            }
        }
    
        /**
         * Merge user-defined ISAM configuration, stored in the user's home
         * directory, into the System properties.
         *
         * @see Launcher.USER_PROPS_FILE
         */
    private static final void
    loadUserConf()
        {
            File userConfFile = new File(
                System.getProperty("user.home")
                + System.getProperty("file.separator")
                + USER_PROPS_FILE);
            
            if ( userConfFile.exists() ) {
                try {
                    System.getProperties().load(
                        new FileInputStream(userConfFile));
                }
                catch ( IOException ioe ) {
                    System.err.println(
                        "Warning: Loading of ~/.isam_config failed." + ioe);
                }
            }
            else {
                System.err.println(
                    "Warning: ~/.isam_config not found. Using default values.");
            }
        }
    
        /**
         * Sets up the ISAM classloader, pointing it to the provided
         * BDA application URL.
         *
         */
    private static final void
    setupClassLoader()
        {
            isamClassLoader = new IsamClassLoader(appCodeBaseURL);
        }
    
        /**
         * TODO: Sets up the ISAM Security Manager.
         *
         */
    private static final void
    setupSecurityManager()
        {
            debug("TO DO: install ISAM security manager");
        }
    
        /**
         * TODO: Sets up local instances/proxys of the ISAM core middleware
         * services.
         *
         */
    private static final void
    setupMiddlewareServices()
        {
            debug("TO DO: start middlware services");
        }

    
        /**
         * Loads and returns the application main class (entry point) based
         * either on the ISAM Launch Descriptor <main-class ... /> tag or the
         * Main-class JAR attribute.
         *
         * @return a <code>Class</code> value
         * @exception ClassNotFoundException if main class could not be loaded
         * @exception IOException if an error occurs
         */
    private static final Class
    loadMainClass()
        throws ClassNotFoundException, IOException
        {
                // if launch defines the app entry point 
            if ( appMainClassName != null ) {
                debug("Loading "+appMainClassName);
                return isamClassLoader.loadClass(appMainClassName);
            }
                // not defined, fall back to JAR Main-class attribute
            else {
                    // access app meta-info
                Manifest m = isamClassLoader.getAppManifest();
                if ( m == null ) {
                    throw new IOException("Manifest file not found");
                }
                    // search for the Main-class attribute
                Attributes attr = m.getMainAttributes();
                String mainClassName = ( attr == null ) ? null
                    : attr.getValue( Attributes.Name.MAIN_CLASS);
                if ( mainClassName != null ) {
                    return isamClassLoader.loadClass(mainClassName);
                }
            }
            throw new ClassNotFoundException("Main-class not defined");   
        }

    
        /**
         * Loads and parses the ISAM Launch descriptor pointed by <code>descUrl</code>,
         * extracting from it information about the application code base and main class
         * (entry point).
         *
         * @param descUrl launch descriptor URL
         * @exception IOException if an error occurs while accessing the
         * descriptor.
         */
    private static final void
    parseLaunchDescriptor( String descUrl )
        throws IOException
        {
                // extract app jar url from descriptor
            BufferedReader inr = new BufferedReader( 
                new InputStreamReader(
                    (new URL(descUrl)).openStream()));
            

            String line = null;
            while ( (line = inr.readLine()) != null ) {

                line = line.trim();

                if ( line.startsWith("<jar href=") ) {
                    appCodeBaseURL = line.substring(11,
                                               line.indexOf('"', 13));
                    if ( "".equals(appCodeBaseURL) ) {
                        System.err.println(
                            PROGRAM_NAME + ": Invalid application descriptor");
                        System.exit(ERR_ARG_MISSING);
                    }
                }
                else if ( line.startsWith("<main class=") ) {
                    appMainClassName = line.substring(13,
                                                      line.indexOf('"', 13));
                }
            }

            inr.close();
        }

        /**
         * Prints help message describing command-line parameters to System.err.
         *
         */
    private static final void
    printHelp()
        {
            System.err.println(
                "\n"
                + "Usage: "+PROGRAM_NAME+" <app.isam> [OPTIONS] [-- params... ]"
                + "\n\n"
                + "Where possible OPTIONS are:\n" 
                + "\n" 
                + "\t -h            \t Show this message.\n" 
                + "\t -i <inurl>    \t Set input URL\n" 
                + "\t -o <outurl>   \t Set output URL\n" 
                + "\t -e <errurl>   \t Set error output URL\n"
                + "\n");
        }    

    
        /**
         * Launches the ISAM application defined by the launch descriptor
         * pointed by <code>appDescUrl</code>, providing <code>params</code> as
         * startup arguments. The standard streams in, out and err are
         * redirected to inUrl, outUrl, and errUrl respectively.
         *
         * @param appDescUrl URL to a valid ISAM Application Launch Descriptor
         * @param inUrl System.in redirection
         * @param outUrl System.out redirection
         * @param errUrl System.err redirection
         * @param params Arguments to be passed to application's main method.
         * @exception Exception if an error occurs
         */
    public static final void
    launch( String appDescUrl, String inUrl, String outUrl,
            String errUrl, String[] params )
        throws Exception
        {
            parseLaunchDescriptor(appDescUrl);
          
            debug(System.getProperty("isam.bdaHost")+"+++++++++++++++++++++++++++++");

            setupClassLoader();
            setupSecurityManager();
            setupMiddlewareServices();

            Class mainClass = loadMainClass();
            Class[] mainArgTypes = new Class[] { String[].class };

            Method mainMethod = mainClass.getMethod("main", mainArgTypes);

            mainMethod.invoke(null, new Object[] { params });
        }
    
        /**
         * Launcher entry-point for stand-alone operation mode.
         *
         * @param args a <code>String[]</code> value
         * @exception Exception if an error occurs
         */
    public static void
    main( String[] args )
        throws Exception
        {
            String inUrl  = "file:/dev/null";
            String outUrl = "file:/dev/null";
            String errUrl = "file:/dev/null";
            String appDescUrl = null;
            
            String[] appArgs = new String[0];
            
            if ( args.length < 1 ) {
                System.err.println(PROGRAM_NAME + ": Argument missing.");
                System.err.println("Try `" + PROGRAM_NAME + " -h` for help.");
                System.exit(ERR_ARG_MISSING);
            }
            else {
                if ( "-h".equals(args[0]) ) {
                    printHelp();
                    System.exit(0);
                }
                else {
                    appDescUrl = args[0].startsWith("avu:")
                        ? args[0]
                        : "file:"+args[0];
                }
                
                for ( int i = 1; i < args.length; i++ ) {
                    if ( "-h".equals(args[i]) ) {
                        printHelp();
                        System.exit(0);
                    }
                    else if ( "-i".equals(args[i]) ) {
                        inUrl = args[++i];
                    }
                    else if ( "-o".equals(args[i]) ) {
                        outUrl = args[++i];
                    }
                    else if ( "-e".equals(args[i]) ) {
                        errUrl = args[++i];
                    }
                    else if ( "--".equals(args[i]) ) {
                        appArgs = new String[args.length - (++i)];
                        System.arraycopy(args, i, appArgs, 0, appArgs.length);
                        break;
                    }
                    else {
                        System.err.println(
                            PROGRAM_NAME + ": Unexpected value " + args[i]
                            + "\n"
                            + "Try `" + PROGRAM_NAME + " -h` for help.");
                        System.exit(ERR_UNEXPECTED);                          
                    }
                }

                launch( appDescUrl, inUrl, outUrl, errUrl, appArgs );
            }
        }
}
