// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda;


/**
 * Uniquely identifies an application in the scope os the distributed system.
 *
 * In order to guarantee unicity, a hierarchical approach is used to generate these
 * identifiers. The id itself is built up of two parts:
 *
 * <ul>
 *
 * <li> host id: a system-wide unique HostId which describes the node where the
 * application execution started.
 *
 * <li> local id: a host-wide, unique, 64 bits number. It's up to each host to control
 * which local ids are currently in use and avoid conflicts in local id allocation.
 *    
 * </ul>
 *
 *
 * Currently, local id is generated by taking a snapshot of the system current time,
 * through System.currentTimeMillis(). These technique should be replaced in future by a
 * strong-random number generation mechanism in order to improve system's security (by
 * making harder for one to forge a valid Id).
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2004-12-16 17:46:04 -0200 (Thu, 16 Dec 2004) $ $Revision: 496 $
 */
public final class ApplicationId
    extends HierarchicalId
{
        /**
         * Creates a new <code>ApplicationId</code> instance using the provided id as
         * application's local id. The base host (parent ID) is set to
         * HostId.getLocalHost().
         *
         * @todo Perhaps localId should by a byte[]
         *
         * @param localId a <code>long</code> value
         */
    public ApplicationId(long localId)
        {
            super(HostId.getLocalHost());
            this.localId = localId;
        }
    
        /**
         * Creates a new <code>ApplicationId</code> instance. Equivalent to using:
         *
         *    <p><code> new ApplicationId(System.currentTimeMillis());</code>
         *
         * <p>Such approach leads to deterministic ApplicationId generation, so it must
         * be used for debugging purposes only. Using the
         * <code>ApplicationId(long)</code> constructor combined with a strong random
         * number generator is highly encouraged.
         *
         * @deprecated Generating local id through System.currentTimeMillis() is not very
         * secure.
         */
    public ApplicationId()
        {
            this(System.currentTimeMillis());
        }

        /**
         * Returns the host where the application referred by these ApplicationId was
         * first started (launched).
         *
         * @return the HostId of the application's base host.
         */
    public HostId
    getBaseHost()
        {
            return (HostId) getParentId();
        }

    
        /**
         * Returns whether this ApplicationId is equals to the given object.
         *
         * Two ApplicationIds are said tobe equal if and only if their base hosts and
         * loca ids are equals.
         *
         * @param o an <code>Object</code> value
         * @return a <code>boolean</code> value
         */
    public boolean
    equals(Object o)
        {           
            if ( o != null && o instanceof ApplicationId ) {
                ApplicationId aid = (ApplicationId) o;
                
                return ( localId == aid.localId)
                    && getParentId().equals(aid.getParentId());
            }
            return false;
        }
    
        /**
         * Returns a hash code for these object suitable for the
         * <code>java.util.Hashtable()</code> semantics. This is currently a bitwise xor
         * between the low 32 bits of the local id and the hash code provided by the base
         * Host.
         * 
         *
         * @return a hash code.
         */
    public int
    hashCode()
        {
            return (int) localId ^ getParentId().hashCode();
        }

        /**
         * @inheritDoc
         */
    protected final String
    getType()
        {
            return "app";
        }
    
        /**
         * @inheritDoc
         */
    protected final Object
    getLocalId()
        {
            return Long.toHexString(localId);
        }
        

        /* A unique ID within that base host. */
    private long   localId; // perhaps this id should be a byte[]
}
