// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda;


/**
 * Abstract super-class of all system-wide unique IDs. In order to guarantee unicity such
 * IDs are build hierarchically.
 *
 * Hierarchical IDs are build up of two parts a parent ID as a local ID. Unicity is
 * ensured if, for a given parent ID, the chosen local ID is unique.
 *
 * @see org.isam.exehda.ApplicationId
 * @see org.isam.exehda.HostId
 * @see org.isam.exehda.CellId
 * @see org.isam.exehda.ObjectId
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2004-12-31 19:17:49 -0200 (Fri, 31 Dec 2004) $ $Revision: 550 $
 */
public abstract class HierarchicalId
    implements java.io.Serializable
{
    
        /**
         * Creates a new <code>HierachicalId</code> instance.
         *
         * @param parentId the parent of the ID been instantiated; null means this is
         * a root ID.
         */
    protected HierarchicalId( HierarchicalId parentId )
        {
            this.parentId = parentId;
        }

    
        /**
         * Returns the parent Id of this Id.
         *
         * @return the parent ID or null if this is a root ID.
         */
    public HierarchicalId
    getParentId()
        {
            return this.parentId;
        }
    
        /**
         * Returns whether this ID is a root ID. Root IDs are those ID's who do not have
         * a parent (i.e. parentId == null).
         *
         * @return true if this is a root ID, false otherwise.
         */
    public boolean
    isRoot()
        {
            return (parentId == null);
        }
    
        /**
         * Returns a <code>String</code> representation of this ID suitable to be used in debug
         * messages.
         *
         * @return this ID formated as a String
         */
    public String
    toString()
        {
            return "ID:{"+getType()+":"+getLocalId()+","+getParentId()+"}";
        }

    
        /**
         * Implements HierarchicalId equality semantics. Two HierarchicalIds are said to
         * be equal if the equality holds for their local IDs, types and parent IDs
         * respectively.
         *
         * <i>This is a reference implementation. Subclasses are highly encouraged to
         * overwrite this method with faster (optimized) implementations.</i>
         *
         * @param o an <code>Object</code> value
         * @return true if the given object is equals to this object, false otherwise.
         */
    public boolean
    equals(Object o)
        {
            if ( o != null && o instanceof HierarchicalId ) {
                HierarchicalId hid = (HierarchicalId) o;

                return getType().equals(hid.getType())
                    && getLocalId().equals(hid.getLocalId())
                    && ( isRoot() ? true : parentId.equals(hid.parentId) );
            }
            return false;
        }

    
        /**
         * Implements the hash code semantics for <code>HierarchicalId</code> objects so
         * that those object would be used as keys in a <code>java.util Hashtable</code>.
         *
         * <i>This is a reference implementation. Subclasses are highly encouraged to
         * overwrite this method with faster (optimized) implementations.</i>
         *
         * @return a hash code for this object
         */
    public int
    hashCode()
        {
            return getLocalId().hashCode() ^
                (isRoot() ? 0 : parentId.hashCode());
        }

    
        /**
         * Subclasses must overwrite this method and return a <code>String</code>
         * representation of the ID's type.
         *
         * @return a <code>String</code> value
         */
    protected abstract String
    getType();

    
        /**
         * Subclasses must overwrite this method and return an Object representation of
         * the local id.
         *
         * Notice that, subclasses may actually store such ID using primitive types, in
         * order to be more efficient, and just wrap such values in an Object when
         * <code>getLocalId()</code> is invoked.
         *
         * @return an <code>Object</code> value
         */
    protected abstract Object
    getLocalId();
    
    private HierarchicalId parentId;
}
