// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services;

import org.isam.exehda.ApplicationId;


/**
 * Context Recognition Service.
 *
 * The ContextManager takes as input a context definition in XML. It then proceeds
 * registering itself within the Collector service and accomplishing any necessary sensor
 * parameter configuration.
 *
 * The raw-level data gathered from the collector's sensors is then filtered in order to
 * produce high-level context information.
 *
 * The whole application execution context is built up from a a set of context elements
 * (<code>Context</code> objects) and its associated states (<code>ContextState</code>
 * objects).
 *
 * @todo describe the context definition language
 *
 * @author last modified by $Author: $
 * @version $Date: 2005-04-29 17:45:59 -0300 (Fri, 29 Apr 2005) $ $Revision: 835 $
 */
public interface ContextManager
{
    public static final String SERVICE_NAME = "ctxmanager";

        /**
         * Represents a context element.
         *
         */
    public static class Context
    {
            /**
             * Returns the name of this context element.
             *
             * @return a <code>String</code> value
             */
        public String
        getName()
            {
                return this.name;
            }
            
            /**
             * Returns the scope (application) where this context element was defined.
             *
             * The scope information is necessary to avoid conflicts and inconsistencies
             * when comparing context elements defined by different applications.
             *
             * @return an <code>ApplicationId</code> value
             */
        public ApplicationId
        getScope()
            {
                return this.scope;
            }
            
            /**
             * Returns a <code>String</code> containing the XML definition used to create
             * this context.
             *
             * @return a <code>String</code> value
             */
        public String
        getDefinition()
            {
                return this.definition;
            }
            
            /**
             * Returns a <code>ContextState</code> object that described the given
             * abstract state of this context element.
             *
             * @param state a <code>String</code> value
             * @return a <code>ContextState</code> value
             */
        public ContextState
        getStateInstance( String state )
            {
                for (int i=0; i<states.length; i++) {
                    if (states[i].getValue().equals(state)) return states[i];
                }
                throw new NullPointerException("No such context state");
            }

        
            /**
             * Returns whether this object and the given object describe the same context
             * element.
             *
             * @param o an <code>Object</code> value
             * @return a <code>boolean</code> value
             */
        public boolean
        equals( Object o )
            {
                if ( o != null && o instanceof Context ) {
                    Context ctx2 = (Context) o;

                        // XXX: should also compare states or comparing definitions is enough/safe?
                    return ctx2.name == this.name
                        && ctx2.definition == this.definition
                        && ((ctx2.scope != null)
                            ? ctx2.scope.equals(this.scope)
                            : ctx2.scope == this.scope);
                }
                return false;
            }

            /**
             * Computes a hash code for this context object.
             *
             * @return an <code>int</code> value
             */
        public int
        hashCode()
            {
                return this.name.hashCode() ^ this.definition.hashCode();
            }
        
            /**
             * Creates a new <code>Context</code> instance.
             *
             * @param name the context element's name
             * @param scope the application where is being defined
             * @param def the raw XML definition
             * @param states the allowed stated for this element
             */
        public
        Context( String name, ApplicationId scope, String def, String[] states )
            {
                    // for fast comparing, use the conical form for String attributes
                this.name = name.intern();
                this.definition = def.intern();
                this.scope = scope;
                this.states = new ContextState[states.length];

                    // TODO: is it worth this.states to be a hashtable for faster lookup?
                for ( int i=0; i<states.length; i++ ) {
                    this.states[i] = new ContextState(this, states[i]);
                }
            }

        private String name;
        private ApplicationId scope;
        private String definition;
        private ContextState[] states;
    }

    
        /**
         * Describes a context element's state.
         *
         */
    public static class ContextState
    {
            /**
             * Returns the context element to which this context state is bound.
             *
             * @return a <code>Context</code> value
             */
        public Context getType()
            {
                return this.type;
            }

            /**
             * Describe <code>getValue</code> method here.
             *
             * @return a <code>String</code> value
             */
        public String getValue()
            {
                return this.value;
            }
        
            /**
             * Returns whether this and the given object both describe the same context
             * state.
             *
             * @param o an <code>Object</code> value
             * @return a <code>boolean</code> value
             */
        public boolean
        equals( Object o )
            {
                if ( o != null && o instanceof ContextState ) {
                    ContextState cs = (ContextState) o;

                    return value == cs.value
                        && type.equals(cs.type);
                }
                return false;
            }

            /**
             * Computes the hash code for this context state object.
             *
             * @return an <code>int</code> value
             */
        public int
        hashCode()
            {
                return value.hashCode();
            }
        
            /**
             * Creates a new <code>ContextState</code> instance.
             *
             * @param type a <code>Context</code> value
             * @param value a <code>String</code> value
             */
        ContextState( Context type, String value )
            {
                if ( type == null ) throw new NullPointerException("Null context element not allowed.");
                if ( value == null ) throw new NullPointerException("Null context state value not allowed.");
                
                this.type = type;
                this.value = value.intern();
            }

        private Context type;
        private String  value;
    }

    
        /**
         * Call-back interface for receiving context element change events.
         *
         */
    public static interface ContextListener
    {
            /**
             * Invoked by the ContextManager when a context element's changed its current
             * state.
             *
             * @param prev the previous state, or null if no pervious state is available
             * @param curr the new state
             */
        public void contextChanged( ContextState prev, ContextState curr );
    }
    
        /**
         * Creates/registers a new context element.
         *
         * The context manager is responsible for allocating resources and enabling the
         * necessary sensor to produce that context information.
         *
         * Sample context definition:
         * <code><pre>
         *
           &lt;context n="context1"&gt;
             &lt;states&gt;
               &lt;state n="stt1"/&gt;
               &lt;state n="stt2"/&gt;
               &lt;state n="stt3"/&gt;
             &lt;/states&gt;

             &lt;index&gt;
               &lt;switch&gt;
                  &lt;sensor n="s1" scale="10.0"/&gt;
                  &lt;sensor n="s2" scale="3.3" /&gt;
                  &lt;composite type="sum"&gt;
                     &lt;sensor n="s1" scale="10.0"/&gt;
                     &lt;sensor n="s2" scale="3.3" /&gt;
                  &lt;/composite&gt;
               &lt;/switch&gt;
             &lt;/index&gt;

             &lt;ranges&gt;
               &lt;range ub="-1" state="stt1"/&gt;
               &lt;range lb="-1" ub="1" state="stt2"/&gt;
               &lt;range lb="1" state="stt3"/&gt;
               &lt;default state="stt1"/&gt;
             &lt;/ranges&gt;
           &lt;/context&gt;
           
         * </pre></code>
         *
         * @param xmlDesc a <code>String</code> value
         * @return a <code>Context</code> value
         */
    public Context createContext( String xmlDesc );
    
        /**
         * Notifies the ContextManager that the given context element is not needed
         * anymore.
         *
         * The ContextManager is allowed to release resources used to produce that
         * context information.
         *
         * @param ctx a <code>Context</code> value
         */
    public void releaseContext( Context ctx );
    
        /**
         * Registers a listener for events on the given context element.
         *
         * @param l the call back
         * @param ctx the context element of interest
         */
    public void addContextListener( ContextListener l, Context ctx );
    
        /**
         * Removes a previously registered context listener.
         *
         * @param l a <code>ContextListener</code> value
         * @param ctx a <code>Context</code> value
         */
    public void removeContextListener( ContextListener l, Context ctx );
}

