// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services;

import java.net.*;
import java.io.*;
import java.util.*;
import org.isam.exehda.Exehda;
import org.isam.exehda.services.Logger.Sev;


/**
 * Base class for HTTP-based ISAM services
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2004-06-16 12:14:16 -0300 (Wed, 16 Jun 2004) $ $Revision: 230 $
 */
public class HttpService
    extends AbstractService
    implements java.lang.Runnable
{
    private final int BUFF_SIZE = 512;
    private final int THR_POOL_SIZE = 3;

    
    protected class ServiceRequest
    {
        public final static int GET  = 0;
        public final static int POST = 1;
        public final static int HEAD = 2;

        private int requestType;
        private HashMap reqProps;
        private Socket  reqSocket;

        ServiceRequest(int type, Socket s)
            {
                this.requestType = type;
                this.reqProps = new HashMap();
                this.reqSocket = s;

                log(Sev.DEBUG, "ServiceRequest()", "New request");
            }
    
        public String
        getProperty( String propName )
            {
                return (String) reqProps.get( propName );
            }

        public int
        getType()
            {
                return requestType;
            }

        public Socket
        getClientSocket()
            {
                return reqSocket;
            }

        public void
        setProperty( String propName, String propValue )
            {
                log(Sev.TRACE, "ServiceRequest.setProperty()", "prop: {0}={1}", propName, propValue);
                reqProps.put(propName, propValue);
            }

        public String
        toString()
            {
                String str = null;

                switch (requestType) {
                    case GET:  str = "GET: "; break;
                    case POST: str = "POST: "; break;
                    default: str = "Unknown: "; break;
                }

                return str + reqProps;
            }
    
    }

    protected class ServiceResponse
    {
        public final static int SUCCESS_MASK         = 200;
        public final static int CLIENT_ERROR_MASK    = 400;
        public final static int SERVER_ERROR_MASK    = 500;
        public final static int OK              = 200;
        public final static int CREATED         = 201;
        public final static int ACCEPTED        = 202;
        public final static int NO_CONTENT      = 203;
        public final static int BAD_REQUEST     = 400;
        public final static int UNAUTHORIZED    = 401;
        public final static int FORBIDDEN       = 403;
        public final static int NOT_FOUND       = 404;
        public final static int NOT_IMPLEMENTED = 501;
        public final static int UNAVAILABLE     = 503;
        private final static String CRLF     = "\r\n";

        private ServiceRequest peerRequest;
        private OutputStream outs;
        private HashMap      headers;
        private boolean      inHeader;

        public ServiceResponse(ServiceRequest req, int responseCode)
            throws IOException
            {
                peerRequest = req;
                outs = req.getClientSocket().getOutputStream();
                headers = null;
                inHeader = true;
                outs.write(("HTTP/1.0 "+responseCode + CRLF).getBytes());
            }

        public void
        putHeader(String attrName, String attrValue)
            {
                if ( inHeader ) {
                    if ( headers == null ){
                        headers = new HashMap();
                    }

                    headers.put(attrName, attrValue);
                }
            }

        private void
        flushHeaders()
            throws IOException
            {
                if ( inHeader ) {
                    if ( headers != null ) {
                        Iterator hi = headers.entrySet().iterator();

                        while ( hi.hasNext() ) {
                            Map.Entry entry = (Map.Entry) hi.next();
                            String key = (String) entry.getKey();
                            String val = (String) entry.getValue();

                            outs.write((key+": "+val+CRLF).getBytes());
                        }
                        
                    }
                    outs.write(CRLF.getBytes());
                    inHeader = false;
                }
            }

        public void
        putData(byte[] data, int offset, int size)
            throws IOException
            {
                flushHeaders();
                outs.write( data, offset, size );
            }

        public void
        close()
            throws IOException
            {
                outs.flush();
                peerRequest.getClientSocket().close();
            }
    }

    private class ServiceThreadPool
    {
        private class ServiceThread
            extends Thread
        {
            public ServiceThread()
                {
                    start();
                }
    
            public synchronized void
            handleRequest( ServiceRequest req )
                {
                    log(Sev.TRACE, "ServiceThreadPool.ServiceThread..handleRequest()", "req={0}", req);
                    this.req = req;
                    notify();
                }

            private synchronized ServiceRequest
            getRequest() throws InterruptedException
                {
                    if ( req == null ) {
                        if ( enterPool() ) {
                            wait();
                            log(Sev.TRACE,
                                "ServiceThreadPool.ServiceThread.getRequest()",
                                "waken up from pool req={0}", req);
                        }
                        else {
                            return null;
                        }
                    }
                    ServiceRequest tmp = req;
                    req = null;
                    return tmp;
                }

            public void
            run()
                {
                    while (! Thread.interrupted() ) {
                        try {
                            ServiceRequest req = getRequest();
                            
                            if ( req == null ) return;

                            HttpService.this.handleRequest(req);
                        }
                        catch ( InterruptedException ie ) {
                            log(Sev.WARNING, "ServiceThread.run()", "interrupted");
                        }
                        catch ( Exception e ) {
                            log(Sev.ERROR, "ServiceThread.run()",
                                "Caught exception {0}", e);
                        }
                    }
                }

            private transient ServiceRequest req;
        }
    
        public ServiceThreadPool( int poolSize )
            {
                this.poolSize = poolSize;
                pool = new java.util.Vector(poolSize);
            }

        public void
        setSize(int sz)
            {
                this.poolSize = sz;
            }
        
        public int
        getSize()
            {
                return this.poolSize;
            }
    
        public void
        handleRequest( ServiceRequest req )
            {
                ServiceThread st = null;

                synchronized (pool) {
                    if ( pool.size() > 0 ) {
                        st = (ServiceThread) pool.lastElement();
                        pool.removeElementAt(pool.size()-1);
                    }
                    else {
                        log(Sev.DEBUG, "ServiceThreadPool.handleRequest()",
                            "increasing thread-pool size");
                            // TODO: perhaps we have an upper bound on the number
                            // of threads created
                        st = new ServiceThread();
                    }
                }
                log(Sev.TRACE, "ServiceThreadPool.handleRequest()", "req={0}", req);
                st.handleRequest(req);
            }

        private boolean enterPool()
            {
                synchronized (pool) {
                    if ( pool.size() < poolSize ) {
                        pool.addElement(Thread.currentThread());
                        return true;
                    }
                    else {
                        return false;
                    }
                }
            }

        private volatile int poolSize;
        private java.util.Vector pool;
    }
    
    public HttpService( int port )
        {
            this.port = port;
            this.serverThread = null;
            this.serverSocket = null;

            setThreadPoolSize(THR_POOL_SIZE);
        }

    protected HttpService()
        {
            this.port = -1;
            this.serverThread = null;
            this.serverSocket = null;

            setThreadPoolSize(THR_POOL_SIZE);
        }

    protected void
    setPort(int port)
        {
            this.port = port;
        }

    protected int
    getPort()
        {
            return port;
        }

    protected void
    setThreadPoolSize(int sz)
        {
            if ( sz > 1 ) {
                if ( svcThrPool != null ) {
                    svcThrPool.setSize(sz);
                }
                else {
                    svcThrPool = new ServiceThreadPool(sz);
                }
            }
            else {
                svcThrPool = null;
            }
        }

    protected int
    getThreadPoolSize()
        {
            if ( svcThrPool != null ) {
                return svcThrPool.getSize();
            }
            else {
                return 1;
            }
        }
    
    private ServiceRequest
    getRequest( Socket clientSocket )
        throws java.io.IOException
        {
            InputStream in = clientSocket.getInputStream();
            
            byte[] buff = new byte[BUFF_SIZE];
            buff[0] = (byte) '\n';
            
            int pos = 0;
            int c = -1;
            int nprops = 0;
            ServiceRequest req = null;

            while ((c=in.read()) != -1) {
                switch (c) {
                    case '\r':
                        break;
                    case '\n':
                        String prop = new String(buff, 1, pos);

                            // if first line
                        if ( nprops == 0 ) {
                            if (prop.startsWith("GET ")) {
                                req = new ServiceRequest( ServiceRequest.GET,
                                                          clientSocket );
                                
                                req.setProperty("Resource",
                                                prop.substring(4,
                                                               prop.lastIndexOf(" HTTP/")));
                            }
                            else if ( prop.startsWith("POST ") ) {
                                req = new ServiceRequest( ServiceRequest.POST,
                                                          clientSocket );
                                
                                req.setProperty("Resource",
                                                prop.substring(5,
                                                               prop.lastIndexOf(" HTTP/")));
                                
                            }
                            else {
                                log(Sev.WARNING, "getRequest()","Unknown request: {0}", prop);
                            }
                        }
                        else {
                            if ( buff[pos] == c ) {
                                return req;
                            }
                            else {
                                int sepIdx = prop.indexOf(':');
                                if ( sepIdx < 0 ) System.out.println(prop);
                                req.setProperty(prop.substring(0, sepIdx),
                                                prop.substring(sepIdx+1).trim());
                            }
                        }
                        pos = 0;
                        nprops++;
                        break;
                    default:
                        buff[++pos] = (byte) c;
                        break;
                }
            }
            return null;
        }

    protected void
    handleRequest( ServiceRequest req )
        throws IOException
        {
            log(Sev.DEBUG, "handleRequest()", "req={0}", req);
        }
    
    public void
    run() {

        try {
            serverSocket = new ServerSocket(port);

            while (! Thread.interrupted() ) {
                try {
                    Socket clientSocket = serverSocket.accept();
                    
                        // if multi-threading is enabled
                    if ( svcThrPool != null ) {
                        log(Sev.DEBUG, "run()", "multi-threading enabled for handleRequest");
                        svcThrPool.handleRequest( getRequest(clientSocket) );
                    }
                    else {
                        handleRequest( getRequest(clientSocket) );
                    }

                        // its responsability of the request handler to
                        // close the socket, otherwise we would inhibit
                        // multi-threading
//                    clientSocket.close();
                }
                catch ( Exception e ) {
                    if ( Thread.interrupted() ) break;
                    log(Sev.ERROR, "run()", "Caught exception {0}", e);
                }
            }
        }
        catch ( Exception e ) {
            log(Sev.ERROR, "run()", "Caught exception {0}", e);
        }
    }

    public void
    start()
        {
            if (serverThread == null) {
                serverThread = new Thread(this);
                serverThread.setDaemon(true);
                serverThread.start();
          //       System.err.println("Service "+this.getClass()
//                                    +" started at port "+port);
            }
            else {
//                 System.err.println("Service already running");
            }
        }

    public void
    stop()
        {
            try {
                if (serverThread != null) {
                    serverThread.interrupt();
                    serverSocket.close();
                    serverThread = null;
                }
                else {
                    log(Sev.ERROR, "stop()", "Shutdown attempted but service is not running");
                }
            }
            catch (Exception e) {
                System.out.println(e);
            }
        }
    
    public static void
    main( String[] args )
        throws Exception
        {
            HttpService service = new HttpService(1977);

            service.start();

            Thread.currentThread().wait();
            
            service.stop();
        }


    private int port;
    private Thread serverThread;
    private ServerSocket serverSocket;
    private ServiceThreadPool svcThrPool;
}

