// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services;

import org.isam.exehda.ObjectId;
import org.isam.exehda.HostId;
import org.isam.exehda.Exehda;
import java.util.Hashtable;
import java.util.Enumeration;



/**
 * Manager for OX's meta-attributes.
 *
 * @author Implemented by frainer, last modified by $Author: lucc $
 * @version $Date: 2005-01-21 10:40:29 -0200 (Fri, 21 Jan 2005) $ $Revision: 561 $
 */
public interface OXManager
{
    public static final String SERVICE_NAME = "oxmanager";
    
    public static final byte ATT_READ = 0;
    public static final byte ATT_WRITE = 1;
    
        /**
         * Abstract base type for all kinds of OX meta-attribute handles.
         *
         */
    public static abstract class OXHandle
        implements java.io.Serializable
    {
        protected final static String ATT_ID = "ox.id";
        protected final static String ATT_OWNER = "ox.owner";
            /**
             * Returns the system-wide ObjectId of the OX refered by this handle.
             *
             * @todo Should be equivalent to getAttribute(ATT_ID)
             * @return an <code>ObjectId</code> value
             */
        public ObjectId getId()
            {
                return this.oid;
            }
        
            /**
             * Shortcut for accessing the "ox.owner" attribute, describing the node that
             * curretly hosts the master OXHandle.
             *
             * @return a <code>HostId</code> value
             */
        public HostId getOwner()
            {
                return (HostId) getAttribute(ATT_OWNER);
            }
        
            /**
             * Retrieves the value of the given OX attribute.
             *
             * @param attr the attribute's name
             * @return the value of the attributed of null if its not defined.
             */
        abstract public Object getAttribute ( String attr );
        
        
            /**
             * Updates the value of the given OX attribute.
             *
             * @param attr the attribute name
             * @param val the new attribute's value
             */
        abstract public void setAttribute ( String attr, Object val );


            /**
             * Calls to OXManager.releaseHandle() and invalidate this handle by setting
             * the ObjectId to null.
             *
             */
        public void finalize()
            {
                    // if it is to be garbage collected it is no longer used in the application
                OXManager oxm = (OXManager)
                    Exehda.getService(OXManager.SERVICE_NAME);
                oxm.releaseHandle(this);
                    // invalidate the handle
                this.oid = null;
            }
        
            /**
             * Creates a new <code>OXHandle</code> instance.
             *
             * @param oid an <code>ObjectId</code> value
             */
        OXHandle( ObjectId oid )
            {
                this.oid = oid;
            }

        private ObjectId oid;
//         private HostId owner;
    }

    
        /**
         * Home OXHandle; it is hosted in the base of the OX's home cell and knows which
         * node currently hosts the master OX Handle.
         *
         */
    public final class HomeOXHandle
        extends OXHandle
    {
            /**
             * Describe <code>getOwner</code> method here.
             *
             * @return a <code>HostId</code> value
             */
        public HostId getOwner()
            {
                throw new NullPointerException("Not yet implemented");
            }
        
            /**
             * Describe <code>getAttribute</code> method here.
             *
             * @param attr a <code>String</code> value
             * @return an <code>Object</code> value
             */
        public Object getAttribute( String attr )
            {
                throw new NullPointerException("Not yet implemented");
            }
            
            /**
             * Describe <code>setAttribute</code> method here.
             *
             * @param attr a <code>String</code> value
             * @param val an <code>Object</code> value
             */
        public void setAttribute( String attr, Object val )
            {
                throw new NullPointerException("Not yet implemented");
            }

        public void finalize()
            {
                super.finalize();
            }
        
            /**
             * Creates a new <code>HomeOXHandle</code> instance.
             *
             * @param ox an <code>ObjectId</code> value
             */
        HomeOXHandle ( ObjectId ox, HostId owner )
            {
                super(ox);
                this.owner = owner;
            }

        private HostId owner;
    }


        /**
         * Master Handle , i.e. the real storage for OX meta-attributes.
         *
         */
    public final class MasterOXHandle
        extends OXHandle
    {
            /**
             * @inheritDoc
             *
             * <p>Always returns the ID of the local host.
             */
        public HostId getOwner (  )
            {
                return HostId.getLocalHost();
            }
        
            /**
             * @inheritDoc
             */
        public Object getAttribute ( String attrName )
            {
                return attrs.get(attrName);
            }
        
            
            /**
             * @inheritDoc
             */
        public void setAttribute ( String attr, Object val )
            {
                attrs.put(attr, val);
            }

        public void finalize()
            {
                super.finalize();
            }
            
            /**
             * Creates a new <code>MasterOXHandle</code> instance.
             *
             * @param ox an <code>ObjectId</code> value
             */
        MasterOXHandle ( ObjectId ox )
            {
                super(ox);
                this.attrs = new Hashtable();
            }

        private Hashtable attrs;
    }

        /**
         * Proxy for remotely accessing the master handle.
         *
         * <p>In order to reduce latency, a caching scheme is used to serve closed read
         * operations. Attributes are read in bulks from the OX Manager in order to
         * reduce the number of messages.
         *
         * <p>In order to reduce cache-coherence related issues, write operation are
         * always propagated instantaneously to the master OX handle.
         *
         * @todo A possible optimitization would be to use per-entry modification time
         * stamps
         */
    public final class ProxyOXHandle
        extends OXHandle
    {
        final long DFLT_UPDATE_THREASHOLD = 1000L;
        
            /**
             * Creates a new <code>ProxyOXHandle</code> instance.
             *
             * @param oid an <code>ObjectId</code> value
             */
        ProxyOXHandle( ObjectId oid )
            {
                super(oid);
                this.attrCache = new Hashtable();
                this.cacheUpdateThreashold = DFLT_UPDATE_THREASHOLD;
                this.cacheLastUpdate = 0;
                this.updateInCourse = false;
            }

        public Object getAttribute(String attr)
            {
                long now = System.currentTimeMillis();

                    // if attr does not exist in cache or is out-dated
                if ( (now - cacheLastUpdate > cacheUpdateThreashold)
                     || !attrCache.containsKey(attr)) {
                        // update cache prior returning
                    attrCache.put(attr, "");
                    fetchAttributes();
                }
                    // serve from cache
                return attrCache.get(attr);
            }

        public synchronized void setAttribute(String attr, Object val)
            {
                    // update cached entry
                attrCache.put(attr, val);
                    // update cache timestamp
                cacheLastUpdate = System.currentTimeMillis();
                    // avoid infinite recursion since it could be a call back update from the OXManager
                if (!updateInCourse) {
                        // propagate update to the Master OX Handle
                    OXManager oxm = (OXManager)
                        Exehda.getService(OXManager.SERVICE_NAME);

                    updateInCourse = true;
                    oxm.updateAttributes(this,
                                         new String[] { attr },
                                         new byte[] { OXAttributeOps.WRITE });
                    updateInCourse = false;
                }
                    
            }

        public void finalize()
            {
                super.finalize();
            }            

            ////////////////////////////////////////////////////////////////////////
            // PRIVATE STUFF
            ////////////////////////////////////////////////////////////////////////

            /**
             * Request to the OX Manager up-to-date values from the attributes cached in
             * this proxy. As an optimitation, attributes never requested from this proxy
             * are not included in the update.
             *
             */
        private final void fetchAttributes()
            {
                    // build request
                String [] attrs = new String[attrCache.size()];
                byte[] ops = new byte[attrCache.size()];
                    // for each cached attribute
                Enumeration ea = attrCache.keys();
                for ( int i=0; ea.hasMoreElements(); i++) {
                        // include that attribute in the update request
                    attrs[i] = (String) ea.nextElement();
                    ops[i] = OXAttributeOps.READ;
                }
                    // get a new snapshot of the attributes from the OXManager
                OXManager oxm = (OXManager)
                    Exehda.getService(OXManager.SERVICE_NAME);

                updateInCourse = true;
                copyAttributes(oxm.updateAttributes(this, attrs, ops));
                updateInCourse = false;
            }
                

            /**
             * Copies attributes from the given OX handle into this handle. Tryies to be
             * smart and avoid attribute-by-attribute copying if possible.
             *
             * @param h an <code>OXHandle</code> value
             */
        private final void copyAttributes(OXHandle h)
            {
                    // avoid copying if this and h refer to the same object
                if ( h != this ) {
                        // fast copy if the given handle is of the same class
                    if ( h instanceof ProxyOXHandle ) {
                        attrCache = (Hashtable) ((ProxyOXHandle) h).attrCache.clone();
                    }
                    else {
                            // otherwise, slower but safe copying is done
                        Enumeration ea = attrCache.keys();
                        for ( int i=0; ea.hasMoreElements(); i++) {
                            String k = (String) ea.nextElement();
                            attrCache.put(k, h.getAttribute(k));
                        }
                    }
                        // after copying, update last-modification time stamp
                    cacheLastUpdate = System.currentTimeMillis();
                }
            }
        
        private Hashtable attrCache;
        private long cacheUpdateThreashold;
        private long cacheLastUpdate;
        private boolean updateInCourse;
    }

    
//     public static class RmiOXHandle
//     {
//             /**
//              * Describe <code>getStub</code> method here.
//              *
//              * @param ox an <code>OXHandle</code> value
//              * @return a <code>java.rmi.Remote</code> value
//              */
//         public static java.rmi.Remote getStub ( OXHandle ox )
//             {
//                 throw new NullPointerException("Not yet implemented");
//             }
  
//             /**
//              * Describe <code>setStub</code> method here.
//              *
//              * @param ox an <code>OXHandle</code> value
//              * @param stub a <code>java.rmi.Remote</code> value
//              */
//         public static void setStub ( OXHandle ox, java.rmi.Remote stub )
//             {
//                 throw new NullPointerException("Not yet implemented");
//             }
//     }

        /**
         * Defines operations over the OX attributed.
         *
         * @todo SET_RO and CLR_RO are not currently supported.
         *
         * @see OXManager.updateAttributes(OXHandle,String[],byte[]);
         */
    public static interface OXAttributeOps
    {
        public static final byte READ = 0;
        public static final byte WRITE = 1;
        public static final byte SET_RO = 2;
        public static final byte CLR_RO = 4;
    }


        /**
         * Thrown to indicate that a OXManager method has been passed an illegal or
         * inappropriate OXHandle.
         *
         */
    public static class IllegalOXHandleException
        extends java.lang.IllegalArgumentException
    {
            /**
             * Creates a new <code>IllegalOXHandleException</code> instance with the
             * specified detail message.
             *
             * @param msg message detailing the exception cause.
             */
        public IllegalOXHandleException(String msg)
            {
                super(msg);
            }
    }
    
        /**
         * Creates a handle for accessing the meta-attributes of the OX identified by the
         * given <code>ObjectId</code>.
         *
         * <p>Whether a new MasterOXHandle, HomeOXHandle or ProxyOXHandle will the
         * created depends on several aspects of the current context.
         *
         * <p>MasteOXHandles and HomeOXHandles are created only once for each existing OX
         * in teh system. On the other hand, there would be several ProxyOXHandles
         * refering to the same OX.
         *
         * <p>Tipically, a MasterOXHandle will be created iif the all following
         * constraints hold:
         *
         * <ol>
         * <li>The namespace where the given ObjectId is defined is that of the local
         *     host (i.e. the local host is the home node for that OX);
         *
         * <li>No other master handle was previously created for that OX;
         * </ol>
         *
         * <p>Besides, a HomeOXHandle is created if:
         *
         * <ol><li>this is the cellular instance of the OX Manager and the local cell is
         * the home cell (i.e. the cells namespace contains the given object id);
         *
         * <li>>No other home handle was previously created for that OX;
         * </ol>
         *
         * Otherwise, a proxy handle will be created. No check is done if the OX still
         * exists by the time the proxy handle is created. OX dead will be detected by
         * the time the first attribute is accessed.
         *
         * @param oid the id on an OX
         * @return an <code>OXHandle</code> for accessing meta-attributes of the given OX.
         */
    public OXHandle createHandle ( ObjectId oid );
    
        /**
         * Notifies the OX Manager the given handle is no longer used. The semanthics of
         * releazing the master handle is that of ceasing the OX existency, meaning
         * trying to access that OX's attributes will result in an exception being
         * throwed.
         *
         * Whether the OX Manager will allow applications to directly release master
         * handles or will use indirect, node-wide, reference counting mechanisms is
         * implementation specific.
         *
         * Typically, the releazing of proxy handles is to be some how back propagated to
         * the node that actually, allowing the system to better position the master
         * handle in order to optimize attribute access.
         *
         * @param ox an <code>OXHandle</code> value
         */
    public void releaseHandle ( OXHandle ox ) throws IllegalOXHandleException;
    
        /**
         * Enables the master handle to the re-located in a subsequent call to
         * <code>requestOwnership()</code>. This call is to be back-propagated untill the
         * home cell where the home handle unbound flag is set. To be used when an
         * OXhandle migrates from one node to another.
         *
         * <p>While the ownership of the master handle is un-set, that handle is mean to be
         * in transit between nodes (i.e. its location is undefined). Therefore, any kind
         * of attribute access eventually back-propagated to the home cell will be queued
         * and delayed until the new master handle location is determined, what will
         * happen by a subsequent call to request ownership. Those queued requests will
         * be then forwarded to the new master handle location.
         *
         * @param ox an <code>OXHandle</code> value, eventually the master handle
         * @return an <code>OXHandle</code> suitable to be used in a subsequent call to
         * <code>requestOwnership()</code>.
         */
    public OXHandle grantOwnership ( OXHandle ox )
        throws IllegalOXHandleException;
    
        /**
         * Requests the master handle to be moved towards this node. The master handle
         * will, eventually, be re-located if a grantOwnership() was previously done for
         * it.  This call is meant to be back propagated until the home cell, so that the
         * home handle updated and made aware of the new master handle location and the
         * "ox.owner" meta-attribute is updated.
         *
         * @param ox an <code>OXHandle</code> value
         * @return an <code>OXHandle</code> value
         */
    public OXHandle requestOwnership ( OXHandle ox )
        throws IllegalOXHandleException;
    
        /**
         * Requests the given list of attributes of the given OXHandle (tipically a
         * proxy) to be updated/propagated to the masted handle.
         *
         * <p>Whether the attributes are read from or written to the master handle depends
         * on the per-attribute defined operation.
         *
         * <p>This update scheme is mean to reduce update latency by allowing several
         * operations to be collapsed in a single message. Reads and writes are allowed
         * to be mixed in a single update request.
         *
         * @param ox the handle that must be updated
         * @param att the subset of the attributes that are meant to be updated
         * @param op (per-attributed) defines whether it is a write ou read update
         * @return an pottencially new handle with the attributes updated as requested.
         */
    public OXHandle updateAttributes ( OXHandle ox, String[] att, byte[] op )
        throws IllegalOXHandleException;
}

