// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.perv.desktop;


import org.isam.perv.IsamClassLoader;
import org.isam.exehda.Exehda;
import org.isam.exehda.services.Executor;
import org.isam.util.xml.*;
import java.lang.reflect.Method;
import java.util.Properties;
import java.util.jar.Attributes;
import java.util.jar.Manifest;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.URL;
import java.net.URLConnection;


/**
 * Helper tool used to launch ISAM applications.
 *
 * @author lucc
 * @version $Id: Launcher.java 215 2004-06-14 21:01:01Z lucc $
 */
public class Launcher
{    
        /** 
         * Set to 'true' if debugging output should be generated.
         */
    static final boolean NDEBUG = false;
    
        /**
         * Exit code if a command-line argument is missing.
         */
    static final int ERR_ARG_MISSING = 65;
    
        /**
         * Exit code if an unexpected error occured.
         */
    static final int ERR_UNEXPECTED  = 66;
    
        /**
         * Program name displayed by the --help command-line option.
         */
    static final String PROGRAM_NAME = "isam-run" ;

        /**
         * Name of the properties file that holds user-specific ISAM
         * configurations.
         */
    static final String USER_PROPS_FILE = ".isam_config";

    
    static {        
            // initialize the environment
        Properties sysprop = System.getProperties();
        
        sysprop.put("java.protocol.handler.pkgs", "org.isam.perv.protocols");
//         sysprop.put("exehda.profile", "default");
        sysprop.put("isam.gatekeeper.host", "localhost");
        sysprop.put("isam.gatekeeper.port", "29901");

        loadUserConf();
    }
    
        /**
         * Merge user-defined ISAM configuration, stored in the user's home
         * directory, into the System properties.
         *
         * @see Launcher.USER_PROPS_FILE
         */
    private static final void
    loadUserConf()
        {
            File userConfFile = new File(
                System.getProperty("user.home")
                + System.getProperty("file.separator")
                + USER_PROPS_FILE);
            
            if ( userConfFile.exists() ) {
                try {
                    System.getProperties().load(
                        new FileInputStream(userConfFile));
                }
                catch ( IOException ioe ) {
                    System.err.println(
                        "Warning: Loading of ~/.isam_config failed." + ioe);
                }
            }
            else {
                System.err.println(
                    "Warning: ~/.isam_config not found. Using default values.");
            }
        }

        /**
         * Prints help message describing command-line parameters to System.err.
         *
         */
    private static final void
    printHelp()
        {
            System.err.println(
                "\n"
                + "Usage: "+PROGRAM_NAME+" <app.isam> [OPTIONS] [-- params... ]"
                + "\n\n"
                + "Where possible OPTIONS are:\n" 
                + "\n" 
                + "\t -h            \t Show this message.\n" 
                + "\t -i <inurl>    \t Set input URL\n" 
                + "\t -o <outurl>   \t Set output URL\n" 
                + "\t -e <errurl>   \t Set error output URL\n"
                + "\n");
        }    

    
        /**
         * Launches the ISAM application defined by the launch descriptor
         * pointed by <code>appDescUrl</code>, providing <code>params</code> as
         * startup arguments. The standard streams in, out and err are
         * redirected to inUrl, outUrl, and errUrl respectively.
         *
         * @param appDescUrl URL to a valid ISAM Application Launch Descriptor
         * @param inUrl System.in redirection
         * @param outUrl System.out redirection
         * @param errUrl System.err redirection
         * @param params Arguments to be passed to application's main method.
         * @exception Exception if an error occurs
         */
    public static final void
    launch( String appDescUrl, String inUrl, String outUrl,
            String errUrl, String[] params )
        throws Exception
        {
            XmlElement launchDesc = XmlElement.parse(
                new InputStreamReader(
                    (new URL(appDescUrl)).openStream()));

                // add application args to launch descriptor
            if ( params.length > 0 ) {
                XmlElement parameters = XmlElement.createElement("parameters");
                for ( int i=0; i<params.length; i++ ) {
                    XmlElement p = XmlElement.createElement("param");
                    p.addChild(XmlElement.createPCData( params[i]));
                    parameters.addChild(p);
                }
		//                launchDesc.addChild(parameters);
                ((XmlElement)launchDesc.getChildren().nextElement()).addChild(parameters);
            }
            
                // contact gatekeeper@localhost
            URLConnection gkconn =
                (new URL("http",
                         System.getProperty("isam.gatekeeper.host"),
                         Integer.getInteger("isam.gatekeeper.port").intValue(),
                         "/executor/startApplication")).openConnection();

            gkconn.setDoOutput(true);

            OutputStreamWriter out = new OutputStreamWriter(
                gkconn.getOutputStream());

            out.write(launchDesc.toString());
            out.close();

            gkconn.getInputStream().read();
        }
    
        /**
         * Launcher entry-point for stand-alone operation mode.
         *
         * @param args a <code>String[]</code> value
         * @exception Exception if an error occurs
         */
    public static void
    main( String[] args )
        throws Exception
        {
            String inUrl  = "file:/dev/null";
            String outUrl = "file:/dev/null";
            String errUrl = "file:/dev/null";
            String appDescUrl = null;
            
            String[] appArgs = new String[0];
            
            if ( args.length < 1 ) {
                System.err.println(PROGRAM_NAME + ": Argument missing.");
                System.err.println("Try `" + PROGRAM_NAME + " -h` for help.");
                System.exit(ERR_ARG_MISSING);
            }
            else {
                if ( "-h".equals(args[0]) ) {
                    printHelp();
                    System.exit(0);
                }
                else {
                    appDescUrl = args[0].startsWith("avu:")
                        ? args[0]
                        : "file:"+args[0];
                }
                
                for ( int i = 1; i < args.length; i++ ) {
                    if ( "-h".equals(args[i]) ) {
                        printHelp();
                        System.exit(0);
                    }
                    else if ( "-i".equals(args[i]) ) {
                        inUrl = args[++i];
                    }
                    else if ( "-o".equals(args[i]) ) {
                        outUrl = args[++i];
                    }
                    else if ( "-e".equals(args[i]) ) {
                        errUrl = args[++i];
                    }
                    else if ( "--".equals(args[i]) ) {
                        appArgs = new String[args.length - (++i)];
                        System.arraycopy(args, i, appArgs, 0, appArgs.length);
                        break;
                    }
                    else {
                        System.err.println(
                            PROGRAM_NAME + ": Unexpected value " + args[i]
                            + "\n"
                            + "Try `" + PROGRAM_NAME + " -h` for help.");
                        System.exit(ERR_UNEXPECTED);                          
                    }
                }

                launch( appDescUrl, inUrl, outUrl, errUrl, appArgs );
            }
        }
}
