// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda;


import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.NetworkInterface;


/**
 * Uniquely identifies a host in the scope of the distributed system.
 *
 * In order to guarantee unicity, a hierarchical approach is used to generate the
 * identifiers. The id itself is built up of two parts:
 *
 * <ul>
 *
 * <li> cell id: a system-wide unique cell id, in the form of a string. The idea is the
 * each cell in the system is assigned a unique name by th cell administrator. In order
 * to guarantee unicity of such name, it would be derived from the internet domain name
 * assigned to the institution that provides a cell. For example, cells in institution
 * abcd.org would be named cell1.abcd.org, cell2.abcd.org etc. Of course, cell1,
 * cell2..., would be changed to something more meaningful and descriptive in the scope
 * of that institution.
 *
 * <li> local id: a cell-wide, unique, 32 bits number. It's up to each cell to fix a
 * suitable policy for assigning local ids to hosts. A possible policy would be to derive
 * the local id from the host's IP address. Since the local id is suppose to be unique
 * only within the cell, the host IP need not to be a system-wide unique IP and so,
 * 10.x.x.x and 192.168.x.x would be used more or less safely without causing conflicts.
 *
 * </ul>
 *
 * <b>Note:</b><i>Currently, methods getHostByName() assumes local id are generated from
 * the host's ip address and that such address is unique (i.e. 10.x and 192.168.x IPs are
 * not used). Such scheme will be changed in future releases to reflect the behavior
 * described above.</i>
 *
 * @see CellId
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-04-25 01:57:38 -0300 (Mon, 25 Apr 2005) $ $Revision: 795 $
 */
public final class HostId
    extends HierarchicalId
{
        /**
         * Look ups a host with the given <code>hostName</code> in the given cell and, if
         * such host exists, returns it's system-wide HostId.
         *
         * <b>Note:</b><i>The current scheme assumes the local id is generated from the
         * host's IP address and only works if the host name is resolvable throught an DNS
         * query. Such scheme will be improved in future to consider data stored in the
         * Cell Information Base as long as other local id association policies.</i>
         *
         * @param hostName a public name of the host (e.g. its internet name) or an raw
         * id in hexadecimal notation prefixed by '0x'.
         * @param cellName the name of the cell in which the search is to be
         * done. <code>null</code> would be used to refer to the local cell. <i> This
         * parameter is currently not considered in lookup process.</i>
         *
         * @return the system-wide host identifier of that host.
         *
         * @exception UnknownHostException if the given hostName do not exists in the
         * given cell.
         *
         */
    public static HostId
    getByName( String hostName, String cellName ) throws UnknownHostException
        {
            if ( hostName == null || "localhost".equals(hostName)) {
                return getLocalHost();
            }
            else {
                if ( cellName == null ) cellName = detectLocalCell();

                    // raw id format
                if ( hostName.startsWith("0x") ) {
                    
                    int localId = (int) Long.parseLong(hostName.substring(2), 16);

                    return new HostId(cellName, localId, null, null); 
                }
                else {
                            //
                            // TODO: lookup in the Cell Information Base
                            //
                        InetAddress addr = InetAddress.getByName(hostName); 
                        return new HostId( cellName,
                                           ipv4ToLocalId(addr.getAddress()),
                                           hostName,
                                           addr);
                }
            }
        }
    
        /**
         * Look ups a host with the given <code>hostName</code> in the local cell and, if
         * such host exists, returns it's system-wide HostId.
         *
         * <b>Note:</b><i>The current scheme assumes the local id is generated from the
         * host's IP address and only works if the host name is resolvable trough an DNS
         * query. Such scheme will be improved in future to consider data stored in the
         * Cell Information Base as long as other local id association policies.</i>
         *
         * @param hostName a public name of the host (e.g. its internet name) or an raw
         * id in hexadecimal notation prefixed by '0x'.
         *
         * @return the system-wide host identifier of that host.
         *
         * @exception UnknownHostException if the given hostName do not exists in the
         * given cell.
         *
         */
    public static HostId
    getByName( String hostName ) throws UnknownHostException
        {
            return getByName( hostName, detectLocalCell() );
        }

        /**
         * Creates a new <code>HostId</code> instance by parsing the given String
         * representation.
         *
         * <p>The string representation is expected to be in the form
         * <i>hexvalue</i>.<i>cellname</i>. The first field (up to the first dot)
         * consists of a local host id represented in hexadecimal notation. The second
         * field consists of textual (ASCII) representation of a cell name.
         *
         * <p>Except for the local host, the fields hostName and inetAddr are left empty.
         *
         * @param s a <code>String</code> representation of a HostId
         */
    public static HostId
    parseId( String s )
        {
            if ( s.startsWith("hostid:") ) {
                s = s.substring(7);
            
                int idx = s.indexOf('.');

                HostId id = new HostId(s.substring(idx+1),
                                       Integer.parseInt(s.substring(0, idx), 16),
                                       null,
                                       null);

                return id.equals(getLocalHost()) ? getLocalHost() : id;
            }
            else {
                throw new IllegalArgumentException("Malformed host ID:"+s);
            }
        }
    
        /**
         * Returns the system-wide HostId that describes the local host.
         *
         * @return a <code>HostId</code> value
         */
    public static HostId
    getLocalHost()
        {
            return (localHost == null) ? detectLocalHost() : localHost;
        }
    
        /**
         * Returns the symbolic name of the host described by this HostId, if such name
         * exists.
         *
         * Currently, this name is the Internet name of the host as would be used in DNS
         * queries.
         *
         * @return Returns the symbolic name of the host if its available or
         * <code>null</ocde> otherwise.
         */
    public String
    getName()
        {
              return hostName;
        }
    
        /**
         * Returns the cell-wide id of this host.
         *
         * @return a 32 bits integer that uniquely identifies this host in the scope of
         * it's containing cell.
         */
    public int
    getId()
        {
            return id;
        }

        /**
         * Returns the ID of the cell to which the host refered by this HostId belongs.
         *
         * @return the cell name.
         */
    public CellId
    getCell()
        {
            return (CellId) getParentId();
        }
    
        /**
         * Returns the name of the cell to which the host refered by this HostId belongs.
         *
         * Shortcut to <code>getCell().getName()</code>.
         *
         * @return the cell name.
         */
    public String
    getCellName()
        {
            return ((CellId) getParentId()).getName();
        }

    
        /**
         * Returns the Internet address of this host.
         *
         * This method will only return a non-null value if the host has a real IP
         * (i.e. a IP not in 10.x, 127.x, or 192.168.x ranges).
         *
         * <b>Note</b>: This method is to be deprecated in future releases of the
         * middleware, since its not guaranteed that it will be able to return an address
         * suitable to contact every host in the system. An alternative mechanism is
         * under development which will be probably based on host attributes stored in
         * the Cell Information Base.
         *
         * @return an <code>InetAddress</code> value, or null if a real IP is not
         * available for this host.
         */
    public InetAddress
    getInetAddress()
        {
            return inetAddr;
        }


        /**
         * Constructs a string representation of this HostId suitable to be used with the
         * method parseId(String).
         *
         * @return a <code>String</code> value
         * @see #parseId(String)
         */
    public String
    toExternalForm()
        {
                // returns AAAXXXDD.cellname
            return "hostid:"+getLocalId().toString()+'.'+getCellName();
        }

    
        /**
         * Returns whether this object is equals to the given object.
         *
         * <p>The two objects are said to be equal iif: both as of class HostId, their cell
         * scope and local id are equals too.
         *
         *
         * @param obj an <code>Object</code> value
         *
         * @return true if this object describes the same host as the given object, false
         * otherwise.
         */
    public boolean
    equals( Object obj ) 
        {
            if ( obj != null && obj instanceof HostId ) {
                HostId hid = (HostId) obj;
                return hid.id == id && hid.getCellName().equals(getCellName());
            }
            else {
                return false;
            }
        }
    
        /**
         * Returns a hash code suitable for the <code>java.util.Hashtable</code> semantics.
         *
         * @return the hash code for this object
         */
    public int
    hashCode()
        {
            return id;
        }


        ////////////////////////////////////////////////////////////////////
        // PROTECTED STUFF
        ////////////////////////////////////////////////////////////////////

        /**
         * @return the String "host"
         */
    protected final String
    getType()
        {
            return "host";
        }
    
        /**
         * @return a String hexadecimal representation of the local (cell-wide) ID of the
         * host.
         */
    protected final Object
    getLocalId()
        {
            return Integer.toHexString(id);
        }
    
        ////////////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        ////////////////////////////////////////////////////////////////////

        //
        // Parameters read from the active profile
        //
    private static final String AUTO_DETECT = "autodetect";
    private static final String PROP_LOCALHOST_ID = "localhost.id";
    private static final String PROP_LOCALHOST_NAME = "localhost.name";
    private static final String PROP_LOCALHOST_INETADDR = "localhost.inetaddr";
    private static final String PROP_LOCALHOST_CELL = "localhost.cell";

        /**
         * Creates a new <code>HostId</code> instance.
         *
         * @param cellName The system-wide unique cell name.
         * @param localId The cell-wide unique id of the host.
         * @param hostName A symbolic name of the host (e.g. its internet name).
         * @param addr An internet contact address of the host (if available).
         */
    private
    HostId( String cellName, int localId, String hostName, InetAddress addr )
        {
            super( new CellId(cellName) );
            this.id = localId;
            this.hostName = hostName;
            this.inetAddr = addr;
        }

        /**
         * Detects the HostId of the local host. Primarly, information is read from the
         * middleware profile. If no such information is provided, it falls-back to an
         * auto-detection procedure which generates the host identification from its
         * internet address.
         *
         * If the internet address information is also not available, it uses a hardcode
         * configuration. Notice that, such hardcoded configuration will be useless
         * except for single-host testing pourposes.
         *
         * @return a <code>HostId</code> value
         */
    private synchronized static HostId
    detectLocalHost()
        {
            if ( localHost == null ) {
                    // import profile definitions
                String sHostID = Exehda.getProperty(PROP_LOCALHOST_ID, AUTO_DETECT);
                String sHostName = Exehda.getProperty(PROP_LOCALHOST_NAME, AUTO_DETECT);
                String sHostAddr = Exehda.getProperty(PROP_LOCALHOST_INETADDR, AUTO_DETECT);

                InetAddress ia = null;
            
                    // check if any autodetection is necessary
                if ( AUTO_DETECT.equals(sHostID) 
                     || AUTO_DETECT.equals(sHostName)
                     || AUTO_DETECT.equals(sHostAddr) ){
                    ia = detectInetAddress();
                }
            
                int hid = ( AUTO_DETECT.equals(sHostID) )
                    ? ( ia == null ? (int) 0x7F000001 : ipv4ToLocalId(ia.getAddress()) )
                    : Integer.parseInt(sHostID);

            
                String hn = ( AUTO_DETECT.equals(sHostName) )
                    ? ( ia == null ? "localhost" : ia.getHostName())
                    : sHostName;

                InetAddress hia = ia;
                if ( !AUTO_DETECT.equals(sHostAddr) ) {
                    try {
                        hia = InetAddress.getByName(sHostAddr);
                    }
                    catch ( java.net.UnknownHostException uhe ) {
                        hia = null;
                    }
                }

                localHost = new HostId ( detectLocalCell(), hid, hn, hia );

                Exehda.logMessage("LOCALHOST is id="+localHost.getLocalId()
                                  +", name="+localHost.getName()
                                  +", cell="+localHost.getCellName()
                                  +", inetaddr="+localHost.getInetAddress());
            }
                
            return localHost;
        }


    
        /**
         * Detects the local-cell's name. Returns the cell name defined in the active
         * profile. If such value is not defined, returns the fake name 'localcell'.
         *
         * @return a <code>String</code> value
         */
    private static String
    detectLocalCell()
        {
                // check if profile defined the local cell name
            
            String localCellName = Exehda.getProperty(PROP_LOCALHOST_CELL, AUTO_DETECT);
            
            if ( AUTO_DETECT.equals(localCellName) ) {
                    //
                    // TO DO: build cell name from dns information
                    //
                return "localcell";
            }
            else {
                return localCellName;
            }
        }

        /**
         * Detects a non-local ip address of this node, if such address is availble.
         *
         * A best-effort approach is used to determine the ip address of the host:
         *
         *   <ol>
         
         *   <li> First, JDK 1.4 NetworkInterface classes are used, if available to
         *   determine all IP addresses assigned to local network interfaces. If it
         *   succeeds, the first non-loopback ip found is used. Note that is may became a
         *   little non-deterministic on multi-interface nodes;
         
         *   <li> Second, if the class NetworkInterface is not available in the system,
         *   e.g. if we as running in a pre-1.4 release of JDK or either in a J2ME
         *   environemnt, a DNS lookup is tried using the registered local host
         *   name. Depending on the local-host name resolving configuration this will
         *   eventually succeed an return a real IP of the node.
         
         *   <li> Third, of nothing works, the address returned by
         *   InetAddress.getLocalHost() is used. Beware that, often, 127.0.0.1 will be
         *   returned and so that HostId will be useless for everything else debugging
         *   things in the local host.
         *
         *   </ol>
         *
         * @return an <code>InetAddress</code> value
         */
    private static InetAddress
    detectInetAddress()
        {
            Exehda.logMessage("Detecting host identification...");
            try {
                    //
                    // XXX: Try the java.net.NetworkInterface approach (requires Java 1.4
                    // or later)
                    //
                if ( Class.forName("java.net.NetworkInterface") != null ) {
                    java.util.Enumeration eni = NetworkInterface.getNetworkInterfaces();
                    while ( eni.hasMoreElements() ) {
                        NetworkInterface ni = (NetworkInterface) eni.nextElement();

                        java.util.Enumeration eia = ni.getInetAddresses();
                        while ( eia.hasMoreElements() ) {
                            InetAddress ia = (InetAddress) eia.nextElement();

                                // stop searching on the first non-loopback address
                            if ( !ia.isLoopbackAddress() ) {
                                    // skip non-ipv4 addresses by now
                                if ( ia.getAddress().length == 4 ) {
                                    return ia;
                                }
                            }
                        }
                    }
                }
            }
            catch (ClassNotFoundException cnfe) {
                Exehda.logMessage("Warning: java.net.NetworkInterface not available");
            }
            catch (Exception e) {
                Exehda.logMessage("Warning: "+e);
            }

            try {
                    //
                    // XXX: Try the DNS search approach (requires a DNS entry for the
                    // host and the host name to be properly set on the local host)
                    //
                InetAddress[] ia =
                    InetAddress.getAllByName(
                        InetAddress.getLocalHost().getHostName());

                for ( int i=0; i < ia.length; i++ ) {
                    
                        // stop searching on the first non-loopback address
                    if ( ! ia[i].getHostAddress().startsWith("127.0") ) {
                        return ia[i];
                    }
                }
            }
            catch (Exception e) {
                Exehda.logMessage("Warning: "+e);
            }

            Exehda.logMessage("Warning: non-loopback address of the localhost could not be determined");
            
            try {
                    // last chance, use the local host directly
                return InetAddress.getLocalHost();
            }
            catch ( Exception e ) {
                Exehda.logMessage("Warning: " + e);
            }

            Exehda.logMessage("Warning: detection failed. Will use hardcoded settings.");
                    // could not detect the local address
            return null;
        }
    

        /**
         * Facility for generating host id's from IPv4 addresses.
         *
         * @param b a <code>byte[]</code> value
         * @return an <code>int</code> value
         */
    private static int
    ipv4ToLocalId(byte[] b)
        {
            return ((b[0] & 0xFF) << 24)
                | ((b[1] & 0xFF) << 16)
                | ((b[2] & 0xFF) << 8)
                | ( b[3] & 0xFF);
        }

        /** cell-wide unique host id */
    private int id;
        /** symbolic host name (if available) */
    private String hostName;
        /** Internet address (IP) os this host (if available) */
    private InetAddress inetAddr;
        /** The id of the local host (cache) */
    private static HostId localHost = null;
}

