// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.cib.backend.ldap;

import org.isam.exehda.ResourceName;
import org.isam.exehda.ResourceName.NameComponent;
import org.isam.exehda.ResourceName.NameSpace;

import javax.naming.NamingException;
import javax.naming.NameNotFoundException;
import javax.naming.NameAlreadyBoundException;


/**
 * Describe class <code>BaseLRTH</code> here.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-05-04 11:32:27 -0300 (Wed, 04 May 2005) $ $Revision: 853 $
 */
abstract class BaseLRTH
    extends LdapResourceTypeHandler
{
    protected BaseLRTH(String[] ct, String[] lp)
        {
            super(ct, lp);
        }

    public void
    deleteNode( LdapSession s, ResourceName rn )
        {
                // try to delete the node
            
                // if it fails because the node has a subtree, tag the node as deleted

                //
                // TO DO:
                //
            throw new NullPointerException("Not yet implemented.");
        }
    
    public String
    getNodeAttribute( LdapSession s, ResourceName rn, String a )
        throws NamingException
        {
            String la = encodeAttrName(rn.getType(), a);
            String lav = super.getNodeAttribute(s, rn, la);
            return decodeAttrValue(a, la, lav);
        }

    public void
    setNodeAttribute( LdapSession s, ResourceName rn, String a, String v)
        throws NamingException
        {
                // pre-process 'a' and 'v'
            String la = encodeAttrName(rn.getType(), a);
            String lav = encodeAttrValue(rn.getType(), a, v);   
            super.setNodeAttribute(s, rn, la, lav);
        }

    public String[]
    listNodeAttributes( LdapSession s, ResourceName rn )
        throws NamingException
        {
                //
                // XXX: force overwritting in sub-classes
                //
            throw new NullPointerException("Must be overwritten in subclasses");
        }

    public DnBuffer
    encode( NameComponent nc, DnBuffer dnb )
        {
            if ( dnb == null ) dnb = new DnBuffer();
            
            String t = encodeAttrName(nc.getType(),
                                      CIB_ATT_RES_INSTANCE);

            String v = encodeAttrValue(nc.getType(),
                                       CIB_ATT_RES_INSTANCE,
                                       nc.getInstance());

            dnb.append(t, v);

            return dnb;
        }

//     public NameComponent
//     decodeNameComponent( String dnc )
//         {
//             System.out.println(dnc);
            
//             int idx = dnc.indexOf('=');
//             if ( idx <= 0 || idx == (dnc.length()-1) ) {
//                 throw new IllegalArgumentException("DN component:"+dnc);
//             }

//             String t = dnc.substring(0,idx);
//             String i = dnc.substring(idx+1);

//             return new NameComponent(t,i);
//         }

        /////////////////////////////////////////////////////////
        // PROTECTED STUFF
        /////////////////////////////////////////////////////////

        /**
         * Returns which LDAP attribute mantains the information corresponding to the
         * given CIB attribute.
         *
         * @param ct a CIB resource type
         * @param ca a CIB resource attribute
         * @return the corresponding LDAP attribute
         */
    protected abstract String
    encodeAttrName( String ct, String ca );
    
        /**
         * Packs a CIB resource attribute value into an LDAP attribute value. Optionally,
         * information about the CIB attribute name, or even the CIB reosurce type, may
         * be packed together with the actual attribute value, e.g. when multiplexing
         * several CIB attributes names over a single LDAP attribute.
         *
         * @param ct a CIB resource type
         * @param ca a CIB resource attribute
         * @param cav a CIB resource attribute value
         * @return the encode representation of the attribute value
         */
    protected abstract String
    encodeAttrValue( String ct, String ca, String cav );

        /**
         * Returns the 'value' part of a LDAP attribute value representation previously
         * built with <code>encodeAttrValue</code>
         *
         * @param ca CIB attribute name
         * @param la LDAP attribute name
         * @param lav LDAP attribute value
         * @return a CIB attribute value
         */
    protected abstract String
    decodeAttrValue( String ca, String la, String lav);
    
        /**
         * Creates an node in the LDAP directory. Recursively creates parent nodes if
         * necessary.
         *
         * @param s a <code>LdapSession</code> value
         * @param rn a CIB resource name
         * @param dn the corresponding distinguished name of the associated LDAP node
         * @param ans initial attributes 
         * @param avs initial attributes values
         * @exception NamingException if an error occurs
         */
    protected void
    internalCreateNode( LdapSession s, ResourceName rn, String dn, String[] ans, String[] avs )
        throws NamingException
        {
            try {
                    // best effort, try to create the node
                s.createNode(dn, ans, avs);
            }
            catch (NameNotFoundException e) {                
                    // if it fails, try creating parent node first
                createParentNode(s, rn);
                s.createNode(dn, ans, avs);

            }
            catch (NameAlreadyBoundException nabe) {
                    //
                    // XXX: node already exists, it should be mostly safe to just ignore
                    // the exception. Perhaps, should check the existing node objectClass.
                    //
            }
        }
    
        /**
         * Recursively creates the non-existing parent nodes of the LDAP node
         * corresponding to the given CIB resource.
         *
         * @param s a <code>LdapSession</code> value
         * @param rn a <code>ResourceName</code> value
         * @exception NamingException if an error occurs
         */
    final private void
    createParentNode( LdapSession s, ResourceName rn)
        throws NamingException
        {
            NameSpace ns = rn.getNameSpace();
            ResourceName rnParent = new ResourceName(
                ns.getCommonName().getType(),
                ns.getCommonName().getInstance(),
                ns.getParent() == null ? NameSpace.DEFAULT_NS : ns.getParent());

            LdapResourceTypeHandler.getRootHandler().createNode(s, rnParent);
        }
}

