// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.cib.backend.ldap;

import org.isam.exehda.ResourceName;
import org.isam.exehda.ResourceName.NameComponent;
import org.isam.exehda.ResourceName.NameSpace;

import javax.naming.NamingException;
import javax.naming.directory.Attributes;
import javax.naming.directory.Attribute;


/**
 * Provides a mapping beteween the CIB type <code>cell</code> and the LDAP type
 * <code>exehdaCell</code>.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-05-04 10:42:20 -0300 (Wed, 04 May 2005) $ $Revision: 849 $
 */
class CellLRTH 
    extends BaseLRTH
{
    private static final int ATT_OBJCLASS = 0;
    private static final int ATT_CN       = 1;
    
    private static final String[] DFLT_ATTR;
    private static final String[] DFLT_ATTR_VAL;

    private static final String[] HANDLED_CIB_TYPES = { "cell" };
    private static final String[] HANDLED_DN_PREFIXES = { "cn=cell:" };

    static {
        
        DFLT_ATTR = new String[] {
            "objectClass",     // ATT_OBJCLASS
            "cn"               // ATT_CN
        };
        
        DFLT_ATTR_VAL = new String[] {
            "exehdaCell",      // ATT_OBJCLASS
            null               // ATT_CN
        };
    }


    public CellLRTH()
        {
            super(HANDLED_CIB_TYPES, HANDLED_DN_PREFIXES);
        }

    public String
    createNode( LdapSession s, ResourceName rn )
        throws NamingException
        {
            if ( !"cell".equals(rn.getType()) ) {
                throw new IllegalArgumentException(
                    "Invalid resource type: "+rn.getType());
            }
            
                // map the given resource name to an ldap DN
            String dn = encode(rn, null).toString();
                // fill in mandatory attributes
            String[] avs = (String[]) DFLT_ATTR_VAL.clone();
                // overwrite default values with user-provided values
            avs[ATT_CN] = encodeAttrValue("cell",
                                          CIB_ATT_RES_INSTANCE,
                                          rn.getSimpleName());
                // commit to LDAP
            internalCreateNode(s, rn, dn, DFLT_ATTR, avs);

            return dn;
        }

    public NameComponent
    decodeNameComponent( String dnc )
        {
            System.out.println(dnc);
            
            int idx = dnc.indexOf('=');
            if ( idx <= 0 || idx == (dnc.length()-1) ) {
                throw new IllegalArgumentException("DN component:"+dnc);
            }

                //
                // FIX ME: not checking if dnc starts with "cn=cell:"
                //
            
            String t = "cell";
            String i = dnc.substring(idx+6);

            return new NameComponent(t,i);
        }


    public String
    getDefaultSearchFilter()
        {
            return "(objectClass=exehdaCell)";
        }

    public LdapResourceInfo
    toLdapResourceInfo( String dn, Attributes atts )
        {
            ResourceName rn = decodeResourceName(dn);            
                
            return new LdapResourceInfo(rn, atts) {
                    public String getAttribute(String a) {
                        try {
                            final String LDAP_ATT_CN = "cn";
                            final String LDAP_ATT_XATTR = "xAttr";
                            
                            if ( CIB_ATT_RES_TYPE.equals(a) ) {
                                return "cell";
                            }
                            else if ( CIB_ATT_RES_INSTANCE.equals(a)) {
                                String cn = (String) atts.get(LDAP_ATT_CN).get();

                                System.out.println(cn); // <<<<<<<<<<<<<<<<<<<===========================
                                    
                                return cn.substring(cn.indexOf(':')+1);
                            }
                            else {
                                if ( this.cache == null ) {
                                    this.cache = new java.util.Hashtable();
                                        
                                        // parse xAttrs only once
                                    Attribute att = atts.get(LDAP_ATT_XATTR);
                                    for ( int i=0; i<att.size(); i++ ) {
                                        String s = (String) att.get(i);
                                        int p = s.indexOf(':');
                                        cache.put(s.substring(0,p),
                                                  s.substring(p+1));
                                    }
                                }
                                return (String) cache.get(a);
                            }
                        }
                        catch ( NamingException ne ) {
                                //
                                // TO DO: use logger
                                //
                            ne.printStackTrace();
                            return null;
                        }
                    }
                };
        }
    
        /////////////////////////////////////////////////////////
        // PROTECTED STUFF
        /////////////////////////////////////////////////////////
    
    protected String
    encodeAttrName( String ct, String ca )
        {
            if ( CIB_ATT_RES_TYPE.equals(ca) ) {
                return "objectClass";
            }
            else if ( CIB_ATT_RES_INSTANCE.equals(ca) ) {
                return "cn";
            }
            else {
                return "xAttr";
            }
        }

    protected String
    encodeAttrValue( String ct, String ca, String cav )
        {
            if ( CIB_ATT_RES_TYPE.equals(ca) ) {
                return "exehdaCell";
            }
            else if ( CIB_ATT_RES_INSTANCE.equals(ca) ) {
                return "cell:"+cav;
            }
            else {        
                return ca+":"+cav;
            }
        }
    
    protected String
    decodeAttrValue( String ca, String la, String lav )
        {
            System.out.println( "Unknown encoding: ca="+ca+" la="+ca+" lav="+lav);
            if ( "xAttr".equals(la) ) {
                return lav.substring(lav.indexOf(':')+1);
            }
            else if ( CIB_ATT_RES_INSTANCE.equals(ca)
                      && "cn".equals(la)) {
                return lav.substring(lav.indexOf(':')+1);
            }
            else if ( CIB_ATT_RES_TYPE.equals(ca) ) {
                return "cell";
            }
            else {
                throw new NullPointerException(
                    "Unknown encoding: ca="+ca+" la="+ca+" lav="+lav);
            }
        }
}
