// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.cib.backend.ldap;

import org.isam.exehda.ResourceName;
import org.isam.exehda.ResourceName.NameComponent;
import org.isam.exehda.ResourceName.NameSpace;
import org.isam.exehda.ResourceConstraint;
import org.isam.exehda.AttributeSimpleMatchConstraint;

import javax.naming.NamingException;
import javax.naming.directory.Attributes;
import javax.naming.directory.Attribute;
import java.util.StringTokenizer;

/**
 * Maps cib resource types to the generic 'exehdaResource' LDAP object class.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-05-04 10:39:29 -0300 (Wed, 04 May 2005) $ $Revision: 848 $
 */
class DefaultLRTH
    extends BaseLRTH
{    
    private static final int ATT_OBJCLASS = 0;
    private static final int ATT_CN = 1;
    
    private static final String[] DFLT_ATTR;
    private static final String[] DFLT_ATTR_VAL;

    private static final String[] HANDLED_CIB_TYPES = null; // ALL
    private static final String[] HANDLED_DN_PREFIXES = null; // ALL

    static {
        
        DFLT_ATTR = new String[] {
            "objectClass",     // ATT_OBJCLASS
            "cn"               // ATT_CN
        };
        
        DFLT_ATTR_VAL = new String[] {
            "exehdaResource",  // ATT_OBJCLASS
            null               // ATT_CN
        };
    }
    

    public DefaultLRTH()
        {
            super(HANDLED_CIB_TYPES, HANDLED_DN_PREFIXES);
        }

    public String
    createNode( LdapSession s, ResourceName rn )
        throws NamingException
        {            
                // map the given resource name to an ldap DN
            String dn = encode(rn, null).toString();
                // fill in mandatory attributes
            String[] avs = (String[]) DFLT_ATTR_VAL.clone();
                // overwrite default values with user-provided values
            avs[ATT_CN] = encodeAttrValue( rn.getType(),
                                           CIB_ATT_RES_INSTANCE,
                                           rn.getSimpleName());
                // commit to LDAP
            internalCreateNode(s, rn, dn, DFLT_ATTR, avs);

            return dn;
        }

    public NameComponent
    decodeNameComponent( String dnc )
        {
            System.out.println(dnc);
            
            int idx = dnc.indexOf('=');
            if ( idx <= 0 || idx == (dnc.length()-1) ) {
                throw new IllegalArgumentException("DN component:"+dnc);
            }

                //
                // FIX ME: not checking if dnc is in the form "cn=type:instance"
                //
            
            String t = dnc.substring(idx+1, dnc.indexOf(':'));
            String i = dnc.substring(dnc.indexOf(':'));

            return new NameComponent(t,i);
        }

        /**
         * Knows about AttributeSimpleMatchingConstraints.
         *
         * @param rc a <code>ResourceConstraint</code> value
         * @return a <code>String</code> value
         */
    public String
    toSearchFilter( ResourceConstraint rc )
        {            
            if ( rc instanceof AttributeSimpleMatchConstraint ) {
                AttributeSimpleMatchConstraint arc =
                    (AttributeSimpleMatchConstraint) rc;

                StringBuffer buff = new StringBuffer(512);

                buff.append("(&");
//                     .append(DEFAULT_FILTER);

                String an = arc.getAttributeName();
                String exp = arc.getMatchingExp();

                    // avoid creating StringTokenizer if possible
                if ( exp.indexOf("|") < 0 ) {
                    if ( CIB_ATT_RES_TYPE.equals(an) ) {
                        buff.append("(cn=").append(exp).append(":*)");
                    }
                    else if ( CIB_ATT_RES_INSTANCE.equals(an) ) {
                        buff.append("(cn=*:").append(exp).append(")");
                    }
                    else {
                        buff.append("(xAttr=").append(an).append(":").append(exp).append(")");
                    }
                }
                else {
                    StringTokenizer st = new StringTokenizer(exp, "|");
                    
                    while ( st.hasMoreTokens() ) {
                        exp = st.nextToken().trim();
                        
                        if ( CIB_ATT_RES_TYPE.equals(an) ) {
                            buff.append("(cn=").append(exp).append(":*)");
                        }
                        else if ( CIB_ATT_RES_INSTANCE.equals(an) ) {
                            buff.append("(cn=*:").append(exp).append(")");
                        }
                        else {
                            buff.append("(xAttr=").append(an).append(":").append(exp).append(")");
                        }
                    }
                }

                buff.append(")");
                    
                return buff.toString();
            }
            else {
                    // default, match anything handled by this handler
                return getDefaultSearchFilter();
            }
        }

    public String
    getDefaultSearchFilter()
        {
            return "(objectClass=exehdaResource)";
        }


    public LdapResourceInfo
    toLdapResourceInfo( String dn, Attributes atts )
        {
            ResourceName rn = decodeResourceName(dn);
                
            return new LdapResourceInfo(rn, atts)
                {    
                    public String getAttribute(String a) {
                        try {
                            final String LDAP_ATT_CN = "cn";
                            final String LDAP_ATT_XATTR = "xAttr";
                            
                            if ( CIB_ATT_RES_TYPE.equals(a) ) {
                                String cn = atts.get(LDAP_ATT_CN).get().toString();
                                return cn.substring(0, cn.indexOf(':'));
                            }
                            else if ( CIB_ATT_RES_INSTANCE.equals(a)) {
                                String cn = atts.get(LDAP_ATT_CN).get().toString();
                                return cn.substring(cn.indexOf(':')+1);
                            }
                            else {
                                if ( this.cache == null ) {
                                    this.cache = new java.util.Hashtable();
                                        
                                        // parse xAttrs only once
                                    Attribute att = atts.get(LDAP_ATT_XATTR);
                                    for ( int i=0; i<att.size(); i++ ) {
                                        String s = (String) att.get(i);
                                        int p = s.indexOf(':');
                                        cache.put(s.substring(0,p),
                                                  s.substring(p+1));
                                    }
                                }
                                return (String) cache.get(a);
                            }
                        }
                        catch (NamingException ne) {
                                //
                                // TO DO: use logger
                                //
                            ne.printStackTrace();
                            return null;
                        }
                    }
                };
        }

    
        /////////////////////////////////////////////////////////
        // PROTECTED STUFF
        /////////////////////////////////////////////////////////
    
    protected String
    encodeAttrName( String ct, String ca )
        {
            if ( CIB_ATT_RES_TYPE.equals(ca) ) {
                return "objectClass";
            }
            else if ( CIB_ATT_RES_INSTANCE.equals(ca) ) {
                return "cn";
            }
            else {
                return "xAttr";
            }
        }

    protected String
    encodeAttrValue( String ct, String ca, String cav )
        {
            if ( CIB_ATT_RES_TYPE.equals(ca) ) {
                return "exehdaResource";
            }
            else if ( CIB_ATT_RES_INSTANCE.equals(ca) ) {
                return ct+":"+cav;
            }
            else {
                return ca+":"+cav;
            }
        }
    
    protected String
    decodeAttrValue( String ca, String la, String lav )
        {
            if ( "xAttr".equals(la) ) {
                return lav.substring(lav.indexOf(':')+1);
            }
            else if ( CIB_ATT_RES_INSTANCE.equals(ca)
                      && "cn".equals(la)) {
                return lav.substring(lav.indexOf(':')+1);
            }
            else if ( CIB_ATT_RES_TYPE.equals(ca)
                      && "cn".equals(la)) {
                return lav.substring(0, lav.indexOf(':'));
            }
            else {
                throw new NullPointerException(
                    "Unknown encoding: ca="+ca+" la="+ca+" lav="+lav);
            }
        }

}
