// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.ctxm;

import org.isam.exehda.ApplicationId;
import org.isam.exehda.services.ContextManager.Context;
import org.isam.exehda.services.ContextManager.ContextState;
import org.isam.exehda.services.ContextManager.ContextListener;
import org.isam.util.xml.XmlElement;
import org.isam.util.xml.XmlParsingException;
import java.util.Vector;

/*
 <CONTEXT n="ctxname">

 <STATES>... </STATES>

 <FILTER syntax="SingleSensorRangeMapping">
    <INDEX>
       <SENSOR ...>
    </INDEX>

    <RANGES>
       <RANGE ...>
       <RANGE ...>
       <DEFAULT ...>
    </RANGES>
 </FILTER>

 </CONTEXT>


*/


/**
 * This class is not thread-safe. Synchronization must be built in the CtxManagerImplNode.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-16 18:19:13 -0300 (Thu, 16 Jun 2005) $ $Revision: 922 $
 */
class ContextChain
{
    public static void
    setFilterFactory( ContextFilterFactory ff )
        {
            filterFactory = ff;
        }
        
    public ContextChain ( ApplicationId scope, String ctxDef, Environment env)
        {
            this.ctxElement = parseContext(scope, ctxDef);
            this.ctxFilter = createFilter(ctxDef);
            this.pendingEvents = new ContextEventQueue();
            this.lastFilteredState = null;
            this.lastDeliveredEvent = null;
            this.consumers = new Vector(3);
            
                // XXX: complete the initialization by generating a fake event
            this.update(env);
            this.deliverEvent();   
        }
    
    public Context
    getContext()
        {
            return this.ctxElement;
        }

    public ContextState
    getCurrentState()
        {
                // returns the last **NOTIFIED** state, so that the view the caller has
                // of the current context state is consistent with that of the registered
                // listeners.
            return this.lastDeliveredEvent.getState();
        }

    public long
    getLastModified()
        {
            return this.lastDeliveredEvent.getTimeStamp();
        }

    public ApplicationId
    getScope()
        {
            return this.ctxElement.getScope();
        }

    public void
    addListener( ContextListener l )
        {
            if ( l != null && ! consumers.contains(l) ) {
                consumers.addElement(l);
                    //
                    // XXX: ensure the listener will be made aware of the current
                    // context state
                l.contextChanged(lastDeliveredEvent.getPreviousState(),
                                 lastDeliveredEvent.getState());
            }
        }

    public void
    removeListener( ContextListener l )
        {
            if ( l != null ) {
                consumers.removeElement(l);
            }
        }

    public boolean
    hasPendingEvent()
        {
            return ! pendingEvents.isEmpty();
        }

    
        /**
         * Delivers *ONE* event in the pendingEvents queue to *ALL* registered listeners.
         *
         */
    public void
    deliverEvent()
        {
                // XXX: should never happen we enter this method with an empyt queue since
                // this condition is checked in the context manager implementation,
                // but..
            if ( pendingEvents.isEmpty() ) { 
                return;
            }
            else {
                ContextEvent ev = pendingEvents.pop();

                for (int i=0; i<consumers.size(); i++) {
                    ((ContextListener) consumers.elementAt(i))
                        .contextChanged(ev.getPreviousState(), ev.getState());
                }
                this.lastDeliveredEvent = ev;
            }   
        }

    public void
    update( Environment env )
        {
                // apply the filter to the current environment
            String state = ctxFilter.filter(env);
            
                // if the context state has changed
            if ( !state.equals(lastFilteredState) ) {
                    // enqueue an event to be later delivered to listeners of this chain

                    //
                    // FIX ME: lastFileteredState == null;
                    //
                ContextState sPrev = (lastFilteredState == null) ?
                    null : ctxElement.getStateInstance(lastFilteredState);

                ContextState sCurr = ctxElement.getStateInstance(state);
                
                pendingEvents.push(
                    new ContextEvent(
                        sPrev, sCurr, env.getSnapshotTimeStamp()));
                
                    // remember the new state
                lastFilteredState = state;
            }                
        }

    
        /**
         * Returns true if the given object is a ContextChain and is associated with the
         * production of the same context element information.
         *
         * @param o an <code>Object</code> value
         * @return a <code>boolean</code> value
         */
    public boolean
    equals( Object o )
        {
            if ( o != null && o instanceof ContextChain ) {
                return ctxElement.equals(((ContextChain) o).ctxElement);
            }
            return false;
        }

    
        /**
         * Uses the associated context element object as key for hashing.
         *
         * @return an <code>int</code> value
         */
    public int
    hashCode()
        {
            return ctxElement.hashCode();
        }

        //////////////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        //////////////////////////////////////////////////////////////////////
    
        /**
         * Builds the context object corresponding to the given XML context definition.
         *
         * @param scope  the scope (application) in which the context is being defined
         * @param ctxDef the XML context definition
         * @return the built Context object
         * @exception IOException if an error occurs
         */
    private static final Context
    parseContext(ApplicationId scope, String ctxDef)
        {
            try {
                    // parse context definition
                XmlElement docroot = XmlElement.parse(
                    new java.io.StringReader(ctxDef));            

                    // get context name from the <CONTEXT> XML element
                Vector tmp = XmlElement.selectByPath(docroot, "/CONTEXT", null);
                String ctxName = ((XmlElement) tmp.elementAt(0)).getAttribute("n", null);
                    // check if context name is broken
                if ( ctxName == null ) {
                    throw new IllegalArgumentException(
                        "Missing context name in context definition");
                }
                else {
                        // get defined context states from <STATES> XML element
                    tmp.removeAllElements();
                    tmp = XmlElement.selectByPath(docroot, "/CONTEXT/STATES/STATE", tmp);
                        // check if context states definition is broken
                    if ( tmp.size() == 0 ) {
                        throw new IllegalArgumentException(
                            "Missing context states in context definition");
                    }
                    else {
                            // ok, at least one state has been defined
                        String[] ctxStates = new String[tmp.size()];
                        for ( int i=0; i<ctxStates.length; i++ ) {
                            ctxStates[i] = ((XmlElement) tmp.elementAt(i))
                                .getAttribute("n", null);
                            
                            if ( ctxStates[i] == null ) {
                                throw new IllegalArgumentException(
                                    "Missing mandatory attribute 'n' in <STATE> element");
                            }
                        }
                        
                        tmp.removeAllElements();
                            // create the corresponding context object
                        return new Context(ctxName, scope, ctxDef, ctxStates);
                    }
                }
            }
            catch (XmlParsingException xpe) {
                throw new IllegalArgumentException(
                    "Malformed context definition:"+xpe);
            }
            catch (java.io.IOException ioe) {
                    //
                    // XXX: should never happen, since we are reading
                    //      from in-memory strings, but...
                throw new IllegalArgumentException(
                    "Exception raised while parsing definition:"+ioe);
            }
        }

    private static final ContextFilter
    createFilter( String ctxdef )
        {
                //
                // FIX ME: not very robust against malformed filters
                //
            String fdef = ctxdef.substring(
                ctxdef.indexOf("<FILTER"),
                ctxdef.lastIndexOf("</FILTER>")+10);

            return filterFactory.createFilter(fdef);
        }

    private Context ctxElement;
    private ContextFilter ctxFilter;
    private String lastFilteredState;
    private ContextEvent lastDeliveredEvent;
    private ContextEventQueue pendingEvents;
    private Vector consumers;

    private static ContextFilterFactory filterFactory = null;
}

