// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.ctxm;

import java.util.Vector;


/**
 * Implements a 'smart' queue (First-In First-Out) for storing context events.
 *
 * By smart we mean the policy of insertion of new events in the queue may be tunned
 * according to the context semantics demand. For example, curretly events would either
 * be appended or cause the oler events to be droped. In future, more advanced policies
 * should also be implemented.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-10 10:47:30 -0300 (Fri, 10 Jun 2005) $ $Revision: 886 $
 */
class ContextEventQueue
{
    public static interface QueuingPolicy
    {
        public int APPEND = 0;
        public int OVERWRITE = 1;
    }
    
        /**
         * Creates a new <code>ContextEventQueue</code> instance with default APPEND
         * policy.
         *
         */
    public ContextEventQueue ()
        {
            this(QueuingPolicy.APPEND);
        }
    
        /**
         * Creates a new <code>ContextEventQueue</code> instance with the given event
         * insertion policy.
         *
         * @param policy an <code>int</code> value
         */
    public ContextEventQueue ( int policy )
        {
            this.policy = policy;
            this.queue = new Vector(4);
        }

    
        /**
         * Appends a new event to the tail of the queue.
         *
         * Depending on the configured queuing policy, older events may be dropped.
         *
         * @param ev a <code>ContextEvent</code> value
         */
    public void
    push( ContextEvent ev )
        {
            if ( policy == QueuingPolicy.OVERWRITE ) queue.setSize(0);
            queue.add(ev);
        }

    
        /**
         * Consumes the event in the queue's head, returning that element.
         *
         * @return the event that was in the head of the queue
         */
    public ContextEvent
    pop()
        {
            ContextEvent ev = (ContextEvent) queue.firstElement();
            queue.removeElementAt(0);
            return ev;
        }
    
        /**
         * Returns whether this queue is empty (size()==0) or not.
         *
         * @return true if the queue is empty, false otherwise.
         */
    public boolean
    isEmpty()
        {
            return queue.size() == 0;
        }

    
        /**
         * Returns the number os events still in the queue.
         *
         * @return a int value >= 0
         */
    public int
    size()
        {
            return queue.size();
        }

    private int policy;
    private Vector queue;
}

