// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.ctxm;

import org.isam.exehda.Exehda;
import org.isam.exehda.ApplicationId;
import org.isam.exehda.services.ContextManager;
import org.isam.exehda.services.ContextManager.Context;
import org.isam.exehda.services.ContextManager.ContextState;
import org.isam.exehda.services.ContextManager.ContextListener;
import org.isam.exehda.services.AbstractService;
import org.isam.exehda.services.Logger.Sev;
import org.isam.exehda.services.Worb;
import org.isam.exehda.services.Executor;
import java.util.Vector;
import java.util.Hashtable;

//
// TO DO: document <CONTEXT> DTD; idiom and syntax <FILTER> attributes.
//
// TO DO: Perhaps the default context state should be defined outside the <FILTER>
// element, for instance in the <STATES> element. The ContextFilter would return null if
// it's not able to determine the current state of the context element and the
// ContextChain would map it to the default state defined in <STATES>.
//
// TO DO: StringMappingFilter
//
// TO DO: SystemPropertiesMonitor
//
// TO DO: SwPkgAvailableMonitor
//
//

public class CtxManagerImplNode
    extends AbstractService
    implements ContextManager, Runnable
{
    private final static String PROP_FITLER_FACTORY = "filterFactory";
    private final static String DFLT_FILTER_FACTORY =
    "org.isam.exehda.services.ctxm.filter.basic.BasicContextFilterFactory";
    
    public CtxManagerImplNode()
        {
            this.env = null;
            this.chains = new Vector();
            this.chainsByContext = new Hashtable();
        }

    public void
    start()
        throws ClassNotFoundException, InstantiationException, IllegalAccessException
        {
                // initialize the environment object
            this.env = new Environment();

                // setup context filter factory
            Class ffclass = Class.forName(
                Exehda.getServiceProperty(ContextManager.SERVICE_NAME,
                                          PROP_FITLER_FACTORY,
                                          DFLT_FILTER_FACTORY));
            ContextChain.setFilterFactory( (ContextFilterFactory) ffclass.newInstance());
            
                // activate service's main thread
            this.thr = new Thread(this);
            this.thr.start();

                // export service to WORB
            Worb worb = (Worb) Exehda.getService(Worb.SERVICE_NAME);            
            worb.exportService(this,
                               ContextManager.class,
                               ContextManager.SERVICE_NAME);
        }

    public void
    stop()
        {
                //
                // TO DO: unexport service
                //
            
                // stop main thread
            thr.interrupt();
            try {
                thr.join(50);
            }
            catch (InterruptedException ie ) {
                log(Sev.WARNING, "stop()", ie);
            }
            thr = null;

                // release allocated monitoring resources
            env.finalize();
        }
    
    public Context
    createContext( String ctxdef )
        {
            log(Sev.TRACE, "createContext()", "entering desc={0}", ctxdef);
            
            try {
                    // determine the scope in which the context element is being defined
                ApplicationId scope = getExecutor().currentApplication();
                    // let the context chain parse the context and build the context object
                ContextChain cc = new ContextChain(scope, ctxdef, env);
                    // if the given context element is already being produced
                Context ctx = cc.getContext();
                if ( chainsByContext.containsKey(ctx) ) {
                        // do nothig, just return the alreay existing object
                    return ((ContextChain) chainsByContext.get(ctx)).getContext();
                }
                else {
                        // register the new chain 
                    chainsByContext.put(ctx, cc);
                        // add it to the iteration list
                    chains.addElement(cc);
                        // return the new context object
                    return ctx;
                }
            }
            finally {
                log(Sev.TRACE, "createContext()", "leaving");
            }
        }
    
        
    public void
    releaseContext( Context ctx )
        {
            log(Sev.TRACE, "releaseContext", "entering ctx={0}", ctx);            
            
                // lookup the corresponding chain
            ContextChain cc = (ContextChain) chainsByContext.get(ctx);

            if ( cc != null ) {
                    // check if the caller scope is allowed to release the given
                    // context element
                ApplicationId currScope = getExecutor().currentApplication();
                
                if ( currScope == null // its a middleware thread
                     || currScope.equals(cc.getContext().getScope()) ) { 

                        //
                        // TO DO: only allow deleting contexts that currently have *NO*
                        // listeners
                        //
            
                        // unregister the chain and remove it from the iteration list
                    chains.remove(chainsByContext.remove(ctx));
                }
                else {
                    throw new IllegalArgumentException(
                        "Context not accessible in the current execution scope");
                }
            }
            
            log(Sev.TRACE, "releaseContext", "leaving");
        }
    
    public void
    addContextListener( ContextListener l, Context ctx )
        {
            log(Sev.TRACE, "addContextListener", "entering ctx={0}", ctx);

            findContextChain(ctx).addListener(l);
            
            log(Sev.TRACE, "addContextListener", "leaving");
        }
    
    public void
    removeContextListener( ContextListener l, Context ctx )
        {
            log(Sev.TRACE, "removeContextListener", "entering ctx={0}", ctx);

            findContextChain(ctx).removeListener(l);
            
            log(Sev.TRACE, "removeContextListener", "leaving");
        }
    
        /**
         * ContextManager's main thread, responsible but updating context chains up on
         * modifications in the Environment and delivering context events to registered
         * listeners.
         *
         */
    public void
    run()
        {
            log(Sev.TRACE, "run()", "entering");
            
            while ( !Thread.interrupted() ) {
                try {
                        // waits until the environement is modified
                    synchronized (env) {
                        env.wait();
                    }
                        // freeze the environment
                    env.lockSnapshot();
                        // for each existing ContextChain 
                    for (int i=0; i < chains.size(); i++ ) {
                            // update it using the new environment state
                        ((ContextChain) chains.elementAt(i)).update(env);
                    }
                        // allow the environment to be updated again
                    env.unlockSnapshot();
                        // deliver events for registered listeners
                    deliverPendingEvents();
                }
                catch (InterruptedException ie) {
                    log(Sev.WARNING, "run()", ie);
                }
            }
            
            log(Sev.TRACE, "run()", "leaving");
        }

        /////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        /////////////////////////////////////////////////////////////
    
        /**
         * Helper singletone class used in context event delivery to ensure event
         * listeneres are invoked in the proper execution context (ApplicationId).
         *
         * @author last modified by $Author: lucc $
         * @version $Date: 2005-06-16 18:21:17 -0300 (Thu, 16 Jun 2005) $ $Revision: 923 $
         */
    private static class DeliverEventAction
        implements java.lang.Runnable
    {
        public static DeliverEventAction
        getInstance() { return singletone; }
    
        public static void
        setChain(ContextChain chain) { singletone.chain = chain; }
    
        public void run() { chain.deliverEvent(); }

        private ContextChain chain;
        private static DeliverEventAction singletone = new DeliverEventAction();
    }

    
        /**
         * Loop through registered context detection chains, delivering pending events.
         *
         */
    private final void
    deliverPendingEvents()
        {
            boolean notDone = true;

                // round-robing event delivering
            do {
                    // pretend it is done
                notDone = false;
                
                    // for each registered context chain
                for (int i=0; i < chains.size(); i++ ) {
                    ContextChain cc = (ContextChain) chains.elementAt(i);

                        // if there are pending events to be delivered to listeners
                    if ( cc.hasPendingEvent() ) {
                            // deliver ONE event 
                            //
                            // XXX: delivery is done through Executor.runAction() in
                            // order to ensure the execution scope of the event listener
                            // is correct
                        DeliverEventAction.setChain(cc);
                        getExecutor().runAction(cc.getScope(),
                                                DeliverEventAction.getInstance());
 
                            // and schedule a new loop
                        notDone = true;
                    }
                }
            }
            while ( notDone ) ;
        }
    
        /**
         * Lookups the context detection chain who produces the state information
         * relative to the given context element.
         *
         * @param ctx a <code>Context</code> value
         * @return a <code>ContextChain</code> value
         */
    private final ContextChain
    findContextChain(Context ctx)
        {
            ContextChain cc = (ContextChain) chainsByContext.get(ctx);

            if ( cc == null ) throw new NullPointerException("No such context.");
            
            return cc;
        }
    
        /**
         * Shortcut for accessing the middleware service "Executor".
         *
         * @return an <code>Executor</code> value
         */
    private final Executor
    getExecutor()
        {
            return (Executor) Exehda.getService(Executor.SERVICE_NAME);
        }

    private Environment env;
    private Thread      thr;
    private Vector      chains;
    private Hashtable   chainsByContext;
}




