// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.ctxm.filter.basic;


import org.isam.exehda.services.ContextManager.ContextState;
import org.isam.exehda.services.ctxm.ContextFilter;
import org.isam.exehda.services.ctxm.Environment;
import org.isam.util.xml.XmlElement;
import java.io.StringReader;
import java.io.IOException;
import java.util.Vector;
import java.util.Enumeration;


/**
 * This context filter that takes as input a value of type <code>double</code> from a
 * single sensor in the Environment and translates it into abstract context states (Strings)
 * according to defined ranges.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-16 15:52:56 -0300 (Thu, 16 Jun 2005) $ $Revision: 918 $
 */
class SingleSensorNumericRangeMappingFilter
    extends ContextFilter
{    
    public SingleSensorNumericRangeMappingFilter( String fdef  ) 
        {
            super(fdef);
            try {
                parseFilter();

                this.currState = this.dfltState;
                    // XXX: force first state update
                this.lastUpdate = Long.MIN_VALUE;
            }
            catch (IllegalArgumentException ie) {
                throw ie;
            }
            catch (Exception e) {
                throw new IllegalArgumentException(
                    "Invalid context filter definition: "+e);
            }            
        }
    
    public String
    filter(Environment env)
        {            
                // check input has changed since last filtering
            if ( env.isModifiedSince(sname, lastUpdate) ) {
                    // save last modification time
                lastUpdate = env.getLastModified(sname);
                    // start at the default state, since either monitoring data may not
                    // be available or the range matching may fail
                this.currState = this.dfltState;
                    // check input data is available
                if ( env.isDefined(sname) ) {
                        // read environment variable
                    double sval = readEnvironment(env);
                        // for each defined range 
                    for (int i=0; i<ranges.length; i++) {
                            // check if it matches the current sensor value
                        if ( ranges[i].contains(sval) ) {
                                // update current state
                            this.currState = ranges[i].getState(); 
                            break;
                        }
                    }
                }
            }
            return currState;
        }

        /////////////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        /////////////////////////////////////////////////////////////////////

    private final String XP_RANGES     = "/FILTER/RANGES/RANGE";
    private final String XP_DFLT_STATE = "/FILTER/RANGES/DEFAULT";
    private final String XP_SENSOR     = "/FILTER/INDEX/SENSOR";
    
        /**
         * Fills this filter with parameters from the context filter definition.
         *
         * @todo Review this code, its not very safe/robust in face of malformed context
         * definitions.
         */
    private final void parseFilter()
        throws IOException
        {
                // parse filter definition
            XmlElement docroot = XmlElement.parse(
                new java.io.StringReader(this.def));
                //
                // A) setup context state index range mappings
                //
            Vector tmp = XmlElement.selectByPath(docroot, XP_RANGES, null);
                // check ranges are really there
            if ( tmp.size() > 0 ) {
                this.ranges = new NumericRange[tmp.size()];
                    // for each <RANGE> element
                for ( int i=0; i<ranges.length; i++) {
                        // build the corresponding NumericRange object
                    this.ranges[i] = makeRange((XmlElement) tmp.elementAt(i));
                }
                    // optimize range order for faster matching
                sortRanges();
                    //
                    // B) setup default context state
                    //
                tmp = XmlElement.selectByPath(docroot, XP_DFLT_STATE, null);
                this.dfltState = ((XmlElement) tmp.elementAt(0))
                    .getAttribute("state", null);
                
                if ( this.dfltState != null ) { 
                        //
                        // C) initialize context index
                        //
                    initIndex(docroot);
                }
                else {
                    throw new IllegalArgumentException(
                        "Missing default context state definition");
                }
            }
            else {
                throw new IllegalArgumentException(
                    "Missing range mappings in context definition");
            }
        }
    
        /**
         * Initialize the index.
         *
         * Notice: sensors will be automatically enabled by the environment up-on the
         * first access.
         *
         * @param docroot a <code>XmlElement</code> value
         */
    private final void
    initIndex(XmlElement docroot)
        {
                // extract context element input sensor
            Vector tmp = XmlElement.selectByPath(docroot, XP_SENSOR, null);
            this.sname = ((XmlElement) tmp.elementAt(0)).getAttribute("n", null);

            if ( this.sname == null ) {
                throw new IllegalArgumentException(
                    "Missing sensor name in context index definition");
            }
            else {
                sname = "sensor:"+sname;
            }
        }
    
        /**
         * Builds a NumericRange object from the given <code>RANGE</code> XML element.
         *
         * @param xeRange a <code>XmlElement</code> value
         * @return a <code>NumericRange</code> value
         */
    private final NumericRange
    makeRange(XmlElement xeRange)
        {
                // handle range upper bound
            boolean ubo = true;
            String vu = xeRange.getAttribute("ubo", null);
            if ( vu == null ) {
                vu = xeRange.getAttribute("ub", null);
                if ( vu != null ) ubo = false;
            }

            double ub = ( vu == null ) ? Double.POSITIVE_INFINITY
                : Double.valueOf(vu).doubleValue();            
    
                // handle range lower bound
            boolean lbo = true;
            String vl = xeRange.getAttribute("lbo", null);
            if ( vl == null ) {
                vl = xeRange.getAttribute("lb", null);
                if ( vl != null ) lbo = false;
            }
            double lb = ( vl == null ) ? Double.NEGATIVE_INFINITY
                : Double.valueOf(vl).doubleValue();

            String vs = xeRange.getAttribute("state", null);
            if ( vs == null ) {
                throw new IllegalArgumentException(
                    "Missing context state in range mapping definition");
            }

            return new NumericRange(ub, ubo, lb, lbo, vs);
        }

    
        /**
         * Optimize ranges array ordering in order to allow a faster range matching in
         * method filter()
         *
         */
    private final void
    sortRanges()
        {
                //
                // TO DO: sort this.ranges considering ranges upper and lower bounds.
                //
        }
    
        /**
         * Read the selected environment variable, usually a sensor, returning its
         * current value as a double value.
         *
         * @param env an <code>Environment</code> value
         * @return a <code>double</code> value
         */
    private double
    readEnvironment(Environment env)
        {
            try {
                return env.getDouble(sname, Double.NaN, true);
            }
            catch (Exception e) {
                return Double.NaN;
            }
        }

    private String fdef;
    private String dfltState;
    private String sname;
    private NumericRange[] ranges;
    private long   lastUpdate;
    private String currState;
}
