// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.oxm;

import org.isam.exehda.Exehda;
import org.isam.exehda.ObjectId;
import org.isam.exehda.HostId;
import org.isam.exehda.CellId;
import org.isam.exehda.ApplicationId;
import org.isam.exehda.ResourceName;
import org.isam.exehda.ResourceName.NameComponent;
import org.isam.exehda.ResourceName.NameSpace;
import org.isam.exehda.services.AbstractService;
import org.isam.exehda.services.Logger.Sev;
import org.isam.exehda.services.OXManager;
import org.isam.exehda.services.OXManager.OXHandle;
import org.isam.exehda.services.OXManager.MasterOXHandle;
import org.isam.exehda.services.OXManager.HomeOXHandle;
import org.isam.exehda.services.OXManager.ProxyOXHandle;
import org.isam.exehda.services.OXManager.IllegalOXHandleException;
import org.isam.exehda.services.CellInformationBase;
import org.isam.exehda.services.Worb;
import java.util.Hashtable;

/**
 * Cellular instance of the OX Manager service. This implementation is for standalone use
 * only. If you intend to allow creation of OX'es in the same node the cellular manager
 * is use the DualOXManagerCell implementation instead.
 *
 *
 * @author Implemented by frainer, last modified by $Author: lucc $
 * @version $Date: 2005-05-04 10:26:19 -0300 (Wed, 04 May 2005) $ $Revision: 844 $ 
 */
public class OXManagerCell
    extends AbstractOXManager
{
    static final CellId LOCAL_CELL = HostId.getLocalHost().getCell();
    static final String CIB_ATT_CONTACT_ADDR = "contactAddress";
    
    public void
    start()
        {
                // export the service to the WORB
            String ca = getServiceWorb().exportService(this,
                                                       OXManager.class,
                                                       OXManager.SERVICE_NAME);

                // register itself as this cell manager
            ResourceName oxmrn = new ResourceName("service",
                                                  OXManager.SERVICE_NAME,
                                                  LOCAL_CELL.toNameSpace());
            
            getServiceCib().addResource(oxmrn);
            getServiceCib().setAttribute(oxmrn, CIB_ATT_CONTACT_ADDR, ca);
            
        }

    public void
    stop()
        {
                //
                // TO DO: unexport service
                //
        }

    public OXManagerCell()
        {
            this.homeHandlesById = new Hashtable();
        }

    public OXHandle
    createHandle(org.isam.exehda.ObjectId oid)
        {
                //
                // TO DO: review this implementation, e.g. synchronization
                //

                // standard ox handle creation semantics
            OXHandle ph = super.createHandle(oid);
            
                // additionally, catch home handle creation requests
                // if local cell is the home cell of the given ox
            if ( isHomeCell(oid) ) {
                    // create the home handle
                createHomeHandle(oid, ph);
            }

            return ph;
        }
    
        /**
         * If this manager is the home cell manager of the given OX, accepts the
         * ownership of the given master OX handle, returning a handle suitable to
         * be used in a subsequent call to requestOwnership(), probably a proxy.
         *
         * Otherwise, forwards the request to the home cell manager of the given OX.
         *
         * @param ox a <code>OXHandle</code> whose ownership is being granted
         * @return a <code>OXhandle</code>
         */
    public OXHandle
    grantOwnership(OXHandle oxh)
        throws IllegalOXHandleException 
        {
                //
                // TO DO: review this implementation, e.g. synchronization
                //

            ObjectId oxID = oxh.getId();
            
                // if local cell is the home cell of the given ox
            if ( isHomeCell(oxID) ) {
                    //
                    // if the master handle is already owned by this manager
                    //
                if ( hasOwnershipOf(oxID))  {
                        // potentially broken request, so log a warning
                    log(Sev.WARNING, "grantOwnership()",
                        "{0} trying to grantOwnership() on already owned {1}",
                        getClientHost(),
                        oxh);
                        // nothing to do
                    return oxh;
                }
                    //
                    // a local client trying to grant
                    //
                else if ( isLocalAccess() ) {
                        // should never happen because local clients just have proxies
                    log(Sev.WARNING, "grantOwnership()",
                        "Local client trying to grantOwnership() on {0}",
                        oxh);
                        //
                        // XXX: do nothing. Should we accept this grant from a local
                        // client if a master handle is provided?
                        //
                    return oxh;
                }
                    //
                    // a remote manager is accessing this manager...
                    //
                else {
                        // check the provided handle type is valid
                    if ( oxh instanceof MasterOXHandle ) {
                            // get the respective home handle
                        HomeOXHandle hh = (HomeOXHandle) homeHandlesById.get(oxID);

                        log(Sev.WARNING, "grantOwnership()",
                            "hh={0}, hh.owner={1}, caller={2}", hh, hh.getOwner(), getClientHost());
                        
                            // check if the caller match the current owner
                        if ( hh.getOwner().equals(getClientHost()) ) {
                                // ok, proceed updating internal master handles' table
                            OXHandle ph = super.grantOwnership(oxh);
                                // update the respective home handle
                            hh.setOwner(LOCAL_HOST);

                                //
                                // XXX: home handle key not used
                                //
//                             hh.setKey(ph.getKey());
                            
                                // return the created proxy
                            return ph;
                        }
                        else {
                                // abort, only the owner is allowed to grant
                            log(Sev.WARNING, "grantOwnership()",
                                "{0} client trying to grantOwnership() on {1} "
                                +"but it does not match the registered owner {2}",
                                getClientHost(),
                                oxh,
                                hh.getOwner());
                                // raise an exception on the caller
                            throw new IllegalOXHandleException("Not the owner!"); 
                        }
                    }
                    else {
                            // log the unexpected condition
                        log(Sev.WARNING, "grantOwnership()",
                            "client {0} trying to grantOwnership() on {1}",
                            getServiceWorb().getClientHost(),
                            oxh);
                            // raise an exception on the caller
                        throw new IllegalOXHandleException(
                            "Expected a MasterOXHandle but found "+oxh); 
                    }
                }
            }
                // otherwise (not the home cell)
            else {
                return remoteHomeGrantOwnership(oxh);
            }
        }

    
        /**
         * If this is the home cell manager of the given OX and owns its master handle,
         * returns the master handle if the caller provides the proper access key,
         * updating the home handle accordingly. Otherwise, if its not the home cell
         * manager, forwards the request to the proper manager.
         *
         * @param ox a <code>OXHandle</code> whose ownership is beign requested
         * @return a <code>OXHandle</code> value
         */
    public OXHandle
    requestOwnership(OXHandle oxh)
        throws IllegalOXHandleException 
        {
                //
                // TO DO: review this implementation, e.g. synchronization,
                // and master.setOwner()
                //

            ObjectId oxID = oxh.getId();

                // if local cell is the home cell of the given ox
            if ( isHomeCell(oxID) ) {
                    // if the ownership of the given master handle belongs to this master
                if ( hasOwnershipOf(oxID) ) {
                        // if the caller is not a remote manager
                    if ( isLocalAccess() ) {
                            // nothing to do
                        log(Sev.DEBUG, "requestOwnership()", "is local access");
                        return oxh;
                    }
                    else {
                        log(Sev.DEBUG, "requestOwnership()",
                            " REMOTE client {0} requesting onwership on {1}",
                            getClientHost(), oxID);
                        
                            // otherwise, if operation is allowed
                        OXHandle mh = super.requestOwnership(oxh);
                            // update the respective home handle
                        HomeOXHandle hh = (HomeOXHandle) homeHandlesById.get(oxID);

                        hh.setOwner(getClientHost());
                            // return the master handle
                        return mh;
                    }
                }
                else {
                        // oops, should not happen, log an warning
                    log(Sev.WARNING, "requestOwnership()",
                        "Host {0} requested ownership on OX {1} but master handle is not here",
                        getClientHost(),
                        oxh);
                        // silently return the same handle
                    return oxh;
                }
            }
            else {
                if ( isLocalAccess() ) {
                        // otherwise, forwards the request to the remote home cell
                    return remoteHomeRequestOwnership(oxh);
                }
                else {
                    throw new NullPointerException("Pass-throught request-ownership not enabled for remote clients");
                }
            }        
        }
    
    
        /**
         * If this manager owns the master handle of the given ox, do a local update of
         * the defined attributes set accordingly to the operations defined in ops.
         *
         * @param oxh a proxy handle of the target ox
         * @param atts set of attributes to be updated
         * @param ops per-attribute merge operation
         * @return a proxy handle with the defined attribute set updated.
         */
    public OXHandle
    updateAttributes(OXHandle oxh, String[] atts, byte[] ops)
        throws IllegalOXHandleException
        {
                //
                // TO DO: review this implementation, e.g. synchronization
                //
            
            log(Sev.DEBUG, "updateAttributes()", "oxh={0}", oxh);

            ObjectId oxID = oxh.getId();

                // if the master handle is owned by this manager
            if ( hasOwnershipOf(oxID) ) {
                return super.updateAttributes(oxh, atts, ops);                
            }
                // master handle is not here 
            else {
                    // if this is the home cell of the given ox
                if ( isHomeCell(oxID) ) {
                        // forward the request to the known current owner
                    oxh = remoteOwnerUpdateAttributes(oxh, atts, ops);
                }
                    // otherwise, we don't know who is the current owner
                else {
                        // forward the request to the home cell's manager
                    oxh = remoteHomeUpdateAttributes(oxh, atts, ops);
                }
                    // check the type of the returned handle
                if ( oxh instanceof MasterOXHandle ) {
                        //
                        // TO DO: check the type of the returned handle; if a master handle
                        // was returned it may require the ownership stuff to be updated.
                        //
                }

                return oxh;
            }
        }

        ///////////////////////////////////////////////////////////////////////
        // PROTECTED STUFF
        ///////////////////////////////////////////////////////////////////////

        /**
         * Implements abstract method defined in the super-class.
         *
         * @param oid an <code>ObjectId</code> value
         * @param ph an <code>OXHandle</code> value
         */
    protected final void
    createHomeHandle(ObjectId oid, OXHandle ph)
        {
                // if home handle has not yet been created
            if ( ! homeHandlesById.containsKey(oid) ) {
                
                log(Sev.DEBUG, "createHomeHandle()",
                    "creating home OXHandle for ox={0} h.key={1}", oid, ph.getKey());
                
                    // create it
                HomeOXHandle hh = new HomeOXHandle(oid);

                    // setup owner
                hh.setOwner(getClientHost());
                
                    //
                    // XXX: home handle key not used
                    //
//                 hh.setKey(ph.getKey());
                
                homeHandlesById.put(oid, hh);
            }
        }
    
        ///////////////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        ///////////////////////////////////////////////////////////////////////
    
        /**
         * Returns a reference to the local WORB service
         *
         * @return a <code>Worb</code> value
         */
    private final Worb
    getServiceWorb()
        {
            return (Worb) Exehda.getService(Worb.SERVICE_NAME);
        }

    
        /**
         * Returns a reference to the local Cell Information Base  service
         *
         * @return a <code>CellInformationBase</code> value
         */
    private final CellInformationBase
    getServiceCib()
        {
            return (CellInformationBase) Exehda.getService(
                CellInformationBase.SERVICE_NAME);
        }

    
        /**
         * Returns a reference to the OX Manager running at the specified host.
         *
         * @param host the target host
         * @return a <code>Worb</code> value
         */
    private final OXManager
    getRemoteOXManager( HostId host )
        {
                //
                // XXX: is it safe?
                //
            String url = "worb://"+host.toExternalForm()+"/"+OXManager.SERVICE_NAME;

            return (OXManager) getServiceWorb().lookupService(url, OXManager.class);
        }

    
        /**
         * Returns a reference to the cellular instance of the WORB service running at
         * the specified cell.
         *
         * @param cell the target cell
         * @return a <code>Worb</code> value
         */
    private final OXManager
    getRemoteOXManager( CellId cell )
        {
            ResourceName oxmrn = new ResourceName("service",
                                                  OXManager.SERVICE_NAME,
                                                  cell.toNameSpace());

            String url = getServiceCib().getAttribute(oxmrn, CIB_ATT_CONTACT_ADDR);

                // check contact address
            if ( url != null ) {
                return (OXManager) getServiceWorb().lookupService(url, OXManager.class);
            }
            else {
                throw new NullPointerException("service contact address undefined for cell="+cell);
            }
        }
    
        /**
         * Detects whether the current method invocation is a local or a remote acess.
         *
         * @return true if its a local access, false if its coming through the Worb
         * service.
         */
    private final boolean
    isLocalAccess()
        {
                //
                // FIX ME: broken for chained calls, e.g.
                // Remote Host->Executor.activateObject()->OXManager.requestOwnership();
                //
            return getServiceWorb().getClientHost() == null;
        }

    
        /**
         * Returns the ID of the host invoking the method.
         *
         * @return a <code>HostID</code> value
         */
    private final HostId
    getClientHost()
        {
            HostId ch = getServiceWorb().getClientHost();
            
            if ( ch == null ) {
                ch = LOCAL_HOST;
            }

            return ch;
        }
    
        /**
         * Returns the home cell of the given OX.
         *
         * @param oxID an <code>ObjectId</code> value
         * @return a <code>CellId</code> value
         */
    private final CellId
    getHomeCell(ObjectId oxID)
        {
            return oxID.getHost().getCell();
        }
    
        /**
         * Checks whether this cell is the home cell of the given OX
         *
         * @param oxID an <code>ObjectId</code> value
         * @return a <code>boolean</code> value
         */
    private final boolean
    isHomeCell(ObjectId oxID)
        {
            return LOCAL_CELL.equals(getHomeCell(oxID));
        }

    
        /**
         * Checks whether this host is the home node of the given OX
         *
         * @param oxID an <code>ObjectId</code> value
         * @return a <code>boolean</code> value
         */
    private final boolean
    isHomeNode(ObjectId oxID)
        {
            return LOCAL_HOST.equals(oxID.getHost());
        }
    
        /**
         * Forwards the grantOwnership() request to the ox's home cell (remote)
         *
         * @param oxh a master handle
         * @return an <code>OXHandle</code> value
         * @exception IllegalOXHandleException if an error occurs
         */
    private final OXHandle
    remoteHomeGrantOwnership(OXHandle oxh)
        throws IllegalOXHandleException
        {
                // if client is not a remote manager proxying through this manager
            if ( isLocalAccess() ) {
                    // replace the proxy by the real master handle
                oxh = super.requestOwnership(oxh);
            }
                // sanity check
            if ( oxh instanceof MasterOXHandle ) {
                    // locate the manager at the home cell
                OXManager homeManager = getRemoteOXManager(getHomeCell(oxh.getId()));
                    // forward the request to the remote manager
                return homeManager.grantOwnership(oxh);

                    //
                    // TO DO: check the type of the returned handle;
                    // Ownership may have been not accepted.
                    //
            }
            else {
                    // log the unexpected condition
                log(Sev.WARNING, "remoteHomeGrantOwnership()",
                    "{0} client trying to grantOwnership() on {1}",
                    getServiceWorb().getClientHost(),
                    oxh);
                    // raise an exception on the caller
                throw new IllegalOXHandleException(
                    "Expected a MasterOXHandle but found "+oxh); 
            }
        }

        /**
         * Forwards the requestOwnership() call to the home cell.
         *
         * @param oxh an <code>OXHandle</code> value
         * @return an <code>OXHandle</code> value
         */
    private final OXHandle
    remoteHomeRequestOwnership(OXHandle oxh)
        {
                // locate the manager at the home cell
            OXManager homeManager = getRemoteOXManager(getHomeCell(oxh.getId()));
                // forward the request to the remote manager
            oxh = homeManager.requestOwnership(oxh);
                // ensure a master handle has been returned
            if ( oxh instanceof MasterOXHandle ) {
                    // if the caller is not a node manager proxying through this manager
                if ( isLocalAccess() ) {
                        // self-grants the master handle, replacing it by a proxy 
                    oxh = super.grantOwnership(oxh);
                }
            }
            else {
                    // should never happen, but...
                log(Sev.WARNING, "requestOwnership()",
                    "Expected a master handle but home cell manager returned {0}", oxh);
            }
            
            return oxh;
        }

    
        /**
         * Forwards the updateAttributes() call to the home cell.
         *
         * @return an <code>OXHandle</code> value
         */
    private final OXHandle
    remoteHomeUpdateAttributes(OXHandle oxh, String[] atts, byte[] ops)
        {
                // locate the manager at the home cell
            OXManager homeManager = getRemoteOXManager(getHomeCell(oxh.getId()));
                // forward the request to the remote manager
            return homeManager.updateAttributes(oxh, atts, ops);
        }


        /**
         * Forwards the updateAttributes() call to the remote owner.
         *
         * @return an <code>OXHandle</code> value
         */
    private final OXHandle
    remoteOwnerUpdateAttributes(OXHandle oxh, String[] atts, byte[] ops)
        {
            HostId ownerHost = null;
            
                // if the home handle has already been created
            HomeOXHandle hh = (HomeOXHandle) homeHandlesById.get(oxh.getId());

            if ( hh != null ) {
                    // ask it who is the current owner
                ownerHost = hh.getOwner();
            }
            else {
                    // otherwise, guess the probable owner from the ID
                ownerHost = oxh.getId().getHost();

                log(Sev.WARNING, "remoteOwnerUpdateAttributes()",
                    "home handle not yet created for {0}", oxh);
            }
            log(Sev.DEBUG, "remoteOwnerUpdateAttributes()", "forwarding request to host={0}", ownerHost);
            
                // locate the manager on that node
            OXManager ownerManager = getRemoteOXManager(ownerHost);
                // forward the request to the remote manager
            return ownerManager.updateAttributes(oxh, atts, ops);
        }

    
//         /**
//          * Returns the contact address of the base OXManger of the given cell.
//          *
//          * @param cell a <code>String</code> with the name of the cell whose OXM contact
//          * address is to be found.
//          * @return a <code>String</code> with the contact address of the desired OXM
//          */
//     private String
//     getOXMContactAddress(String cell)
//         {
//             Worb worb = (Worb) Exehda.getService(Worb.SERVICE_NAME);
//             String cibCellUri = Exehda.getServiceProperty(CellInformationBase.SERVICE_NAME,
//                                                           PROP_CIB_URI,
//                                                           "//localhost:1980/cib");

//             CellInformationBase cibCell = (CellInformationBase)
//                 worb.lookupService(cibCellUri, CellInformationBase.class);
//             ResourceName[] results =
//                 cibCell.selectByName("oxm",new NameSpace("cell",cell,null),-1);
            
//             if (results.length > 0) {
//                 return(cibCell.getAttribute(results[0],"contactAddress"));
//             }
//             return(null);
//         }


    

        /**
         * Table of home handles managed by this instance, indexed by OX ID.
         *
         */
    private Hashtable homeHandlesById;
}
