// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.oxm;

import org.isam.exehda.Exehda;
import org.isam.exehda.HostId;
import org.isam.exehda.ObjectId;
import org.isam.exehda.CellId;
import org.isam.exehda.ApplicationId;
import org.isam.exehda.services.AbstractService;
import org.isam.exehda.services.Logger.Sev;
import org.isam.exehda.services.OXManager;
import org.isam.exehda.services.OXManager.OXHandle;
import org.isam.exehda.services.OXManager.MasterOXHandle;
import org.isam.exehda.services.OXManager.HomeOXHandle;
import org.isam.exehda.services.OXManager.ProxyOXHandle;
import org.isam.exehda.services.OXManager.IllegalOXHandleException;
import org.isam.exehda.services.CellInformationBase;
import org.isam.exehda.ResourceName;
import org.isam.exehda.ResourceName.NameComponent;
import org.isam.exehda.ResourceName.NameSpace;
import org.isam.exehda.services.Worb;
import java.util.Hashtable;



/**
 * OX Manager node instance.
 *
 * @author frainer, last modified by $Author: lucc $
 * @version $Date: 2005-05-04 10:24:51 -0300 (Wed, 04 May 2005) $ $Revision: 843 $
 */
public class OXManagerNode
    extends AbstractOXManager
{    
    final static String PROP_OX_URI = "contactAddress";
    
    public void
    start() throws Exception
        {
                // export the service to the WORB
            getServiceWorb().exportService(this,
                                           OXManager.class,
                                           OXManager.SERVICE_NAME);

                //
                // FIX ME: need to register in the CIB??
                //
        }

    public void
    stop()
        {
                //
                // TO DO: unexport service
                //
        }
    
    
        /**
         * Creates a new instance of OXManagerClientImpl
         */
    public OXManagerNode()
        {
            this.pendingHomeHandles = new Hashtable();
            this.oxmCell = null;
        }

    
        /**
         * Moves the master handle towards the cell's OX manager and returns a handle
         * suitable to be used in a subsequent call to request ownership.
         *
         * <p>Curretly, raises an exception if the given ox is not owned by this manager.
         *
         * @param ox a <code>OXHandle</code> whose ownership is being granted
         * @return a <code>OXhandle</code> suitable to be used in a subsequent call to
         * requestOwnership()
         */
    public OXHandle
    grantOwnership(OXHandle oxh)
        throws IllegalOXHandleException 
        {
                //
                // TO DO: review this implementation, e.g synchronization
                //

            ObjectId oxID = oxh.getId();

                // if ownership of the given master handle already belongs to this manager
            if ( hasOwnershipOf(oxID) ) {
                    // check if the given handle 'knows' the secret key
                OXHandle masterHandle = super.requestOwnership(oxh);
                    // if handle ownership is effectivelly changing
                if ( masterHandle instanceof MasterOXHandle ) {
                        // commit creation of home handle if still pending
                    commitCreateHomeHandle(oxh.getId());
                        // update cell's ox manager
                    return forwardGrantOwnership(masterHandle);
                }
                else {
                        //
                        // XXX: "to throw exception or to be silent": that's the question.
                        //
                    throw new IllegalOXHandleException(
                        "Trying to grant onwership on un-owned ox: "+oxh.getId());
                }
            }
            else {
                    //
                    // TO DO: check if the caller is a trusted manager or is someone
                    // trying to forge a MasterHandle
                    //
                
                    //
                    // FIX ME: what if the given ox is not native to the local cell?
                    //
                
                    // update master handles table (insert handle)
                return  super.grantOwnership(oxh);
            }
        }

    
        /**
         * Requests to the local cell manager the master handle of the given ox to be
         * moved towards the local host. If the master handle is already owned by this
         * manager, it is a no-op call.

         * <p> The local cell manager will either return the master handle itself or a
         * proxy suitable for grantOwnership(). In the former, this manager will keep the
         * master handle with itself and return a proxy to calller. In the later, the
         * proxy returned by the cell manager will be returned to the caller.
         *
         * @return Always a proxy handle
         */
    public OXHandle
    requestOwnership(OXHandle oxh)
        throws IllegalOXHandleException 
        {
                //
                // TO DO: review this implementation, e.g synchronization
                //

            ObjectId oxID = oxh.getId();

                // if ownership of the given master handle already belongs to this manager
            if ( hasOwnershipOf(oxID) ) {
                    // nothing to do
                return oxh;
            }
            else {
                    // forward request to the cell's manager
                oxh = forwardRequestOwnership(oxh);
                    // check if the master handle has be returned
                if ( oxh instanceof MasterOXHandle ) {
                        // if so, keep the master handle (self-grants) and return a proxy
                    oxh = super.grantOwnership(oxh);
                }
                    // return a proxy to the client
                return oxh;
            }
        }
    
    
        /**
         * OXhandle updates the attributes of the given ox. The attributes to be updated
         * are defined in a String array. What type of update is to be done (read or write)
         * is defined in an byte array.
         *
         * @param ox a <code>OXHanlde</code> whose attributes are to be updated
         * @param att an <code>String[]</code> defining what attributes are to be updated
         * @return a <code>OXHandle</code> that is the OXHandle passed as parameter with
         * it's attributes updated
         */
    public OXHandle
    updateAttributes(OXHandle oxh, String[] atts, byte[] ops)
        throws IllegalOXHandleException 
        {
                //
                // TO DO: review this implementation, e.g. synchronization
                //
            
            log(Sev.DEBUG, "updateAttributes()", "oxh={0}", oxh);

            ObjectId oxID = oxh.getId();

                // if the master handle is owned by this manager
            if ( hasOwnershipOf(oxID) ) {
                return super.updateAttributes(oxh, atts, ops);                
            }
            else {
                    // forward request to the cell's manager
                return forwardUpdateAttributes(oxh, atts, ops);
            }
        }

        //////////////////////////////////////////////////////////////
        // PROTECTED STUFF
        //////////////////////////////////////////////////////////////

        /**
         * Actually, just schedules the creation of the home handle
         *
         * @param oid an <code>ObjectId</code> value
         */
    protected final  void
    createHomeHandle(ObjectId oid, OXHandle ph)
        {
            log(Sev.DEBUG, "createHomeHandle()",
                "scheduling home handle creation for ox={0}", oid);
            
                // schedule creation of the home handle
            pendingHomeHandles.put(oid,oid);
        }

        //////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        //////////////////////////////////////////////////////////////

        /**
         * Returns a reference to the local WORB service
         *
         * @return a <code>Worb</code> value
         */
    private final Worb
    getServiceWorb()
        {
            return (Worb) Exehda.getService(Worb.SERVICE_NAME);
        }

    
        /**
         * Returns a reference to the local Cell Information Base  service
         *
         * @return a <code>CellInformationBase</code> value
         */
    private final CellInformationBase
    getServiceCib()
        {
            return (CellInformationBase) Exehda.getService(
                CellInformationBase.SERVICE_NAME);
        }
    
        /**
         * Commits the creation of home handle (if still pending)
         *
         * @param oid an <code>ObjectId</code> value
         */
    private final void
    commitCreateHomeHandle(ObjectId oxID)
        {
            if ( pendingHomeHandles.containsKey(oxID) ) {
                log(Sev.DEBUG, "commitCreateHomeHandle()",
                    "commiting home handle creation for ox={0}", oxID);
                
                getCellularManager().createHandle(oxID);
                    //
                    // FIX ME: perhaps, should keep the returned handle for using later
                    // in grant ownership?
                    //
                
                pendingHomeHandles.remove(oxID);
            }
        }
    
        /**
         * Forwards a requestOwnership() call the the local cell's OX Manager.
         *
         * @param oxh an <code>OXHandle</code> value
         * @return an <code>OXHandle</code> value
         */
    private final OXHandle
    forwardRequestOwnership(OXHandle oxh)
        {
                //
                // XXX: currently need to forward ownership requests directly to the home
                // cell
                //
            return getHomeCellularManager(oxh).requestOwnership(oxh);
        }

        /**
         * Forwards a grantOwnership() call the the local cell's OX Manager.
         *
         * @param oxh an <code>OXHandle</code> value
         * @return an <code>OXHandle</code> value
         */
    private final OXHandle
    forwardGrantOwnership(OXHandle oxh)
        {
                //
                // XXX: curretly need to forward ownership requests directly to the home
                // cell
                //
            return getHomeCellularManager(oxh).grantOwnership(oxh);
        }

        /**
         * Forwards a updateAttributes() call the the local cell's OX Manager.
         *
         * @param oxh an <code>OXHandle</code> value
         * @param atts a <code>String[]</code> value
         * @param ops a <code>byte[]</code> value
         * @return an <code>OXHandle</code> value
         */
    private final OXHandle
    forwardUpdateAttributes(OXHandle oxh, String[] atts, byte[] ops)
        {
            log(Sev.DEBUG, "forwardUpdateAttributes()",
                "forwarding updateAttributes() request to the cellular manager");
            return getCellularManager().updateAttributes(oxh, atts, ops);
        }
    
        /**
         * Returns a reference to the cellular instance of the OX Manager service at
         * the local cell.
         *
         * @param cell the target cell
         * @return a <code>Worb</code> value
         */
    private final OXManager
    getCellularManager()
        {
            if ( oxmCell == null ) {
                synchronized (this) {

                    final String CIB_ATT_CONTACT_ADDR = "contactAddress";
                    final CellId LOCAL_CELL = HostId.getLocalHost().getCell();

                    ResourceName oxmrn = new ResourceName("service",
                                                          OXManager.SERVICE_NAME,
                                                          LOCAL_CELL.toNameSpace());

                    String url = getServiceCib().getAttribute(oxmrn,
                                                              CIB_ATT_CONTACT_ADDR);

                        // check contact address
                    if ( url != null ) {
                        this.oxmCell = (OXManager)
                            getServiceWorb().lookupService(url, OXManager.class);
                    }
                    else {
                        throw new NullPointerException("service contact address is undefined");
                    }
                }
            }

            return oxmCell;
        }

    
        /**
         * Returns a reference to the cellular instance of the OX Manager service at the
         * home cell of the given OX.
         *
         * @param h an <code>OXHandle</code> value
         * @return an <code>OXManager</code> value
         */
    private final OXManager
    getHomeCellularManager(OXHandle h)
        {
            final String CIB_ATT_CONTACT_ADDR = "contactAddress";

                // XXX: assume id.host.cell
            NameSpace nsHomeCell = h.getId().toNameSpace().getParent().getParent();
            
            ResourceName oxmrn = new ResourceName("service",
                                                  OXManager.SERVICE_NAME,
                                                  nsHomeCell);

            String url = getServiceCib().getAttribute(oxmrn,
                                                      CIB_ATT_CONTACT_ADDR);

                // check contact address
            if ( url != null ) {
                return (OXManager) getServiceWorb().lookupService(url, OXManager.class);
            }
            else {
                throw new NullPointerException("service contact address is undefined");
            }
        }
    
    
        /**
         * Reference to the cellular instance (this cell) of OXManager
         */
    private OXManager oxmCell;

        /**
         * Set of ox ID whose homeHandle creation is pending (delayed)
         */
    private Hashtable pendingHomeHandles;

    
//     private long timeBetweenUpdates;
}
