// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.worb;

import java.net.URL;
import java.net.URLConnection;
import java.net.ProtocolException;
import java.net.UnknownServiceException;
import java.net.HttpURLConnection;
import java.net.Socket;
import java.io.IOException;
import java.io.EOFException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.BufferedInputStream;


/**
 * "worb:" URL connection.
 *
 * <p>Worb URLs are the form: 
 * <pre>
 *      worb:<transport>://<targethost>/<service>
 * </pre>
 *
 * <p>Where <targethost> follows a notation that is consistent with the specificed transport.
 *
 * <p>Curretly, the only supported transport is 'tcp'. For this transport the target host
 * is described using the standard internet URL notation, i.e., ip:port.
 *
 * <p>Another planned transport to the suported is the 'ssl' transport.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-04-12 16:54:50 -0300 (Tue, 12 Apr 2005) $ $Revision: 743 $
 */
public class WorbURLConnection
    extends URLConnection
    implements WorbProtocolConstants
//
// TO DO: extends HttpURLConnection
//
{
    public WorbURLConnection(URL url)
        {
            super(url);

//             setDoOutput(true);

            this.sock = null;
            this.outs = null;

            if ( ! "worb".equals(this.url.getProtocol()) ) {
                throw new IllegalArgumentException(
                    "Unsupported protocol: "+ this.url.getProtocol());
            }

//             System.out.println("WorbURLConnection: url="+url);
//             System.out.println("url.file="+url.getFile());
//             System.out.println("url.host="+url.getHost());
//             System.out.println("url.port="+url.getPort());
//             System.out.println("url.ref="+url.getRef());
//             System.out.println("url.proto="+url.getProtocol());
        }

    public void
    connect()
        throws IOException
        {
            if ( ! connected ) {
                    // open socket
                this.sock = createSocket(url.getHost(), url.getPort());

                    // set connection state
                this.connected = true;
            }
        }

    public void
    setRequestProperty( String k, String v )
        {
            if ( keysMatch(WORB_REQ_PROP_OPNUM, k) ) {
                this.rpOpnum = v;
            }
            else if ( keysMatch(WORB_REQ_PROP_OP_HASH, k)) {
                this.rpOpHash = v;
            }
            else if ( keysMatch(WORB_REQ_PROP_OP_SESSION, k) ) {
                this.rpOpSession = v;
            }
            else if ( keysMatch(WORB_REQ_PROP_OP_RESUME, k) ) {
                this.rpOpResume = v;
            }
            else if ( keysMatch(WORB_REQ_PROP_PROTO_VERSION, k)) {
                this.rpProtoVersion = v;
            }
        }

    public String
    getRequestProperty( String k )
        {
            if ( keysMatch(WORB_REQ_PROP_OPNUM, k) ) {
                return this.rpOpnum;
            }
            else if ( keysMatch(WORB_REQ_PROP_OP_HASH, k)) {
                return this.rpOpHash;
            }
            else if ( keysMatch(WORB_REQ_PROP_OP_SESSION, k) ) {
                return this.rpOpSession;
            }
            else if ( keysMatch(WORB_REQ_PROP_OP_RESUME, k) ) {
                return this.rpOpResume;
            }
            else if ( keysMatch(WORB_REQ_PROP_PROTO_VERSION, k)) {
                return this.rpProtoVersion;
            }
            else {
                return null;
            }
        }

    public InputStream
    getInputStream()
        throws IOException
        {
            connect();

            OutputStream out = sock.getOutputStream();
            
                // commit output stream data
            outs.flush();
            byte[] reqData = outs.toByteArray();
                // compute req content length
            this.rpContentLength = reqData.length;
                // write out protocol headers
            writeHeaders(out);
                // write out the request itself
            out.write(reqData);
            out.flush();
            
                // read result code
            return checkResponseCode(sock.getInputStream());
        }

    public OutputStream
    getOutputStream()
        throws IOException
        {
            if ( doOutput ) {
                if ( outs == null ) {
                    outs = new ByteArrayOutputStream(BUFF_SIZE);
                }

                return outs;
            }
            else {
                throw new ProtocolException("Try to write to a read-only stream.");
            }
        }

    public String
    getContentType()
        {
            return this.rpContentType;
        }

    public int
    getContentLength()
        {
            return this.rpContentLength;
        }

        //////////////////////////////////////////////////
        // PRIVATE STUFF
        //////////////////////////////////////////////////

    
        /**
         * Creates a Socket connected to the specified contact address.
         *
         * @param ca a <code>String</code> value
         * @param port an <code>int</code> value
         * @return a <code>Socket</code> value
         * @exception IOException if an error occurs
         */
    private final Socket
    createSocket( String ca, int port ) throws IOException
        {
                //
                // TO DO: select alternative Socket implementations based on the
                // specified transport. E.g., use WorbSecureSocket if transport is 'ssl'.
                //

                // if a supported transport was specified, create a WorbSocket
            if ( ca.startsWith("tcp:") ) {
                return new WorbSocket(ca.substring(4), port);
            }
                // otherwise (transport was not specified in the contact address)
            else {
                    //
                    // XXX: fall-back to stardard Java sockets (just for debugging purposes)
                    //
//             return new WorbSocket(ca, port);
//                throw new ProtocolException("Unsupported transport protocol:"+ca);
                return new Socket(ca, port);                
            }
        }

    
        /**
         * Write Worb protocol headers to the given output stream.
         *
         * @param out an <code>OutputStream</code> value
         * @exception IOException if an error occurs
         */
    private final void
    writeHeaders( OutputStream out ) throws IOException
        {
                //
                // write headers
                //
            final String CRLF = "\r\n";
            final String PROP_SEP = ": ";
            final String HTTP_CHARSET = "ISO8859-1";
            
            StringBuffer hbuf = new StringBuffer(512);

                //
                // General http headers;
                //
            
                // resource being accessed
            hbuf.append("POST ")
                .append(this.url.getFile())
                .append(' ')
                .append(HTTP_VERSION_1_0)
                .append(CRLF);
            
                // request's encoding
            hbuf.append(HTTP_PROP_CONTENT_TYPE)
                .append(PROP_SEP).append(rpContentType).append(CRLF);
            
                // how many bytes in the request
            hbuf.append(HTTP_PROP_CONTENT_LENGTH)
                .append(PROP_SEP).append(rpContentLength).append(CRLF);
            
                //
                // worb-specific headers
                //
            
            if ( rpOpnum != null ) {
                hbuf.append(WORB_REQ_PROP_OPNUM)
                    .append(PROP_SEP).append(rpOpnum).append(CRLF);
            }

            if ( rpOpHash != null ) {
                hbuf.append(WORB_REQ_PROP_OP_HASH)
                    .append(PROP_SEP).append(rpOpHash).append(CRLF);
            }

            if ( rpOpSession != null ) {
                hbuf.append(WORB_REQ_PROP_OP_SESSION)
                    .append(PROP_SEP).append(rpOpSession).append(CRLF);
            }

            if ( rpOpResume != null ) {
                hbuf.append(WORB_REQ_PROP_OP_RESUME)
                    .append(PROP_SEP).append(rpOpResume).append(CRLF);
            }

                // end-of-header (empty line)
            hbuf.append(CRLF);

            out.write(hbuf.toString().getBytes(HTTP_CHARSET));
        }

        /**
         * Parses (consumes) HTTP response headers. Up-on a successfull call, return an
         * input stream from which the resquest result could be read, otherwise aborts
         * the request (closing the underlaying socket) by raising an exception
         * describing the error occured.
         *
         * @param in an <code>InputStream</code> value
         * @return an <code>InputStream</code> value
         * @exception IOException if an error occurs
         */
    private final InputStream
    checkResponseCode( InputStream in ) throws IOException
        {
            final int BUFF_SIZE = 512;
            
                // wrap it in a buffered stream
            in = new BufferedInputStream(in);
            
                //
                // read HTTP response code
                //

            int c = -1;
            int pos = 0;
            byte[] buff = new byte[BUFF_SIZE];
            int respCode = -1;
            
            while ((c=in.read()) != -1) {

                    // if read a CR
                if ( '\r' == c ) {
                        // expecting a LF
                    if ( '\n' == in.read() ) {
                            // check for end-of-header mark (empty line)
                        if ( pos == 0 ) {
                                // stop parsing
                            return in;
                        }
                        else {
                            String line = new String(buff, 1, pos);
                                // if first line
                            if ( respCode < 0 ) {
                                respCode = parseResponseCode(line);
                            }
                            else {
                                parseResponseHeader(line);                                
                            }
                                // reset buffer
                            pos = 0;
                        }
                    }
                    else {
                        buff[++pos] = (byte) c;
                            // shutdown the connection
                        try { sock.close(); } catch (IOException ioe) {}

                        throw new ProtocolException(
                            "Malformed response: "+ new String(buff, 1, pos));
                    }
                }
                else {
                    buff[++pos] = (byte) c;
                }
            }
                // shutdown the connection
            try { sock.close(); } catch (IOException ioe) {}
            throw new EOFException("reached end-of-stream while parsing response headers");
        }

        /**
         * Checks whether the given keys are equal. It used a smart (fail-fast) key
         * comparing strategy.
         *
         * @param k1 a <code>String</code> value
         * @param k2 a <code>String</code> value
         * @return a <code>boolean</code> value
         */
    private final boolean
    keysMatch( String k1, String k2 )
        {
                //
                // XXX: use reverse matching since it would be faster given the keys
                // usually share the same prefix
                //

            return ( k1.length() == k2.length() ) ? k1.endsWith(k2) : false;
        }

        /**
         * Parses the first HTTP response line (HTTP/<ver> <code> <msg>.
         *
         * @param line a <code>String</code> value
         * @return an <code>int</code> value
         */
    private final int
    parseResponseCode(String line)
        throws IOException
        {
                //
                // FIX ME: hardcoded string positions, need to check against the HTTP
                // spec
                //
            
                // get the protocol version 
            String pv = line.substring(0, 8);
                //
                // XXX: check the protocol version is supported. Do
                // we need this?
                //
            if ( HTTP_VERSION_1_0.equals(pv) || HTTP_VERSION_1_1.equals(pv) ) {
                    // parse the response code
                int respCode = Integer.parseInt(line.substring(9, 12));
                
                    // check req ok
                if (  respCode >= HttpURLConnection.HTTP_OK
                      && respCode < HttpURLConnection.HTTP_OK+99 ) {
                    return respCode;
                }
                else {
                        // shutdown the connection / release resources
                    try { sock.close(); } catch (IOException ioe) {}
                        // signal failure to the caller
                    throw new ProtocolException(line.substring(9));
                }
            }
            else {
                    // shutdown the connection / release resources
                try { sock.close(); } catch (IOException ioe) {}
                    // and raise an exception
                throw new ProtocolException(
                    "Unsupported protocol version: <"+pv+">");
            }
        }

    
        /**
         * Parses the given HTTP response reader (<prop>: <value>. Filling the respective
         * object attributes. Headers that are senseless concerning to a worb call are
         * silently ignored.
         *
         * @param line a <code>String</code> value
         * @exception IOException if an error occurs
         */
    private final void
    parseResponseHeader(String line)
        throws IOException
        {                                            
            final String HEADER_CTYPE = HTTP_PROP_CONTENT_TYPE + ": ";
            final String HEADER_CLEN = HTTP_PROP_CONTENT_LENGTH + ": ";
                                
                // check for content type
            if ( line.startsWith(HEADER_CTYPE) ) {
                this.rpContentType = line.substring(HEADER_CTYPE.length());
            }
                // check for content length
            else if ( line.startsWith(HEADER_CLEN) ) {
                this.rpContentLength = Integer.parseInt(
                    line.substring(HEADER_CLEN.length()));
            }
            else {
                    // ignore other headers
            }
        }
    
    
    private static final int BUFF_SIZE = 4096;
    private static final String HTTP_VERSION_1_0 = "HTTP/1.0";
    private static final String HTTP_VERSION_1_1 = "HTTP/1.1";
    private static final String HTTP_PROP_CONTENT_TYPE = "Content-Type";
    private static final String HTTP_PROP_CONTENT_LENGTH = "Content-Length";
            
    private Socket sock;
    private ByteArrayOutputStream outs;
    private int rpContentLength = -1;
    private String rpProtoVersion = null;
    private String rpOpnum = null;
    private String rpOpHash = null;
    private String rpOpSession = null;
    private String rpOpResume = null;
    private String rpContentType = "application/x-java-serialized-object";


        /////////////////////////////////////////////////////////////
        // MAIN (only for testing pourposes)
        /////////////////////////////////////////////////////////////
    
    public static void
    main(String[] args )
        throws Exception
        {
            System.setProperty("java.protocol.handler.pkgs",
                               "org.isam.perv.protocols");

            String url = ( args.length > 0 ) ? args[0] : "worb:tcp://127.0.0.1:1980/obj/mtd";
            
            URLConnection wcon = new URL(url).openConnection();
            wcon.setDoOutput(true);

            System.out.println(WORB_REQ_PROP_OP_HASH+"="
                               + wcon.getRequestProperty(WORB_REQ_PROP_OP_HASH));
            
            wcon.setRequestProperty(WORB_REQ_PROP_OP_HASH, "AABBDD");

            System.out.println(WORB_REQ_PROP_OP_HASH+"="
                               + wcon.getRequestProperty(WORB_REQ_PROP_OP_HASH));

            OutputStream out = wcon.getOutputStream();

            out.write("teste teste teste teste\n".getBytes());
            
            InputStream in = wcon.getInputStream();

            System.out.println("Content-Length="+ wcon.getContentLength());
            System.out.println("Content-Type="+ wcon.getContentType());
            System.out.println("----------");

//             for (int c=in.read(); c > 0; c=in.read()) {
                System.out.print((char) in.read());
//             }
        }
}

