// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.primos.exec;

import org.isam.perv.IsamClassLoader;
import org.isam.exehda.Exehda;
import org.isam.exehda.ResourceName;
import org.isam.exehda.HostId;
import org.isam.exehda.ObjectId;
import org.isam.exehda.ApplicationId;
import org.isam.exehda.services.Worb;
import org.isam.exehda.services.Executor;
import org.isam.exehda.services.CellInformationBase;
import org.isam.exehda.services.ObjectSeed;
import org.isam.exehda.services.OXManager;
import org.isam.exehda.services.OXManager.OXHandle;
import org.isam.exehda.services.ObjectSeed.Activator;
import org.isam.exehda.services.ObjectSeed.MarshaledOX;
import org.isam.exehda.services.Logger.Sev;
import org.isam.exehda.services.AbstractService;
import org.isam.util.xml.XmlElement;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.io.StringReader;
import java.io.IOException;
import java.net.UnknownHostException;



/**
 * PRIMOS' Executor implementation
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-04-25 02:45:32 -0300 (Mon, 25 Apr 2005) $ $Revision: 801 $
 */
public class ExecutorImpl
    extends AbstractService
    implements Executor
{
    private final static String DFLT_SCHED_HEURISTIC = SysSchedulerSchedHeuristic.class.getName();
    private final static String DFLT_MAX_RETRY = "5";

    private final static String CIB_ATT_APP_CODEBASE = "codeBase";
    private final static String CIB_ATT_APP_LDESC = "lauchingDescriptor";

    
        /**
         * Creates a new <code>ExecutorImpl</code> instance.
         *
         */
    public ExecutorImpl()
        {
            this.appByThrGrp = new Hashtable();
            this.appById = new Hashtable();
            this.thisHost = HostId.getLocalHost();
        }
    
    public void
    start()
        throws ClassNotFoundException, InstantiationException,
        IllegalAccessException, UnknownHostException
        {
                // load parameters from profile
            String pSchedHeuristic = Exehda.getServiceProperty(
                SERVICE_NAME, "defaultHeuristic", DFLT_SCHED_HEURISTIC);
            
            String pMaxRetryies = Exehda.getServiceProperty(
                SERVICE_NAME, "maxRetryies", DFLT_MAX_RETRY);
            
                // parse params and initialize instance attributes
            this.defaultHeuristic = (SchedulingHeuristic)
                Class.forName(pSchedHeuristic).newInstance();
                
            this.maxRetryies = Integer.parseInt(pMaxRetryies);

                // export service to WORB
            Worb worb = (Worb) Exehda.getService(Worb.SERVICE_NAME);
            worb.exportService(this, Executor.class, Executor.SERVICE_NAME);

                //
                // XXX: remeber thread group of kernel threads
                //
            this.kernelThrGroup = Thread.currentThread().getThreadGroup();
        }
    
    public void
    stop() throws java.lang.Exception
        {
                // TODO: unexport service
        }

    
        /**
         * Describe <code>startApplication</code> method here.
         *
         * @param launchDescriptor a <code>XmlElement</code> value
         */
    public void
    startApplication(XmlElement launchDescriptor)
        {
            log(Sev.TRACE, "startApplication", "entering");
            try {
                ApplicationId newID = nextApplicationId();
                Application app = new Application(newID, launchDescriptor, false);
                
                log(Sev.INFO, "startApplication",
                    "Launching application codebase={0}", app.getCodeBase());

                    //
                    // XXX: register the launched application before calling start()
                    // otherwise currentApplication() would be inconsistent for a few
                    // moments
                    //
                registerApplication(app, true);

                    // start main thread
                app.start();
            }
            catch (Exception e) {
                log(Sev.ERROR, "startApplication", e);

                    //
                    // TO DO: cancel registration of the app and release allocated resources
                    //
            }
            log(Sev.TRACE, "startApplication", "leaving");
        }

    

    public ObjectId
    createObject( Class cls, Object[] params, Activator a, Object hint)
        {
            try {
                log(Sev.TRACE, "createObject()", "entering");


                if ( thisHost.equals(hint) ) {
                    return localCreateObject(cls, params, a);
                }
                else {                
                    Vector avoidedHosts = new Vector();
                    int n = maxRetryies;

                    do {
                        HostId targetHost = getHeuristic().chooseCreationHost(
                            cls.getName(), params, hint, avoidedHosts);

                        log(Sev.DEBUG, "createObject()",
                            "Selected host is {0}", targetHost);

                        try {
                            if ( targetHost != null ) {
                                return thisHost.equals(targetHost)
                                    ? localCreateObject(cls, params, a)
                                    : remoteCreateObject(cls, params, a, targetHost);
                            }
                        }
                        catch (Exception e) {
                            log(Sev.ERROR, "createObject()",
                                "failed on {0} due to exception:", targetHost);
                            log(Sev.ERROR, "createObject()", e);
                        }
                    }
                    while ( --n > 0 );

                        //
                        // TO DO: throw exception
                        // 
                }
            }            
            catch ( Exception e ) {
                log(Sev.ERROR, "createObject()", e);
            }
            finally {
                log(Sev.TRACE, "createObject()", "leaving");
            }

                //
                // TO DO: throw exception
                // 
            return null;
        }
    

    public void
    moveObject( ObjectId oxID, Object placementHint )
        {
            try {
                log(Sev.DEBUG, "moveObject()", "entering");
            
                    // where is it?
                HostId loc = getOXLocation(oxID);
                            
                    // check the ox is hosted in this node
                if ( thisHost.equals(loc) ) {

                    log(Sev.DEBUG, "moveObject()", "target OX is HERE! Deactivating it...");
                       
                        // deactivate the ox
                    MarshaledOX mox = deactivateObject(oxID);

                    log(Sev.DEBUG, "moveObject()", "OX {0} deactivated.", oxID);

                    Vector avoidedHosts = new Vector();
                    int n = maxRetryies;
                    do {
                            // select the target host for receiving it
                        HostId targetHost = getHeuristic().chooseMigrationHost(
                            oxID,  placementHint, avoidedHosts);

                        log(Sev.DEBUG, "moveObject()",
                            "Selected host is {0}", targetHost);

                        try {
                            if ( targetHost != null ) {

                                if ( thisHost.equals(targetHost) ) {
                                        // just reactivate the OX
                                    activateObject(mox);
                                }
                                else {
                                        // contact remote Executor for re-activating the OX
                                    remoteActivateObject(mox, targetHost);
                                }
                                    // exit (success)
                                return;
                            }
                            else {
                                    // scheduling heuristic had given up
                                break;
                            }
                        }
                        catch (Exception e) {
                            log(Sev.ERROR, "moveObject()",
                                "failed on {} due to exception:", targetHost);
                            log(Sev.ERROR, "createObject()", e);
                        }
                    }
                    while ( --n > 0 );
                
                    log(Sev.ERROR, "moveObject()",
                        "max retries exceeded. Restoring the object in the local host.:");

                        //
                        // XXX: restore locally or throw exception?
                        //
                    activateObject(mox);

                }
                    // the ox is not here
                else {
                        // forward the request to the remote node
                    getRemoteExecutor(loc).moveObject(oxID, placementHint);
                }
            }
            finally {
                log(Sev.DEBUG, "moveObject()", "leaving");
            }
        }


    public MarshaledOX
    deactivateObject( ObjectId oxID )
        {
                //
                // TO DO: check the OX is hosted in this node
                //
            
                // just forward the request to the local objectseed
            MarshaledOX oxState = new MarshaledOX();

            getObjectSeed().deactivateObject(oxID, oxState);

            return oxState;
        }


    public void
    activateObject( MarshaledOX oxState )
        {
                // just forward the request to the local objectseed
            getObjectSeed().activateObject(oxState);
        }


    public void
    exitApplication()
        {
                //
                // TODO: add implemention
                //
            throw new NullPointerException("Not yet implemented.");
        }

    public void
    setHeuristic( SchedulingHeuristic h )
        {
            ThreadGroup tg = Thread.currentThread().getThreadGroup();

            if ( tg instanceof Application ) {
                log(Sev.DEBUG, "setHeuristic", "current thread group is instance of Application");
                ((Application) tg).setHeuristic(h);
            }
            else {
                    // TODO: follow up tg hierarchy
            }
        }

    
        /**
         * Returns the ID of the current application according to the current execution
         * context (formelly, the current thread).
         *
         * @return an <code>ApplicationId</code> value
         *
         * @exception NullPointerException if the current execution context could not
         * determined because the caller thread is an un-managed thread.
         */
    public ApplicationId
    currentApplication()
        {
            ThreadGroup tg = Thread.currentThread().getThreadGroup();

                //
                // TO DO: cache detected applications for faster access; only significant
                // if the thread group hierachy has too many levels.
                //
            
            while ( tg != null ) {
                if ( tg instanceof Application ) {
                    return ((Application) tg).getId();
                }
                else if ( tg.equals(kernelThrGroup) ) {
                    return null;
                }
                tg = tg.getParent();
            }

            throw new NullPointerException(
                "Could not determine current execution context because"
                +" the current thread is not actally managed by EXEHDA");

        }

    
        /**
         * Synchronously run the provided action in behalf of the given application.
         *
         * @param aid an <code>ApplicationId</code> value
         * @param action a <code>Runnable</code> value
         */
    public void
    runAction(ApplicationId aid, Runnable action)
        {
            log(Sev.TRACE, "runAction", "entering appid={0}", aid);
            try {
                if ( aid == null ) {
                        // run as middleware
                    Thread t = new Thread(action);
                    t.start();
                    t.join();
                }
                else {
                    Application app = getApplicationById(aid);

                    app.runAction(action);
                }
            }
            catch (Exception e) {
                log(Sev.ERROR, "runAction", e);
            }
            finally {
                log(Sev.TRACE, "runAction", "leaving");
            }
        }

        ////////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        ////////////////////////////////////////////////////////////////

    private final Executor
    getRemoteExecutor(HostId targetHost)
        {
            Worb worb = (Worb) Exehda.getService(Worb.SERVICE_NAME);

                //
                // XXX: is it safe or should we use contactAddress and CIB to build URI?
                //
            String executorUri = "worb://"+targetHost.toExternalForm()+"/"+SERVICE_NAME;

            return (Executor) worb.lookupService(executorUri, Executor.class);
        }

    private final SchedulingHeuristic
    getHeuristic()
        {
                //
                // TODO: proceed with parent thread group if app not found
                //
            Application app = (Application)
                appByThrGrp.get(Thread.currentThread().getThreadGroup());

            if ( app == null ) {
                log(Sev.DEBUG, "getHeuristic()", "thread group not registered");
                return defaultHeuristic;
            }
            
            SchedulingHeuristic h = app.getHeuristic();

            if ( h != null ) {
                log(Sev.DEBUG, "getHeuristic()", "using app configured heuristic");
                return  h;
            }
            else {
                log(Sev.DEBUG, "getHeuristic()", "using default heuristic");                
                return defaultHeuristic;
            }
        }

    private Application
    getApplicationById( ApplicationId appID ) throws Exception
        {
            log(Sev.TRACE, "getApplicationById()", "entering id={0}", appID);
            
            Application app = (Application) appById.get(appID);
            
            if ( app == null ) {
                    // the given application is not yet running on this node
                
                log(Sev.INFO, "getApplicationById",
                    "Localhost is joining the execution of application {0}", appID);

                app = startApplicationWorker(appID);
            }
            log(Sev.TRACE, "getApplicationById()", "leaving");
            
            return app;
        }
    
    private final ObjectId
    localCreateObject(Class cls, Object[] params, Activator a)
        throws Exception
        {
            log(Sev.DEBUG, "localCreateObject()",
                "invoking newInstance for class={0}", cls);

            return getObjectSeed().newInstance(cls, params, a);
        }
    
    private final ObjectId
    remoteCreateObject(Class cls, Object[] params, Activator a, HostId targetHost)
        {
            log(Sev.DEBUG, "remoteCreateObject()",
                "Selected target host is {0}", targetHost);
            
            Executor remExecutor = getRemoteExecutor(targetHost);
            
            return remExecutor.createObject(cls, params, a, targetHost);
        }

    private final void
    remoteActivateObject(MarshaledOX mox, HostId targetHost)
        {
            log(Sev.DEBUG, "remoteActivateObject()",
                "Selected target host is {0}", targetHost);
            
            Executor remExecutor = getRemoteExecutor(targetHost);
            
            remExecutor.activateObject(mox);
        }

    private final void
    registerApplication(Application app, boolean master)
        {
            appByThrGrp.put(app, app);
            appById.put(app.getId(), app);

            if ( master ) {
                log(Sev.DEBUG, "registerApplication()", "updating CIB");
                
                CellInformationBase cib = getCib();

                    //
                    // TO DO: include cell id in app resource name
                    //
                ResourceName apprn = app.getId().toResourceName();
                cib.addResource(apprn);
                cib.setAttribute(apprn, CIB_ATT_APP_CODEBASE, app.getCodeBase());
                cib.setAttribute(apprn, CIB_ATT_APP_LDESC, app.getLaunchingDescriptor());
                
            }
            else {
                    //
                    // TO DO: register worker node
                    //
            }
        }
    
    private final Application
    startApplicationWorker( ApplicationId appID )
        throws Exception
        {
                // fetch the application attributes from the CIB
            CellInformationBase cib = getCib();
                //
                // FIX ME: just toResourceName(), what if the application is not from the localcell?
                //
            ResourceName apprn = appID.toResourceName();
                
            String codeBase = cib.getAttribute(apprn, CIB_ATT_APP_CODEBASE);
            String ldesc = cib.getAttribute(apprn, CIB_ATT_APP_LDESC);
            
            Application app = new Application(appID, ldesc, true);
            registerApplication(app, false);
            
                // setup the new environment
            log(Sev.DEBUG, "startApplicationWorker",
                "Worker joined app {0}, codebase={1}", appID, app.getCodeBase());

            return app;
        }

    private final ApplicationId
    nextApplicationId()
        {
                // FIX ME: need stronger ID generation strategy
            return new ApplicationId(System.currentTimeMillis());
        }


    private final CellInformationBase
    getCib()
        {
            return (CellInformationBase)
                Exehda.getService(CellInformationBase.SERVICE_NAME);
        }

    private final ObjectSeed
    getObjectSeed()
        {
            return (ObjectSeed)
                Exehda.getService(ObjectSeed.SERVICE_NAME);
        }

    private final HostId
    getOXLocation( ObjectId oxID )
        {
                // if OX Manager available, ask it where the given OX is
            OXManager oxm = (OXManager) Exehda.getService(OXManager.SERVICE_NAME);
            if ( oxm != null ) {
                    //
                    // XXX: need to create a new handle every time? Well, leave any
                    // handle creation optimizations to the OX Manager itself
                    //
                OXHandle oxh = oxm.createHandle(oxID);

                return oxh.getLocation();
            }
                // otherwise, try to guess from the ObjectId
            else {
                return oxID.getHost();
            }
        }


    private int maxRetryies;
    private HostId thisHost;
    private SchedulingHeuristic defaultHeuristic;
    private final Hashtable appByThrGrp;
    private final Hashtable appById;
    private ThreadGroup kernelThrGroup;


}

// class Application
//     extends ThreadGroup
//     implements Runnable
// {
//     private ApplicationId id;
//     private ClassLoader appClassLoader;
//     private String codeBaseURL;
//     private String mainClassName;
//     private String[] args;
//     private Executor.SchedulingHeuristic heuristic;
//     private Thread mainThread;

//     public Application( ApplicationId id, String codeBaseURL, String mainClassName, String[] args)
//         throws Exception
//         {
//             super("Application");
//             this.codeBaseURL = codeBaseURL;
//             this.appClassLoader = IsamClassLoader.newInstance(codeBaseURL, this);
//             this.mainClassName = mainClassName;
//             this.args = args;
//             this.id = id;
//             this.heuristic = null;
//             this.mainThread = null;
//         }
    
//     public Application( String codeBaseURL, String mainClassName, String[] args)
//         throws Exception
//         {
//             this (new ApplicationId(), codeBaseURL, mainClassName, args);
// //             super("Application");
// //             this.codeBaseURL = codeBaseURL;
// //             this.appClassLoader = IsamClassLoader.newInstance(codeBaseURL, this);
// //             this.mainClassName = mainClassName;
// //             this.args = args;
// //             this.id = new ApplicationId();
// //             this.heuristic = null;
// //             this.mainThread = null;
//         }

//     public void
//     start()
//         {
//             if (mainThread == null) {
//                 mainThread = new Thread(this, this, "main");
//                 mainThread.setDaemon(false);
//                 mainThread.start();
//             }
//         }
    
//     public void
//     run()
//         {
//             try {
                
//                 Class mainClass = appClassLoader.loadClass(mainClassName);
//                 Class[] mainArgTypes = new Class[] { String[].class };

//                 Method mainMethod = mainClass.getMethod("main", mainArgTypes);
                
//                 mainMethod.invoke(null, new Object[] { args });
//             }
//             catch (Exception e) {
//                 e.printStackTrace();
//             }
//         }

//     public ApplicationId
//     getId()
//         {
//             return id;
//         }

//     public String
//     getCodebase()
//         {
//             return this.codeBaseURL;
//         }
    
//     public void
//     setHeuristic( Executor.SchedulingHeuristic h)
//         {
//             if ( heuristic == null ) {
//                 heuristic = h;
//             }
//         }

//     public  Executor.SchedulingHeuristic
//     getHeuristic()
//         {
//             return heuristic;
//         }

//     public void
//     runAction( Runnable action )
//         {
//             try {
//                 Thread t = new Thread(this, action, "actionThread");
//                 t.start();
//                 t.join();
//             }
//             catch (InterruptedException ie)
//             {
//                     // empty
//             }
//         }
// }

// class CreateObjectThread
//     extends Thread
// {
//     private Class cls;
//     private Object[] p;
//     private Activator a;
//     private ObjectHandle h;
            
//     public CreateObjectThread(ThreadGroup tg, Class cls, Object[] p, Activator a)
//         {
//             super(tg, "createObject");
//             this.cls = cls;
//             this.p = p;
//             this.a = a;
//             this.h = null;
//             start();
//         }

//     public void run()
//         {
//             ObjectSeed objectseed = (ObjectSeed) Exehda.getService("objectseed");

//             synchronized (this) {
//                 h = objectseed.newInstance(cls, p, a);
//             }
//         }

//     public synchronized ObjectHandle getHandle() throws InterruptedException
//         {
//             this.join();
//             return this.h;
//         }
// }

// class HostIdHeuristic
//     implements Executor.SchedulingHeuristic
// {
//     public HostId
//     chooseCreationHost(String clsName,
//                        Object[] params,
//                        Object hint,
//                        java.util.Vector avoidedHosts)
//         {
//             try {
//                 if ( hint == null ) {
//                     return HostId.getLocalHost();
//                 }
//                 else if ( hint instanceof HostId ) {
//                     return (HostId) hint;
//                 }
//                 else
//                     return null;
//             }
//             catch (Exception e) {
//                 return null;
//             }
//         }

//     public HostId
//     chooseMigrationHost(ObjectHandle oh,
//                        Object hint,
//                        java.util.Vector avoidedHosts)
//         {
//             throw new NullPointerException("Not yet implemented");
//         }
// }

// class LocalHostHeuristic
//     implements Executor.SchedulingHeuristic
// {
//     public HostId
//     chooseCreationHost(String clsName,
//                        Object[] params,
//                        Object hint,
//                        java.util.Vector avoidedHosts)
//         {
//             try {
//                 return HostId.getLocalHost();
//             }
//             catch (Exception e) {
//                 return null;
//             }
//         }

//     public HostId
//     chooseMigrationHost(ObjectHandle oh,
//                        Object hint,
//                        java.util.Vector avoidedHosts)
//         {
//             throw new NullPointerException("Not yet implemented");
//         }

// }
