//  
// PRIMOS - Primitives for Object Scheduling 
//  
// Copyright (C) 2002 Luciano Cavalheiro da Silva  <lucc@inf.ufrgs.br> 
//                    Claudio Fernando Resin Geyer <geyer@inf.ufrgs.br>
//  
// This file is part of PRIMOS. 
//  
// PRIMOS is free software; you can redistribute it and/or 
// modify it under the terms of the GNU General Public License 
// as published by the Free Software Foundation; either version 2 
// of the License, or any later version. 
//  
// PRIMOS is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
// GNU General Public License for more details. 
//  
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
//

/**
* @file  SolarisMonitorImpl.icc
* @brief kstat based implementation of the MonitorImpl interface.
*
* @cvsid $Id: SolarisMonitorImpl.cc 170 2004-03-19 20:13:46Z lucc $
*/

#include <MonitorImpl.hh>

#ifdef HAVE_CONFIG_H
#  include <acconfig.h>
#endif

#ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
#endif

#ifdef HAVE_SYS_UTSNAME_H
#  include <sys/utsname.h>
#endif

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#ifdef HAVE_UTMPX_H
#  include <utmpx.h>
#endif

#ifdef HAVE_OSTREAM
#  include <ostream>
#elif defined HAVE_OSTREAM_H
#  include <ostream.h>
#endif

#ifdef HAVE_FSTREAM
#  include <fstream>
#endif

#ifdef HAVE_VECTOR
#  include <vector>
#endif

#ifdef HAVE_SSTREAM
#  include <sstream>
#endif

#ifdef HAVE_CCTYPE
#  include <cctype>
#endif

#ifdef HAVE_KSTAT_H
#  include <kstat.h>
#  ifdef  HAVE_SYS_SYSINFO_H
#    include <sys/sysinfo.h>
#  else
#    error sys/sysinfo.h not found.
#  endif
#  ifdef HAVE_SYS_PARAM_H
#    include <sys/param.h>
#  else
#    error sys/param.h not found.
#  endif
#endif

#include <SensorData.hh>

#include "SolarisToolkit.hh"

using primos::kernel::Sensor;
using primos::kernel::SensorSet;
using primos::kernel::SensorData;
using primos::kernel::SensorParameter;
using primos::kernel::SensorParameterSet;

// need to fix a bug in ctype: in gcc 3.2 isspace() belongs to
// namespace std::, but in 2.95.3 it does not.
using namespace std;

/**
 * @struct NetDeviceData
 * Used internally by the monitor to store network interface stats
 */
struct NetInterfaceData
{
    std::string  name; // system name of the interface
    std::string  type; // description e.g. Ethernet 100Mbits
    kstat_t*     ks;     // peer kernel stat
    struct netstat
    {
        uint32_t rb; // received bytes
        uint32_t rp; // received packets
        uint32_t re; // receive  errors
        uint32_t rd; // receive  drops
        uint32_t xb; // xmit bytes
        uint32_t xp; // xmit packets
        uint32_t xe; // xmit errors
        uint32_t xd; // xmit drops
        uint32_t xc; // xmit collisions
    } curr, prev;
    int64_t maxByteThroughput;
    int64_t maxPacketThroughput;
    int64_t bytesSentPerSec;
    int64_t bytesRecvPerSec;
    int64_t packetsSentPerSec;
    int64_t packetsRecvPerSec;
    int64_t collisions;

    NetInterfaceData( const string& n, const string& t, kstat_t* ks )
            :name(n),type(t),maxByteThroughput(-1),maxPacketThroughput(0),
             bytesSentPerSec(0),bytesRecvPerSec(0),
             packetsSentPerSec(0),packetsRecvPerSec(0),collisions(0)
        {
            this->ks = ks;
            memset( &curr, 0, sizeof(curr) );
            memset( &prev, 0, sizeof(prev) );
        }
};

/**
 * @brief Storage for per processor stats
 */
struct CpuData
{
    std::string   type;     // textual reprensetation of the cpu type
    bool     online;      // true if this cpu is online (usable)
    kstat_t* ks;          // peer kernel stat
    struct cpustat {
        int64_t u;  // time spent in user computations 
        int64_t s;  // time spent in system(kernel) computations
        int64_t i;  // idle time 
    } curr, prev;

    CpuData()
        {
            curr.u = curr.s = curr.i = 0;
        }
};


/**
 * @class intSensorData
 *
 */
class IntSensorData : public SensorData
{
public:
    IntSensorData(Integer* data)
            : SensorData(INTEGER_DATA, data)
        {}
    
    void
    toString(std::ostream& out) const
        {
            out << *static_cast<Integer*>(data);
        }    
};

/**
 * @class FloatSensorData
 *
 */
class FloatSensorData : public SensorData
{
public:
    FloatSensorData(Float* data)
            : SensorData(FLOAT_DATA, data)
        {}
    
    void
    toString(std::ostream& out) const
        {
            out << *static_cast<Float*>(data);
        }
};


/**
 * @class StringSensorData
 *
 */
class StringSensorData : public SensorData
{
public:
    StringSensorData(String* data)
            : SensorData(STRING_DATA, data)
        {}
    
    
    void
    toString(std::ostream& out) const
        {
            out << *static_cast<String*>(data);
        }
};

/**
 * @class StringSensorData
 *
 */
template <class TData, class TRawData, class TMonitor>
class FakeSensor : public primos::kernel::Sensor
{
private:
    bool      isenabled;
    bool      isstatic;    
    int       index;
    SensorParameterSet params;
    TMonitor* monitor;
    TData     data;    
public:
    FakeSensor(std::string name,
               SensorData::DataTypes
               type, int index,
               TRawData* rawdata,
               TMonitor* monitor,
               bool isenabled=false,
               bool isstatic=false)
            : Sensor(name, type), data(rawdata)
        {
            this->index     = index;
            this->isenabled = isenabled;
            this->isstatic  = isstatic;
            this->monitor   = monitor;
        }
    
    void
    setEnabled(bool enable)
        {
            if (enable) {
                monitor->enableSensor(index);
            } else {
                monitor->disableSensor(index);
            }
            isenabled = enable;
        }

    bool
    isEnabled() const
        {
            return isenabled;
        }

    bool
    isStatic() const
        {
            return isstatic;
        }
    
    bool
    isNative() const
        {
            return true;
        }
    
    const SensorData* 
    probe()
        {
            return &data;    
        }

    void
    toString(std::ostream& out) const
        {
            out << "s=\"" << name << "\" v=\"" << data << '\"';
        }

    SensorParameterSet&
    Parameters() const 
        {
            return const_cast<SensorParameterSet&>(params);
        }

    void
    setParameter( const std::string name, const std::string val )
        {}

    std::string
    getParameter( const std::string name ) const
        {
            throw std::string("Not Defined");
        }
    

};

/**
 * @class SolarisMonitorImpl
 *
 */
class SolarisMonitorImpl
    : public primos::kernel::MonitorImpl
{
public:
        /**
         * @brief Constructor
         *
         */
    SolarisMonitorImpl();

        /**
         * @brief Returns the set of native sensors suported by this monitor implementation.
         */
    const SensorSet&
    availableSensors()
        {
            return nativeSensors;
        }

    void
    quantumExpired();
    

    
    void
    enableSensor(int idx)
        {
                // do nothing in this implementation
        }
    
    void
    disableSensor(int idx)
        {
                // do nothing in this implementation
        }


protected:
        /**
         * @brief Add a sensor to the set of available sensors.
         *
         */
    typedef FakeSensor<StringSensorData, SensorData::String,  SolarisMonitorImpl> SolarisStringSensor;
    typedef FakeSensor<FloatSensorData,  SensorData::Float,   SolarisMonitorImpl> SolarisFloatSensor;
    typedef FakeSensor<IntSensorData,    SensorData::Integer, SolarisMonitorImpl> SolarisIntSensor;

    void
    addSensor(std::string name, int idx, SensorData::Integer* data, bool isStatic = false)
        {
            
            Sensor* s = new SolarisIntSensor(name, SensorData::INTEGER_DATA,
                                          idx, data, this, false, isStatic);
            nativeSensors.insert(s);
        }

    void
    addSensor(std::string name, int idx, SensorData::Float* data, bool isStatic = false)
        {
            Sensor* s = new SolarisFloatSensor(name, SensorData::FLOAT_DATA,
                                          idx, data, this, false, isStatic);
            nativeSensors.insert(s);
        }

    void
    addSensor(std::string name, int idx, SensorData::String* data, bool isStatic = false)
        {
            Sensor* s = new SolarisStringSensor(name, SensorData::STRING_DATA,
                                          idx, data, this, false, isStatic);
            nativeSensors.insert(s);
        }
    
        /**
         * @brief detect the host type
         *
         */
    void
    detectArchitecture();
    
        /**
         * @brief detect available processors
         *
         */
    void 
    detectAvailableProcessors();
    
        /**
         * @brief detect the types of each processor
         */
    void
    detectProcessorTypes();

        /**
         * @brief detect existing network interfaces
         * 
         * TODO: skip the ppp inerface also?
         */
    void
    detectNetworkInterfaces();
    
        /**
         * @brief initialize availableSensors
         *
         */
    void
    initSensors();

    void
    parseSysMisc();

    void
    parseSysInfo();

    void
    parseVmInfo();

    void
    parseCpuStat();

    void
    parseNetDevices();

    void
    parseUtmpx();
    
            
private:
    int64_t    pageSize;
    int        nsensors;
    int64_t    elapsedTime;
    SensorSet  nativeSensors;
    kstat_ctl_t*  pKsChain;
    kstat_t*   pKsSysInfo;
    kstat_t*   pKsVmInfo;
    kstat_t*   pKsCpuInfo;
    kstat_t*   pKsCpuStat;
    kstat_t*   pKsSysMisc;
    vminfo_t   prevVmInfo;
    sysinfo_t  prevSysInfo;
    cpu_stat_t prevCpuStat;
    long clockFreqHz;
    std::vector<NetInterfaceData> netDevs;
    std::vector<CpuData> configuredCpus;
    std::string arch;
    int64_t    nCpus;
    int64_t    nNetInterfaces;
    double     cpuOccupIdle;
    double     cpuOccupUser;
    double     cpuOccupSys;
    double     loadAvg1Min;
    int64_t    totalProcs;
    int64_t    rnblProcs;
    int64_t    totalPhysMem;
    int64_t    freePhysMem;
    int64_t    totalVirtualMem;
    int64_t    freeVirtualMem;
    int64_t    nInterUsers;
    int64_t    upTimeSeconds;
};

std::istream&
skipWord( std::istream& in)
{
    while ( isspace(in.peek()) ) {
        in.ignore();
    }
    while ( !isspace(in.peek()) ) {
        in.ignore();
    }
    return in;
}

std::istream&
skipLine( std::istream& in)
{
    while ( in.peek() != '\n' ) {
        in.ignore();
    }
    in.ignore();
    return in;
}
/**
 *
 *
 */
SolarisMonitorImpl::SolarisMonitorImpl()
{
    nsensors = 0;
    detectArchitecture();
    detectAvailableProcessors();
    detectNetworkInterfaces();
    detectProcessorTypes();
    initSensors();
}

/**
 * @brief detect the host type
 *
 */
inline void
SolarisMonitorImpl::detectArchitecture()
{
        // storage for uname result
    struct utsname undata;
        // issue uname syscall
    if ( uname(&undata) == -1) {
            // on uname failure, fills host type with "unknown"
        arch = "{ unknown }-{ unknown }-{ unknown }";
    } else {
        arch = "{ ";
        arch += undata.machine;
        arch += " }-{ ";
        arch += undata.sysname;
        arch += " }-{ ";
        arch += undata.release;
        arch += " }";
    }

        // create a sensor for this data
    addSensor("ARCHITECTURE", nsensors++, &arch, true );
}

/**
 * @brief detect available processors
 *
 */
inline void 
SolarisMonitorImpl::detectAvailableProcessors()
{
        // discover the total number of cpus
    nCpus = sysconf(_SC_NPROCESSORS_CONF);
        // create the respective sensor
    addSensor("N_PROCESSORS", nsensors++, &nCpus, false);
        // detect the type of each available processor

}

/**
 * @brief detect the types of each processor
 */
inline void
SolarisMonitorImpl::detectProcessorTypes()
{
        // constant to compare to the cpu state
    const std::string CPU_STATE_ONLINE = "on-line";
        // constant to compare to the cpu state
    configuredCpus.reserve(nCpus);
        // get information for each existing processor from kstat
    for (int i=0; i<nCpus; i++) {
            // get the kstat for the given processor
        kstat_t* cpuinfo = kstat_lookup(pKsChain, "cpu_info", i, NULL);
            // get cputype and clock from the kstat
        kstat_named_t& cputype  =
            *(kstat_named_t*) kstat_data_lookup(cpuinfo, "cpu_type");
        
        kstat_named_t& cpuclock =
            *(kstat_named_t*) kstat_data_lookup(cpuinfo, "clock_MHz");
        
        kstat_named_t& cpustate =
            *(kstat_named_t*) kstat_data_lookup(cpuinfo, "state");
        
            // build cputype string
        ostringstream os;
        os << "{ " << cputype.value.c << " }-{ "
           << cpuclock.value.ui32 << " MHz }";
            // fill CpuData
        CpuData currCpu;
        currCpu.type    = os.str();
        currCpu.online  = (CPU_STATE_ONLINE == cpustate.value.c);
        currCpu.ks      = cpuinfo;

        configuredCpus.push_back(currCpu);
            // create sensor
        std::ostringstream sensorName;
        sensorName << "PROCESSOR_TYPE[" << i << "]";

        addSensor(sensorName.str(), nsensors++, &configuredCpus[i].type, true );
    } // for
}

/**
 * @brief detect existing network interfaces
 * 
 * TODO: skip the ppp inerface also?
 */
inline void
SolarisMonitorImpl::detectNetworkInterfaces()
{
        // aux. constants
    const std::string NET      = "net";     // class net
    const std::string STREAMS  = "streams"; // module streams
    const std::string ETHERNET = "hme";     // ethernet module
    
        // alloc initial storage for net dev. data 
    vector<NetInterfaceData*> devs;
    devs.reserve(16);
        // devices to skip in detection
    std::set<std::string> skipDevs;
    skipDevs.insert(std::string("lo"));     
    skipDevs.insert(std::string("streams"));
    
        // initially there are no detected interfaces
    nNetInterfaces = 0;
    
        // parse kstast looking for net. devices
    for (kstat_t* ksp = pKsChain->kc_chain; ksp != NULL; ksp=ksp->ks_next) {
            // skip non network related stats
        if ( NET != ksp->ks_class ) continue;
            // check skip interfaces
        else if ( skipDevs.find(ksp->ks_module) != skipDevs.end() ) continue;
            // assume that all other names are physical interface names
        else {
                // last check, see if in/out stats are available
            kstat_read(pKsChain, ksp, NULL);
            if ( kstat_data_lookup(ksp, "obytes") != NULL ) {
                std::string ifname = ksp->ks_name;
                std::string iftype;
                iftype += "{ ";
                iftype += ifname;
                iftype += " }-{ ";
                    // try to indentify the interface type
                if ( ETHERNET == ksp->ks_module ) {
                    uint32_t ifspeed = ((kstat_named_t*)
                                        kstat_data_lookup( ksp,"ifspeed"))->value.ui32;
                    ostringstream os;
                    os << "ethernet }-{ " << ifspeed << "Mb/s ";
                    iftype = os.str();
                } else {
                    iftype = "unknown";
                }
                iftype += " }";
                
                devs.push_back(new NetInterfaceData(ifname, iftype, ksp));
                    // incrememt the number of detected interfaces
                nNetInterfaces++;
            }
        }
    }
        // shrink the net devices array
    this->netDevs.reserve(devs.size());
    for ( int ni = 0; ni < (int) devs.size(); ni++ ) {
        netDevs.push_back(*devs[ni]);
        delete (devs[ni]);
            // create sensors for this network interface
        std::ostringstream oss;
        oss << "[" << ni << "]";
        std::string sensorSuffix = oss.str();
        addSensor("NI_TYPE"+sensorSuffix, nsensors++,&netDevs[ni].type, true );
        addSensor("NI_MAX_BYTE_THROUGHPUT"+sensorSuffix, nsensors++,&netDevs[ni].maxByteThroughput);
        addSensor("NI_MAX_PACKET_THROUGHPUT"+sensorSuffix, nsensors++,&netDevs[ni].maxPacketThroughput);
        addSensor("NI_BYTES_SENT_PER_SEC"+sensorSuffix, nsensors++,&netDevs[ni].bytesSentPerSec);
        addSensor("NI_BYTES_RECV_PER_SEC"+sensorSuffix, nsensors++,&netDevs[ni].bytesRecvPerSec);
        addSensor("NI_PACKETS_SENT_PER_SEC"+sensorSuffix, nsensors++,&netDevs[ni].packetsSentPerSec);
        addSensor("NI_PACKETS_RECV_PER_SEC"+sensorSuffix, nsensors++,&netDevs[ni].packetsRecvPerSec);
        addSensor("NI_PACKET_COLLISIONS"+sensorSuffix, nsensors++,&netDevs[ni].collisions);
    }
}

void
SolarisMonitorImpl::quantumExpired()
{
    elapsedTime = 1; // actiually it must be equals to the quantum size
    parseSysMisc();
//    parseCpuInfo();
    parseSysInfo();
    parseVmInfo();
    parseCpuStat();
    parseNetDevices();
    parseUtmpx();
}

/**
 *
 *
 */
inline void
SolarisMonitorImpl::parseSysMisc()
{

    kstat_read(pKsChain, pKsSysMisc,  NULL);
    //
    // STAT: Up time
    //
        // gets the number of clock interrupts since boot
    kstat_named_t& pClockTicks =
        *(kstat_named_t*) kstat_data_lookup(pKsSysMisc, "clk_intr");
        // converts the number of interrupts to seconds
    upTimeSeconds = pClockTicks.value.ul / clockFreqHz;

    //
    // STAT: Load average in the last minute
    //
        // get the raw index value (number of processes)
    kstat_named_t& pRawLoadAvg =
        *(kstat_named_t*) kstat_data_lookup( pKsSysMisc, "avenrun_1min" );
        // compute the real value using the FSCALE factor.
        // FSCALE correspond to the number of time-slices per second
    loadAvg1Min = ( (double)pRawLoadAvg.value.ul ) / FSCALE;
    //
    // STAT: Total number of processes
    //
    kstat_named_t& pNProc =
        *(kstat_named_t*) kstat_data_lookup(pKsSysMisc, "nproc");
    totalProcs =  pNProc.value.ui32;

}

// /**
//  *
//  *
//  */
// inline void
// SolarisMonitorImpl::parseCpuInfo()
// {
// //         // read the new stats from kstat
// //     kstat_read(pKsChain, pKsCpuInfo, NULL);
// }

/**
 *
 *
 */
inline void
SolarisMonitorImpl::parseSysInfo()
{
        // save previous kstat values
    prevSysInfo = *(sysinfo_t*) pKsSysInfo->ks_data;
        // read the new stats from kstat
    kstat_read(pKsChain, pKsSysInfo, NULL);
        // get references for the current stats data
    sysinfo_t&  currSysInfo = *(sysinfo_t*)  pKsSysInfo->ks_data;
        //
        // STAT: Number of runnable processes
        //
    rnblProcs = (currSysInfo.runque - prevSysInfo.runque) / elapsedTime;
}

/**
 *
 *
 */
inline void
SolarisMonitorImpl::parseVmInfo()
{
        // save previous kstat values
    prevVmInfo  = *(vminfo_t*) pKsVmInfo->ks_data;
        // read the new stats from kstat
    kstat_read(pKsChain, pKsVmInfo,  NULL);
        // get a reference for the current stats data
    vminfo_t&   currVmInfo  = *(vminfo_t*)   pKsVmInfo->ks_data;
        // re-compute free physical memory
    freePhysMem =  ((currVmInfo.freemem-prevVmInfo.freemem)
                    * pageSize) / elapsedTime;
    
        // re-compute free virtual memory
    freeVirtualMem = ((currVmInfo.swap_avail-prevVmInfo.swap_avail)
                      * pageSize / elapsedTime) + freePhysMem;
                      
        // re-compute total VirtualMemory
    totalVirtualMem =
        ((currVmInfo.swap_resv - prevVmInfo.swap_resv)
         * pageSize / elapsedTime) + freeVirtualMem + totalPhysMem;
}

/**
 *
 *
 */
inline void
SolarisMonitorImpl::parseCpuStat()
{
        // save previous kstat values
    prevCpuStat = *(cpu_stat_t*) pKsCpuStat->ks_data;
        // read the new stats from kstat
    kstat_read(pKsChain, pKsCpuStat, NULL);
        // get references for the current stats data
    cpu_stat_t& currCpuStat = *(cpu_stat_t*) pKsCpuStat->ks_data;

    //
    // STAT: Processor occupations (user, sys, idle)
    //
        // compute the raw cpu times
    cpu_sysinfo_t& cpuInfo = currCpuStat.cpu_sysinfo;
    cpu_sysinfo_t& oldInfo = prevCpuStat.cpu_sysinfo;

    int64_t i = cpuInfo.cpu[CPU_IDLE]   - oldInfo.cpu[CPU_IDLE];
    int64_t u = cpuInfo.cpu[CPU_USER]   - oldInfo.cpu[CPU_USER];
    int64_t k = cpuInfo.cpu[CPU_KERNEL] - oldInfo.cpu[CPU_KERNEL];
    int64_t w = cpuInfo.cpu[CPU_WAIT]   - oldInfo.cpu[CPU_WAIT];
    int64_t t = i+u+k+w;
        // compute the percentual occupation indexes
    cpuOccupUser = 100.0 * u/t;
    cpuOccupSys  = 100.0 * (k+w)/t;
    cpuOccupIdle = 100.0 - (cpuOccupUser + cpuOccupSys);
}

/**
 *
 *
 */
inline void
SolarisMonitorImpl::parseUtmpx()
{
        // reset utmpx database
    setutxent();
    struct utmpx* ent;
        // reset the interactive processes counter
    nInterUsers = 0;
        // parse utmp entries looking for interactive processes
    while ((ent = getutxent()) != NULL) {
            // increment the counter on each found interactive process
        if (ent->ut_type == USER_PROCESS) {
            nInterUsers++;
        } 
    }
}

/**
 *
 *
 */
inline void
SolarisMonitorImpl::parseNetDevices()
{
    vector<NetInterfaceData>::iterator ni;
    
    for ( ni = netDevs.begin(); ni != netDevs.end(); ni++) {
            // save previous state
        ni->prev = ni->curr;
        kstat_t* ksp = ni->ks;
            // request an update of the stat chain to kernel
        kstat_read(pKsChain, ni->ks, NULL);
            // update current stats
        ni->curr.rb = ((kstat_named_t*) kstat_data_lookup(ksp, "rbytes"))->value.ui32;
        ni->curr.rp = ((kstat_named_t*) kstat_data_lookup(ksp, "ipackets"))->value.ui32;
//        ni->curr.re = ((kstat_named_t*) kstat_data_lookup(ksp, "ierrors"))->value.ui32;
        ni->curr.xb = ((kstat_named_t*) kstat_data_lookup(ksp, "obytes"))->value.ui32;
        ni->curr.xp = ((kstat_named_t*) kstat_data_lookup(ksp, "opackets"))->value.ui32;
//        ni->curr.xe = ((kstat_named_t*) kstat_data_lookup(ksp, "oerrors"))->value.ui32;
        ni->curr.xc = ((kstat_named_t*) kstat_data_lookup(ksp, "collisions"))->value.ui32;

        if ( ni->maxByteThroughput == -1 ) {
                // do nothing, just calibrate the mechanism
            ni->maxByteThroughput = 0;
        } else {
            ni->bytesSentPerSec   = (ni->curr.xb - ni->prev.xb) / elapsedTime;
            ni->packetsSentPerSec = (ni->curr.xp - ni->prev.xp) / elapsedTime;
            ni->collisions        = (ni->curr.xc - ni->prev.xc) / elapsedTime;
            ni->bytesRecvPerSec   = (ni->curr.rb - ni->prev.rb) / elapsedTime;
            ni->packetsRecvPerSec = (ni->curr.rp - ni->prev.rp) / elapsedTime;

            int64_t byteThroughput = ni->bytesSentPerSec + ni->bytesRecvPerSec;
            int64_t packetThroughput = ni->packetsSentPerSec + ni->packetsRecvPerSec;

            if ( ni->maxByteThroughput < byteThroughput ) {
                ni->maxByteThroughput = byteThroughput;
            }
            
            if ( ni->maxPacketThroughput < packetThroughput ) {
                ni->maxPacketThroughput = packetThroughput;
            }
        }
    } // for
}




/**
 * @brief initialize availableSensors
 *
 */
inline void
SolarisMonitorImpl::initSensors()
{
    int idx=nsensors;
    addSensor("CPU_OCCUP_USER",idx++, &cpuOccupUser);
    addSensor("CPU_OCCUP_SYS",idx++, &cpuOccupSys);
    addSensor("CPU_OCCUP_IDLE",idx++, &cpuOccupIdle );
//     addSensor("CPU_OCCUP_BG",idx++, &cpuOccupBg);
    addSensor("LOAD_AVG",idx++, &loadAvg1Min);
    addSensor("N_TOTAL_PROCESSES",idx++, &totalProcs);
    addSensor("N_RNBL_PROCESSES",idx++, &rnblProcs);
    addSensor("TOTAL_PHYS_MEMORY",idx++, &totalPhysMem);
    addSensor("FREE_PHYS_MEMORY",idx++, &freePhysMem);
    addSensor("TOTAL_VIRTUAL_MEMORY",idx++, &totalVirtualMem);
    addSensor("FREE_VIRTUAL_MEMORY",idx++, &freeVirtualMem);
    addSensor("N_INTERACTIVE_USERS",idx++, &nInterUsers);
    addSensor("UP_TIME",idx++, &upTimeSeconds);
    nsensors = idx;
}


/**
 *
 *
 */
primos::kernel::MonitorImpl*
SolarisToolkit::createMonitor( void )    
{
    return new SolarisMonitorImpl();
}
