// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.primos.objseed;

import org.isam.exehda.Exehda;
import org.isam.exehda.ObjectId;
import org.isam.exehda.services.AbstractService;
import org.isam.exehda.services.Logger.Sev;
import org.isam.exehda.services.ObjectSeed;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Hashtable;
import java.util.Vector;



/**
 * Implementation of the ObjectSeed service interface.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-04-25 02:47:38 -0300 (Mon, 25 Apr 2005) $ $Revision: 802 $
 */
public class ObjectSeedImpl
    extends AbstractService
    implements ObjectSeed
{
    private final static String DFLT_ACTIVATOR_FACTORY = DefaultActivatorFactory.class.getName();
    private final static String PROP_ACTIVATOR_FACTORY = "activatorFactory";
    
    public void
    start()
        throws ClassNotFoundException, InstantiationException, IllegalAccessException
        {
                // load parameters from profile
            String pActivatorFactory = Exehda.getServiceProperty(
                SERVICE_NAME, PROP_ACTIVATOR_FACTORY, DFLT_ACTIVATOR_FACTORY);

                // parse params and initialize instance attributes
            this.activatorFactory = (ActivatorFactory)
                Class.forName(pActivatorFactory).newInstance();
        }
    
    public void
    stop() 
        {
                // empty
        }

    
    public ObjectId
    newInstance( Class cls, Object[] params, Activator ca )
        {
            MetaOX mox = null;
            
            try {
                    //
                    // TO DO: once stable set severity level to TRACE
                    //
                log(Sev.DEBUG, "newInstance()", "entering");

                    // instantiate the OX
                Object obj = (params == null)
                    ? cls.newInstance()
                    : findConstructor(cls, params).newInstance(params);

                log(Sev.DEBUG, "newInstance()", "instantiated {0}", cls);
                
                    // generate na ObjectId for the just created OX
                ObjectId oxID = getNextID();
                    // create an activator for the OX
                Activator oxa = activatorFactory.createActivator(obj, ca);
                    // create the system-wrapper for the OX
                mox = new MetaOX( oxID, obj, ca );
                    //  register the OX
                register(mox);

                log(Sev.DEBUG, "newInstance()", "activating it...");
                                
                    // proceed activation
                mox.activate();

                log(Sev.DEBUG, "newInstance()", "activation concluded.");
                
                    // on success, return the ID of the OX
                return oxID;
            }
            catch (Exception e) {
                    // in case of failure, clean-up management tables
                if ( mox != null ) unregister(mox);
                    // log the unexpected condition occurred
                log(Sev.ERROR, "newInstance()", e);
                    // notify the caller
                
                    //
                    // XXX: need a more specific exception type?
                    //
                throw new IllegalArgumentException(
                    "newInstance() failed due to "+e);
            }
            finally {
                    //
                    // TO DO: once stable set severity level to TRACE
                    //
                log(Sev.DEBUG, "newInstance()", "leaving");
            }
        }
    

    public MarshaledOX
    deactivateObject( ObjectId oxID, MarshaledOX oxState )
        throws OXActivationException
        {
            if ( oxState == null ) {
                oxState = new MarshaledOX();
            }
                // lookup the corresponding OxInfo for the given ID
            MetaOX mox = lookup(oxID);

            if ( mox != null ) {
                try {
                        //
                        // FIX ME: do we need check if the caller is allowed to
                        // deactivate the given OX?
                        //
                    
                        // proceed ox de-activation
                    mox.deactivate(oxState);

                        // return the persistent OX representation
                    return oxState;
                }
                catch (Exception e) {
                        // log the unexpected condition occurred
                    log(Sev.WARNING, "deactivateObject()", e);
                        // notify the caller
                    throw new OXActivationException(e.toString());
                }
            }
            else {
                throw new IllegalArgumentException("No such ox:"+oxID);
            }
        }
    
    public void
    activateObject( MarshaledOX oxState )
        throws OXActivationException
        {
                //
                // TO DO: are there any side-effects when restoring OXes coming from
                // other nodes? e.g. what about the remotely generated ObjectIDs?
                //
            
            MetaOX mox = null;
            
            try {
                    // create the meta-ox and register it
                mox = new MetaOX(oxState);
                register(mox);
                    // re-activate the OX
                mox.activate(oxState);
            }
            catch (Exception e) {
                    // in case of failure, clean-up management tables
                if ( mox != null ) unregister(mox);
                    // log the unexpected condition occurred
                log(Sev.WARNING, "activateObject()", e);
                    // notify the caller
                throw new OXActivationException(e.toString());
            }
        }


        //////////////////////////////////////////////////////////
        // PRIVATE STUFF
        //////////////////////////////////////////////////////////
    
        /**
         * Find a constructor defined in the given object class suiltable to the used
         * with the given parameters list.
         *
         * @param cls the target classed
         * @param params the user-provided parameters list
         * @return a <code>Constructor</code> value
         * @exception NoSuchMethodException if no suitable constructor is avaiable in the
         * given class
         */
    private final Constructor
    findConstructor( Class cls, Object[] params )
        throws NoSuchMethodException
        {
            Class paramTypes[] = new Class [params.length];
            for (int i=0; i<params.length; i++){
                paramTypes[i] = params[i].getClass();
            }
                // find the matching constructor for the given parameter list
            Constructor[] candidate = cls.getConstructors();
            Constructor   matched    = null;
            for (int c=0; c<candidate.length && (matched== null); c++) {
                
                Class[] formalParamTypes = candidate[c].getParameterTypes();
                    // check against number of parameters
                if (formalParamTypes.length == paramTypes.length ) {
                    
                    matched = candidate[c];
                    
                        // check against parameter types
                    for (int p=0; p<formalParamTypes.length; p++) {
                        
                            //
                            // TO DO: match primitive types
                            //
                        
                            // if parameter matching fails clear matched
                        if (!formalParamTypes[p].isAssignableFrom(
                                paramTypes[p])) {
                            matched = null;;
                        }
                    }
                }
            }

            if ( matched != null ) {
                return matched;
            }
            else {
                throw new NoSuchMethodException(
                    "No suitable constructor found for the given parameter list");
            }
        }
    
        /**
         * Starts managing the given OX.
         *
         * @param mOX a <code>MetaOX</code> value
         */
    private final void
    register( MetaOX mox )
        {
                //
                // TO DO: should use per-application tables
                //
            moxById.put( mox.getId(), mox );
        }

    
        /**
         * Stops managing the given OX
         *
         * @param mOX a <code>MetaOX</code> value
         */
    private final void
    unregister( MetaOX mox )
        {
                //
                // TO DO: should use per-application tables
                //
            moxById.remove(mox.getId());
        }

    
        /**
         * Returns the MetaOX corresponding to the given ox ID.
         *
         * @param oxID an <code>ObjectId</code> value
         * @return a <code>MetaOX</code> value
         */
    private final MetaOX
    lookup( ObjectId oxID )
        {
                //
                // TO DO: should use per-application tables
                //
            return (MetaOX) moxById.get(oxID);
        }
    
        /**
         * Generates a new (not yet allocated) ObjectId
         *
         * @return an <code>ObjectId</code> value
         */
    private final ObjectId
    getNextID()
        {
            return new ObjectId();
        }


        /**
         * Table of managed OX's
         *
         * @todo TO DO:Use a distinct table for each running application
         *
         */
    private final Hashtable moxById = new Hashtable();
    private ActivatorFactory activatorFactory = null;
}

