// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda;

import java.util.Enumeration;
import java.util.Vector;



/**
 * A resource constraint that is actually a "three-state AND" of other constraints.
 *
 *
 * <p> BROKEN 3and * = BROKEN
 * <p> UNDEFINED 3and OK = UNDEFINED
 * <p> OK 3and OK = OK;
 *
 * 
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-02-24 02:30:27 -0300 (Thu, 24 Feb 2005) $ $Revision: 585 $
 */
public class CompositeResourceConstraint
        extends ResourceConstraint
    {
        public static class CRCElement
        {
            public CRCElement(ResourceConstraint rc, boolean undefIsOk)
                {
                    this.undefIsOk = undefIsOk;
                    this.rc = rc;
                }
            
            public boolean
            getUndefinedIsOk()
                {
                    return this.undefIsOk;
                }

            public ResourceConstraint
            getConstraint()
                {
                    return this.rc;
                }
            
            public int
            check( ResourceInfo r )
                {
                    return rc.check(r);
                }

            public ResourceInfo[]
            select( ResourceInfo[] c )
                {
                    return rc.select(c, this.undefIsOk);
                }

            private boolean undefIsOk;
            private ResourceConstraint rc;
        }

        
        public CompositeResourceConstraint()
            {
                this.rcs = new Vector(8);
            }

        
            /**
             * Successively applies each of the sub-constraints to the resource. The
             * final result is defined as a 3-state AND of the sub-results.
             *
             * @param r a <code>ResourceInfo</code> value
             * @return an <code>int</code> value
             */
        public int
        check ( ResourceInfo r )
            {
                int res = ResourceConstraint.OK;
                
                for (int i=0; i<rcs.size(); i++) {
                    CRCElement rc = (CRCElement) rcs.elementAt(i);

                    switch (rc.check(r)) {
                        case ResourceConstraint.OK: {
                            continue;
                        }
                        case ResourceConstraint.BROKEN: {
                            return ResourceConstraint.BROKEN;
                        }
                        default: {
                            res = ResourceConstraint.UNDEFINED;
                        }
                    }
                }
                return res;
            }

            /**
             * Filters the given set of candidate resources keeping only those that
             * satisfy all the constraints that form this composite constraint.
             *
             * <p> <b>Note:</b> if multiple sorting criterias are applied by the
             * subconstraints the final ordering of elements is undetermined (probably it
             * will be that of the last constraint added to the composite).
             *
             * @param c the candidate resources
             * @param undefIsOk ignored, each sub constraint defines its own undef mode
             * @return resources that satisfy the criteria
             */
        public ResourceInfo[]
        select( ResourceInfo[] c, boolean undefIsOk )
            {
                throw new NullPointerException("Not yet implemented");
            }

        
            /**
             * Append a constraint to this composite.
             *
             * @param c a <code>ResourceConstraint</code> value
             * @param undefIsOk a <code>boolean</code> value
             */
        public void
        addConstraint( ResourceConstraint c,  boolean undefIsOk )
            {
                rcs.addElement(new CRCElement(c, undefIsOk));
            }

        
            /**
             * Returns an enumeration of the CRCElements holding the constraints that
             * form this composite.
             *
             * @return an <code>Enumeration</code> value
             */
        public Enumeration
        constraints()
            {
                return rcs.elements();
            }

            //////////////////////////////////////////////////////
            // PRIVATE STUFF
            //////////////////////////////////////////////////////
        private Vector rcs;
    }
    
