// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services;

import org.isam.exehda.HostId;
import org.isam.exehda.ObjectId;
import org.isam.exehda.ApplicationId;
import org.isam.exehda.services.ObjectSeed.Activator;
import org.isam.exehda.services.ObjectSeed.MarshaledOX;
import java.util.Vector;


/**
 * Executor's service interface. It defines methods for:
 *
 * <ul>

 * <li> launching and finishing applications;
 
 * <li> executing actions in behalf of a given application;
 
 * <li> creating and migrating objects (formelly OXes);

 * <li> binding per-application heuristics which will guide host selection in both remote
 * intantiation and migration of objects.

 * </ol>

 * The Executor service defines the notion of per-object Activator, and instantiation of
 * the Strategy Design Pattern. Activator are responsible to "connect" the object to it's
 * execution environment, by loading and initializing any extra-object resources
 * (e.g. libraries or physical devices) prior to the object effectivelly starts its
 * operation (e.g. prior the start() method of a thread is called).

 * The user specifies an activator to be with some OX by the time the OX is being
 * instantiated though the createObject() operation. Such activator is bound to that
 * object and will be carried with the object even through migration.

 * Activators may be chainned. Actually, a default Activator which initializes registers
 * the OX within the OXManager service is implicitly invoked prior to any user-provided
 * activators.

 * If the OX Activator is omitted, the system will try to guess an suitable activator
 * between the system pre-defined activator based on the objects type. Towards that goal,
 * the following rationale is used:
 *
 * @todo to be continued...
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-04-25 02:34:14 -0300 (Mon, 25 Apr 2005) $ $Revision: 796 $
 */
public interface Executor
{
    public static final String SERVICE_NAME = "executor";

    public static class MalformedAppLaunchingDescriptorException
        extends IllegalArgumentException
    {
        public MalformedAppLaunchingDescriptorException(String msg) {
            super(msg);
        }
    }


        /**
         * Describe interface <code>Migrable</code> here.
         *
         */
    public static interface Migrable
        extends java.io.Serializable
    {
            /**
             * Captures the object state for later restoring after the object is
             * migrated. This method is invoked before migration takes place, but after
             * the object is de-activated. This method would be used to record additional
             * state information else that included by the standard serialization.
             *
             * @param to a <code>java.io.ObjectOutput</code> value
             */
        public void captureState ( java.io.ObjectOutput to );
        
            /**
             * Restores the object's state. This method is invoked after migration takes
             * place, but before the object's re-activation takes place. This method is
             * used to recover extra state information (i.e. stack state) not usually
             * included by the standart serialization.
             *
             * @param from a <code>java.io.ObjectInput</code> value
             */
        public void restoreState ( java.io.ObjectInput from );
    }

    
        /**
         * Per-application scheduling strategies are bound to the Executor operation by
         * implementing the SchedulingHeuristic interface and registering such component
         * through a call to Executor.setHeuristic().
         *
         */
    public static interface SchedulingHeuristic
        extends java.io.Serializable
    {
            /**
             * This method is invoked by the Executor for selecting a node suitable to host
             * an OX being instantiated through createObject(). The information provided
             * to createObject() is made available to the SchedulingHeuristic and would
             * be take in account when selecting the best suitable node.
             *
             * @param cls the name of the class being instantiated
             * @param params parameters passed to the object's construtor
             * @param hint an (optional) hint provided in the call to createObject()
             * @param toAvoid nodes that must be avoided when doing the selection.
             * @return the heuristic's choice or null if there is no suitable host for
             * the request
             */
        public HostId chooseCreationHost ( String cls, Object[] params, Object hint, Vector toAvoid );
  
            /**
             * This method is invoked by the Executor for selection a node suitable to
             * host an OX being migrated (through moveObject()).
             *
             * @param o the ID of ox being migrated
             * @param hint an (optional) placement hint provided in the call to moveObject().
             * @param toAvoid nodes that must be avoided when doing the selection.
             * @return the heuristic's choice or null if there is no suitable host for
             * the request
             */
        public HostId chooseMigrationHost ( ObjectId o, Object hint, Vector toAvoid );
    }

    
        /**
         * Creates an OX in behalf of the calling application, by instantiating the given
         * class invoking the constructor apropriated for the given parameter list.
         *
         * <p>If the providaded activator is non-null, it will be used to complete the OX
         * actiovation process (e.g. starting threads).
         *
         * <p>The exact meaning of the last parameter depends on the SchedulingHeuristic in
         * used. Although, its guaranteed that the passed value is equals to the ID of
         * the local host, the OX whill be created in the local host.
         *
         * <p>The SchedulingHeuristic would select the target host itself or would delegate
         * such decision to another middleware service, the Scheduler for example.
         *
         * @see org.isam.exehda.HostId#getLocalHost
         *
         * @param c the class to be instantiated
         * @param p the arguments to be passed to the constructor
         * @param a an optional, custom activator
         * @param hint placement hint used to guide the target host selection process.
         * @return the ID of the instantiaded OX
         */
    public ObjectId createObject( Class c, Object[] p, Activator a, Object hint );

    
        /**
         * Migrate anexisting OX between nodes.
         *
         * The meaning of the placement hint is the same that in createObject(). If a
         * custom activatior has been provided by the time the OX was instantiaded, it
         * will be used to pause the OX execution in its current host and later to resume
         * its execution on the new host.
         *
         * @param ox an <code>ObjectId</code> value
         * @param placementHint an <code>Object</code> value
         */
    public void moveObject( ObjectId ox, Object placementHint );

    
        /**
         * Deactivates the identified OX, storing its state and other management info in
         * the returned MarshaledOX. The OX execution would later be resumed by invoking
         * activateObject() on that MarshaledOX.
         *
         * @param ox an <code>ObjectId</code> value
         * @return a <code>MarshaledOX</code> value
         */
    public MarshaledOX deactivateObject( ObjectId ox );

        /**
         * Resumes the exeuction of an OX previously pause by deactivateObject().
         *
         * @param ox an <code>ObjectId</code> value
         * @return a <code>MarshaledOX</code> value
         */
    public void activateObject( MarshaledOX oxState );
    
        /**
         * Launches the execution of the application described by the given lauching
         * descriptor document.
         *
         * @param launchingDesc an <code>org.isam.util.xml.XmlElement</code> value
         * @exception MalformedAppLaunchingDescriptorException if an error occurs
         */
    public void startApplication( org.isam.util.xml.XmlElement launchingDesc )
        throws MalformedAppLaunchingDescriptorException;
    
        /**
         * Causes the calling application to abort its execution by stopping all the
         * threads running in behalf of the given applicaiton. This effect is to be
         * propagated to every node in the system which currently takes part of the
         * application execution (i.e. which hosts some component (OX) of the aplication).
         *
         */
    public void exitApplication();
    
        /**
         * Returns the ID of the application to which the calling thread belongs.
         *
         * @return an <code>ApplicationId</code> value
         */
    public ApplicationId currentApplication();
    
        /**
         * Runs the given action in behalf of the given application, waiting its
         * completion prior to returning.
         *
         * The method run() of the specified action will be invoked in a thread such that
         * if currentApplication() is invoked inside the given action it will return the
         * ID specified in app.
         *
         * This method would used to isolate the calling chain from the that started by
         * the action. For instance, it would be used to force the local execution
         * context despite of what Worb.getClientHost() would normally return.
         *
         * @see org.isam.exehda.services.Worb#getClientHost()
         *
         * @param app an <code>ApplicationId</code> value
         * @param action a <code>Runnable</code> value
         */
    public void runAction(ApplicationId app, Runnable action);

        /**
         * This method is deprecated, avoid using it in new code.
         *
         * Sets the per-application scheduling heuristic. 
         *
         * @param sh a <code>SchedulingHeuristic</code> value
         * @deprecated As an application attribute, the schedheuristic is going
         *             to be configured through the application launching descriptor,
         *             but this feature is not yet implemented.
         */
    public void setHeuristic( SchedulingHeuristic sh );
}

