// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda;


/**
 * Identifies a resource in the system. It's used when working with resource discovery
 * and allocation in the pervasive environment.
 *
 * Resource names are arranged hierarchically using the concept of typed name spaces.
 *
 * @see org.isam.exehda.services.CellInformationBase
 * @see org.isam.exehda.services.Discoverer
 * @see org.isam.exehda.services.ResourceBroker
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-19 15:51:38 -0300 (Sun, 19 Jun 2005) $ $Revision: 941 $
 */
public class ResourceName
    implements java.io.Serializable
{
        /**
         * A resource name component. All name components are type, i.e. they include a
         * type identifier as long as the instance identifier.
         *
         */
    public static class NameComponent
        implements java.io.Serializable
    {    

        public NameComponent(String type, String instance)
            {
                if ( type == null
                     || instance == null
                     || "".equals(type) || "".equals(instance) ) {
                    throw new IllegalArgumentException(
                        "Both type and instance identifiers must be non-null, non-empty strings");
                }
                this.type = type.trim();
                this.instance = instance.trim();
            }

        public String
        getType()
            {
                return this.type;
            }
        
        public String
        getInstance()
            {
                return this.instance;
            }
        
        public boolean
        equals( Object o )
            {
                if ( (o != null ) && (o instanceof NameComponent) ) {
                    NameComponent nc = (NameComponent) o;

                    return  type.equals(nc.type) && instance.equals(nc.instance);
                }
                return false;
            }
        
        public int
        hashCode()
            {
                return type.hashCode() ^ instance.hashCode();
            }
        
        public String
        toString()
            {
                return "NC{" + type +':' + instance + '}';
            }

            ////////////////////////////////////////////////////
            // PRIVATE STUFF
            ////////////////////////////////////////////////////
        
        private String type;
        private String instance;
    }
    
        /**
         * Defines a scope for a  resource name. Name spaces are hierarchical and typed, meaning a
         * name space may be defined inside another name space and all name space
         * components also include a type information.
         *
         * @see org.isam.exehda.ResourceName.NameComponent
         *
         */
    public static class NameSpace
        implements java.io.Serializable
    {
        public static final NameSpace DEFAULT_NS = new NameSpace("DEFAULT", "DEFAULT" , null);
        
            /**
             * Creates a new <code>NameSpace</code> instance with the provided name
             * inside the given parent name space.
             *
             * @param cn the space name (i.e. type and instance identifiers)
             * @param parent the parent name space
             */
        public NameSpace(NameComponent cn, NameSpace parent)
            {
                if ( cn == null ) {
                    throw new IllegalArgumentException("Name component must be non-null.");
                }
                
                this.cn = cn;
                this.parent = parent;
            }
        
            /**
             * Creates a new <code>NameSpace</code> with the provided type and instance
             * identifiers inside the given parent name space.
             *
             * @param t a String defining the name space type
             * @param i a String defining the name space instance
             * @param parent the parent name space
             */
        public NameSpace(String t, String i, NameSpace parent)
            {                
                this.cn = new NameComponent(t,i);
                this.parent = parent;
            }
        
            /**
             * Returns the parent name space of this name space.
             *
             * @return a <code>NameSpace</code> value
             */
        public NameSpace getParent()
            {
                return parent;
            }

            /**
             * Returns the common name of this name space, i.e. its inner-most name
             * component.
             *
             * @return a <code>NameComponent</code> value
             */
        public NameComponent getCommonName()
            {
                return this.cn;
            }

            /**
             * Shortcut for getName().getType().
             *
             * @return a <code>String</code> value
             */
        public String getType()
            {
                return this.cn.getType();
            }
        
            /**
             * Shortcut for getName().getInstance().
             *
             * @return a <code>String</code> value
             */
        public String getSimpleName()
            {
                return this.cn.getInstance();
            }

            /**
             * Checks whether the given name space is defined inside this name space.
             *
             * @param ns a <code>NameSpace</code> value
             * @return a <code>boolean</code> value
             */
        public boolean contains(NameSpace ns)
            {
                while ( ns != null ) {
                    ns = ns.getParent();
                    
                    if ( this.equals(ns) ) return true;
                }
                return false;
            }
        
            /**
             * Checks whether the given object describes the same name space as this
             * NameSpace instance.
             *
             * @param o an <code>Object</code> value
             * @return true if both describe the same name space, false otherwise
             */
        public boolean equals( Object o )
            {
                if ( ( o != null ) && (o instanceof NameSpace) ) {
                    NameSpace ns2 = (NameSpace) o;
                    
                    if (cn == null) {
                        if (ns2.cn != null) {
                            return false;
                        }
                    }
                    else if ( ! cn.equals(ns2.cn) ) {
                        return false;
                    }

                    if (parent == null) {
                        if (ns2.parent != null ) {
                            return false;
                        }
                    }
                    else if ( ! parent.equals(ns2.parent) ) {
                        return false;
                    }

                    return true;
                }
                return false;
            }
        
            /**
             * Computes a hash code for this NameSpace instance suitable to be used in
             * hash tables.
             *
             * @return an <code>int</code> value
             */
        public int hashCode()
            {
                return cn.hashCode();
            }
        
            /**
             * Returns a String representation of this NameSapce object.
             *
             * @return a <code>String</code> value
             */
        public String toString()
            {
                return (cn == null)
                    ? ""
                    : (parent == null) ? cn.toString() : cn.toString()+','+parent.toString();
            }
        
            ////////////////////////////////////////////////////
            // PRIVATE STUFF
            ////////////////////////////////////////////////////
        
        private NameComponent cn;
        private NameSpace parent;
    }

        /**
         * Creates a new <code>ResourceName</code> with the provided type and instance
         * identifiers inside the given name space.
         *
         * @param type a <code>String</code> value
         * @param instance a <code>String</code> value
         * @param ns a <code>NameSpace</code> value, defaults to NameSpace.DEFAULT_NS if
         * null is provided
         */
    public ResourceName(String type, String instance, NameSpace ns)
        {
            if ( ns == null ) {
                ns = NameSpace.DEFAULT_NS;
            }
            this.commonName = new NameComponent(type, instance);
            this.nameSpace = ns;
        }

        /**
         * Shortcut for getCommonName().getType().
         *
         * @return a <code>String</code> value
         */
    public String
    getType()
        {
            return commonName.getType();
        }
    
        /**
         * Returns the instance identifier component of this resource name.
         *
         * @return a <code>String</code> value
         */
    public String
    getSimpleName()
        {
            return commonName.getInstance();
        }
        
        /**
         * Returns the common name (type and instance identifiers, without namespace
         * information) of the resource described by this resource name).
         *
         * @return a <code>NameComponent</code> value
         */
    public NameComponent
    getCommonName()
        {
            return commonName;
        }
        
        /**
         * Returns the container name space where this resource name is defined.
         *
         * @return a <code>NameSpace</code> value
         */
    public NameSpace
    getNameSpace()
        {
            return nameSpace;
        }
        
        /**
         * Returns the full distinguished name represented by this ResourceName formated
         * as comma separated list of name components. This is essentially the same
         * information included in toString(), but in a more compact form.
         *
         * @return a <code>String</code> value
         *
         * @see #toString()
         */
    public String getDN()
        {
            return commonName.toString()+','+nameSpace.toString();
        }
        
        /**
         * Checks whether this instance and the given object describe the same resource
         * name.
         *
         * @param o an <code>Object</code> value
         * @return true if both instances describe the same resource name, false
         * otherwise.
         */
    public boolean
    equals(Object o)
        {
            if ( o instanceof ResourceName ) {
                ResourceName rn = (ResourceName) o;
            
                return ( this.commonName.equals(rn.commonName)
                         && this.nameSpace.equals(rn.nameSpace) );
            }
            return false;
        }
        
        /**
         * Returns a String representation of this resource name useful for debugging.
         *
         * @return a <code>String</code> value
         */
    public String
    toString()
        {
            return "{cn="+commonName+", ns="+nameSpace+'}';
        }
        
        /**
         * Computes a hash code for this resource name instance suitable for use with
         * hash tables.
         *
         * @return an <code>int</code> value
         */
    public int
    hashCode()
        {
            return commonName.hashCode();
        }

        ////////////////////////////////////////////////////
        // PRIVATE STUFF
        ////////////////////////////////////////////////////
    private NameComponent commonName;
    private NameSpace nameSpace;
}


