// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
package org.isam.exehda.services;

import org.isam.exehda.services.ContextManager.ContextListener;



/**
 * The AdaptEngine service is responsible by the controlling of <i>functional</i>
 * adaptations, i.e. those that affect the algorithm/modify the behavior implemented by
 * the software component being adapted.
 *
 * <p>This service provides a higher level interface for applications make use of context
 * information, releasing application programmers from the task of explicitly interacting
 * with the ContextManager service for defining and releasing context elements.
 *
 * <p>Moreover, it provides features not directly addressed by the ContextManager
 * service, but that builds on the functionality provided by that service. Among such
 * functionalities are the scheduling of actions to be executed once a context element
 * assumes a given state and the context-aware loading of component implementations
 * through <i>functional</i> adaptation policies.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-27 17:36:36 -0300 (Mon, 27 Jun 2005) $ $Revision: 961 $
 */
public interface AdaptEngine
{
        /**
         * Standard name of this service.
         */
    String SERVICE_NAME = "adaptengine";

        /**
         * The code adaptation heuristic is responsible by selecting the specilization of
         * a given base type that is most suitable to the current context state.
         *
         * <p>It gains access to the context state information through an object that
         * implements the ContextInfo interface.
         *
         */
    static interface CodeAdaptHeuristic
    {
            /**
             * Provides a map-like access to context state information.
             *
             */
        public static interface ContextInfo
        {
                /**
                 * Returns the state of the given context element.
                 *
                 * @param ctxName the context element of interest
                 * @return the state of the context element or null if such information
                 * is not currently available
                 */
            String getContext(String ctxName);
        }

            /**
             * Selects an implementation class for the given base type considering the
             * relevant current context elements states available in ctxInfo.
             *
             * <p>The returned type must be a subtype of the given base type, i.e. it
             * must accept a typecast to the given baseType.
             *
             * @param ctxInfo map-like interface provinding access to context information
             * @param baseType the base type for the component being loaded
             * @return the name of the base type specialization class more suitable to be
             * used in considering the current context state.
             */
        String chooseImpl(ContextInfo ctxInfo, String baseType);
    }

        /**
         * Returns the implementation of the given base type suitable to be used
         * considering the current context state.
         *
         * <p>In fact, the selection process is delegated to the installed code
         * adaptation heuristic. The default code adaptation heuristic trys to load a
         * table of adapter from a file named <code>adapters.xml</code> in the
         * application's classpath. That file defines the context elements of interest
         * for the application, as long as the code alternatives for adaptaion enables
         * components, binding such alternatives to especific context element states.
         *
         * Sample adapters.xml:
         *
          <ADAPTENGINE heuristic="default">
            <CONTEXTS>
              <CONTEXT n="ctx1">
                <STATES>...</STATES>
                <FILTER>...</FILTER>
              </CONTEXT>
              <CONTEXT n="ctx2">
                <STATES>...</STATES>
                <FILTER>...</FILTER>
              </CONTEXT>
            <CONTEXTS>
            <CLASS name="BaseType" context="ctx1>
              <ADAPTER state="pda" impl="BaseTypeImpl_awt"/>
              <ADAPTER state="desktop" impl="BaseTypeImpl_swing"/>
            </CLASS>
            <CLASS name="BaseType2" context="ctx2>
              <ADAPTER state="s1" impl="BaseType2Impl_s1"/>
              <ADAPTER state="s2" impl="BaseType2Impl_s2"/>
            </CLASS>
          </ADAPTENGINE>
         *
         * @param baseType The base type (a class fully qualified name) for which a
         * specialized implementation should be loaded.
         * @return Object
         */
    Object getInstance(String baseType);

        /**
         * Returns the current state of the given context element.
         *
         * @param ctxName context element name
         * @return String
         */
    String checkContextState(String ctxName);

        /**
         * Blocks until the context element assumes the given state.
         *
         * @param ctxState the context element state it should wait for.
         */
    void waitForContext(String ctxState);

        /**
         * Schedules an action to be executed when the context element assumes the given
         * state.
         *
         * @param action the action to be scheduled
         * @param ctxState the context state up on occurence the action should be run
         */
    void runOnContext(Runnable action, String ctxState);

        /**
         * Register a listener for changes in state of the given context element.
         *
         * @param l a context listener to be registered
         * @param ctxName the context of interest
         */
    void addContextListener(ContextListener l, String ctxName);

        /**
         * Operation
         *
         * @param l a context listener to be removed
         */
    void removeContextListener(ContextListener l);
}

