// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services;

import org.isam.exehda.ApplicationId;


/**
 * Context Recognition Service.
 *
 * <p>The ContextManager takes as input a context definition in XML. It then proceeds
 * registering itself within the Collector service and accomplishing any necessary sensor
 * parameter configuration.
 *
 * <p>The raw-level data gathered from the collector's sensors is then filtered in order to
 * produce high-level context information.
 *
 * <p>The whole application execution context is built up from a a set of context elements
 * (<code>Context</code> objects) and its associated states (<code>ContextState</code>
 * objects).
 *
 * @todo describe the context definition language
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-19 20:02:57 -0300 (Sun, 19 Jun 2005) $ $Revision: 942 $
 */
public interface ContextManager
{
    public static final String SERVICE_NAME = "ctxmanager";

        /**
         * Represents a context element.
         *
         */
    public static class Context
    {
            /**
             * Returns the name of this context element.
             *
             * @return a <code>String</code> value
             */
        public String
        getName()
            {
                return this.name;
            }
            
            /**
             * Returns the scope (application) where this context element was defined.
             *
             * <p>The scope information is necessary to avoid conflicts and inconsistencies
             * when comparing context elements defined by different applications.
             *
             * @return an <code>ApplicationId</code> value
             */
        public ApplicationId
        getScope()
            {
                return this.scope;
            }
            
            /**
             * Returns a <code>String</code> containing the XML definition used to create
             * this context.
             *
             * @return a <code>String</code> value
             */
        public String
        getDefinition()
            {
                return this.definition;
            }
            
            /**
             * Returns a <code>ContextState</code> object that described the given
             * abstract state of this context element.
             *
             * @param state a <code>String</code> value
             * @return a <code>ContextState</code> value
             */
        public ContextState
        getStateInstance( String state )
            {
                for (int i=0; i<states.length; i++) {                    
                    if (states[i].getValue().equals(state)) return states[i];
                }
                throw new NullPointerException("No such context state: "+state);
            }

        
            /**
             * Returns whether this object and the given object describe the same context
             * element.
             *
             * @param o an <code>Object</code> value
             * @return a <code>boolean</code> value
             */
        public boolean
        equals( Object o )
            {
                if ( o != null && o instanceof Context ) {
                    Context ctx2 = (Context) o;

                        // XXX: should also compare states or comparing definitions is
                        // safe enough?
                    return ctx2.name == this.name
                        && ctx2.definition == this.definition
                        && ((ctx2.scope != null)
                            ? ctx2.scope.equals(this.scope)
                            : ctx2.scope == this.scope);
                }
                return false;
            }

            /**
             * Computes a hash code for this context object.
             *
             * @return an <code>int</code> value
             */
        public int
        hashCode()
            {
                return this.name.hashCode() ^ this.definition.hashCode();
            }
        
            /**
             * Creates a new <code>Context</code> instance.
             *
             * @param name the context element's name
             * @param scope the application where is being defined
             * @param def the raw XML definition
             * @param states the allowed stated for this element
             */
        public
        Context( String name, ApplicationId scope, String def, String[] states )
            {
                    // XXX: in order to allow for faster comparing in equals(), use the
                    // canonical form for String attributes
                this.name = name.intern();
                this.definition = def.intern();
                this.scope = scope;
                this.states = new ContextState[states.length];

                    //
                    // TODO: is it worth this.states to be a hashtable for faster lookup?
                    //
                for ( int i=0; i<states.length; i++ ) {
                    this.states[i] = new ContextState(this, states[i]);
                }
            }

        private String name;
        private ApplicationId scope;
        private String definition;
        private ContextState[] states;
    }

    
        /**
         * Describes a context element's state.
         *
         */
    public static class ContextState
    {
            /**
             * Returns the context element to which this context state is bound.
             *
             * @return a <code>Context</code> value
             */
        public Context getType()
            {
                return this.type;
            }

            /**
             * Describe <code>getValue</code> method here.
             *
             * @return a <code>String</code> value
             */
        public String getValue()
            {
                return this.value;
            }
        
            /**
             * Returns whether this and the given object both describe the same context
             * state.
             *
             * @param o an <code>Object</code> value
             * @return a <code>boolean</code> value
             */
        public boolean
        equals( Object o )
            {
                if ( o != null && o instanceof ContextState ) {
                    ContextState cs = (ContextState) o;

                    return value == cs.value
                        && type.equals(cs.type);
                }
                return false;
            }

            /**
             * Computes the hash code for this context state object.
             *
             * @return an <code>int</code> value
             */
        public int
        hashCode()
            {
                return value.hashCode();
            }

            /**
             * Returns a string representation of this context state object.
             *
             * @return a <code>String</code> value
             */
        public String
        toString()
            {
                return type.getName() + "::" + value;
            }
        
            /**
             * Creates a new <code>ContextState</code> instance.
             *
             * @param type a <code>Context</code> value
             * @param value a <code>String</code> value
             */
        ContextState( Context type, String value )
            {
                if ( type == null )
                    throw new NullPointerException("Null context element not allowed.");
                
                if ( value == null )
                    throw new NullPointerException("Null context state value not allowed.");
                
                this.type = type;
                this.value = value.intern();
            }

        private Context type;
        private String  value;
    }

    
        /**
         * Call-back interface for receiving context element change events.
         *
         */
    public static interface ContextListener
    {
            /**
             * Invoked by the ContextManager when a context element's changed its current
             * state.
             *
             * @param prev the previous state, or null if no pervious state is available
             * @param curr the new state
             */
        public void contextChanged( ContextState prev, ContextState curr );
    }
    
        /**
         * Creates/registers a new context element.
         *
         * <p>The context manager is responsible for allocating resources and enabling the
         * necessary sensor to produce that context information.
         *
         * <p>Context are defined through XML documents whose outermost element is
         * <code>CONTEXT</code>. Inside the <code>CONTEXT</code> element, two elements
         * are accepted:
         *
         * <p><ul>
         
         * <li> a <code>STATES</code> element, which defines the allowed states of the
         * context element; and
         
         * <li> a <code>FILTER</code> element, which defines inside of it how the context
         * information should produced.

         * </ul>

         * <p>The way how the filter definition given in <code>FILTER</code> must be
         * interpreted is controlled through the attributes <code>syntax</code>. It's
         * conceived that at some time filters would be expressed in different syntaxes
         * and with idioms specialized for some scenarios. Later, even script languages
         * would be used to produce more sophisticated filters.
         *
         * <p>The sample context definition below uses a filter defined with the XML syntax,
         * using the NumericRangeMapping idiom, which maps numeric values read from
         * sensors to abstract values using ranges for that task:
         * <pre>

           &lt;CONTEXT n="context1"&gt;
           
             &lt;STATES&gt;
               &lt;STATE n="stt1"/&gt;
               &lt;STATE n="stt2"/&gt;
               &lt;STATE n="stt3"/&gt;
             &lt;/STATES&gt;

             &lt;FILTER syntax="xml:NumericRangeMapping"&gt; 
               &lt;INDEX&gt;
                 &lt;SWITCH&gt;
                    &lt;SENSOR n="s1" scale="10.0"/&gt;
                    &lt;SENSOR n="s2" scale="3.3" /&gt;
                    &lt;COMPOSITE type="sum"&gt;
                       &lt;SENSOR n="s1" scale="10.0"/&gt;
                       &lt;SENSOR n="s2" scale="3.3" /&gt;
                    &lt;/COMPOSITE&gt;
                 &lt;/SWITCH&gt;
               &lt;/INDEX&gt;
               &lt;MAPPINGS&gt;
                 &lt;RANGE ub="-1" state="stt1"/&gt;
                 &lt;RANGE lb="-1" ub="1" state="stt2"/&gt;
                 &lt;RANGE lb="1" state="stt3"/&gt;
                 &lt;DEFAULT state="stt1"/&gt;
               &lt;/MAPPINGS&gt;
             &lt;/FILTER&gt;
             
           &lt;/CONTEXT&gt;
           
         * </pre>
         *
         * <p>Filter syntaxes shall be standardized sometime in future. Until then, check
         * your ContextManager implementation for information about the filter syntaxes it
         * supports.
         *
         * @param xmlDesc a <code>String</code> value
         * @return a <code>Context</code> value
         */
    public Context createContext( String xmlDesc );
    
        /**
         * Notifies the ContextManager that the given context element is not needed
         * anymore.
         *
         * <p>The ContextManager is then allowed to release resources used to produce that
         * context information (e.g., disabling no longer needed sensors).
         *
         * @param ctx a <code>Context</code> value
         */
    public void releaseContext( Context ctx );
    
        /**
         * Registers a listener for events on the given context element.
         *
         * @param l the call back
         * @param ctx the context element of interest
         */
    public void addContextListener( ContextListener l, Context ctx );
    
        /**
         * Removes a previously registered context listener.
         *
         * @param l a <code>ContextListener</code> value
         * @param ctx a <code>Context</code> value
         */
    public void removeContextListener( ContextListener l, Context ctx );
}

