// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.ctxm.filter.basic;

import org.isam.exehda.services.ctxm.ContextFilter;
import org.isam.exehda.services.ctxm.Environment;
import org.isam.util.xml.XmlElement;
import java.io.StringReader;
import java.io.IOException;
import java.util.Vector;


/**
 * Abstract super-class of all single-sensor filters.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-18 23:03:54 -0300 (Sat, 18 Jun 2005) $ $Revision: 936 $
 */
abstract class AbstractSingleSensorFilter
    extends ContextFilter
{
    public String
    filter( Environment env )
        {
                // check input has changed since last filtering
            if ( env.isModifiedSince(sname, lastUpdate) ) {
                    // save last modification time
                lastUpdate = env.getLastModified(sname);
                    // start at the default state, since either monitoring data may not
                    // be available or the range matching may fail
                this.currState = this.dfltState;
                    // check input data is available
                if ( env.isDefined(sname) ) {
                        // update current state using filter-specific
                        // mapping techniques
                    updateCurrentState(env);
                }
            }
            return currState;
        }


        //////////////////////////////////////////////////////
        // PROTECTED STUFF
        //////////////////////////////////////////////////////

        /**
         * Creates a new <code>AbstractSingleSensorFilter</code> instance.
         *
         * @param fdef a filter definition
         */
    protected AbstractSingleSensorFilter(String fdef)
        {
            super(fdef);
            
            try {
                parseFilter();

                this.currState = this.dfltState;
                    // XXX: force first state update
                this.lastUpdate = Long.MIN_VALUE;
            }
            catch (IllegalArgumentException ie) {
                throw ie;
            }
            catch (Exception e) {
                throw new IllegalArgumentException(
                    "Invalid context filter definition: "+e);
            }            
        }
    
        /**
         * Fills this filter with parameters from the context filter definition.
         *
         * @todo Review this code, its not very safe/robust in face of malformed context
         * definitions.
         */
    private final void parseFilter()
        throws IOException
        {
                // parse filter definition
            XmlElement docroot = XmlElement.parse(
                new java.io.StringReader(this.def));
                // setup mapping rules
            initMappings(docroot);
                // initialize input index
            initIndex(docroot);
        }
    
        /**
         * Initialize the input index.
         *
         * Notice: sensors will be automatically enabled by the environment up-on the
         * first access.
         *
         * Accepts the default <code>INDEX</code> syntax:
         *
           <pre>
           <INDEX><SENSOR n="sensorName"/><INDEX>
           </pre>
         *
         * @param docroot a <code>XmlElement</code> value
         */
    protected void 
    initIndex( XmlElement docroot )
        {
            final String XP_SENSOR     = "/FILTER/INDEX/SENSOR";
            
                // extract context element input sensor
            Vector tmp = XmlElement.selectByPath(docroot, XP_SENSOR, null);
            this.sname = ((XmlElement) tmp.elementAt(0)).getAttribute("n", null);

            if ( this.sname == null ) {
                throw new IllegalArgumentException(
                    "Missing sensor name in context index definition");
            }
            else {
                sname = "sensor:"+sname;
            }
        }
    
        /**
         * Initialize filter-specific mapping rules.
         *
         * @param docroot a <code>XmlElement</code> value
         */
    protected abstract void
    initMappings( XmlElement docroot );
    
        /**
         * Applys filter-specific mapping rules to the input environment variable
         * <code>sname</code>, updating currState accordingly.
         *
         * @param env an <code>Environment</code> value
         */
    protected abstract void
    updateCurrentState( Environment env );
    

    protected String sname;
    protected String dfltState;
    protected long   lastUpdate;
    protected String currState;
}
