/*
 * NeighborhoodBroker.java
 *
 * Created on November 11, 2004, 3:30 PM
 */

package org.isam.exehda.services.perdis;

import org.isam.exehda.services.Logger.Sev;
import org.isam.exehda.services.Logger;
import org.isam.exehda.services.Worb;
import org.isam.exehda.services.DiscovererService;
import org.isam.exehda.Exehda;
import org.isam.exehda.services.CellInformationBase;
import org.isam.exehda.services.CellInformationBase.NameSpace;
import org.isam.exehda.services.CellInformationBase.ResourceName;
import org.isam.util.xml.XmlElement;

import java.util.StringTokenizer;
import java.util.Random;
import java.util.Vector;

/**
 *
 * @author  egon
 */
public class NeighborhoodBroker {
    
    private CellInformationBase cib;
    private NameSpace nameSpace;
    private Vector recentRequests;
    private Logger logger;
    private Worb worb;
    
    private final double PROBABILITY_BROADCAST = 0.8;
    
    /** 
     * Creates a new instance of NeighborhoodBroker 
     */
    public NeighborhoodBroker(CellInformationBase cib, NameSpace nameSpace) {
        
        this.cib = cib; 
        this.nameSpace = nameSpace;
        this.logger = ((Logger) Exehda.getService(Logger.SERVICE_NAME));
        this.worb = ((Worb) Exehda.getService(Worb.SERVICE_NAME));
    }
    
    /**
     * propagateRequest
     */
    public XmlElement propagateRequest(XmlElement xmlRequest, XmlElement xmlResults) throws Exception {
    
        long requestTTL = this.getAndUpdateRequestTTL(xmlRequest);
        
        if (requestTTL > 1) {            
        
            Vector propagators = new Vector();            
            DiscovererService[] neighbors = this.getNeighbors();
            
            for (int i = 0; i < neighbors.length; i++ ) {    
                
                NeighborhoodPropagator neighborhoodPropagator = new NeighborhoodPropagator(neighbors[i], this, xmlRequest, xmlResults);
                neighborhoodPropagator.start();
                propagators.add(neighborhoodPropagator);
            }
            
            for (int i = 0; i < propagators.size(); i++ ) {    
                
                NeighborhoodPropagator neighborhoodPropagatorTmp = (NeighborhoodPropagator) propagators.get(i);
                neighborhoodPropagatorTmp.join();
            }
        }
        return xmlResults;
    }
    
    /**
     * getAndUpdateRequestTTL
     */
    private long getAndUpdateRequestTTL(XmlElement xmlRequest) {
    
        Vector vecTTL = XmlElement.selectByPath(xmlRequest, "/QUERY/TTL", null);
        XmlElement xmlTTL = (XmlElement) vecTTL.firstElement();
        long requestTTL = (new Long(xmlTTL.getContents().toString())).longValue();
        
        this.logger.log(Logger.Sev.DEBUG,"NeighborhoodBroker","getAndUpdateRequestTTL","######### TTL = " + requestTTL, null);
        
        long newRequestTTL = requestTTL - 1;
        xmlTTL.clearContents();
        xmlTTL.addChild(XmlElement.createPCData(new Long(newRequestTTL).toString()));
        
        return requestTTL;
    }
    
    /**
     * getNeighbors
     */
    private DiscovererService[] getNeighbors() {
                
        //Do initial select on CIB looking for neighbors
        ResourceName[] rnCells = this.cib.selectByType("neighbor", NameSpace.DEFAULT_NS, -1);
        
        //Get only the cells selected by the probabilistic protocol
        ResourceName[] rankedNeighborCells = this.rankNeighborCells(rnCells);
        
        //Create the array which will be returned
        DiscovererService[] rankedNeighborDBs = new DiscovererService[rankedNeighborCells.length];        
        
        //For each ranked cell, it needs to get a Discoverer reference
        for (int i = 0; i < rankedNeighborCells.length; i++) {
            
            this.logger.log(Logger.Sev.DEBUG,"NeighborhoodBroker","getNeighbors","######### NeighborCell (DN) = " + rankedNeighborCells[i].getDN(), null);
            
            //Get the reference for the remote CIB
            String remoteCibContactAddressTmp = this.cib.getAttribute(rankedNeighborCells[i], "contactAddress");
            CellInformationBase cibTmp = (CellInformationBase) this.worb.lookupService(remoteCibContactAddressTmp, CellInformationBase.class);
            
            //Get the remote Discoverer on remote CIB
            ResourceName[] rnDiscoverers = cibTmp.selectByName(DiscovererService.SERVICE_NAME, NameSpace.DEFAULT_NS, -1);
            
            //Select just one instance of DB on each cell
            if (rnDiscoverers.length > 0) {
            
                String contactAddressList = cibTmp.getAttribute(rnDiscoverers[0], "contactAddress");
                StringTokenizer st = new StringTokenizer(contactAddressList, ";");
                while(st.hasMoreTokens()) {
        
                    String token = st.nextToken(); 
                    try {
                        DiscovererService discovererTmp = (DiscovererService) worb.lookupService(token, DiscovererService.class);
                        rankedNeighborDBs[i] = discovererTmp;
                        this.logger.log(Logger.Sev.DEBUG,"NeighborhoodBroker","getNeighbors","######### Looked up for = " + token, null);
                        break;
                    } catch(Exception e) {
                        this.logger.log(Logger.Sev.ERROR,"NeighborhoodBroker", "Failed trying to lookup informed contact address: " + token, null);
                        rankedNeighborDBs[i] = null;
                        e.printStackTrace();
                    }
                }
            } else {
                this.logger.log(Logger.Sev.DEBUG,"NeighborhoodBroker","getNeighbors","######### DiscovererBases not found on " + remoteCibContactAddressTmp, null);
                rankedNeighborDBs[i] = null;
            }
        }
        
        return rankedNeighborDBs;
    }
    
    /**
     * rankNeighbors
     */
    private ResourceName[] rankNeighborCells(ResourceName[] rnList) {
        
        int totalNeighborAmount = rnList.length;
        int rankedNeighborAmount = (int) Math.ceil(totalNeighborAmount * this.PROBABILITY_BROADCAST);
        int randomValue = (new Random(System.currentTimeMillis())).nextInt(rankedNeighborAmount);
        
        this.logger.log(Logger.Sev.DEBUG,"NeighborhoodBroker","rankNeighborCells","######### rankedNeighborAmount = " + rankedNeighborAmount, null);
        
        ResourceName[] rankedNeighborCells = new ResourceName[rankedNeighborAmount];
        for (int i = 0; i < rankedNeighborAmount; i++) {
            rankedNeighborCells[i] = rnList[(i + randomValue) % totalNeighborAmount];
        }
            
        return rankedNeighborCells;
    }    
    
    /**
     * mergeResults
     */
    protected synchronized XmlElement mergeResults(XmlElement xmlResult, XmlElement xmlResultTmp) {
    
        if (xmlResultTmp == null) {
            return xmlResult;
        }
        
        Vector vecSrc = XmlElement.selectByPath(xmlResultTmp,"/RESOURCE",null);
        
        for (int i = 0; i < vecSrc.size(); i++){
            xmlResult.addChild((XmlElement) vecSrc.elementAt(i));
        }
        
        return xmlResult;
    }
}


/**
 *
 * @author  egon
 */
class NeighborhoodPropagator extends Thread {
    
    private DiscovererService discoverer;
    private NeighborhoodBroker neighborhoodBroker;
    private XmlElement xmlRequest;
    private XmlElement xmlResult;
    
    /** 
     * Creates a new instance of NeighborhoodPropagator 
     */
    public NeighborhoodPropagator(DiscovererService discoverer, NeighborhoodBroker neighborhoodBroker, XmlElement xmlRequest, XmlElement xmlResult) {
        
        this.discoverer = discoverer;
        this.neighborhoodBroker = neighborhoodBroker;
        this.xmlRequest = xmlRequest;
        this.xmlResult = xmlResult;
    }
    
    /**
     * run
     */
    public void run() {
        
        try {
            XmlElement xmlResultTmp = this.discoverer.requestResource(this.xmlRequest);
            this.neighborhoodBroker.mergeResults(this.xmlResult, xmlResultTmp);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
}