// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.primos;

import org.isam.exehda.Exehda;
import org.isam.exehda.HostId;
import org.isam.exehda.ResourceName;
import org.isam.exehda.services.Collector;
import org.isam.exehda.services.Collector.Sensor;
import org.isam.exehda.services.Collector.SensorParameter;
import org.isam.exehda.services.Collector.MonitoringConsumer;
import org.isam.exehda.services.Collector.MonitoringData;
import org.isam.exehda.services.Collector.ConsumerId;
import org.isam.exehda.services.Collector.Monitor;
import org.isam.exehda.services.AbstractService;
import org.isam.exehda.services.Worb;
import org.isam.exehda.services.Logger.Sev;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Node implementantion of the collector service.
 *
 * @todo handle add/remove consumers
 * @todo probeSensor
 * @todo threashold meta parameter
 * @todo create views of each sensor for each consumer
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-19 15:51:38 -0300 (Sun, 19 Jun 2005) $ $Revision: 941 $
 */
public class CollectorImplNode
    extends AbstractService
    implements Collector, Runnable
{
        /** maximum number of monitors the collector will try to load at bootstraptime */
    private static final int MAX_BOOTSTRAP_MONITORS=32;
        /** collector update cicle period (in milliseconds) */
    private static final long DFLT_QUANTUM_MS = 2000;
    
        /**
         * Creates a new <code>CollectorImplNode</code> instance.
         *
         */
    public CollectorImplNode()
        {
            super();
            this.monitorsSet = new Hashtable();
            this.sensorsByKey = new Hashtable();
            this.consumersById = new Hashtable();
        }
    
        /**
         * Starts the service, exporting it to the WORB.
         *
         * @todo Load collector name from service properties
         */
    public void
    start()
        {
                // load monitors configured in the profile
            for (int i=0; i<MAX_BOOTSTRAP_MONITORS; i++) {
                String mclass = Exehda.getServiceProperty(
                    Collector.SERVICE_NAME, "monitor."+i, null);

                if ( mclass != null ) {
                    try {
                        Monitor m = (Monitor) Class.forName(mclass).newInstance();
                        addMonitor(m);
                        log(Sev.INFO, "start()", "Loaded monitor {0}", mclass);
                    }
                    catch (Exception e) {
                        log(Sev.ERROR, "start()", "Failed to load monitor #"+i);
                        log(Sev.ERROR, "start()", e);
                    }
                }
                else {
                    break;
                }
            }
            
            Worb worb = (Worb) Exehda.getService(Worb.SERVICE_NAME);
            worb.exportService(this,
                               Collector.class,
                               Collector.SERVICE_NAME);
            
            this.poolingTimer = new Timer(this, DFLT_QUANTUM_MS);
            this.poolingTimer.start();
        }
    
        /**
         * Describe <code>stop</code> method here.
         *
         * @todo Unexport service
         */
    public void
    stop()
        {
                // stop timer
            this.poolingTimer.interrupt();
            
                // TO DO: unexport service
        }

        /**
         * Timer callback
         *
         * @todo update meta sensors.
         *
         */
    public void
    run()
        {
//             log(Sev.DEBUG, "run()",
//                 "quantum expired time={0}", Long.toString(System.currentTimeMillis()/100));

                // trigger sensor pooling for underlaying monitors
            Enumeration emon = monitors();
            while ( emon.hasMoreElements() ) {
                Monitor m = (Monitor) emon.nextElement();
                m.quantumExpired();
            }
                // time stamp of this update step
            long tsNow = System.currentTimeMillis();
            
                // avoid allocating new MonitoringData array each run
            if ( mdataBuff.length < sensorsByKey.size()+1 ) {
                mdataBuff = new MonitoringData[sensorsByKey.size()+1];
            }            
                // build data update snapshot
            Enumeration esen = sensorsByKey.elements();
            int i = 0;
            while (esen.hasMoreElements()) {
                MetaSensor s = (MetaSensor) esen.nextElement();

                if ( s.isEnabled() ) {                
                    MonitoringData d = s.probe(null);
                    if ( s.checkStateChanged() ) {
                        s.update(tsNow);
                            //
                            // XXX: cached monitoring data knows nothing about current
                            // time stamp, since it was sampled prior to the call to
                            // s.update(tsNow)
                            //
                        d.setTimeStamp(tsNow);
                        
                        mdataBuff[i++] = d;
                    }
                }
            }
            mdataBuff[i] = null;

                // if there is data to be delivered to consumers
            if ( i > 0 ) {
                    // notify consumers
                    //
                    // FIX ME: should be a separated thread
                    //
                Enumeration econ = consumersById.elements();
                while ( econ.hasMoreElements() ) {
                    MetaConsumer c = (MetaConsumer) econ.nextElement();
                    c.update(tsNow, mdataBuff);
                    log(Sev.DEBUG, "run", "invoking update on MetaConsumer {0}", c);
                }
            }
        }
    
        /**
         * Returns the name of this collector
         *
         * @return a <code>ResourceName</code> value
         */
    public ResourceName getName()
        {
            return new ResourceName("service",
                                    Collector.SERVICE_NAME,
                                    HostId.getLocalHost().toNameSpace());
        }
    
        /**
         * Returns an array containing the sensors currently provided by this collector.
         *
         * @return an array of Sensor objects
         */
    public synchronized Sensor[]
    getSensors()
        {
//             log(Sev.DEBUG, "getSensors()", "entering");
//             log(Sev.DEBUG, "getSensors()", "sensorsByKey={0}",sensorsByKey);
            
            Sensor[] result = new Sensor[sensorsByKey.size()];

            Enumeration enum = sensorsByKey.elements();
            
            for (int i=0; enum.hasMoreElements(); i++) {
                result[i] = ((MetaSensor) enum.nextElement()).getPublicSensor();
            }
//             log(Sev.DEBUG, "getSensors()", "leaving");
            return result;
        }
    
        /**
         * Look ups a sensor with the specified name.
         *
         * The provided name is compared to the sensor simple name.
         *
         * @todo what if two sensors have the same simplename? perhaps, should use a
         *        ResourceName instead of an String.
         *
         * @todo optimize search;
         *
         * @param name sensor name 
         * @return a sensor object or null if the sensor does not exist.
         */
    public Sensor
    getSensor( String name )
        {
            Enumeration senum = sensorsByKey.elements();

            while ( senum.hasMoreElements() ) {
                Sensor s = ((MetaSensor) senum.nextElement()).getPublicSensor();

                if (name.equals(s.getName().getSimpleName())) {
                    return s;
                }
            }

            return null;
        }

        /**
         * Returns a list of the parameters supported by the given sensor.
         *
         * @param s a <code>Sensor</code> value
         * @return a <code>SensorParameter[]</code> value
         */
    public SensorParameter[]
    getSensorParameters( Sensor s )
        {
            return findMetaSensor(s).getParameters();
        }
    
        /**
         * Returns the current value of the given parameter.
         *
         * @param p a <code>SensorParameter</code> value
         * @return an <code>Object</code> value
         */
    public Object
    getSensorParameter( SensorParameter p )
        {
            return findMetaSensor(p.getSensor()).getParameter(p);
        }

        /**
         * Applyies a new value to the given sensor parameter.
         *
         * @todo check if operation allowed for the given consumer
         *
         * @param c a <code>ConsumerId</code> value
         * @param p a <code>SensorParameter</code> value
         * @param v an <code>Object</code> value
         */
    public void
    setSensorParameter( ConsumerId c, SensorParameter p, Object v )
        {
            if ( c != null ) {
                // TODO: check consumer is allowed to do this operation

                findMetaSensor(p.getSensor()).setParameter(c,p,v);
            }
            else throw new IllegalArgumentException("Invalid consumer id");
        }

        /**
         * Enables or disables the given sensor.
         *
         * @param cid a <code>ConsumerId</code> value
         * @param s a <code>Sensor</code> value
         * @param enable a <code>boolean</code> value
         */
    public void
    setSensorEnabled( ConsumerId cid, Sensor s, boolean enable )
        {
            if ( cid != null ) {
                    // TO DO: check consumer is allowed to do this operation
                
                    // FIX ME:
                MetaConsumer c = (MetaConsumer) consumersById.get(cid);
                if ( c != null ) {
                    MetaSensor ms = findMetaSensor(s);
                    
                    if ( enable ) {
                        c.addEnabledSensor(ms);
                    }
                    else {
                        c.removeEnabledSensor(ms);
                    }
                }
            }
            else throw new IllegalArgumentException("Invalid consumer id");
        }

        /**
         * Returns whether the given sensors is currently enabled.
         *
         * @param c a <code>ConsumerId</code> value
         * @param s a <code>Sensor</code> value
         * @return a <code>boolean</code> value
         */
    public boolean
    isSensorEnabled( ConsumerId c, Sensor s )
        {
            if ( c != null ) {
                    // TODO: check consumer is allowed to do this operation

                    // FIX ME: consumerid is not used
                return findMetaSensor(s).isEnabled();
            }
            else throw new IllegalArgumentException("Invalid consumer id");
        }

        /**
         * Returns the current sensor value.
         *
         * @param c a <code>ConsumerId</code> value
         * @param s a <code>Sensor</code> value
         * @param d a <code>MonitoringData</code> value
         * @return a <code>MonitoringData</code> value
         */
    public MonitoringData
    probeSensor( ConsumerId c, Sensor s, MonitoringData d )
        {
            if ( c != null ) {
                    // TODO: check consumer is allowed to do this operation

                    // FIX ME: consumerid is not used
                return findMetaSensor(s).probe(d);
            }
            else throw new IllegalArgumentException("Invalid consumer id");
        }

        /**
         * Describe <code>installSensor</code> method here.
         *
         * @param c a <code>ConsumerId</code> value
         * @param name a <code>String</code> value
         * @param impl a <code>String</code> value
         * @param params a <code>Hashtable</code> value
         */
    public void
    installSensor( ConsumerId c, String name, String impl, Hashtable params )
        {
            throw new NullPointerException("Not yet implemented");
        }

        /**
         * Describe <code>addMonitor</code> method here.
         *
         * @param m a <code>Monitor</code> value
         */
    public synchronized void
    addMonitor( Monitor m )
        {
            log(Sev.DEBUG, "addMonitor", "entering");
                // TODO: check synchronization
            if ( monitorsSet.get(m) == null ) {
                monitorsSet.put(m,m);

                Sensor[] sensors = m.getSensors();
                
                log(Sev.DEBUG, "addMonitor", "sensors.length={0}", ""+sensors.length);
                            
                for (int i=0; i<sensors.length; i++) {
                    registerSensor(m,sensors[i]);
                }

                log(Sev.DEBUG, "addMonitor", "sensorsByKey.size()={0}", ""+sensorsByKey.size());
            }

                // FIX ME:
            m.quantumExpired();
            
            log(Sev.DEBUG, "addMonitor", "leaving");
        }

        /**
         * Describe <code>removeMonitor</code> method here.
         *
         * @param m a <code>Monitor</code> value
         */
    public synchronized void
    removeMonitor( Monitor m )
        {
                // TODO: check synchronization
            if ( monitorsSet.get(m) != null ) {
                
                    // delete registered sensors
                Vector toRemove = new Vector();
                
                Enumeration senum = sensorsByKey.elements();
                while (senum.hasMoreElements()) {

                    MetaSensor sh = (MetaSensor) senum.nextElement();
                        // FIX ME:
                    if ( m == sh.getProvider() ) {
                        toRemove.add(sh);
                    }
                }
                    // commit deletion
                senum = toRemove.elements();
                while (senum.hasMoreElements()) {
                    sensorsByKey.remove(
                        ((MetaSensor) senum.nextElement()).getKey());
                }
                    // delete the monitor itself
                monitorsSet.remove(m);
            }
        }

        /**
         * Describe <code>addConsumer</code> method here.
         *
         * @todo complete the implementation
         *
         * @param callback a call-back for receiving monitoring data
         * @return a <code>ConsumerId</code> value
         */
    public ConsumerId
    addConsumer( MonitoringConsumer callback )
        {
            log(Sev.DEBUG, "addConsumer", "consumersById.size()={0}", ""+consumersById.size());
            MetaConsumer c = new MetaConsumer(genConsumerId(), callback);
            consumersById.put(c.getId(), c);
            
            log(Sev.DEBUG, "addConsumer", "consumersById.size()={0}", ""+consumersById.size());
            log(Sev.DEBUG, "addConsumer", "consumersById={0}", consumersById);
            
            return c.getId();
        }

        /**
         * Describe <code>removeConsumer</code> method here.
         *
         * @param c a <code>ConsumerId</code> value
         */
    public void
    removeConsumer( ConsumerId c )
        {
                // FIX ME: need better clean up, i.e. realease sensors
            consumersById.remove(c);
        }

        // PRIVATE METHODS //////////////////////////

    private final Enumeration
    monitors()
        {
            return monitorsSet.keys();
        }

    private final MetaSensor
    findMetaSensor(Sensor s)
        {
            MetaSensor ms = (MetaSensor) sensorsByKey.get(s.getKey());

            if ( ms == null )
                throw new IllegalArgumentException("No such sensor");
            else return  ms;
        }

    private final void
    registerSensor(Monitor m, Sensor s)
        {            
            MetaSensor ms = new MetaSensor(s,m);
            sensorsByKey.put(ms.getKey(), ms);
        }


    private final ConsumerId
    genConsumerId()
        {
            return new ConsumerId(new Long(nextConsumerID++), getName(), null);
        }

    private final boolean
    checkParameterAssignAllowed(SensorParameter p, Object v)
        {
                // TO DO:
            throw new NullPointerException("Not yet implemented");
        }

        // ATTRIBUTES ///////////////////////////////
    private Hashtable monitorsSet;
    private Hashtable sensorsByKey;
    private Hashtable consumersById;
    private Timer     poolingTimer;

        // buffer for monitoring data delivery
    private MonitoringData[] mdataBuff = new MonitoringData[1];

    private static long nextConsumerID = System.currentTimeMillis();
}


/**
 * A Timer to control the frequency of monitor pooling operations.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-19 15:51:38 -0300 (Sun, 19 Jun 2005) $ $Revision: 941 $
 */
class Timer extends Thread
{
    public Timer(Runnable callback, long quantumMillis)
        {
            this.callback = callback;
            this.quantum = quantumMillis;
        }

    public void
    run()
        {
            while (! Thread.interrupted() ) {
                long next = System.currentTimeMillis() + quantum;
                
                callback.run();

                long delay = next - System.currentTimeMillis();

                if ( delay > 0 ) {
                    try { Thread.sleep(delay); }
                    catch (InterruptedException ie) {}
                }
            }
        }

    private Runnable callback;
    private long quantum;
}




