// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.cc;

import org.isam.exehda.services.CCManager.Tuple;
import java.util.HashMap;
import java.util.LinkedList;


/**
 * This class implements a tuple space using a hash tree for the first 
 * <i>hashlevel</i> levels and then linked lists to store the tuples.
 *
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2004-06-16 12:16:23 -0300 (Wed, 16 Jun 2004) $ $Revision: 231 $
 */
public class HashTreeSpace
{
        /** this tuple space name (unique id) */
    private String name;
        /** The root node of the hashtree used by this space */
    private HashTreeSpaceNode root;
        /** number of hashtree levels to be used */
    int hashlevel;

    
        /**
         * Builds a hash tree space that will be at most <code>hashlevel</code>
         * height, above which linked lists will be used for leaf nodes;
         *
         * @param hashlevel an <code>int</code> value
         * @param name a <code>String</code> value
         */
    public HashTreeSpace( int hashlevel, String name)
        {
            this.name = name;
            this.hashlevel = hashlevel;
            this.root = new HashTreeSpaceNode(null, 0, this);
        }

        /**
         * Gets a tuple that matches the given pattern from the TupleSpace. The returned
         * tuple is removed from the object space.
         *
         * This call would block until a tuple that matches the given pattern becames
         * available.
         *
         * @param pattern the pattern to which the returned tuple must adhere.
         * @return a tuple that matches the given pattern.         
         */
    public synchronized Tuple take( Tuple pattern, long timeout )
        {
            while (true) {
                HashTreeSpaceNode parent = root.findParent(pattern);
                if ( parent != null ){
                    Tuple t = parent.findTuple(pattern, true);
                    if ( t != null ) {
                        return t;
                    }
                }
                if ( timeout > 0 ) {
                    try { 
                        wait(timeout);
                        timeout = 0;
                    }
                    catch (InterruptedException ie) {}
                }
                else {
                    return null;
                }
            }
        }

        /**
         * Gets a tuple that matches the given pattern from the tuple space.
         *
         * A copy of the tuple is returned instead of removing the original tuple. If
         * block is set to false and no such tuple is found, null is returned instead.
         *
         * @param pattern the pattern to which the returned tuple must adhere.
         * @param block operation mode
         * @return a tuple that matches the given pattern or null if none is available
         * and non-blocking operation mode was selected.
         */
    public Tuple read( Tuple pattern, long timeout )
        {
            while (true) {
                HashTreeSpaceNode parent = root.findParent(pattern);
                if ( parent != null ){
                    Tuple t = parent.findTuple(pattern, false);
                    if ( t != null ) {
                        return t;
                    }
                }
                if ( timeout > 0 ) {
                    try { 
                        wait(timeout);
                        timeout = 0;
                    }
                    catch (InterruptedException ie){}
                }
                else {
                    return null;
                }
            }
        }

    
        /**
         * Inserts the given tuple into the space.
         *
         * @param t a <code>Tuple</code> value
         */
    public synchronized void put(Tuple t)
        {
            root.insertTuple(t);
            notifyAll();
        }
}



/**
 * Describe class <code>HashTreeSpaceNode</code> here.
 *
 * This class need java Collection API. It will not compile on J2ME systems.
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2004-06-16 12:16:23 -0300 (Wed, 16 Jun 2004) $ $Revision: 231 $
 */
class HashTreeSpaceNode
{
        /** tag associated with this node */
    private Object tag;
        /** childre nodes of this node, uses hash for faster lookup */
    private HashMap children;
        /** height of this node in the tuple space hash tree */
    private int level;
        /** space who owns this node */
    private HashTreeSpace space;
        /** tuples associated with this node */
    private LinkedList tupleList;

    public HashTreeSpaceNode( Object tag, int level, HashTreeSpace space )
        {
            this.tag = tag;
            this.level = level;
            this.space = space;
            this.children = new HashMap();
            this.tupleList = new LinkedList();
        }

        /**
         * Follows down this node subtree looking for the node that is the parent for tuples
         * that match the given pattern.
         *
         * @param pattern the pattern to be matched
         * @return the HashTreeSpaceNode that owns at least one tuple that
         * matches the pattern, or null if such node does not exist.
         */
    public HashTreeSpaceNode findParent( Tuple pattern )
        {
                //test against end of pattern match
            if ( level >= pattern.length()) { 
                    // in this case any Tuple of the list, or of the
                    // children nodes satisfies the pattern
                if ( !tupleList.isEmpty() ) {
                    return (this);
                }
                else if ( !children.isEmpty() ) {
                        // for each child 
                    java.util.Iterator chldIterator = children.values().iterator();
                    while ( chldIterator.hasNext()) {
                        HashTreeSpaceNode child = (HashTreeSpaceNode) chldIterator.next();
                            // repeats the search 
                        HashTreeSpaceNode result = child.findParent(pattern);
                            //until a node matches the pattern
                        if ( result != null) {
                                // pattern matches
                            return(result);
                        }
                    } //while
                }
                    // pattern not found
                return(null);
            }
            else {
                    // pattern is not fully evaluated yet
                    // if it's a leaf node, 
                if ( level == space.hashlevel ) {
                        // matching has been partialy completed, return this leaf.
                    return (this);
                }
                else { 
                        // not a leaf
                        // get the next tag to be matched    
                    Object nextTag = pattern.get(level);
                        // check if there is a child node for that tag
                    HashTreeSpaceNode child = (HashTreeSpaceNode)children.get(nextTag);
                    if ( child != null) {
                            // continues the matching in the children
                        return (child.findParent(pattern));
                    }
                    else {
                            // no child for the given tag, matching failed!
                        return (null);
                    }
                } //else (not a leaf)
            } //else (partial matching)
        }

        /**
         * Find a tuple in this node's tupleList that matches the given pattern.
         *
         * Typically, two cases would happen:
         *  <ul>
         * 	<li> the current node completely matches the pattern, so any tuple
         * 	     in the nodes list would be returned.
         * 	<li> the current node is a leaf, and it just partially matches
         * 	     the given pattern. Is this case, the remaining tags in the
         * 	     pattern must be checked against the tuples in the tuplesList
         * 	     until we have a complete matching.
         * 	</ul>
         *
         * In both cases, if the flag parameter remove is set to true,
         * the tuple is removed from the tupleList before return. Otherwise,
         * the tuple is kept in the tupleList.
         *
         * This node should be previously selected through findParent. If a pattern that
         * do not match the current node prefix is passed to findTuple() the behavior of
         * this method is unspecified, possibly returning a false matching tuple.
         *
         *
         * @param pattern Pattern to be matched
         * @param remove If true, the tuple is removed, otherwise it is not
         * @return A tuple that matches the pattern or null if no one
         * could be found.
         */
    public Tuple findTuple( Tuple pattern, boolean remove )
        {
                // if the matching is already completed, just return the first tuple
            if (this.level == pattern.length()) {
                return (Tuple) ((remove) ? tupleList.removeFirst() : tupleList.getFirst());
            }
                // for each tuple in the tupleList
            java.util.ListIterator tuplesIter = tupleList.listIterator();
            while ( tuplesIter.hasNext() ) {
                Tuple currentTuple = (Tuple) tuplesIter.next();
                    // complete the pattern matching process starting from the next tag
                if ( currentTuple.matches(pattern, this.level+1) ) {
                        // successfull matching
                        // if flaged to remove, remove the tuple
                    if (remove) {
                        tuplesIter.remove();
                    }
                        // return the matching tuple
                    return currentTuple;
                }//if
            }//while
                // no tuple in the list matches the given pattern
            return (null);
        }

    public void insertTuple( Tuple t)
        {
                // if the tuples size matches this level of the tree or
                // this node is a leaf node, just insert the tuple in the tupleList
            if ( (t.length()==this.level) || (this.level == space.hashlevel) ){
                tupleList.addFirst(t);
            } else {
                    // otherwise, find the proper child 
                HashTreeSpaceNode child = (HashTreeSpaceNode) children.get(t.get(level));
                    // if the child does not exist, create it
                if ( child == null ) {
                    child = new HashTreeSpaceNode(t.get(level), level+1, this.space);
                    children.put(t.get(level), child);
                }// if
                    // delegates the insertion to the child
                child.insertTuple(t);
            }// else (delegate to child)
        }

    public String toString()
        {
            return "HashTreeNode{level="+level+",tag="+tag+"}";
        }
}
