package dimi;

import java.util.Map;
import java.util.List;
import java.util.Hashtable;
import java.util.Iterator;

import java.rmi.RemoteException;
import java.rmi.NotBoundException;
import java.net.MalformedURLException;
import java.rmi.Naming;

import dimi.configuration.ChannelConfiguration;
import dimi.exception.CantCreateChannelException;
import dimi.exception.CantFindChannelException;
import dimi.exception.CantFindDBaseException;
import dimi.exception.CantEnterChannelException;
import dimi.channel.exception.ChannelAlreadyExistsException;
import dimi.DisseminationUnit;
import dimi.channel.connection.DisseminationUnitConnection;
import dimi.name.ChannelName;
import dimi.name.DNodeName;
import dimi.name.DBaseName;
import dimi.name.BadformedDimiNameException;
import dimi.channel.NeighbourhoodManager;
import dimi.channel.neighbour.NeighbourPointer;
import dimi.DBase;
import dimi.ProducerEntryPointImpl;
import dimi.name.DisseminationUnitName;
import dimi.channel.*;

/**
 * Gerente de canais de um disseminador.
 * @author mcmoraes
 */
public abstract class ChannelManager{
    
    //atributos - INICIO
    private Map channels = null;
    private DisseminationUnitImpl disseminationUnitImpl = null;
    private DisseminationUnitName disseminationUnitName = null;
    private long channelCounter = 0;
    //atributos - FIM
    
    /**
     * Creates a new instance of DisseminatorChannels
     * @param disseminationUnit a unidade disseminadora deste ChannelManager
     */
    protected ChannelManager(DisseminationUnitImpl disseminationUnitImpl){
        this.channels = new Hashtable();
        this.disseminationUnitImpl = disseminationUnitImpl;
        this.channelCounter = 0;
        try{this.disseminationUnitName = (DisseminationUnitName)disseminationUnitImpl.getName();}catch(RemoteException re){}
    }

    //construtores - INICIO
    /**
     * Protocolo de inicio de consumo em um canal.
     * @param desiredPeerEntryPointsNumber o numero desejado de pais no canal
     * @param disseminator o disseminador que se tornara filho no canal
     * @param channelName o canal do qual se deseja consumir dados
     * @throws CantFindChannelException se nao for possivel encontrar o canal
     * @throws CantFindPeerException se nao for possivel encontrar o ponto de entrada no canal
     * @throws CantEnterChannelException se nao for possivel entrar no canal
     * @throws RemoteException indicate that either a communication failure or a protocol error has occurred
     */
    public abstract ChannelData enterChannel(ChannelName channelName,int desiredPeerEntryPointsNumber,Disseminator disseminator) throws RemoteException,CantFindChannelException,CantFindDBaseException,CantEnterChannelException;
    //construtores - FIM
    

    //getters e setters - INICIO
    /**
     * Getter for property disseminatorImpl.
     * @return Value of property disseminatorImpl.
     */
    public final DisseminationUnitImpl getDisseminationUnitImpl() {
        return disseminationUnitImpl;
    }
    /**
     * Getter for property channels.
     * @return Value of property channels.
     */
    public final DisseminationUnitChannel[] getChannels() {
        return (DisseminationUnitChannel[])channels.values().toArray();
    }
    /**
     * Getter for property disseminationUnitName.
     * @return Value of property disseminationUnitName.
     */
    public final DisseminationUnitName getDisseminationUnitName() {
        return disseminationUnitName;
    }    
    //getters e setters - FIM
    
    /**
     * Obtem um determinado canal.
     * @param channelName o nome do canal desejado
     * @return o canal desejado
     */    
    public final synchronized DisseminationUnitChannel getChannel(ChannelName channelName) {
        return (DisseminationUnitChannel)channels.get(channelName);
    }

    protected final synchronized DisseminationUnitChannel createChannel(ChannelData channelData,NeighbourhoodManager neighbourhoodManager) throws CantCreateChannelException,ChannelAlreadyExistsException,RemoteException{
        return createChannel(channelData.getChannelRootName(),channelData.getChannelConfiguration(),neighbourhoodManager);
    }
    /**
     * Adiciona um canal ao disseminador
     * @return o canal recem-criado
     * @param channelRootName o nome de raiz do canal
     * @param channelNeighbourhoodManager o gerente de vizinhanca do canal
     * @param channelConfiguration a configuracao do novo canal
     * @throws RemoteException indicate that either a communication failure or a protocol error has occurred
     * @throws CantCreateChannelException se nao foi possivel criar o canal
     */    
    protected final synchronized DisseminationUnitChannel createChannel(ChannelName channelRootName,ChannelConfiguration channelConfiguration,NeighbourhoodManager neighbourhoodManager) throws CantCreateChannelException,ChannelAlreadyExistsException,RemoteException{
        if(channels.containsKey(channelRootName)){
            throw new ChannelAlreadyExistsException();
        }
        ChannelName channelName = null;
        try{
            channelName = new ChannelName(disseminationUnitName + "/" + nextChannelLocalName());
        }
        catch(BadformedDimiNameException bfdne){bfdne.printStackTrace();}

        if(channelRootName == null){
            channelRootName = channelName;
        }
        DisseminationUnitChannel channel = new DisseminationUnitChannel(new ChannelName[]{channelName,channelRootName},channelConfiguration,this,neighbourhoodManager);
        channels.put(channelName,channel);
        channels.put(channelRootName,channel);
        return channel;
    }

    /**
     * Destroi um canal no disseminador.
     * @param channelName o nome do canal a ser destruido
     */    
    public final synchronized void destroyChannel(ChannelName channelName){
        DisseminationUnitChannel disseminatorChannel = getChannel(channelName);
        if(disseminatorChannel != null){
            disseminatorChannel.destroy();
            channels.remove(disseminatorChannel.getNames()[Channel.LOCAL_NAME]);
            channels.remove(disseminatorChannel.getNames()[Channel.ROOT_NAME]);
        }
    }
    
    /**
     * Destroi um canal no disseminador.
     */    
    public final synchronized void destroyAllChannels(){
        Iterator it = channels.values().iterator();
        while(it.hasNext()){
            ((Channel)it.next()).destroy();
        }
    }

    protected final synchronized String nextChannelLocalName(){
        return "channel" + String.valueOf(channelCounter++);
    }

}
