package dimi;

import dimi.message.MessageContent;
import dimi.message.MessageHeader;
import dimi.message.Message;
import dimi.channel.EntryPointChannel;
import dimi.channel.Channel;
import dimi.exception.CantFindChannelException;
import dimi.channel.DisseminationUnitChannel;

import java.rmi.RemoteException;
import java.util.Date;
import java.util.Timer;
import java.util.TimerTask;

/**
 * Uma implementacao de ponto de entrada ao usuario produtor de um canal.
 * @author mcmoraes
 */
public class ProducerEntryPointImpl extends UserEntryPointImpl implements ProducerEntryPoint{
    
    //atributos - INICIO
    private Timer timer = null;
    private long messageSequenceNumber = 0;
    private TimerTask producerAlivenessTimerTask = null;
    private TimerTask hearBeatTimerTask = null;
    //atributos - FIM
    
    //construtores - INICIO
    /**
     * Cria um novo ponto de entrada para um produtor
     * @param name o nome desse ponto de entrada de producao. O nome e o mesmo do respectivo canal.
     * @param disseminatorChannel o canal no qual existe este ponto de entrada
     * @throws RemoteException if failed to export object
     */    
    ProducerEntryPointImpl(DisseminationUnitChannel disseminationUnitChannel) throws RemoteException{
        super(disseminationUnitChannel.getNames()[Channel.ROOT_NAME]);
        setEntryPointChannel(new EntryPointChannel(disseminationUnitChannel,this));
    }
    //construtores - FIM
    
    //interface UserEntryPoint - INICIO
    /**
     * Encerra a producao em um canal. Este metodo deve ser chamado quando
     * o usuario produtor desejar parar a producao.
     * @throws RemoteException indicate that either a communication failure or a protocol error has occurred
     */
    public final void destroy() throws RemoteException{
        //tenta enviar uma mensagem destroy_channel
        MessageHeader messageHeader = mountMessageHeader();
        messageHeader.setType(MessageHeader.DESTROY_CHANNEL);
        try{
            disseminateMessage(new Message(messageHeader));
        }
        catch(CantFindChannelException cfce){
            //se nao conseguir enviar a mensagem destroy_channel, nao faz nada. Deixa que a unidade de disseminacao se vire.
        }
        finally{
            //desativa este ponto de entrada
            hearBeatTimerTask.cancel();
            timer.cancel();
            setActive(false);
        }
    }
    //interface UserEntryPoint - FIM
    
    //interface Disseminator - INICIO
    /**
     * Dissemina dados no respectivo canal, atraves de uma mensagem.
     * @param message a mensagem a ser disseminada
     * @throws CantFindChannelException se nao for possivel encontrar o canal
     * @throws RemoteException indicate that either a communication failure or a protocol error has occurred
     */
    public final synchronized void disseminate(Message message) throws RemoteException,CantFindChannelException{
        //se nao esta ativo
        if(!isActive()){
            //avisa que nao encontrou o canal
            throw new CantFindChannelException();
        }
        //reinicia o temporizador de vida do produtor
        restartProducerAlivenessTimer();
        MessageHeader messageHeader = mountMessageHeader();
        messageHeader.setSequenceNumber(messageSequenceNumber++);
        messageHeader.setType(MessageHeader.DATA);
        messageHeader.setContentSize(message.getContent().getData().length);
        message.setHeader(messageHeader);
        disseminateMessage(message);
    }
    //interface Disseminator - FIM
    
    /** Metodo auxiliar para a utilizacao do ponto de entrada*/
    final void activate(){
        setActive(true);
        this.timer = new Timer();
        hearBeatTimerTask = new HeartBeatMessageProductionTask();
        timer.schedule(hearBeatTimerTask,0,getEntryPointChannel().getConfiguration().getHeartBeatPeriod());
        startProducerAlivenessTimer();
    }

    private void startProducerAlivenessTimer(){
        producerAlivenessTimerTask = new ProducerIsDeadMessageProductionTask();
        timer.schedule(producerAlivenessTimerTask,getEntryPointChannel().getConfiguration().getMaximumSilentInterval());
    }

    private void restartProducerAlivenessTimer(){
        producerAlivenessTimerTask.cancel();
        startProducerAlivenessTimer();
    }

    private void disseminateMessage(Message message) throws RemoteException,CantFindChannelException{
        getEntryPointChannel().disseminate(message);
    }
    
    private class HeartBeatMessageProductionTask extends TimerTask{
        public void run(){
            MessageHeader messageHeader = mountMessageHeader();
            messageHeader.setType(MessageHeader.HEART_BEAT);
            Message message = new Message(messageHeader);
            try{
                disseminateMessage(message);
            }
            catch(RemoteException re){
                try{
                    destroy();
                }
                catch(RemoteException re1){}
            }
            catch(CantFindChannelException cfce){
                try{
                    destroy();
                }
                catch(RemoteException re){}
            }
        }
    }   

    private class ProducerIsDeadMessageProductionTask extends TimerTask{
        public void run(){
            try{
                destroy();
            }
            catch(RemoteException re){}
        }
    }   

    private MessageHeader mountMessageHeader(){
        MessageHeader messageHeader = new MessageHeader();
        try{
            messageHeader.setChannelName(getEntryPointChannel().getNames()[Channel.ROOT_NAME]);
            messageHeader.setOriginName(getName());
            messageHeader.setOriginDateTime(new Date());
            messageHeader.setProductionDateTime(new Date());
        }
        catch(RemoteException re){re.printStackTrace();}
        return messageHeader;
    }

}


