// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.primos;

import org.isam.perv.IsamClassLoader;
import org.isam.exehda.Exehda;
import org.isam.exehda.HostId;
import org.isam.exehda.ObjectId;
import org.isam.exehda.ApplicationId;
import org.isam.exehda.services.Worb;
import org.isam.exehda.services.Executor;
import org.isam.exehda.services.CellInformationBase;
import org.isam.exehda.services.CellInformationBase.ResourceName;
import org.isam.exehda.services.ObjectSeed;
import org.isam.exehda.services.ObjectSeed.Activator;
import org.isam.exehda.services.ObjectSeed.ObjectHandle;
import org.isam.exehda.services.Logger.Sev;
import org.isam.exehda.services.AbstractService;
import org.isam.util.xml.XmlElement;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.io.StringReader;
import java.io.IOException;
import java.net.UnknownHostException;

public class ExecutorImpl
    extends AbstractService
    implements Executor
{
    private final static String DFLT_HEURISTIC = "org.isam.exehda.services.primos.HostIdHeuristic";
    private final static String DFLT_MAX_RETRY = "5";
    private final static boolean NDEBUG = false;
    private int maxRetryies;
    private HostId thisHost;
    private Executor.SchedulingHeuristic defaultHeuristic;
    private final Hashtable appByThrGrp;
    private final Hashtable appById;

    public ExecutorImpl()
        {
            appByThrGrp = new Hashtable();
            appById = new Hashtable();
        }
    
    public void
    start()
        throws ClassNotFoundException, InstantiationException,
        IllegalAccessException, UnknownHostException
        {
            defaultHeuristic = (SchedulingHeuristic) Class.forName(
                Exehda.getServiceProperty(
                    SERVICE_NAME,
                    "defaultHeuristic",
                    DFLT_HEURISTIC)).newInstance();


            maxRetryies = Integer.parseInt(
                Exehda.getServiceProperty(SERVICE_NAME, "maxRetryies", DFLT_MAX_RETRY));
            
            Worb worb = (Worb) Exehda.getService(Worb.SERVICE_NAME);
            worb.exportService(this, Executor.class, Executor.SERVICE_NAME);

            thisHost = HostId.getLocalHost();
        }
    
    public void
    stop() throws java.lang.Exception
        {
                // TODO: unexport service
        }
    
    public void
    startApplication(XmlElement launchDescriptor)
        {
            log(Sev.TRACE, "startApplication", "entering");
            
            String codeBaseURL = null;
            String mainClassName = null;
            String[] appArgs = new String[0];
            String ownerId = null;
            

            Enumeration docEnum = launchDescriptor.getChildren();
                //initial docRoot is not an element of the xml, it 
                //involves the xml.
            docEnum = ((XmlElement) docEnum.nextElement()).getChildren();

            
            while ( docEnum.hasMoreElements() ) {
                XmlElement e = (XmlElement) docEnum.nextElement();
                String tag = e.getTag();
                if ( "code".equals(tag) ) {
                    Enumeration codeEnum = e.getChildren();
                    while ( codeEnum.hasMoreElements() ) {
                        XmlElement ce = (XmlElement) codeEnum.nextElement();
                        if ( "main".equals(ce.getTag()) ) {
                            mainClassName = ce.getAttribute("class", null);
                        }
                        else if ( "jar".equals(ce.getTag()) ) {
                            codeBaseURL = ce.getAttribute("href", null);
                        }
                    }
                }
                else if ( "parameters".equals(tag) ) {
                    Enumeration paramEnum = e.getChildren();
                    Vector v = new Vector();

                    while ( paramEnum.hasMoreElements() ) {
                        XmlElement p = (XmlElement) paramEnum.nextElement();


                        if ( "param".equals(p.getTag()) ) {
                            v.addElement((String) p.getContents());                    
                        }
                        else {
                            log(Sev.ERROR,
                                "createObject",
                                "unexpected tag {0} in launch descriptor",
                                p.getTag());

                                // TODO: ignore or abort, which one is safer?
                        }
                    }

                    appArgs = new String[v.size()];
                    for ( int i=0; i<appArgs.length; i++ ) {
                        appArgs[i] = (String) v.elementAt(i);
                    }
                }
                else if ( "owner".equals(tag) ) {
                    ownerId = e.getAttribute("id", "no-owner");
                }
            }
            
	  
            launch(codeBaseURL, mainClassName, appArgs);
            log(Sev.TRACE, "startApplication", "leaving");
        }

    public ObjectHandle
    createObject( ApplicationId app, Class cls, Object[] params, Activator a, Object hint)
        {
            try {
                log(Sev.TRACE, "createObject", "entering");

                if ( thisHost.equals(hint) ) {
                    log(Sev.DEBUG, "createObject", "Selected host is LOCALHOST");
                    return localCreateObject(app, cls, params, a);
                }
                else {
                
                    Vector avoidedHosts = new Vector();
                    int n = maxRetryies;

                    do {
                        HostId targetHost = getHeuristic().chooseCreationHost(
                            cls.getName(), params, hint, avoidedHosts);

                        log(Sev.DEBUG, "createObject", "Selected host is {0}", targetHost);
                    
                        if ( targetHost != null ) {
                            return remoteCreateObject(app, cls, params, a, targetHost);
                        }
                    }
                    while ( --n > 0 );

                        //
                        // TODO: throw exception
                        // 
                }
            }            
            catch ( Exception e ) {
                log(Sev.ERROR, "createObject", e);
            }
            finally {
                log(Sev.TRACE, "createObject", "leaving");
            }

            return null;
        }

        //
        // TODO:
        //
    public void
    moveObject( ObjectId ox, Object placementHint )
        {
            throw new NullPointerException("Not yet implemented.");
        }

        //
        // TODO:
        //
    public void
    deactivateObject( ObjectId ox, MarshaledOX oxState )
        {
            throw new NullPointerException("Not yet implemented.");
        }

        //
        // TODO:
        //
    public void
    activateObject( MarshaledOX oxState )
        {
            throw new NullPointerException("Not yet implemented.");
        }

        //
        // TODO:
        //
    public void
    exitApplication()
        {
            throw new NullPointerException("Not yet implemented.");
        }

    public void
    setHeuristic( SchedulingHeuristic h )
        {
            ThreadGroup tg = Thread.currentThread().getThreadGroup();

            if ( tg instanceof Application ) {
                log(Sev.DEBUG, "setHeuristic", "is instance of Application");
                ((Application) tg).setHeuristic(h);
            }
            else {
                    // TODO: follow up tg hierarchy
            }
        }

    public ApplicationId
    currentApplication()
        {
            ThreadGroup tg = Thread.currentThread().getThreadGroup();

            while ( tg != null ) {
                if ( tg instanceof Application ) {
                    return ((Application) tg).getId();
                }
                tg = tg.getParent();
            }
            return null;
        }

    public void
    runAction(ApplicationId aid, Runnable action)
        {
            log(Sev.TRACE, "runAction", "entering appid={0}", aid);
            try {
                if ( aid == null ) {
                        // run as middleware
                    Thread t = new Thread(action);
                    t.start();
                    t.join();
                }
                else {
                    Application app = getApplicationById(aid);

                    app.runAction(action);
                }
            }
            catch (Exception e) {
                log(Sev.ERROR, "runAction", e);
            }
            finally {
                log(Sev.TRACE, "runAction", "leaving");
            }
        }



    private final SchedulingHeuristic
    getHeuristic()
        {
                // TODO: follow parent if app not found
            Application app = (Application)
                appByThrGrp.get(Thread.currentThread().getThreadGroup());

            if ( app == null ) {
                log(Sev.DEBUG, "getHeuristic", "thread group not registered");
                return defaultHeuristic;
            }
            
            SchedulingHeuristic h = app.getHeuristic();

            if ( h != null ) {
                log(Sev.DEBUG, "getHeuristic", "using app configured heuristic");
                return  h;
            }
            else {
                log(Sev.DEBUG, "getHeuristic", "using default heuristic");                
                return defaultHeuristic;
            }
        }
        
    private ObjectHandle
    localCreateObject(ApplicationId aid, Class cls, Object[] params, Activator a)
        throws Exception
        {
            Application app = getApplicationById(aid);

            
            return (new CreateObjectThread(app, cls, params, a)).getHandle(); 
        }



    private Application
    getApplicationById( ApplicationId id ) throws Exception
        {
            log(Sev.TRACE, "getApplicationById", "entering id={0}", id);
            Application app = (Application) appById.get(id);
            
            if ( app == null ) {
                log(Sev.DEBUG, "getApplicationById", "app is not running", id);
                
                CellInformationBase cib = (CellInformationBase)
                    Exehda.getService(CellInformationBase.SERVICE_NAME);

                ResourceName rn = new ResourceName("app:"+id);
                String codeBase = cib.getAttribute(rn, "codeBase");

                log(Sev.DEBUG, "getApplicationById", "fetched from CIB codebase={0}", codeBase);
                app = new Application(codeBase, null, null);

                registerApplication(app, false);
//                 appById.put(id, app);
//                 appByThrGrp.put(app,app);
            }
            log(Sev.TRACE, "getApplicationById", "leaving");
            return app;
        }

    private final ObjectHandle
    remoteCreateObject(ApplicationId app, Class cls, Object[] params, Activator a, HostId targetHost)
        {
            Executor remExecutor = (Executor) 
                ((Worb) Exehda.getService("worb")).lookupService(
                    "//"+targetHost.getName()+"/executor",Executor.class);
            
            return remExecutor.createObject(app, cls, params, a, targetHost);
        }

    private final void
    registerApplication(Application app, boolean updateCib)
        {
            appByThrGrp.put(app, app);
            appById.put(app.getId(), app);

            if ( updateCib ) {
                log(Sev.DEBUG, "registerApplication", "updating CIB");
                
                CellInformationBase cib = (CellInformationBase)
                    Exehda.getService(CellInformationBase.SERVICE_NAME);
                
//                 cib.addApplication(app.getId(), app.getCodebase());

                ResourceName rn = new ResourceName("app:"+app.getId());
                cib.addResource(rn);
                cib.setAttribute(rn, "codeBase", app.getCodebase());
                
            }
        }
    
    private void
    launch( String codeBaseURL, String mainClassName, String[] args)
        {
            log(Sev.TRACE, "launch", "entering");
            try {
                log(Sev.DEBUG, "launch", "Launching application {0}", codeBaseURL);
                Application app = new Application(codeBaseURL,mainClassName,args);
                
                app.start();

                registerApplication(app, true);
//                 appByThrGrp.put(app, app);
//                 appById.put(app.getId(), app);
            }
            catch (Exception e) {
                log(Sev.ERROR, "launch", e);
            }
            finally {
                log(Sev.TRACE, "launch", "leaving");
            }
        }

}

class Application
    extends ThreadGroup
    implements Runnable
{
    private ApplicationId id;
    private ClassLoader appClassLoader;
    private String codeBaseURL;
    private String mainClassName;
    private String[] args;
    private Executor.SchedulingHeuristic heuristic;
    private Thread mainThread;
    
    public Application( String codeBaseURL, String mainClassName, String[] args)
        throws Exception
        {
            super("Application");
            this.codeBaseURL = codeBaseURL;
            this.appClassLoader = IsamClassLoader.newInstance(codeBaseURL, this);
            this.mainClassName = mainClassName;
            this.args = args;
            this.id = new ApplicationId();
            this.heuristic = null;
            this.mainThread = null;
        }

    public void
    start()
        {
            if (mainThread == null) {
                mainThread = new Thread(this, this, "main");
                mainThread.setDaemon(false);
                mainThread.start();
            }
        }
    
    public void
    run()
        {
            try {
                
                Class mainClass = appClassLoader.loadClass(mainClassName);
                Class[] mainArgTypes = new Class[] { String[].class };

                Method mainMethod = mainClass.getMethod("main", mainArgTypes);
                
                mainMethod.invoke(null, new Object[] { args });
            }
            catch (Exception e) {
                e.printStackTrace();
            }
        }

    public ApplicationId
    getId()
        {
            return id;
        }

    public String
    getCodebase()
        {
            return this.codeBaseURL;
        }
    
    public void
    setHeuristic( Executor.SchedulingHeuristic h)
        {
            if ( heuristic == null ) {
                heuristic = h;
            }
        }

    public  Executor.SchedulingHeuristic
    getHeuristic()
        {
            return heuristic;
        }

    public void
    runAction( Runnable action )
        {
            try {
                Thread t = new Thread(this, action, "actionThread");
                t.start();
                t.join();
            }
            catch (InterruptedException ie)
            {
                    // empty
            }
        }
}

class CreateObjectThread
    extends Thread
{
    private Class cls;
    private Object[] p;
    private Activator a;
    private ObjectHandle h;
            
    public CreateObjectThread(ThreadGroup tg, Class cls, Object[] p, Activator a)
        {
            super(tg, "createObject");
            this.cls = cls;
            this.p = p;
            this.a = a;
            this.h = null;
            start();
        }

    public void run()
        {
            ObjectSeed objectseed = (ObjectSeed) Exehda.getService("objectseed");

            synchronized (this) {
                h = objectseed.newInstance(cls, p, a);
            }
        }

    public synchronized ObjectHandle getHandle() throws InterruptedException
        {
            this.join();
            return this.h;
        }
}

class HostIdHeuristic
    implements Executor.SchedulingHeuristic
{
    public HostId
    chooseCreationHost(String clsName,
                       Object[] params,
                       Object hint,
                       java.util.Vector avoidedHosts)
        {
            try {
                if ( hint == null ) {
                    return HostId.getLocalHost();
                }
                else if ( hint instanceof HostId ) {
                    return (HostId) hint;
                }
                else
                    return null;
            }
            catch (Exception e) {
                return null;
            }
        }

    public HostId
    chooseMigrationHost(ObjectHandle oh,
                       Object hint,
                       java.util.Vector avoidedHosts)
        {
            throw new NullPointerException("Not yet implemented");
        }
}

class LocalHostHeuristic
    implements Executor.SchedulingHeuristic
{
    public HostId
    chooseCreationHost(String clsName,
                       Object[] params,
                       Object hint,
                       java.util.Vector avoidedHosts)
        {
            try {
                return HostId.getLocalHost();
            }
            catch (Exception e) {
                return null;
            }
        }

    public HostId
    chooseMigrationHost(ObjectHandle oh,
                       Object hint,
                       java.util.Vector avoidedHosts)
        {
            throw new NullPointerException("Not yet implemented");
        }

}
