//  
// PRIMOS - Primitives for Object Scheduling 
//  
// Copyright (C) 2002 Luciano Cavalheiro da Silva  <lucc@inf.ufrgs.br> 
//                    Claudio Fernando Resin Geyer <geyer@inf.ufrgs.br>
//  
// This file is part of PRIMOS. 
//  
// PRIMOS is free software; you can redistribute it and/or 
// modify it under the terms of the GNU General Public License 
// as published by the Free Software Foundation; either version 2 
// of the License, or any later version. 
//  
// PRIMOS is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
// GNU General Public License for more details. 
//  
// You should have received a copy of the GNU General Public License 
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
//

/**
* @file  LinuxMonitorImpl.icc
* @brief /proc based implementation of the MonitorImpl interface.
*
* @cvsid $Id: LinuxMonitorImpl.cc 175 2004-03-22 19:42:58Z lucc $
*/

#include <MonitorImpl.hh>

#ifdef HAVE_CONFIG_H
#  include <acconfig.h>
#endif

#ifdef HAVE_INTTYPES_H
#  include <inttypes.h>
#endif

#ifdef HAVE_SYS_UTSNAME_H
#  include <sys/utsname.h>
#endif

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#else
#  error "unistd.h not found"
#endif

#ifdef HAVE_UTMPX_H
#  include <utmpx.h>
#endif

#ifdef HAVE_OSTREAM
#  include <ostream>
#elif defined HAVE_OSTREAM_H
#  include <ostream.h>
#endif

#ifdef HAVE_FSTREAM
#  include <fstream>
#endif

#ifdef HAVE_VECTOR
#  include <vector>
#endif

#ifdef HAVE_SSTREAM
#  include <sstream>
#endif

#ifdef HAVE_CCTYPE
#  include <cctype>
#endif


#include <SensorData.hh>


using primos::Sensor;
using primos::SensorSet;
using primos::SensorData;
using primos::SensorParameter;
using primos::SensorParameterSet;

// need to fix a bug in ctype: in gcc 3.2 isspace() belongs to
// namespace std::, but in 2.95.3 it does not.
using namespace std;



/**
 * @struct NetDeviceData
 * Used internali by the monitor to store network interface stats
 */
struct NetInterfaceData
{
    std::string     ifname; // system name of the interface
    std::string     iftype; // description e.g. Ethernet 100Mbits
    struct netstat
    {
        uint32_t rb; // received bytes
        uint32_t rp; // received packets
        uint32_t re; // receive  errors
        uint32_t rd; // receive  drops
        uint32_t xb; // xmit bytes
        uint32_t xp; // xmit packets
        uint32_t xe; // xmit errors
        uint32_t xd; // xmit drops
        uint32_t xc; // xmit collisions
    } curr, prev;
    int64_t maxByteThroughput;
    int64_t maxPacketThroughput;
    int64_t bytesSentPerSec;
    int64_t packetsSentPerSec;
    int64_t bytesRecvPerSec;
    int64_t packetsRecvPerSec;
    int64_t collisions;

    NetInterfaceData( const std::string& n, const std::string& t )
            :ifname(n),iftype(t),maxByteThroughput(-1),maxPacketThroughput(0),
             bytesSentPerSec(0),packetsSentPerSec(0),
             bytesRecvPerSec(0),packetsRecvPerSec(0),collisions(0)
        {
            memset( &curr, 0, sizeof(curr) );
            memset( &prev, 0, sizeof(prev) );
        }
};

/**
 * @brief Storage for pre processor stats
 */
struct CpuData
{
    std::string   type;   // textual reprensetation of the cpu type
#if 0    
   bool     online;    // true if this cpu is online (usable)
    struct cpustat {
        int64_t u;      // time spent in user computations 
        int64_t s;      // time spent in system(kernel) computations
        int64_t i;      // idle time 
    } curr, prev;

    CpuData()
        {
            curr.u = curr.s = curr.i = 0;
        }
#endif
};

/**
 * @class intSensorData
 *
 */
class IntSensorData : public SensorData
{
public:
    IntSensorData(Integer* data)
            : SensorData(INTEGER_DATA, data)
        {}
    
    void
    toString(std::ostream& out) const
        {
            out << *static_cast<Integer*>(data);
        }    
};

/**
 * @class FloatSensorData
 *
 */
class FloatSensorData : public SensorData
{
public:
    FloatSensorData(Float* data)
            : SensorData(FLOAT_DATA, data)
        {}
    
    void
    toString(std::ostream& out) const
        {
            out << *static_cast<Float*>(data);
        }
};


/**
 * @class StringSensorData
 *
 */
class StringSensorData : public SensorData
{
public:
    StringSensorData(String* data)
            : SensorData(STRING_DATA, data)
        {}
    
    
    void
    toString(std::ostream& out) const
        {
            out << *static_cast<String*>(data);
        }
};

/**
 * @class StringSensorData
 *
 */

template <class TData, class TRawData, class TMonitor>
class FakeSensor : public primos::Sensor
{
private:
    bool      isenabled;
    bool      isstatic;    
    int       index;
    SensorParameterSet params;
    TMonitor* monitor;
    TData     data;    
public:
    FakeSensor(std::string name,
               SensorData::DataTypes
               type, int index,
               TRawData* rawdata,
               TMonitor* monitor,
               bool isenabled=false,
               bool isstatic=false)
            : Sensor(name, type), data(rawdata)
        {
            this->index     = index;
            this->isenabled = isenabled;
            this->isstatic  = isstatic;
            this->monitor   = monitor;
        }
    
    void
    setEnabled(bool enable)
        {
            if (enable) {
                monitor->enableSensor(index);
            } else {
                monitor->disableSensor(index);
            }
            isenabled = enable;
        }

    bool
    isEnabled() const
        {
            return isenabled;
        }

    bool
    isStatic() const
        {
            return isstatic;
        }
    
    bool
    isNative() const
        {
            return true;
        }
    
    const SensorData* 
    probe()
        {
            return &data;    
        }

    void
    toString(std::ostream& out) const
        {
            out << "s=\"" << name << "\" v=\"" << data << '\"';
        }

    SensorParameterSet&
    Parameters() const 
        {
            return const_cast<SensorParameterSet&>(params);
        }

    void
    setParameter( const std::string name, const std::string val )
        {}

    std::string
    getParameter( const std::string name ) const
        {
            throw std::string("Not Defined");
        }
    

};

/**
 * @class LinuxMonitorImpl
 *
 */
class LinuxMonitorImpl
    : public primos::MonitorImpl
{
public:
        /**
         * @brief Constructor
         *
         */
    LinuxMonitorImpl()
        {
            nsensors = 0;
            detectArchitecture();
            detectAvailableProcessors();
            detectNetworkInterfaces();
            detectProcessorTypes();
            initSensors();
        }

        /**
         * @brief Returns the set of native sensors suported by this monitor implementation.
         */
    const SensorSet&
    availableSensors()
        {
            return nativeSensors;
        }

    void
    quantumExpired()
        {
            elapsedTime = 1; // actually it must be equals to the quantum size
            parseProcMeminfo();
            parseProcStat();
            parseProcUptime();
            parseProcLoadavg();
            parseProcNetDev();
            parseUtmpx();
        }
    
    void
    enableSensor(int idx)
        {
                // do nothing in this implementation
        }
    
    void
    disableSensor(int idx)
        {
                // do nothing in this implementation
        }


protected:
        /**
         * @brief Add a sensor to the set of available sensors.
         *
         */
    typedef FakeSensor<StringSensorData, SensorData::String,  LinuxMonitorImpl> LinuxStringSensor;
    typedef FakeSensor<FloatSensorData,  SensorData::Float,   LinuxMonitorImpl> LinuxFloatSensor;
    typedef FakeSensor<IntSensorData,    SensorData::Integer, LinuxMonitorImpl> LinuxIntSensor;

    void
    addSensor(std::string name, int idx, SensorData::Integer* data, bool isStatic = false)
        {
            
            Sensor* s = new LinuxIntSensor(name, SensorData::INTEGER_DATA,
                                          idx, data, this, false, isStatic);
            nativeSensors.insert(s);
        }

    void
    addSensor(std::string name, int idx, SensorData::Float* data, bool isStatic = false)
        {
            Sensor* s = new LinuxFloatSensor(name, SensorData::FLOAT_DATA,
                                          idx, data, this, false, isStatic);
            nativeSensors.insert(s);
        }

    void
    addSensor(std::string name, int idx, SensorData::String* data, bool isStatic = false)
        {
            Sensor* s = new LinuxStringSensor(name, SensorData::STRING_DATA,
                                          idx, data, this, false, isStatic);
            nativeSensors.insert(s);
        }
    
        /**
         * @brief detect the host type
         *
         */
    void
    detectArchitecture();
    
        /**
         * @brief detect available processors
         *
         */
    void 
    detectAvailableProcessors();
    
        /**
         * @brief detect the types of each processor
         */
    void
    detectProcessorTypes();

        /**
         * @brief detect existing network interfaces
         * 
         * TODO: skip the ppp inerface also?
         */
    void
    detectNetworkInterfaces();
    
        /**
         * @brief initialize availableSensors
         *
         */
    void
    initSensors();

        /**
         * @brief update memory stats
         *
         */
    void
    parseProcMeminfo();

        /**
         * @brief Parses /proc/loadavg
         */
    void
    parseProcLoadavg();

        /**
         * @brief Parses /proc/uptime
         */
    void
    parseProcUptime();
    
        /**
         * @brief Extrects the number of logged in user from the utmpx
         * interface
         */
    void
    parseUtmpx();

        /**
         * @brief re-compute percentual cpu occupations (user, nice, sys, idle)
         *
         */
    void
    parseProcStat();
    
        /**
         * @brief update network interface data
         */
    void
    parseProcNetDev();    
            
private:
        // helper data
    int nsensors;
    int64_t    elapsedTime;
    SensorSet  nativeSensors;
//    CpuData*   configuredCpus;
    std::vector<CpuData>   configuredCpus;
    std::vector<NetInterfaceData> netDevs;
    int64_t   prevCpuStat[4];
    int64_t   currCpuStat[4];
    std::string   arch;
    int64_t  nCpus;
    int64_t  nNetInterfaces;
    double   cpuOccupIdle;
    double   cpuOccupUser;
    double   cpuOccupSys;
    double   cpuOccupBg;
    double   loadAvg1Min;
    int64_t  totalProcs;
    int64_t  rnblProcs;
    int64_t  totalPhysMem;
    int64_t  freePhysMem;
    int64_t  totalVirtualMem;
    int64_t  freeVirtualMem;
    int64_t  nInterUsers;
    int64_t  upTimeSeconds;
};

std::istream&
skipWord( std::istream& in)
{
    while ( isspace(in.peek()) ) {
        in.ignore();
    }
    while ( !isspace(in.peek()) ) {
        in.ignore();
    }
    return in;
}

std::istream&
skipLine( std::istream& in)
{
    while ( in.peek() != '\n' ) {
        in.ignore();
    }
    in.ignore();
    return in;
}

/**
 *
 *
 */
inline void
LinuxMonitorImpl::parseProcNetDev()
{
        // open /proc
    std::ifstream sNetDev("/proc/net/dev");

    std::string tag;
    bool        skipDevice;
    long        skip;
    
    std::vector<NetInterfaceData>::iterator ni;
    for ( ni = netDevs.begin(); ni != netDevs.end(); ) {
            // reset the file
        sNetDev.seekg(0, std::ios::beg);
        sNetDev.clear();
        skipDevice = true;
        
            // skip two first lines
        getline(sNetDev, tag);
        getline(sNetDev, tag);
            // for each entry in the file
        while ( getline(sNetDev,tag, ':') ){
            skipDevice = false;
            
                // get the interface name
            std::string devName = tag.substr(tag.find_first_not_of(" \t"));
                // if current entry matches current NI name
            if ( devName == ni->ifname ) {
                    // save previous state
                ni->prev = ni->curr;
                    // parse the line 
                sNetDev >> ni->curr.rb >> ni->curr.rp
                        >> ni->curr.re >> ni->curr.rd 
                        >> skip >> skip >> skip >> skip
                        >> ni->curr.xb >> ni->curr.xp
                        >> ni->curr.xe >> ni->curr.xd >> skip
                        >> ni->curr.xc >> skip >> skip;
                    // update NI data structure
                if ( ni->maxByteThroughput == -1 ) {
                        // do nothing, just calibrate the mechanism
                    ni->maxByteThroughput = 0;
                } else {
                    ni->bytesSentPerSec   = (ni->curr.xb - ni->prev.xb) / elapsedTime;
                    ni->packetsSentPerSec = (ni->curr.xp - ni->prev.xp) / elapsedTime;
                    ni->collisions        = (ni->curr.xc - ni->prev.xc) / elapsedTime;
                    ni->bytesRecvPerSec   = (ni->curr.rb - ni->prev.rb) / elapsedTime;
                    ni->packetsRecvPerSec = (ni->curr.rp - ni->prev.rp) / elapsedTime;

                    int64_t byteThroughput = ni->bytesSentPerSec + ni->bytesRecvPerSec;
                    int64_t packetThroughput = ni->packetsSentPerSec + ni->packetsRecvPerSec;

                    if ( ni->maxByteThroughput < byteThroughput ) {
                        ni->maxByteThroughput = byteThroughput;
                    }
            
                    if ( ni->maxPacketThroughput < packetThroughput ) {
                        ni->maxPacketThroughput = packetThroughput;
                    }
                }                
                    // next NI
                ni++;
                break;
            }
                // skip the remaing caracters in the line
            getline(sNetDev,tag);
        } // while
        
            // eof means the respective entry for current NI in the incremental
            // search mode.  reset the file and try once again.
        if ( skipDevice ) ni++;


    } // for
    
        // release /proc resources
    sNetDev.close();
}

/**
 *
 *
 */
inline void
LinuxMonitorImpl::parseProcStat()
{
            
    prevCpuStat[0] = currCpuStat[0];
    prevCpuStat[1] = currCpuStat[1];
    prevCpuStat[2] = currCpuStat[2];
    prevCpuStat[3] = currCpuStat[3];
            
        // open /proc
    std::ifstream sCpuStat("/proc/stat");
        // get running totals from /proc
    sCpuStat.ignore(3,' ')
        >> currCpuStat[0] >> currCpuStat[1]
        >> currCpuStat[2] >> currCpuStat[3];
        // compute the difference fromm the last update
    int64_t u = currCpuStat[0]-prevCpuStat[0];
    int64_t n = currCpuStat[1]-prevCpuStat[1];
    int64_t s = currCpuStat[2]-prevCpuStat[2];
    int64_t i = currCpuStat[3]-prevCpuStat[3];
    int64_t t = u+s+i+n;

    if ( t != 0 ) {
            // compute the effective occupations
        cpuOccupUser  = (100 * u)/t;
        cpuOccupSys   = (100 * s)/t;
        cpuOccupBg    = (100 * n)/t;
        cpuOccupIdle  = 100 - (cpuOccupUser+cpuOccupSys+cpuOccupBg);
    }
        // close /proc
    sCpuStat.close();
}

/**
 *
 *
 */      
inline void
LinuxMonitorImpl::parseUtmpx()
{
        // reset utmpx database
    setutxent();
    struct utmpx* ent;
        // reset the interactive processes counter
    nInterUsers = 0;
        // parse utmp entries looking for interactive processes
    while ((ent = getutxent()) != NULL) {
            // increment the counter on each found interactive process
        if (ent->ut_type == USER_PROCESS) {
            nInterUsers++;
        } else {
//             char spc = '\t';
                    
//             cout << ent->ut_type << spc
//                  << ent->ut_pid  << spc                        
//                  << ent->ut_line << spc
//                  << ent->ut_id   << spc
//                  << ent->ut_user << spc
//                  << ent->ut_host << spc
//                  << ent->ut_session << spc
//                  << endl;
        }
    }
}

/**
 *
 *
 */
inline void
LinuxMonitorImpl::parseProcUptime()
{
    std::ifstream sUpTime("/proc/uptime");
    sUpTime  >> upTimeSeconds;
    sUpTime.close();
}

/**
 * @brief update memory stats
 *
 */
inline void
LinuxMonitorImpl::parseProcMeminfo()
{
    std::ifstream sMemInfo("/proc/meminfo");
        // input data from /proc
//    sMemInfo.scan("%*[^\n] %*s")
    sMemInfo >> skipLine
             >> skipWord
             >> totalPhysMem
//    sMemInfo.scan("%*d")
             >> skipWord
             >> freePhysMem
             >> skipWord
             >> skipWord
             >> skipWord
             >> skipWord
//    sMemInfo.scan("%*d").scan("%*d").scan("%*d");
//    sMemInfo.scan("%*s")
             >> totalVirtualMem
//    sMemInfo.scan("%*d")
             >> skipWord
             >> freeVirtualMem;
    
    totalVirtualMem += totalPhysMem;
    freeVirtualMem  += freePhysMem;
    sMemInfo.close();
                
}

/**
 *
 *
 */
inline void
LinuxMonitorImpl::parseProcLoadavg()
{
    double skipDouble;
    char   skipChar;
            
    std::ifstream sLoadAvg("/proc/loadavg");
    sLoadAvg >> loadAvg1Min >> skipDouble >> skipDouble
             >> rnblProcs >> skipChar >> totalProcs;
    sLoadAvg.close();

    if (rnblProcs > 0) rnblProcs--;
            
}

/**
 * @brief detect existing network interfaces
 * 
 * TODO: skip the ppp inerface also?
 */
inline void
LinuxMonitorImpl::detectNetworkInterfaces()
{
        // alloc initial storage for net dev. data 
    std::vector<NetInterfaceData*> devs;
    devs.reserve(16);
        // devices to skip in detection
    std::set<std::string> skipDevs;
    skipDevs.insert(std::string("lo"));

        // initially there are no detected interfaces
    nNetInterfaces = 0;

        // open /proc
    std::ifstream sNetDev("/proc/net/dev");
    std::string tag;                
        // skip two first lines
    getline(sNetDev, tag);
    getline(sNetDev, tag);
        // each line descries an network interface
            
    while ( getline(sNetDev,tag, ':') ){
            // get the interface name
        std::string devName = tag.substr(tag.find_first_not_of(" \t"));
            // skip unwanted interfaces, accept all other interfaces
        if ( skipDevs.find(devName) == skipDevs.end() ) {
                // build ni type description
            std::string iftype;
            iftype += "{ ";
            iftype += devName;

                // try to detect the type of the interface
            if ( devName.find("eth") == 0 ) {
                    // currently works only for ethernet
                iftype +=  " }-{ Ethernet }";
            } else {
                iftype +=  " }-{ unknown }";
            }
                
            devs.push_back(new NetInterfaceData(devName, iftype));
                // incrememt the number of detected interfaces
            nNetInterfaces++;
        }
            // skip the remaing caracters in the line
        getline(sNetDev,tag);
    }
        // release /proc structures
    sNetDev.close();
        // shrink the net devices array
    this->netDevs.reserve(devs.size());
    for ( std::vector<NetInterfaceData*>::size_type ni = 0; ni < devs.size(); ni++ ) {
        netDevs.push_back(*devs[ni]);
        delete (devs[ni]);
            // create sensors for this network interface
        std::ostringstream oss;
        oss << "[" << ni << "]";
        std::string sensorSuffix = oss.str();
        addSensor("NI_TYPE"+sensorSuffix, nsensors++,&netDevs[ni].iftype, true );
        addSensor("NI_MAX_BYTE_THROUGHPUT"+sensorSuffix, nsensors++,&netDevs[ni].maxByteThroughput);
        addSensor("NI_MAX_PACKET_THROUGHPUT"+sensorSuffix, nsensors++,&netDevs[ni].maxPacketThroughput);
        addSensor("NI_BYTES_SENT_PER_SEC"+sensorSuffix, nsensors++,&netDevs[ni].bytesSentPerSec);
        addSensor("NI_BYTES_RECV_PER_SEC"+sensorSuffix, nsensors++,&netDevs[ni].bytesRecvPerSec);
        addSensor("NI_PACKETS_SENT_PER_SEC"+sensorSuffix, nsensors++,&netDevs[ni].packetsSentPerSec);
        addSensor("NI_PACKETS_RECV_PER_SEC"+sensorSuffix, nsensors++,&netDevs[ni].packetsRecvPerSec);
        addSensor("NI_PACKET_COLLISIONS"+sensorSuffix, nsensors++,&netDevs[ni].collisions);
    }
}

/**
 * @brief detect available processors
 *
 */
inline void 
LinuxMonitorImpl::detectAvailableProcessors()
{
        // constant to compare to the cpu state
    const std::string CPU_STATE_ONLINE = "on-line";
        // discover the total number of cpus
    nCpus = sysconf(_SC_NPROCESSORS_CONF);
        // create the respective sensor
    addSensor("N_PROCESSORS", nsensors++, &nCpus, false);
        // detect the type of each available processor

}
    
/**
 * @brief detect the types of each processor
 */
inline void
LinuxMonitorImpl::detectProcessorTypes()
{
        // get information for each existing processor from /proc/cpuinfo
    std::ifstream sCpuinfo("/proc/cpuinfo");
        // alloc cpu data array for the given number of processors
    configuredCpus.reserve(nCpus);
        // parse the /proc/cpuinfo  pseudo file
    std::string key;
    std::string  cputype;
    std::string  cpuclock;
    std::string  value;
    int     cpunumber;
    while ( getline(sCpuinfo,key,':') ){
            // consume separator
        sCpuinfo.get();

            // checks if the line starts with the interest key
            // 2 is used because \n may not be consumed by getline
        if ( key.find("processor") < 2 ) {
                // parse processor number
            sCpuinfo >> cpunumber ; 
        } else if ( key.find("model name") < 2 ) {
                // parse processor type
            getline(sCpuinfo, cputype);
        } else if ( key.find("cpu MHz") < 2 ) {
                // parse processor clock
            sCpuinfo >> cpuclock;
                // build cputype string
            std::ostringstream os;
            os << "{ " << cputype << " }-{ " << cpuclock << " MHz }";

                // fill CpuData            
            CpuData currCpu;
            currCpu.type = os.str();
            configuredCpus.push_back(currCpu);
            
                // build sensor name
            std::ostringstream sensorName;
            sensorName << "PROCESSOR_TYPE[" << cpunumber << "]";
                // create the respective sensor
           addSensor(sensorName.str(), nsensors++, &configuredCpus.back().type, true );
        } else {
                // not recognzed, just consumes to the end of the line
            getline(sCpuinfo, value);
            continue;
        }
    }
        // release /proc resources
    sCpuinfo.close();
}

/**
 * @brief detect the host type
 *
 */
inline void
LinuxMonitorImpl::detectArchitecture()
{
        // storage for uname result
    struct utsname undata;
        // issue uname syscall
    if ( uname(&undata) == -1) {
            // on uname failure, fills host type with "unknown"
        arch = "{ unknown }-{ unknown }-{ unknown }";
    } else {
        arch = "{ ";
        arch += undata.machine;
        arch += " }-{ ";
        arch += undata.sysname;
        arch += " }-{ ";
        arch += undata.release;
        arch += " }";
    }

        // create a sensor for this data
    addSensor("ARCHITECTURE", nsensors++, &arch, true );
}

/**
 * @brief initialize availableSensors
 *
 */
inline void
LinuxMonitorImpl::initSensors()
{
    int idx=nsensors;
    addSensor("CPU_OCCUP_USER",idx++, &cpuOccupUser);
    addSensor("CPU_OCCUP_SYS",idx++, &cpuOccupSys);
    addSensor("CPU_OCCUP_IDLE",idx++, &cpuOccupIdle );
    addSensor("CPU_OCCUP_BG",idx++, &cpuOccupBg);
    addSensor("LOAD_AVG",idx++, &loadAvg1Min);
    addSensor("N_TOTAL_PROCESSES",idx++, &totalProcs);
    addSensor("N_RNBL_PROCESSES",idx++, &rnblProcs);
    addSensor("TOTAL_PHYS_MEMORY",idx++, &totalPhysMem);
    addSensor("FREE_PHYS_MEMORY",idx++, &freePhysMem);
    addSensor("TOTAL_VIRTUAL_MEMORY",idx++, &totalVirtualMem);
    addSensor("FREE_VIRTUAL_MEMORY",idx++, &freeVirtualMem);
    addSensor("N_INTERACTIVE_USERS",idx++, &nInterUsers);
    addSensor("UP_TIME",idx++, &upTimeSeconds);
    nsensors = idx;
}


primos::MonitorImpl*
primos::MonitorImpl::newInstance()
{
    return new LinuxMonitorImpl();
}
