// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda;

/**
 * Root class for all kinds of resource constraint. Resource constraints are widely used
 * in resource selection, discovery and allocation in the pervasive environment.
 *
 * @see org.isam.exehda.services.CellInformationBase
 * @see org.isam.exehda.services.ResourceBroker
 * @see org.isam.exehda.services.Discoverer
 * @see org.isam.exehda.services.Scheduler
 *
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-19 15:51:38 -0300 (Sun, 19 Jun 2005) $ $Revision: 941 $
 */
public abstract class ResourceConstraint
    implements java.io.Serializable
{
    public static final int OK = 2;
    public static final int UNDEFINED = 0;
    public static final int BROKEN = -2;
    
        /**
         * Provides access to Resource attributes.
         *
         */
    public static interface ResourceInfo
    {
        public String
        getAttribute( String attrName );
        
     
//         public Object
//         getIndex( DynamicIndex idx );
    }
    
        /**
         * Checks whether the resource represented by the given ResourceInfo satisfies
         * this constraint.
         *
         * @param r a <code>ResourceInfo</code> value
         * @return OK is the resource satisfies the criteria, BROKEN if it does not or
         * UNDEFINED if the criteria could not be evaluated for this resource.
         */
    public abstract int
    check ( ResourceInfo r );
    
        /**
         * Returns the subset of the candidates who satisfy this constraint.
         *
         * <p> If the constraint is subject to ordering, the best candidate is returned
         * at position 0, the second best at position 1 etc. Although, for such behavior
         * this method must be overwritten in subclasses. The default implementation do
         * not modify the ordering of candidates.
         *
         * <p> <b>Notice</b> The returned array is the same cadidate array, just with the
         * candidates that do not satisfy the criteria removed. The resulting elements
         * are packed from the beginning of the array (i.e. someone scanning the results
         * would stop at the first null).
         *
         * @param candidates a <code>ResourceInfo[]</code> value
         * @param undefinedIsOk wheter if the criteria could not be applyed for a given
         * candidate that candidate should be included in the result se.
         * @return a cleaned-up version of the candidate array
         */
    public ResourceInfo[]
    select ( ResourceInfo[] candidates, boolean undefinedIsOk )
        {
            for (int i=0,j=0; i<candidates.length; i++) {
                int r = check(candidates[i]) + (undefinedIsOk ? 1 : -1);
                
                if ( r > 0 ) {
                        // OK, cp to next free slot
                    candidates[i] = candidates[j++];
                }
                else {
                        // BROKEN, remove from array
                    candidates[i] = null;
                }
            }
            return candidates;
        }
}



//     public static abstract class CompositeResourceConstraint
//         extends ResourceConstraint
//     {
// //         public void addConstraint ( ResourceConstraint c,  undefinedIsOk ){
// //         }

// //         public Enumeration constraints (  ){
// //         }
//     }
    
//     public static abstract class StaticConstraint
//         extends ResourceConstraint {}

//     public static abstract class AttributeInRangeConstraint
//         extends StaticConstraint {
// //         public AttributeInRangeConstraint ( String attName, Number lb, Number up ){
// //         }
//     }

//     public static abstract class PreferredSetConstraint
//         extends StaticConstraint {
// //         public PreferredSetConstraint ( ResourceName[] resources ){
// //         }
//     }

//     public static abstract class ExclusiveAccessConstraint
//         extends StaticConstraint {
// //         public ExclusiveAccessConstraint ( long leaseTimeMinutes ){
// //         }
//     }

//     public static abstract class AvoidedSetConstraint
//         extends StaticConstraint {
// //         public AvoidedSetConstraint ( ResourceName[] resources ){
// //         }
//     }

//     public static abstract class AttributeMatchConstraint
//         extends StaticConstraint
//     {
// //         public AttributeMatchConstraint ( String attName, String regexp ){
// //         }
//     }
    
//     public static abstract class DynamicConstraint
//         extends ResourceConstraint {}
