// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda;

import org.isam.exehda.services.Service;
import org.isam.perv.IsamClassLoader;
import org.isam.util.xml.*;
import java.util.Hashtable;
import java.util.Enumeration;


public class ServiceManager
{
    private String activeProfile;
    private ProfileManager profileManager;
    private Hashtable servicesTbl;
    
    public ServiceManager( ProfileManager profileManager )
        {
            this.profileManager = profileManager;
            this.activeProfile = null;
            this.servicesTbl = new Hashtable();
        }

    public boolean
    bootstrapServices( String profileName )
        {
            activeProfile = profileName;
            Enumeration servicesEnum = profileManager.getBootstrapServices(profileName);

            while ( servicesEnum.hasMoreElements() ) {
                String svcName = (String) servicesEnum.nextElement();
                    // abort if some boot service do not start properly
                if ( !startService( svcName )) {
                    return false;
                }
            }
            return true;
        }

    public boolean shutdownServices()
        {
                //
                // TO DO: should we follow reverse activation order when stoping services?
                //
            Enumeration senum = servicesTbl.elements();

            while ( senum.hasMoreElements() ) {
                ServiceDescriptor svc = (ServiceDescriptor) senum.nextElement();

                if ( svc.isStarted() ) {
                    Exehda.logInfoMessage("Stopping service <"+svc.getName()+">...");
                    try {
                        svc.stop();
                        Exehda.logMessage("Service <"+svc.getName()+"> stoped.\n");
                    }
                    catch (Exception e) {
                        Exehda.logMessage("Warning: <"+svc.getName()+"> shutdown failed!");
                        e.printStackTrace();
                        return false;
                    }
                }
            }

            return true;
        }
    
    public boolean
    startService( String serviceName )
        {            
            try {
                ServiceDescriptor svcDesc = (ServiceDescriptor) servicesTbl.get(serviceName);
                
                if ( svcDesc != null && svcDesc.isStarted()) {
                        // do nothing
                }
                else {
                    Exehda.logInfoMessage("Starting service <"+serviceName+">...");
                        // load implementation
                    if ( svcDesc == null ) {
                        svcDesc = loadService(serviceName);
                    }
                        // start the service
                    svcDesc.start();
                    Exehda.logMessage("Service <"+serviceName+"> started.");
                }

                return true;
            }
            catch (ClassNotFoundException cnfe) {
                Exehda.logMessage("Class not found: "+cnfe.getMessage());
                Exehda.logMessage("Warning: <"+serviceName+"> start-up failed!");
            }
            catch (Exception e) {
                Exehda.logMessage("Warning: <"+serviceName+"> start-up failed!");
                Exehda.logMessage("         Exception was "+e);
            }
            return false;
        }

    public boolean
    stopService( String serviceName )
        {
            try {
                ServiceDescriptor svcDesc = (ServiceDescriptor) servicesTbl.get(serviceName);

                if ( (svcDesc != null) && (svcDesc.isStarted()) ) {
                    Exehda.logInfoMessage("Stopping service <"+serviceName+">...");
                    svcDesc.stop();
                    Exehda.logMessage("Service <"+serviceName+"> stoped.\n");
                }

                return true;
            }
            catch (Exception e) {
                Exehda.logMessage("Warning: <"+serviceName+"> shutdown failed!");
                e.printStackTrace();
                return false;
            }
        }

    public Object
    getServiceProxy( String serviceName )
        {
                // TODO: fix this implementation!!!
            
            ServiceDescriptor svcDesc = (ServiceDescriptor) servicesTbl.get(serviceName);

            return svcDesc == null ? null : svcDesc.getImpl();
        }

    public String
    getActiveProfile()
        {
            return activeProfile;
        }

    private ServiceDescriptor
    loadService( String serviceName ) throws Exception
        {
            ClassLoader cl = IsamClassLoader.getInstance();

            String serviceImpl = Exehda.getServiceProperty(serviceName, "impl", null);

            if ( serviceImpl == null ) {
                throw new ClassNotFoundException(
                    "Implementation not defined for service "
                    + serviceName
                    + " in profile " + activeProfile);
            }

            Exehda.logMessage("Loading <"+serviceImpl+">...");            
            Service svc = (Service) cl.loadClass(serviceImpl).newInstance();

            ServiceDescriptor sd = new ServiceDescriptor(serviceName, svc);
            servicesTbl.put(serviceName, sd);

            return sd;
        }
}


class ServiceDescriptor
{
    String serviceName;
    boolean started;
    Service impl;

    ServiceDescriptor( String serviceName, Service impl )
        {
            this.serviceName = serviceName;
            this.impl = impl;
            started = false;
        }

    public String
    getName()
        {
            return serviceName;
        }
    
    public Service
    getImpl()
        {
            return impl;
        }

    public boolean
    isStarted() 
        {
            return started;
        }

    public void
    start() throws Exception
        {
            impl.start();
            started = true;
        }

    public void
    stop() throws Exception
        {
            impl.stop();
            started = false;
        }
}
