// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services;

import org.isam.exehda.Exehda;
import org.isam.exehda.ResourceName;
import org.isam.exehda.CompositeResourceConstraint;
import org.isam.exehda.AttributeSimpleMatchConstraint;
import org.isam.util.xml.XmlElement;
import java.io.StringReader;
import java.util.Enumeration;
import java.util.Hashtable;



/**
 * Resource Discovery Service
 *
 * @author frainer, last modified by $Author: lucc $
 * @version $Date: 2005-07-15 10:12:18 -0300 (Fri, 15 Jul 2005) $ $Revision: 1003 $
 */
public interface Discoverer
{
    public static final String SERVICE_NAME = "discoverer";

        /**
         * A resource discovery query.
         *
         * @author frainer, last modified by $Author: lucc $
         * @version $Date: 2005-07-15 10:12:18 -0300 (Fri, 15 Jul 2005) $ $Revision: 1003 $
         */
    public static class Query
        extends CompositeResourceConstraint
    {
            /**
             * Creates a new instance of Query
             */
        public Query(String xmlQuery) {
                // remembers the provided query
            this.xmlQuery = xmlQuery;
                // complete the initialization by parsing the xml query
            parseQuery(xmlQuery);
        }
            /**
             * Returns the scope of this query
             *
             * @return  a <code>int</code> representing the scope of this query
             */    
        public int getScope() {
            return scope;
        }
            /**
             * Sets the scope of this query
             *
             * @param scope a <code>int</code> containing the desired scope
             * @return  void
             */   
        public void setScope(int scope){
            this.scope = scope;
        }
            /**
             * Returns the xml version of the query
             *
             * @return  a <code>String</code> representing the query in xml
             */    
        public String getXmlQuery(){
            return xmlQuery;
        }
    
            /**
             * Initializes this query object by parsing the given XML query specification.
             *
             * @param xmlQuery an XML query specification
             */
        private void parseQuery(String xmlQuery) {
                //
                // TO DO: improve this parser, it does not check any tags
                //
            try {
                XmlElement xml = XmlElement.parse(new StringReader(xmlQuery));
                xml = (XmlElement)xml.getChildren().nextElement();

                    // get que query scope
                this.scope = Integer.parseInt(xml.getAttribute("scope", "0"));
            
                    // import inner query constraints
                for ( Enumeration ec = xml.getChildren(); ec.hasMoreElements();) {
                        // each children has a name and a value.
                        // We use those to create an inner SimpleRegexpConstraint
                    XmlElement attribute = (XmlElement) ec.nextElement();
                    String n = attribute.getAttribute("name","null");
                    String v = attribute.getAttribute("value","void");
                    addConstraint(new AttributeSimpleMatchConstraint(n,v),false);
                }
            }
            catch (Exception e) { e.printStackTrace(); }
        }

        private String xmlQuery;
        private int scope;
    }    

    public static class ResourceDescriptor
        implements java.io.Serializable
    {    
            /**
             * Creates a new instance of ResourceDescriptor
             */
        public ResourceDescriptor(ResourceName r, ResourceLease lease) {
            rn = r;
            rl = lease;
            attributes = new Hashtable();
        }
            /**
             * Return the ResourceName associated with the resource
             *
             * @return a  <code>ResourceName</code> .
             */
        public ResourceName getName() {
            return rn ;
        }
            /**
             * Return the ResourceLease associated with the resource
             *
             * @return a  <code>ResourceLease</code> .
             */
        public ResourceLease getLease() {
            return rl;
        }
            /**
             * Recovers the value of the given attribute from the resource. If that attribute
             * is not set in the resource the defaultValue is returned 
             *
             * @param a a <code>String</code> with the name of the attribute to be recovered 
             * @param defaultValue a <code>String</code> with the value to be returned if the attribute is not defined
             * @return  a <code>String</code> with the value of the attribute
             */    
        public String getAttribute(String a, String defaultValues) {
            String attr = (String)attributes.get(a);
            if(attr == null){
                CellInformationBase cib;
                cib = (CellInformationBase)
                    Exehda.getService(CellInformationBase.SERVICE_NAME);
                attr = cib.getAttribute(rn,a);
                if ( attr == null ){
                    attr = defaultValues;
                }
                else{
                    attributes.put(a,attr);
                }
            }
            return attr;
        }
            /**
             * Sets the value of one of the resources attributes
             *
             * @param a a <code>String</code> with the name of the attribute to be set
             * @param v a <code>String</code> with the new value of the attribute
             */  
        public void setAttribute(String a, String v){
            attributes.put(a, v);
        }

        private ResourceName rn;
        private ResourceLease rl;
        private Hashtable attributes;
    }

    public static class ResourceLease
        implements java.io.Serializable
    {
//         public ResourceLease ( Object key, String[] perms );
//         public boolean checkPermission ( String p );
//         public Date setExpirationTime();
//         public Date getExpirationTime();
        
            /**
             * Creates a new instance of ResourceLease
             */
        public ResourceLease(Object key) {
            this.key = key;
            expirationTime = System.currentTimeMillis() + 120000;
        }        
            /**
             * Recovers the expiration time of this lease
             *
             * @return  a <code>long</code> representing the millisecond when the lease expires
             */    
        public long getExpirationTime(){
            return expirationTime;
        }
            /**
             * Recovers the key to this lease
             *
             * @return  a <code>Object</code> that is the key to this lease
             */   
        public Object getKey(){
            return key;
        }
            /**
             * Sets a new expiration time for this lease
             *
             * @return  a <code>long</code> with the new expiration time for this lease
             */   
        public long setExpirationTime(){
            expirationTime = System.currentTimeMillis() + 120000;
            return expirationTime;
        }

        private Object key;
        private long expirationTime;
    }

        /**
         * Registers a new resource that is equivalent to the one
         * described in the xml passed as a parameter.
         *
         * @param xmlDesc a <code>String</code> containing a xml document that describes
         * the attributes of a resource
         * @return a new <code>ResourceName</code> that represents the resource added.
         */
    public ResourceDescriptor register(String xmlDesc);

        /**
         * Removes a previously added resource.
         *
         * @param r a <code>ResourceDescriptor</code> that represents the resource that is to be removed
         * @return void
         */
    public void remove(ResourceDescriptor r);

        /**
         * Renews a previously added resource.
         *
         * @param r a <code>ResourceDescriptor</code> that represents the resource that is to be renewed
         * @return void
         */
    public void renew(ResourceDescriptor r);

        /**
         * Finds a resource that matches the given query.
         *
         * @param xmlQuery a <code>String</code> containing a xml document that describes the query
         * @return a array of <code>ResourceDescriptor</code>, representing the matching resources found.
         */
    public ResourceDescriptor[] find(Query q);
    

//     public ResourceDescriptor[] find( String xmlQuery );
}






