// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.cib.backend.ldap;

import org.isam.exehda.ResourceName;
import org.isam.exehda.ResourceName.NameComponent;
import org.isam.exehda.ResourceName.NameSpace;
import org.isam.exehda.ResourceConstraint;
import org.isam.exehda.ResourceConstraint.ResourceInfo;

import javax.naming.NamingException;
import javax.naming.directory.Attributes;


abstract class LdapResourceTypeHandler
{
        //
        // FIX ME: 'standard' un-mutable attributes, should be in the CellInformationBase
        // interface definition
        //
    protected static final String CIB_ATT_RES_TYPE = "TYPE";
    protected static final String CIB_ATT_RES_INSTANCE = "ID";
    
    public static String[] ALL_CIB_TYPES = null;
    
    public static LdapResourceTypeHandler
    getRootHandler()
        {
            return rootHandler;
        }

    public static void
    setRootHandler(LdapResourceTypeHandler rh)
        {
            rootHandler = rh;
        }

    public abstract DnBuffer
    encode ( NameComponent nc, DnBuffer dnb );

    public abstract NameComponent
    decodeNameComponent( String dnc );
    
    public abstract String
    createNode( LdapSession s, ResourceName rn ) throws NamingException;

    public abstract void
    deleteNode( LdapSession s, ResourceName rn ) throws NamingException;

    
        /**
         * Wrap the given directory attribute set into a LdapResourceInfo object.
         *
         * @param dn distinguished name of the resource
         * @param atts an <code>Attributes</code> value
         * @return a <code>LdapResourceInfo</code> value
         */
    public abstract LdapResourceInfo
    toLdapResourceInfo( String dn, Attributes atts );
    
        /**
         * Encode the given resource constraints into an LDAP search filter string. It is
         * expected the searching the LDAP server with the resulting filter will return a
         * super-set of the set defined by at all the resources handled by this handler
         * that satisfy the given resource constraint.
         *
         * By default, it returns the same value as <code>getDefaultSearchFilter()</code>.
         *
         * Notice that the closer the returned result set is to the real (intended) real
         * result set the better, since it will demand less post-processing. Althougth,
         * it is expected that not always it will be possible to build an exact match
         * filter. So, subclasses are highly encouraged to, at least, overwrite this
         * method to handle type AttributeSimpleMatchConstraint.
         *
         * @param rc a <code>ResourceConstraint</code> value
         * @return a <code>String</code> value
         */
    public String
    toSearchFilter( ResourceConstraint rc )
        {
            return getDefaultSearchFilter();
        }

    
        /**
         * Returns a filter that would match resources from all the CIB types handled by
         * this handler.
         *
         * @return a <code>String</code> value
         */
    public abstract String
    getDefaultSearchFilter();

    
    public String
    getNodeAttribute( LdapSession s, ResourceName rn, String a)
        throws NamingException
        {
                // 
                // FIX ME: performance improvement issue: this method should work on a
                // set of attributes instead of a single attribute
                //
            String dn = encode(rn, null).toString();
            
            return s.getNodeAttribute(dn, a);
        }

    public String[]
    listNodeAttributes( LdapSession s, ResourceName rn )
        throws NamingException
        {
            String dn = encode(rn, null).toString();

            String[] atts = s.listNodeAttributes(dn);

            return atts;
        }

    public void
    setNodeAttribute(  LdapSession s, ResourceName rn, String a, String v)
        throws NamingException
        {
                //
                // FIX ME: performance improvement issue: this method should work on a
                // set of attributes instead of a single attribute
                // 
            String dn = encode(rn, null).toString();

            s.setNodeAttribute(dn, a, v);
        }

    public DnBuffer
    encode( ResourceName rn, DnBuffer dnb )
        {
            if ( dnb == null ) dnb = new DnBuffer();
            
                // handle first name component
            encode(rn.getCommonName(), dnb);
                // delegate handling of namespace to the root handler
            getRootHandler().encode(rn.getNameSpace(), dnb);

            return dnb;
        }

    public DnBuffer
    encode( NameSpace ns, DnBuffer dnb )
        {
            if ( dnb == null ) dnb = new DnBuffer();

            if ( ns != null && !ns.equals(NameSpace.DEFAULT_NS) ) {
                    // handle first name component
                encode(ns.getCommonName(), dnb);
                    // if there is a parent name space
                if ( ns.getParent() != null ) {
                        // delegate handling of the parent name space to the root handler
                    getRootHandler().encode(ns.getParent(), dnb);
                }
            }

            return dnb;
        }

    public ResourceName
    decodeResourceName( String dn )
        {            
            NameSpace pns = null;
            NameComponent nc = null;
            
            int idx = dn.indexOf(',');
            if ( idx > 0 ) {
                    // handle parent namespace first
                pns = getRootHandler().decodeNameSpace(dn.substring(idx+1));
                    // handle this name component
                nc = decodeNameComponent(dn.substring(0,idx));
            }
            else {
                nc = decodeNameComponent(dn);
            }            
            
            return new ResourceName(nc.getType(), nc.getInstance(), pns);
        }

    public NameSpace
    decodeNameSpace( String dn )
        {
                //
                // TO DO: need parameter checking?
                //

            NameSpace pns = null;
            int idx = dn.indexOf(',');
            
            if ( idx > 0 ) {
                    // handle parent namespace first
                pns = getRootHandler().decodeNameSpace(dn.substring(idx+1));
            }
            else {
                idx = dn.length();
            }            
                // handle this name component
            NameComponent nc = decodeNameComponent(dn.substring(0,idx));
            
            return new NameSpace( nc, pns);
        }
    

    public String[]
    getHandledCibTypes()
        {
            return this.handledCibTypes;
        }

    public String[]
    getHandledDnPrefixes()
        {
            return this.handledDnPrefixes;
        }

    
        /**
         * Creates a new <code>LdapResourceTypeHandler</code> instance.
         *
         * @param cibTypes a <code>String[]</code> value, null mean ALL types
         * @param dnPrefixes a <code>String[]</code> value, null means ANY prefixes
         */
    protected LdapResourceTypeHandler(String[] cibTypes, String[] dnPrefixes)
        {
            this.handledCibTypes = cibTypes;
            this.handledDnPrefixes = dnPrefixes;
        }

        /////////////////////////////////////////////////////
        // PRIVATE STUFF
        /////////////////////////////////////////////////////
    
    private static LdapResourceTypeHandler rootHandler = null;

    private String[] handledCibTypes;
    private String[] handledDnPrefixes;
}
