// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.ctxm.filter.basic;

import org.isam.exehda.services.ctxm.ContextFilter;
import org.isam.exehda.services.ctxm.ContextFilterFactory;
import java.lang.reflect.Constructor;

/**
 * Factory for context filtersr that are able to filter data from a single raw sensor.
 *
 * Supported context filter syntaxes:
 *
 * <ul>
 *
 * <li> <code>xml:NumericRangeMapping</code>: maps a double floating point value to an
 * String based on defined floating point ranges.
 *
 * <li> <code>xml:StringMapping</code>: maps strings to string using suffix, prefix,
 * infix or equality match checkings.
 *
 * </ul>
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-19 15:51:38 -0300 (Sun, 19 Jun 2005) $ $Revision: 941 $
 */
public class BasicContextFilterFactory
    implements ContextFilterFactory
{
    final static String FILTER_PREFIX = "org.isam.exehda.services.ctxm.filter.basic.SingleSensor";
    final static String FILTER_SUFFIX = "Filter";
    final static Class[] CONSTRUCTOR_PARAMS = new Class[] { String.class };

       /**
         * Create a context filter able to produce the context state
         * information acordingly to the given filter definition.
         *
         * @todo Throw InvallidContextDefinitionException
         * @todo Document supported filter definition sintaxes/idioms
         *
         * @param fdef a filter definition
         * @return a <code>ContextFilter</code> value
         */
    public ContextFilter
    createFilter ( String fdef )
        {
            String syntax = getFilterSyntax(fdef);

            if ( syntax.regionMatches( false, 0, "xml:", 0, 4) ) {

                syntax = syntax.substring(4);
                
                try {
                        // dinamically load the filter corresponding to the selected syntax
                    Constructor c = Class.forName(FILTER_PREFIX + syntax + FILTER_SUFFIX)
                        .getDeclaredConstructor(CONSTRUCTOR_PARAMS);
            
                    return (ContextFilter) c.newInstance(new Object[] {fdef});
                }
                catch (Exception e) {
                    throw new IllegalArgumentException(
                        "Could not create the context filter for \n"
                        +fdef
                        +"\n due to "+e);
                }
            }
            else {
                throw new IllegalArgumentException(
                    "Unsupported filter syntax: "+syntax);
            }
        }

    private final String
    getFilterSyntax(String fdef)
        {
                //
                // FIX ME: need stronger filter DTD checking/parsing
                //

            int idx0 = fdef.indexOf(" syntax");
            if ( idx0 > 0 ) {
                idx0 = fdef.indexOf('=', idx0);
                if ( idx0 > 0 ) {
                    idx0 = fdef.indexOf('"', idx0) + 1;
                }
            }
            
            int idx1 = (idx0 > 0 ) ? fdef.indexOf('"', idx0) : -1;

            if ( idx0>0 && idx1>0 ) {
                return fdef.substring(idx0, idx1);
            }
            else {
                throw new IllegalArgumentException(
                    "Malformed filter definition: missing 'syntax' attribute");
            }
        }
        
}
