// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.ctxm.filter.basic;


import org.isam.exehda.services.ctxm.Environment;
import org.isam.util.xml.XmlElement;
import java.util.Vector;


/**
 * This context filter that takes as input a value of type <code>double</code> from a
 * single sensor in the Environment and translates it into abstract context states
 * (Strings) according to defined ranges.
 *
 * <p>Accepted syntax: 
 *
   <pre> 
   &lt;FILTER syntax="xml:NumericRangeMapping"&gt;
 
     &lt;INDEX&gt;
        &lt;SENSOR n="sensorName"/&gt;
     &lt;/INDEX&gt;

     &lt;MAPPINGS&gt;
        &lt;RANGE ubo="0.3" state="state1"/&gt;
        &lt;RANGE lb="1.0" ub="2.0" state="state2"/&gt;
        &lt;RANGE lbo="2.0" ub="3.0" state="state3"/&gt;
        &lt;RANGE lbo="50" state="state4"/&gt;
        &lt;DEFAULT state="state5"/&gt;
     &lt;/MAPPINGS&gt;
 
   &lt;/FILTER&gt;
   </pre>
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-19 20:02:57 -0300 (Sun, 19 Jun 2005) $ $Revision: 942 $
 */
class SingleSensorNumericRangeMappingFilter
    extends AbstractSingleSensorFilter
{    
    public SingleSensorNumericRangeMappingFilter( String fdef  ) 
        {
            super(fdef);
        }

        ////////////////////////////////////////////////////////////////////
        // PROTECTED STUFF
        ////////////////////////////////////////////////////////////////////

    
        /**
         * Checks the new input value against the defined numeric ranges, updating the
         * attribute <code>currState</code> accordingly.
         *
         * @param env an <code>Environment</code> value
         */
    protected final void
    updateCurrentState( Environment env )
        {
                // read environment variable
            double sval = readEnvironment(env);
                // for each defined range 
            for (int i=0; i<ranges.length; i++) {
                    // check if it matches the current sensor value
                if ( ranges[i].contains(sval) ) {
                        // update current state
                    this.currState = ranges[i].getState(); 
                    break;
                }
            }
        }

        /**
         * Setup context state index range mappings.
         *
         * @todo Review this code, its not very safe/robust in face of malformed context
         * definitions.
         */
    protected final void
    initMappings( XmlElement docroot )
        {
            final String XP_RANGE     = "/FILTER/MAPPINGS/RANGE";
            final String XP_DFLT_STATE = "/FILTER/MAPPINGS/DEFAULT";
    
                // select <RANGE> eleements
            Vector tmp = XmlElement.selectByPath(docroot, XP_RANGE, null);
                // sanity check
            if ( tmp.size() == 0 ) {
                throw new IllegalArgumentException(
                    "Missing range mappings in context definition");
            }
            
            this.ranges = new NumericRange[tmp.size()];
                // for each <RANGE> element
            for ( int i=0; i<ranges.length; i++) {
                    // build the corresponding NumericRange object
                this.ranges[i] = makeRange((XmlElement) tmp.elementAt(i));
            }
                // optimize range order for faster matching
            sortRanges();
            
                // setup default context state
            tmp = XmlElement.selectByPath(docroot, XP_DFLT_STATE, null);
            this.dfltState = ((XmlElement) tmp.elementAt(0))
                .getAttribute("state", null);
            
                // last sanity check
            if ( this.dfltState == null ) { 
                throw new IllegalArgumentException(
                    "Missing default context state definition");
            }
        }

        /////////////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        /////////////////////////////////////////////////////////////////////

    
        /**
         * Builds a NumericRange object from the given <code>RANGE</code> XML element.
         *
         * @param xeRange a <code>XmlElement</code> value
         * @return a <code>NumericRange</code> value
         */
    private final NumericRange
    makeRange(XmlElement xeRange)
        {
                // handle range upper bound
            boolean ubo = true;
            String vu = xeRange.getAttribute("ubo", null);
            if ( vu == null ) {
                vu = xeRange.getAttribute("ub", null);
                if ( vu != null ) ubo = false;
            }

            double ub = ( vu == null ) ? Double.POSITIVE_INFINITY
                : Double.valueOf(vu).doubleValue();            
    
                // handle range lower bound
            boolean lbo = true;
            String vl = xeRange.getAttribute("lbo", null);
            if ( vl == null ) {
                vl = xeRange.getAttribute("lb", null);
                if ( vl != null ) lbo = false;
            }
            double lb = ( vl == null ) ? Double.NEGATIVE_INFINITY
                : Double.valueOf(vl).doubleValue();

            String vs = xeRange.getAttribute("state", null);
            if ( vs == null ) {
                throw new IllegalArgumentException(
                    "Missing context state in range mapping definition");
            }

            return new NumericRange(ub, ubo, lb, lbo, vs);
        }

    
        /**
         * Optimize ranges array ordering in order to allow a faster range matching in
         * method updateCurrentState()
         *
         */
    private final void
    sortRanges()
        {
                //
                // TO DO: sort this.ranges considering ranges upper and lower bounds.
                //
        }
    
        /**
         * Read the selected environment variable, usually a sensor, returning its
         * current value as a double value.
         *
         * @param env an <code>Environment</code> value
         * @return a <code>double</code> value
         */
    private final double
    readEnvironment(Environment env)
        {
            try {
                return env.getDouble(sname, Double.NaN, true);
            }
            catch (Exception e) {
                return Double.NaN;
            }
        }


    private NumericRange[] ranges;
}
