// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.ctxm.filter.basic;


import org.isam.exehda.services.ctxm.Environment;
import org.isam.util.xml.XmlElement;
import java.util.Vector;


/**
 * This context filter that takes as input a value of type <code>String</code> from a
 * single sensor in the Environment and translates it into abstract context states
 * according to defined string matching rules.
 *
 * <p>Accepted syntax:
 *
   <pre>
   &lt;FILTER syntax="xml:StringMapping"&gt;

     &lt;INDEX&gt;
        &lt;SENSOR s="sensorName"/&gt;
     &lt;/INDEX&gt;

     &lt;MAPPINGS&gt;
        &lt;MATCH op="suffix" v="valueToMatch" ignoreCase="true"  state="s1"/&gt;
        &lt;MATCH op="prefix" v="valueToMatch" ignoreCase="true"  state="s2"/&gt;
        &lt;MATCH op="infix"  v="valueToMatch" ignoreCase="false" state="s3"/&gt;
        &lt;MATCH op="equals" v="valueToMatch" ignoreCase="true"  state="s4"/&gt;
        &lt;DEFAULT state="s5"/&gt;
     &lt;/MAPPINGS&gt;
     
   &lt;/FILTER&gt;
   </pre>
 *
 * @author last modified by $Author: lucc $
 * @version $Date: 2005-06-19 20:02:57 -0300 (Sun, 19 Jun 2005) $ $Revision: 942 $
 */
class SingleSensorStringMappingFilter
    extends AbstractSingleSensorFilter
{
    public SingleSensorStringMappingFilter( String fdef )
        {
            super(fdef);
        }

        ///////////////////////////////////////////////////////////////
        // PROTECTED STUFF
        ///////////////////////////////////////////////////////////////
    
    protected final void
    updateCurrentState( Environment env )
        {
                // read environment variable
            String sval = readEnvironment(env);
                // XXX: should never happend, but...
            if ( sval == null ) return;                    
                // for each defined range 
            for (int i=0; i<rules.length; i++) {
                    // check if it matches the current sensor value
                if ( rules[i].match(sval) ) {
                        // update current state
                    this.currState = rules[i].getState(); 
                    break;                    
                }
            }
        }

    protected final void
    initMappings( XmlElement docroot )
        {
            final String XP_MATCH      = "/FILTER/MAPPINGS/MATCH";
            final String XP_DFLT_STATE = "/FILTER/MAPPINGS/DEFAULT";
    
                // select <MATCH> eleements
            Vector tmp = XmlElement.selectByPath(docroot, XP_MATCH, null);
                // sanity check
            if ( tmp.size() == 0 ) {
                throw new IllegalArgumentException(
                    "Missing matching rule mappings in context definition");
            }
            
            this.rules = new MatchingRule[tmp.size()];
                // for each <MATCH> element
            for ( int i=0; i<rules.length; i++) {
                    // build the corresponding MatchingRule object
                this.rules[i] = makeRule((XmlElement) tmp.elementAt(i));
            }
                // optimize rule order for faster matching
            sortRules();
            
                // setup default context state
            tmp = XmlElement.selectByPath(docroot, XP_DFLT_STATE, null);
            this.dfltState = ((XmlElement) tmp.elementAt(0))
                .getAttribute("state", null);
            
                // last sanity check
            if ( this.dfltState == null ) { 
                throw new IllegalArgumentException(
                    "Missing default context state definition");
            }
        }
    
        ///////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        ///////////////////////////////////////////////////////////////

        /**
         * Read the selected environment variable, usually a sensor, returning its
         * current value as a String.
         *
         * @param env an <code>Environment</code> value
         * @return a <code>String</code> value
         */
    private final String
    readEnvironment(Environment env)
        {
            try {
                return env.getString(sname, null, true);
            }
            catch (Exception e) {
                return null;
            }
        }

        /**
         * Builds a MatchingRule object from the given <code>MATCH</code> XML element.
         *
         * @param xeMatch a <code>XmlElement</code> value
         * @return a <code>MatchingRule</code> value
         */
    private final MatchingRule
    makeRule(XmlElement xeMatch)
        {
                // get selected matching operation
            String tmp = xeMatch.getAttribute("op", null);
            int op = "prefix".equals(tmp) ? MatchingRule.OP_PREFIX
                : "suffix".equals(tmp) ? MatchingRule.OP_SUFFIX
                : "infix".equals(tmp) ? MatchingRule.OP_INFIX
                : "equals".equals(tmp) ? MatchingRule.OP_EQUALS
                : -1;
                // check matching operation is valid
            if ( op >= 0 ) {            
                    // get defined matching expression
                String v  = xeMatch.getAttribute("v", null);
                    // check matching value is valid
                if ( v != null ) {
                        // get the selected case sensitivity strategy
                    tmp = xeMatch.getAttribute("ignoreCase", null);
                    int ignoreCase = "false".equals(tmp) ? 0
                        : "true".equals(tmp) ? 1
                        : -1;
                        // check ignoreCase is a boolean value
                    if ( ignoreCase >= 0 ) {
                        String state = xeMatch.getAttribute("state", null);
                            // check state is valid
                            // FIX ME: currently just check it's a non-null value, what
                            // is not enough to ensure it is valid
                        if ( state != null ) {
                            return new MatchingRule(v, op, ignoreCase!=0, state);
                        }
                    }
                }
            }
                // if it got here, something bad occured during parsing of the rule
            throw new IllegalArgumentException(
                "Malformed matching rule: "+xeMatch);
        }

        /**
         * Optimize rules array ordering in aiming at faster range matching in method
         * updateCurrentState().
         *
         */
    private final void
    sortRules()
        {
                // TO DO: Is there a general way of optimizing rules order? If you can
                // figure out something, implement it here! :-)
        }
    
    
        /**
         * Helper class representing a string matching rule.
         *
         */
    private static class MatchingRule
    {
        static final int OP_ALLWAYS = 0;
        static final int OP_PREFIX  = 1;
        static final int OP_SUFFIX  = 2;
        static final int OP_INFIX   = 3;
        static final int OP_EQUALS  = 4;

        public MatchingRule(String e, int op, boolean ignoreCase, String state)
            {
                    // XXX: except for infix matching, preserving orignial case *would*
                    // improve matching performance
                this.exp = ( op==OP_INFIX && ignoreCase ) ? e.toUpperCase() : e;
                this.op = op;
                this.ignoreCase = ignoreCase;
                this.state = state;
            }

        public String
        getState()
            {
                return this.state;
            }

        public boolean match(String v)
            {
                switch(op) {
                    case OP_PREFIX:
                        return exp.regionMatches(ignoreCase,
                                                 0,
                                                 v,
                                                 0,
                                                 exp.length());
                    case OP_SUFFIX:
                        return exp.regionMatches(ignoreCase,
                                                 0,
                                                 v,
                                                 Math.max(v.length() - exp.length(), 0),
                                                 exp.length());
                    case OP_INFIX:
                            // XXX: exp is already in uppercase if ignoreCase is true
                        return ((ignoreCase) ? v.toUpperCase() : v ).indexOf(exp) > 0;
                        
                    case OP_EQUALS:
                        return (ignoreCase) ? exp.equalsIgnoreCase(v) : exp.equals(v);
                        
                    case OP_ALLWAYS:
                        return true;
                    default:
                        return false;
                }
            }

        private String exp;
        private int op;
        private boolean ignoreCase;
        private String state;
    }

    private MatchingRule[] rules;
}



