// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.oxm;

import org.isam.exehda.Exehda;
import org.isam.exehda.HostId;
import org.isam.exehda.ObjectId;
import org.isam.exehda.ApplicationId;
import org.isam.exehda.services.AbstractService;
import org.isam.exehda.services.Logger.Sev;
import org.isam.exehda.services.OXManager;
import org.isam.exehda.services.OXManager.OXHandle;
import org.isam.exehda.services.OXManager.MasterOXHandle;
import org.isam.exehda.services.OXManager.HomeOXHandle;
import org.isam.exehda.services.OXManager.ProxyOXHandle;
import org.isam.exehda.services.OXManager.IllegalOXHandleException;
import org.isam.exehda.services.OXManager.OXAttributeOps;
import org.isam.exehda.services.CellInformationBase;
import org.isam.exehda.ResourceName;
import org.isam.exehda.ResourceName.NameComponent;
import org.isam.exehda.ResourceName.NameSpace;
import org.isam.exehda.services.Worb;
import java.util.Hashtable;



public abstract class AbstractOXManager
    extends AbstractService
    implements OXManager
{
    protected final static HostId LOCAL_HOST = HostId.getLocalHost();

    public AbstractOXManager()
        {
            this.masterHandlesById = new Hashtable();
            this.grantedHandles = new Hashtable();
        }
        
        /**
         * If this is the Home node of the given ox, triggers the creation of master and
         * home OX handles up on the first call. Always returns a proxy ox handle.
         *
         * @param oid the ID of the target ox
         * @return a proxy ox handle for accessing meta-attributes of the given ox.
         */
    public OXHandle
    createHandle(org.isam.exehda.ObjectId oid)
        {
                //  always return a proxy
                //
                // XXX: no cheking is done if the remote ox actually exists
                //
            ProxyOXHandle ph = new ProxyOXHandle(oid);
            
                // if this node is the home node for the given ox
            if ( LOCAL_HOST.equals(oid.getHost()) ) {
                    // create the master handle (if needed)
                if ( createMasterHandle(oid, ph) ) {
                        // master was created, so need also to create the home handle
                    createHomeHandle(oid, ph);
                }
            }

//             log(Sev.DEBUG, "createHandle()", "ph.key={0}", ph.getKey());
                
            return ph;
        }

        /**
         * Release handle takes away the given handle from the handle list.
         * Only to be uswed when the OX is no longer necessary.
         *
         * @param oxh a <code>OXHandle</code> that is to be realeased
         * @return void
         */
    public void
    releaseHandle(OXHandle oxh)
        {
                //
                // TO DO: should delete the master handle and home handle
                //
            throw new NullPointerException("Not yet implemented.");
        }
    
        /**
         * If the given ox is owned by this manager, flag the ox management as granted
         * and returns the master handle. Otherwise, returns the same handle.
         *
         * Formelly, it returns the master handle after removind it from the internal tables.
         *
         * @param oxh a <code>OXHandle</code> whose ownership is being granted
         * @return a <code>OXhandle</code> suitable to be used in a subsequent call to
         * grantOwnership()
         */
    public OXHandle
    requestOwnership(OXHandle oxh)
        throws IllegalOXHandleException 
        {
            ObjectId oid = oxh.getId();

            OXHandle masterHandle = (OXHandle) masterHandlesById.remove(oid);
            if ( masterHandle != null ) {
                
//                 log(Sev.DEBUG, "super.requestOwnership()",
//                     "master key={0}, proxy key={1}",                    
//                     masterHandle.getKey(), oxh.getKey());
                    
                    // check if the given handle 'knows' the secret
                if ( masterHandle.getKey().equals(oxh.getKey()) ) {
                        // mark it as granted and return the master handle
                    grantedHandles.put(oid,oid);
                    return masterHandle;
                }
                else {
                        // rollback, restoring the master handles table
                    masterHandlesById.put(oid, masterHandle);
                        // abort, the operation could not be completed
                    throw new IllegalOXHandleException(
                        "Operation not allowed for the given handle: "+oxh);
                }
            }
            else {
                    // do nothing, already granted
                return oxh;
            }
        }
    

        /**
         * If the given ox handle is a master han>dle and is not owned by this manager,
         * assigns its ownership to this manager.
         *
         * Formelly, inserts the given master handle in the internal tables.
         *
         * @param oxh a master ox handle
         * @return an <code>OXHandle</code> suitable to be used in a subsequent call to
         * requestOwnership()
         * @exception IllegalOXHandleException if an error occurs
         */
    public OXHandle
    grantOwnership(OXHandle oxh)
        throws IllegalOXHandleException 
        {
                //
                // TO DO: review synchronization
                //
            
            if ( oxh instanceof MasterOXHandle ) {
                ObjectId oxID = oxh.getId();
                    // if already owned
                if ( masterHandlesById.containsKey(oxID) ) {
                        // nothing to do
                    return oxh;
                }
                else {
                        // reset the grant information, since the ox may be 
                        // returning to this node for the second time
                    grantedHandles.remove(oxID);
                        // insert the master handle in the table
                    masterHandlesById.put(oxID, oxh);
                        //
                        // TO DO: update the 'owner' attribute
                        //
//                     oxh.setOwner(LOCAL_HOST);
                        // create a new proxy bound to the master handle
                    OXHandle ph = new ProxyOXHandle(oxID);
                    Object key = genHandleKey();
                    oxh.setKey(key);
                    ph.setKey(key);
                        // returns the new proxy
                    return ph; 
                }
            }
            else {
                throw new IllegalOXHandleException(
                    "Expected a MasterOXHandle but found "+oxh);
            }
        }
    
    
        /**
         * OXhandle updates the attributes of the given ox. The attributes to be updated
         * are defined in a String array. What type of update is to be done (read or write)
         * is defined in an byte array.
         *
         * @param oxh a <code>OXHandle</code> whose attributes are to be updated
         * @param atts an array defining what attributes are to be updated
         * @param ops an array defining the type op update operation for eche selected attribute
         * @return a <code>OXHandle</code> that is the OXHandle passed as parameter with
         * it's attributes updated
         */
    public OXHandle
    updateAttributes(OXHandle oxh, String[] atts, byte[] ops)
        throws IllegalOXHandleException 
        {
            log(Sev.TRACE, "super.updateAttributes()", "ox={0}", oxh);
            
                // expects a proxy handle
            if ( oxh instanceof ProxyOXHandle ) {
                ObjectId oxID = oxh.getId();
                    // get the master handle
                MasterOXHandle mh = (MasterOXHandle) masterHandlesById.get(oxID);
                if ( mh != null ) {
                        // merge attributes
                    mergeAttributes(mh, (ProxyOXHandle) oxh, atts, ops);
                    return oxh;
                }
                else {
                    throw new IllegalOXHandleException(
                        "Not owner of "+oxID);
                }
            }
            else {
                throw new IllegalOXHandleException(
                    "Expected a ProxyOXHandle but found "+oxh);
            }
        }

        //////////////////////////////////////////////////////////////
        // PROTECTED STUFF
        //////////////////////////////////////////////////////////////

        /**
         * Must be overwritten in subclasses since home handle creation depends on the ox
         * manager hole.
         *
         * <p><b>pos-condition</b>: the created home handle is to be bound to the same
         * key as the given proxy handle
         *
         * @param oid an <code>ObjectId</code> value
         * @param ph a proxy handle
         */
    protected abstract void
    createHomeHandle(ObjectId oid, OXHandle ph);


        /**
         * Generate a key to be bound to an OX handle instance. Such key is used to
         * implement access control schemes.
         *
         * @return an <code>Object</code> value
         */
    protected Object
    genHandleKey()
        {
                //
                // FIX ME: need stronger key generation scheme
                //
            return new Double( Math.random() );
        }
    
        /**
         * Checks whether the master handle of the given ox is owned by this manager.
         *
         * @param oid an <code>ObjectId</code> value
         * @return true if the master handle ownership belongs to this manager, false
         * otherwise.
         */
    protected final boolean
    hasOwnershipOf(ObjectId oid)
        {
            return masterHandlesById.containsKey(oid);
        }
    
        /**
         * Checks wheter the master handle of the given ox had been owned by this manager
         * but have been granted to another manager.
         *
         * @param oid an <code>ObjectId</code> value
         * @return true if the ownership of the given master handle have been granted by
         * this manager to another manager.
         */
    protected final boolean
    grantedOwnershipOf(ObjectId oid)
        {
            return grantedHandles.containsKey(oid);
        }


        //////////////////////////////////////////////////////////////
        // PRIVATE STUFF
        //////////////////////////////////////////////////////////////
    
        /**
         * Creates the Master OXHandle for the given ox if it was not yet created
         *
         * @param oid an <code>ObjectId</code> value
         * @param ph a proxy handle that will be bound to the same key as the master
         * allowing it to be later used in grantOwnership().
         * @return true if the a new master handle has been created, false if it already
         * existed
         */
    private final boolean
    createMasterHandle(ObjectId oid, OXHandle ph)
        {
                // if the master handle has already been  created
            if ( masterHandlesById.containsKey(oid)
                 || grantedHandles.containsKey(oid) ) {
                    // nothing to do
                return false;
            }
            else {
                    // create the master handle
                MasterOXHandle mh = new MasterOXHandle(oid);
                masterHandlesById.put(oid, mh);

                    // bind the master handle and the given proxy
                Object key = genHandleKey();
                mh.setKey(key);
                ph.setKey(key);

                log(Sev.DEBUG, "createMasterHandle()",
                    "ph.key={0}, mh.key={1}", ph.getKey(), mh.getKey());
                    
                return true;
            }
        }

        /**
         * Merge the set of attributes defined in atts from both handles following the
         * operations defined in ops.
         *
         * @param mh a <code>MasterOXHandle</code> value
         * @param ph a <code>ProxyOXHandle</code> value
         * @param atts a <code>String[]</code> value
         * @param ops a <code>byte[]</code> value
         */
    private final void
    mergeAttributes(MasterOXHandle mh, ProxyOXHandle ph, String[] atts, byte[] ops)
        {
                //
                // TO DO: check array sizes match
                //
            
            for (int i=0; i<atts.length; i++) {
                String a = atts[i];
                
                switch ( ops[i] ) {
                    case OXAttributeOps.WRITE:
                            // ph -> mh
                        mh.setAttribute(a, ph.getAttribute(a));
                        break;
                    case OXAttributeOps.READ:
                            // mh -> ph
                        ph.setAttribute(a, mh.getAttribute(a));
                        break;
                    default:
                            //
                            // TO DO: support other operations defined in OxAttributeOps
                            //
                        log(Sev.DEBUG, "mergeAttributes()",
                            "Unsupported attribute operation "+ops[i]+" on att {0}", a);
                        break;
                }
            }
        }


        //
        // FIX ME: use per-application master handle tables
        //

        /**
         * Map from ox ID to active (owned) master handles
         */
    private Hashtable masterHandlesById;
        /**
         * Set of existing but unmanaged master handles. The ownership of this handles
         * have been granted to another oxmanager.
         */
    private Hashtable grantedHandles;

}
