package dimi.channel;

import dimi.configuration.ChannelConfiguration;
import dimi.ChannelManager;
import dimi.channel.connection.DisseminationUnitConnection;
import dimi.name.ChannelName;
import dimi.message.Message;
import dimi.message.MessageHeader;
import dimi.exception.CantEnterChannelException;
import dimi.channel.connection.exception.CantCreateConnectionException;
import dimi.Disseminator;
import dimi.name.DBaseName;

import java.rmi.RemoteException;
import java.util.Timer;
import java.util.TimerTask;

/**
 * Um canal de disseminador. Se diferencia de um canal de ponto de entrada
 * de usuario pois pode possuir muitas conexoes.
 * @author mcmoraes
 */
public class DisseminationUnitChannel extends Channel{

    //atributos - INICIO
    protected ConnectionManager connectionManager = null;
    protected ChannelManager channelManager = null;
    private TopologyManager topologyManager = null;
    private DBaseName[] partialAscendenceList = null;
    private NeighbourhoodManager neighbourhoodManager = null;
    private int desiredPeerEntryPointsNumber = 1;
    private Timer timer = null;
    //atributos - FIM
    
    //construtores - INICIO    
    /**
     * Creates a new instance of DisseminatorChannel
     * @param names os nomes do canal. Primeiro o nome local, depois o nome de raiz.
     * @param channelNeighbourhoodManager o NeighbourhooManager deste canal
     * @param configuration a configuracao do novo canal
     * @param channelManager o gerente de canais do novo canal
     */
    public DisseminationUnitChannel(ChannelName[] names, ChannelConfiguration configuration, ChannelManager channelManager,NeighbourhoodManager neighbourhoodManager) {
        super(names,configuration);
        this.connectionManager = new ConnectionManager(this);
        this.partialAscendenceList = new DBaseName[0];
        if(neighbourhoodManager != null){
            this.neighbourhoodManager = neighbourhoodManager;
        }
        else{
            this.neighbourhoodManager = new NeighbourhoodManager();
        }
        this.timer = new Timer();
        this.topologyManager = new TopologyManager();
        this.timer.schedule(topologyManager,2000,2000);
    }
    //construtores - FIM

    //setters e getters - INICIO
    /**
     * Getter for property connections.
     * @return Value of property connections.
     */
    public ConnectionManager getConnectionManager() {
        return connectionManager;
    }    
    
    /**
     * Getter for property channels.
     * @return Value of property channels.
     */
    public ChannelManager getChannelManager() {
        return channelManager;
    }
    
    /**
     * Getter for property partialAscendenceList.
     * @return Value of property partialAscendenceList.
     */
    public DBaseName[] getPartialAscendenceList() {
        return partialAscendenceList;
    }

    /**
     * Getter for property neighbourhoodManager.
     * @return Value of property neighbourhoodManager.
     */
    public NeighbourhoodManager getNeighbourhoodManager() {
        return neighbourhoodManager;
    }
    
    /**
     * Getter for property desiredPeerEntryPointsNumber.
     * @return Value of property desiredPeerEntryPointsNumber.
     */
    public int getDesiredPeerEntryPointsNumber() {
        return desiredPeerEntryPointsNumber;
    }

    /**
     * Setter for property desiredPeerEntryPointsNumber.
     * @param desiredPeerEntryPointsNumber New value of property desiredPeerEntryPointsNumber.
     */
    public void setDesiredPeerEntryPointsNumber(int desiredPeerEntryPointsNumber) {
        if(desiredPeerEntryPointsNumber>this.desiredPeerEntryPointsNumber){
            this.desiredPeerEntryPointsNumber = desiredPeerEntryPointsNumber;
        }
    }
    //setters e getters - FIM

   /**
     * Monta a lista parcial de ascendencia deste canal, a partir de outra
     * lista parcial de ascendencia.
     * @param partialAscendenceList a lista parcial de ascendencia a servir de base
     */    
    public void mountPartialAscendenceListFromParent(DBaseName[] partialAscendenceList) {
        this.partialAscendenceList = new DBaseName[partialAscendenceList.length+1];
        try{
            this.partialAscendenceList[0] = (DBaseName)channelManager.getDisseminationUnitImpl().getName();
        }catch(RemoteException re){}
        for(int i=0;i<partialAscendenceList.length;i++){
            this.partialAscendenceList[i+1] = partialAscendenceList[i];
        }
    }

    public void addDisseminator(Disseminator disseminator,int desiredEntryPointsNumber) throws CantEnterChannelException{
        try{
            connectionManager.createConnection(disseminator,false);
        }catch(RemoteException re){
            throw new CantEnterChannelException();
        }catch(CantCreateConnectionException ccce){
            throw new CantEnterChannelException();
        }
        setDesiredPeerEntryPointsNumber(desiredEntryPointsNumber);
    }

    /**
     * Destroi este canal.
     */    
    public void destroy(){
        super.destroy();
        connectionManager.destroyAllConnections();
        topologyManager.cancel();
        timer.cancel();
    }

    /**
     * Protocolo de disseminacao de dados em um canal
     * @param message a mensagem contendo os dados de um canal
     * @throws CantFindChannelException se nao for possivel encontrar o canal
     * @throws RemoteException indicate that either a communication failure or a protocol error has occurred
     */
    public final synchronized void disseminate(Message message) throws RemoteException{
        switch(message.getHeader().getType()){
            case (MessageHeader.DATA):{
                disseminateMessage(message);
                break;
            }
            case (MessageHeader.HEART_BEAT):{
                processHeartBeatMessage(message);
                disseminateMessage(message);
                break;
            }
            case (MessageHeader.DESTROY_CHANNEL):{
                disseminateMessage(message);
                processDestroyChannelMessage(message);
                break;
            }
            case (MessageHeader.LEAVE):{
                processLeaveMessage(message);
                break;
            }
        }
    }
    
    protected final synchronized void disseminateMessage(Message message){
        if(message.getHeader().getPath().contains(channelManager.getDisseminationUnitName())){
System.out.println("cicle found.");
            connectionManager.destroyConnection(message.getHeader().getOriginName());
System.out.println("connection removed: " + message.getHeader().getOriginName());
        }
        else{
            message.getHeader().getPath().add(channelManager.getDisseminationUnitName());
            connectionManager.disseminate(message);
        }
    }    
    protected final synchronized void processHeartBeatMessage(Message message){
    }
    
    protected final synchronized void processDestroyChannelMessage(Message message){
        channelManager.destroyChannel(getNames()[Channel.ROOT_NAME]);
    }
    
    protected final synchronized void processLeaveMessage(Message message){
        connectionManager.destroyConnection(message.getHeader().getOriginName());
    }

    private class TopologyManager extends TimerTask{

        public void run() {

        }

    }

}
