// Copyright 2003 ISAM Team <http://www.inf.ufrgs.br/~isam/>
//
// This file is part of ISAM.
//
// ISAM is free software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the Free Software
// Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// ISAM is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ISAM; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

package org.isam.exehda.services.primos.objseed;

import org.isam.exehda.Exehda;
import org.isam.exehda.ObjectId;
import org.isam.exehda.HostId;
import org.isam.exehda.services.ObjectSeed.Activator;
import org.isam.exehda.services.ObjectSeed.MarshaledOX;
import org.isam.exehda.services.OXManager;
import org.isam.exehda.services.OXManager.OXHandle;

/**
 * System-view of an OX. Wraps all OX management information into a single
 * structure.
 */
final class MetaOX
{
    MetaOX( MarshaledOX oxState )
        throws java.io.IOException, ClassNotFoundException
        {
                // restore ox from oxState
            this.id = (ObjectId)
                oxState.getAttribute(Activator.OXATT_ID);
                    
            this.instance = (Object)
                oxState.getAttribute(Activator.OXATT_STATE);
                    
            this.handle = (OXHandle)
                oxState.getAttribute(Activator.OXATT_HANDLE);
                    
            this.customActivator = (Activator)
                oxState.getAttribute(Activator.OXATT_ACTIVATOR);

            this.isActive = false;
        }
        
    MetaOX( ObjectId oxID, Object oxInstance, Activator customActivator )
        {
            this.id = oxID;
            this.instance = oxInstance;
            this.handle = null;
            this.customActivator = customActivator;
            this.isActive = false;
        }
        
    final ObjectId
    getId()
        {
            return this.id;
        }
        
    final void
    activate()
        throws Exception
        {
            if ( isActive ) {
                return;
            }
            else {
                    //
                    // first proceed general OX activation...
                    //

                    // trigger MasterOXHandle creation
                this.handle = getOXManager().createHandle(this.id);
                    // update meta-attribute 'ox.location'
                this.handle.setLocation(HostId.getLocalHost());

                    //
                    // ...then invoke custom activator, if one was defined
                    //
                if ( customActivator != null ) {
                    customActivator.activate(this.id, this.instance, null);
                }

                this.isActive = true;
            }
        }

    final void
    activate( MarshaledOX oxState )
        throws Exception
        {
            if ( this.isActive ) {
                    //
                    // XXX: Oops something is wrong if it get here
                    //
                return;
            }
            else {
                    //
                    // first proceed general OX activation...
                    //

                    // request ownership on the ox handle
                this.handle = getOXManager().requestOwnership(this.handle);
                this.handle.setLocation(HostId.getLocalHost());

                    // clean-up standard attributes from oxState since it may contain
                    // sensitive information
                oxState.setAttribute(Activator.OXATT_ID, null);
                oxState.setAttribute(Activator.OXATT_STATE, null);
                oxState.setAttribute(Activator.OXATT_HANDLE, null);
                oxState.setAttribute(Activator.OXATT_ACTIVATOR, null);

                    //
                    // ...then invoke custom activator, if one was defined
                    //
                if ( this.customActivator != null ) {
                    this.customActivator.activate(this.id, this.instance, oxState);
                }

                this.isActive = true;
            }
        }

    final void
    deactivate( MarshaledOX oxState )
        throws Exception
        {
            if ( this.isActive ) {
                    //
                    // Reverse order: first custom de-activation if defined...
                    //
                if ( this.customActivator != null ) {
                    this.customActivator.deactivate(this.id, this.instance, oxState);
                }

                    //
                    // ... then, general deactivation
                    //

                    // grant ownership on master handle
                this.handle = getOXManager().grantOwnership(this.handle);

                    // save standard attributes
                oxState.setAttribute(Activator.OXATT_ID, this.id);
                oxState.setAttribute(Activator.OXATT_STATE, this.instance);
                oxState.setAttribute(Activator.OXATT_HANDLE, this.handle);
                    //
                    // XXX: review this line, will standard activators (e.g. thread,
                    // rmi) be availble when the OX is reactivated?
                    //
                oxState.setAttribute(Activator.OXATT_ACTIVATOR, this.customActivator);
                    

                this.isActive = false;
            }
            else {
                    //
                    // FIX ME: Oops, should not get here
                    //
                return;
            }
        }

    public final int
    hashCode()
        {
            return id.hashCode();
        }
        
    public final boolean
    equals( Object o )
        {
            if ( o != null && o instanceof MetaOX ) {
                return this.id.equals( ((MetaOX) o).id ) ;
            }
            return false;
        }

        ////////////////////////////////////////////////
        // PRIVATE STUFF
        ////////////////////////////////////////////////

    private final OXManager
    getOXManager()
        {
            return (OXManager) Exehda.getService(OXManager.SERVICE_NAME);
        }
        
            
        /** ID of the ox */
    private ObjectId  id;
        /** OX instance (effective implementation)  */
    private Object    instance;
        /** Used to interact with the OXManager*/
    private OXHandle  handle;
        /** Custom (user-provider) activator */
    private Activator customActivator;
        /** Activation state */
    private boolean   isActive;
}
